/* -*-c++-*- */
#ifndef ACCUMULATORS_H
#define ACCUMULATORS_H
/* Accumulators : Helper classes for implementing argmax/inside in the
   same function.  They have an interface that allows "accumulating"
   values and then performing some "finishing" computations. */

#include "Macros.h"
#include "MathUtil.h"
#include "O3Signature.h"
#include "O3Chart.h"

#include <math.h>

/* accumulator that implements argmax */
class accum_argmax {
public:
  accum_argmax();
  void accum(const double sc, const backp bp);
  scobkp finish();
  static double get(const achart& restrict C, const o3sig s);
private:
  scobkp _best;
};


/* accumulator that implements partition-function */
class accum_partition {
public:
  accum_partition();
  void accum(const double sc, const backp bp);
  double finish();
  static double get(const dchart& restrict C, const o3sig s);
private:
  /* represent a summation in log-space with two numbers:
     _log_max = the largest value in the summation, in log-space
     _sum_over_max = the summation divided by max, in normal-space */
  double _log_max;
  double _sum_over_max;
};


/* accumulator for marginals(?) */
class accum_marginals {
};




/*** inlined implementation ***/

/* argmax accumulator */
inline accum_argmax::accum_argmax() : _best() {}
inline void accum_argmax::accum(const double sc, const backp bp) {
  if(sc > _best.sc) { _best.sc = sc; _best.bp = bp; }
}
inline scobkp accum_argmax::finish() {
  const scobkp ret = _best;
  _best = scobkp(); /* reinitialize with invalid score/backp */
  return ret;
}
inline double accum_argmax::
get(const achart& restrict C, const o3sig s) { return C.get(s).sc; }

/* partition-function accumulator */
inline accum_partition::accum_partition()
  : _log_max(-HUGE_VAL), _sum_over_max(0) {
  CHECK(assert(exp(_log_max) == 0));
}
inline void accum_partition::accum(const double sc, const backp bp) {
  _log_max = (sc > _log_max ? sc : _log_max);
  _sum_over_max = (sc > _log_max
                   ? 1.0 + _sum_over_max*exp(_log_max - sc)
                   : exp(sc - _log_max));
}
inline double accum_partition::finish() {
  const double ret = _log_max + log(_sum_over_max);
  _log_max = -HUGE_VAL; _sum_over_max = 0;
  return ret;
}
inline double accum_partition::
get(const dchart& restrict C, const o3sig s) { return C.get(s); }

#endif /* ACCUMULATORS_H */
