#include "ActiveDeps.h"

#include "Mappings.h"
#include "Example.h"

#include <iostream>
#include <iomanip>
#include <map>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "GZIO.h"

using std::map;
using std::cerr;
using std::endl;
using std::flush;
using std::fixed;
using std::setprecision;

#define LINELEN 65536

active_deps::active_deps(const char* const fname,
                         const double beam,
                         const int L,
                         corpus& restrict data,
                         const bool truth_marg,
                         const bool reattach_comma) {
  const bool labeled = (L > 1);
  bool warned = false;
  cerr << "ActiveDeps : reading active deps from \"" << fname
       << "\" " << flush;
  FILE* const f = gzio::gzopen(fname, "r");
  char buf[LINELEN];
  uint64_t nheads = 0;
  uint64_t nwords = 0;
  uint64_t ndeps = 0;
  vector<const int*> curL;
  vector<const int*> curH;
  vector<int> curN;
  data.start();
  const ex* e = data.next();
  while(true) {
    buf[0] = '\0';
    char* const line = fgets(buf, LINELEN, f);
    if(line == NULL) { break; }

    /* handle a finished sentence */
    if(line[0] == '\n') {
      assert(e->id() == (int)_sizes.size());
      const int N = (int)curL.size();
      assert(N > 0);
      assert(N == (int)e->heads.size());
      ndeps += N*N;
      _sizes.push_back(N);
      const int** actL = (const int**)malloc(N*sizeof(int*));
      const int** actH = (const int**)malloc(N*sizeof(int*));
      int* actN = (int*)malloc(N*sizeof(int));
      for(int i = 0; i < N; ++i) {
        actL[i] = curL[i];
        actH[i] = curH[i];
        actN[i] = curN[i];
      }
      _activeL.push_back(actL);
      _activeH.push_back(actH);
      _activeN.push_back(actN);
      curL.clear();
      curH.clear();
      curN.clear();
      if((_sizes.size() % 1000) == 0) {
        if((_sizes.size() % 10000) == 0) {
          cerr << "(" << _sizes.size()/1000 << "k)" << flush;
        } else {
          cerr << "." << flush;
        }
      }
      /* get the next sentence */
      e = data.next();
      continue;
    }

    char* p = line;
    char* oldp = p;

    /* check the sentence id */
    const int id = (int)strtol(p, &p, 10);
    assert(p != oldp); oldp = p;
    if(!warned && id != (int)_sizes.size()) {
      warned = true;
      cerr << "ActiveDeps : *warning: index mismatch "
           << id << " v " << _sizes.size() << endl;
    }

    /* check the word id */
    const int N = (int)e->heads.size();
    const int m = (int)strtol(p, &p, 10) - 1;
    assert(p != oldp); oldp = p;
    assert(m == (int)curN.size());
    assert(m < N);

    /* collect triples of head, label, marginal */
    vector<int> heads;
    vector<int> labels;
    vector<double> margs;
    double maxmarg = -1;

    /* deterministically reattach commas if requested */
    if(reattach_comma &&
       mappings::iscomma(WCONV(e->words[m]))) {
      /* reattach to left */
      const int h = m - 1;
      assert(h >= 0);
      const int l = 0;
      const double marg = 1.0;
      heads.push_back(h);
      labels.push_back(l);
      margs.push_back(marg);
    } else { /* general case */
      const int true_h = e->heads[m];
      const int true_l = e->labels[m];
      while(true) {
        const int h = (int)strtol(p, &p, 10) - 1;
        assert(h >= -1); assert(h < N);
        if(p == oldp) { break; }
        const int ll = (int)strtol(p, &p, 10);
        assert(p != oldp); oldp = p;
        const int l = (labeled ? ll : 0);
        assert(l >= 0); assert(l < L);
        const double marg = strtod(p, &p);
        assert(p != oldp); oldp = p;
        assert(marg >= 0);
        if(marg > maxmarg) { maxmarg = marg; }
        heads.push_back(h);
        labels.push_back(l);
        /* override actual score if truth is requested */
        if(truth_marg && h == true_h && l == true_l) {
          margs.push_back(1.0); /* should pass the beam */
        } else {
          margs.push_back(marg);
        }
      }
    }
    assert(heads.size() > 0);

    /* keep heads and labels that survive the beam */
    vector<int> pheads;
    vector<int> plabels;
    map<int,bool> known;
    const size_t Ntot = heads.size();
    const double thresh = maxmarg*beam; /* NB: relative beam */
    for(size_t i = 0; i < Ntot; ++i) {
      if(margs[i] >= thresh) {
        if(!labeled) {
          /* in the unlabeled case, duplicate heads can arise; check
             them with the map */
          if(known.find(heads[i]) != known.end()) { continue; }
          known[heads[i]] = true;
        }
        pheads.push_back(heads[i]);
        plabels.push_back(labels[i]);
      }
    }
    assert(pheads.size() > 0);

    /* convert to an array and save */
    const int Nact = (int)pheads.size();
    int* actH = (int*)malloc(Nact*sizeof(int));
    int* actL = (int*)malloc(Nact*sizeof(int));
    for(int i = 0; i < Nact; ++i) {
      actH[i] = pheads[i]; actL[i] = plabels[i];
    }
    curH.push_back(actH);
    curL.push_back(actL);
    curN.push_back(Nact);
    nheads += Nact;
    ++nwords;
  }
  gzio::gzclose(fname, f);

  ndeps *= L;
  assert(curN.size() == 0);
  cerr << " done" << endl;
  cerr << fixed << setprecision(2);
  cerr << "ActiveDeps : " << nheads << " heads / "
       << nwords << " words = "
       << ((double)nheads)/((double)nwords) << " average"
       << endl;
  cerr << "ActiveDeps : " << nheads << " heads / "
       << ndeps << " deps = "
       << 100.0*((double)nheads)/((double)ndeps) << "% retention"
       << endl;
}

active_deps::~active_deps() {
  const int total = (int)_sizes.size();
  for(int i = 0; i < total; ++i) {
    const int N = _sizes[i];
    free((void*)_activeN[i]);
    const int** actl = _activeL[i];
    const int** acth = _activeH[i];
    for(int j = 0; j < N; ++j) {
      free((void*)actl[j]); free((void*)acth[j]);
    }
    free((void**)_activeL[i]); free((void**)_activeH[i]);
  }
}
