#include "AuxCorpus.h"

#include "GZIO.h"

#include <stdio.h>
#include <string.h>

#include <iostream>

using std::cerr;
using std::endl;
using std::flush;

#define LINELEN 65536


void auxcorpus::load(const char* const fname, const int L) {
  assert(_aux.empty());

  cerr << "AuxCorpus : reading auxiliary deps from \"" << fname
       << "\" " << flush;


  FILE* const f = gzio::gzopen(fname, "r");
  _Naux = -1;

  char line[LINELEN];
  { /* read a line containing the number of auxdeps */
    line[0] = '\0';
    fgets(line, LINELEN, f);
    const int nread = sscanf(line, " %d", &_Naux);
    assert(nread == 1);
    assert(_Naux > 0);
  }

  while(1) { /* read auxdeps until EOF */
    vector<vector<int> > aheads;  /* aheads[aux][m] */
    vector<vector<int> > alabels;

    /* populate with empty vectors */
    for(int i = 0; i < _Naux; ++i) {
      aheads.push_back(vector<int>());
      alabels.push_back(vector<int>());
    }

    while(1) { /* read lines until blank-line */
      line[0] = '\0';
      const char* const stret = fgets(line, LINELEN, f);
      const size_t len = strlen(line);
      if(stret == NULL && !ferror(f) && feof(f)) { break; }

      /* check overflow */
      if(line[len - 1] != '\n') {
        cerr << "detected line overflow when reading data" << endl;
        assert(0);
      }

      /* check for empty or whitespace-only lines */
      if(len == 0) { break; } /* empty line */
      { /* whitespace-only lines */
        bool allspace = true;
        for(size_t i = 0; i < len; ++i) {
          if(!isspace(line[i])) { allspace = false; break; }
        }
        if(allspace) { break; }
      }

      /* read pairs of head,label for each aux-parser */
      char* p = line;
      for(int i = 0; i < _Naux; ++i) {
        int hh = -1, ll = -1, nchars = -1;
        const int nsc = sscanf(p, " %d %d%n", &hh, &ll, &nchars);
        assert(nsc == 2);
        assert(hh >= 0);
        assert(ll >= 0);
        assert(ll < L);
        assert(nchars > 0);
        p += nchars; /* advance past the pair */
        aheads[i].push_back(hh);
        alabels[i].push_back(ll);
      }
    }

    /* check for an empty sentence; indicating EOF */
    if(aheads[0].empty()) { break; }

    /* check the heads */
    const int NW = (int)aheads[0].size();
    for(int i = 0; i < _Naux; ++i) {
      assert(NW == (int)aheads[i].size());
      assert(NW == (int)alabels[i].size());
      const vector<int>& restrict hh = aheads[i];
      for(int m = 0; m < NW; ++m) { assert(hh[m] <= NW); }
    }

    /* create a new auxdeps */
    auxdeps auxd;
    for(int i = 0; i < _Naux; ++i) { auxd.add(aheads[i], alabels[i]); }
    _aux.push_back(auxd);

    if((_aux.size() % 1000) == 0) {
      if((_aux.size() % 10000) == 0) {
        cerr << "(" << _aux.size()/1000 << "k)" << flush;
      } else {
        cerr << "." << flush;
      }
    }
  }

  gzio::gzclose(fname, f);
  cerr << " done" << endl;
}
