/* -*-c++-*- */
#ifndef AVG_PARAM_H
#define AVG_PARAM_H
/* AvgParam : An object that implements fast parameter averaging via
   the usual perceptron timestamping trick */

#include "ParallelUtil.h"
#include "Macros.h"

/* forward declaration */
class o3params_avg;
class avg_param_util;

class avg_param {
public:
  /* construct with given value and sum, zero timestamp */
  avg_param(const double v, const double s);
  /* the big three */
  avg_param();
  avg_param(const avg_param& that);
  avg_param& operator=(const avg_param& that);
  /* equality of values */
  bool operator==(const avg_param& that) const;

  /* accessor parameter value/summed parameter */
  double val() const;
  double sum() const;

  /* make this consistent time t */
  void update(const uint64_t t);
  /* make this parameter consistent with time t and add d */
  void add(const uint64_t t, const double d);
  /* combine this with another averaged parameter */
  void add(const avg_param& restrict p);

private:
  friend class o3params_avg;
  friend class avg_param_util;
  double _val;           /* parameter value */
  double _sum;           /* summed parameter value */
  uint64_t _timestamp;   /* time of last update to this parameter */
};


/*** inlined implementation ***/

/* construct a zeroed object */
inline avg_param::avg_param() : _val(0), _sum(0), _timestamp(0) {}
/* construct with given value and sum, zero timestamp */
inline avg_param::avg_param(const double v, const double s)
  : _val(v), _sum(s), _timestamp(0) {}
/* copy constructor and assignment operator */
inline avg_param::avg_param(const avg_param& that)
  : _val(that._val), _sum(that._sum), _timestamp(that._timestamp) {}
inline avg_param& avg_param::operator=(const avg_param& that) {
  _val = that._val; _sum = that._sum; _timestamp = that._timestamp;
  return *this;
}

/* accessors */
inline double avg_param::val() const { return _val; }
inline double avg_param::sum() const { return _sum; }

/* equality testing: ignore timestamp */
inline bool avg_param::operator==(const avg_param& that) const {
  return (_val == that._val && _sum == that._sum);
}

/* make this parameter consistent with time t */
inline void avg_param::update(const uint64_t t) {
  _sum += _val*((double)(t - _timestamp)); _timestamp = t;
}
/* make this parameter consistent with time t and add d */
inline void avg_param::add(const uint64_t t, const double d) {
  update(t); _val += d;
}
/* combine this with another averaged parameter */
inline void avg_param::add(const avg_param& restrict p) {
  CHECK(assert(_timestamp == p._timestamp));
  _val += p._val;
  _sum += p._sum;
}



/* utility class: all methods public inline */
class avg_param_util {
public:
  static void fprint(FILE* const f, const avg_param& restrict p) {
    fprintf(f, "%.17g %.17g", p.val(), p.sum());
  }
  static avg_param fscan(FILE* const f) {
    double val = 0, sum = 0;
    const int nscanned = fscanf(f, " %lf %lf", &val, &sum);
    assert(nscanned == 2);
    return avg_param(val, sum);
  }
  static bool equals(const avg_param& restrict p1,
                     const avg_param& restrict p2,
                     const double tol) {
    const double diff_val = p1._val - p2._val;
    const double diff_sum = p1._sum - p2._sum;
    const double absdiff_val = (diff_val < 0 ? -diff_val : diff_val);
    const double absdiff_sum = (diff_sum < 0 ? -diff_sum : diff_sum);
    return (absdiff_val <= tol && absdiff_sum <= tol);
  }
  static void send(const avg_param& restrict p,
                   const int write_fd, const double scale) {
    CHECK(assert(p._timestamp == 0));
    double tmp = (scale*p._val);
    safe_write(write_fd, &tmp, sizeof(double));
    tmp = (scale*p._sum);
    safe_write(write_fd, &tmp, sizeof(double));
  }
  static void recv(avg_param& restrict p, const int read_fd) {
    CHECK(assert(p._timestamp == 0));
    double tmp = 0;
    safe_read(read_fd, &tmp, sizeof(double));
    p._val += tmp;
    safe_read(read_fd, &tmp, sizeof(double));
    p._sum += tmp;
  }
  static void add(avg_param& restrict p, const avg_param& restrict p2) {
    p.add(p2);
  }
};

#endif /* AVG_PARAM_H */
