/* -*-c++-*- */
#ifndef EGSTRA_BINHEAP_H
#define EGSTRA_BINHEAP_H
/* BinHeap : A binary heap implemented as an array.

   The implementation is templated on a key type Key, a value type
   Val, and a Key-comparator KeyLess.  KeyLess should have a static
   method less(k1, k2) that return true if k1 belongs "before" k2
   (i.e., k1 < k2) and false otherwise.  The heap is a minimum-heap
   with respect to the (partial) ordering defined by KeyLess.  The
   heap supports the following interface:

     const Key& mink()
     const Val& minv()
        Obtain the (key,val) pair that would have been returned by
        pop(), but do not alter the heap.  Time O(1).

     void pop()
        Remove a minimal (key,val) pair from the heap and discard it.
        Here, "minimal" means that the returned Key k* is such that
        there is no other Key k for which KeyLess::less(k, k*) is
        true.  Time O(log(N)), where N is the number of elements in
        the heap.

     void ins(const Key&, const Val&)
        Push the (key,val) pair onto the heap.  Time O(log(N))
        amortized (due to possible array doubling).

     void clear()
        Remove all elements from the heap.  Time O(1).


     const Key* kbegin()
     const Val* vbegin()
        Facilities for enumerating the elements of the heap in
        read-only fashion.  Time O(1) to return these pointers.

   This heap does not eliminate identical keys, so groups of identical
   keys will be stored separately and returned (in arbitrary order).
   The key and value types are assumed to be small and therefore easy
   to copy; thus literal Keys and Vals are passed, rather than
   references. */


#include "Macros.h"

#include <assert.h>
#include <stdlib.h>


#define BINHEAP_DEFSIZE 8

/* basic min/max heaps that use operator</operator> */
template<class Key, class Val> class bmin_heap;
template<class Key, class Val> class bmax_heap;

template<class Key, class Val, class KeyLess>
class bin_heap {
public:
  bin_heap(const int initsize = BINHEAP_DEFSIZE);
  /* bin_heap will be subclassed, but only in a trivial way; thus it
     is unnecessary to use a virtual destructor.  cf. bmax_heap and
     bmin_heap. */
  ~bin_heap();

  /* accessors */
  const Key& mink() const;    /* minimal key */
  const Val& minv() const;    /* associated value */
  int size() const;           /* number of elements */

  /* element enumeration: these pointers are valid from
     kbegin()[0] to kbegin()[n() - 1]
     vbegin()[0] to vbegin()[n() - 1] */
  const Key* kbegin() const;
  const Val* vbegin() const;

  /* insert a (key,val) pair */
  void ins(const Key& restrict k, const Val& restrict v);
  void pop();      /* discard minimal key,val pair */
  void clear();    /* remove all elements */

private:
  Key* _keys;    /* array of keys */
  Val* _vals;    /* array of values */
  size_t _size;  /* size of the arrays */
  size_t _N;     /* number of (key,val) pairs in the heap */
};




/*** inlined implementation ***/

/* we use an implicitly-defined heap array.  since the array is
   0-origin, the parent of index i is floor((i-1)/2), and the
   children of index i are 2*i + 1 and 2*i + 2. */

template<class Key, class Val, class KeyLess> inline
bin_heap<Key,Val,KeyLess>::bin_heap(const int initsize)
  : _keys((Key*)malloc(initsize*sizeof(Key))),
    _vals((Val*)malloc(initsize*sizeof(Val))),
    _size(initsize), _N(0), _less() {
  assert(_keys != NULL); assert(_vals != NULL);
}

template<class Key, class Val, class KeyLess> inline
bin_heap<Key,Val,KeyLess>::~bin_heap() { free(_keys); free(_vals); }


/* get a minimal key */
template<class Key, class Val, class KeyLess> inline const Key&
bin_heap<Key,Val,KeyLess>::mink() const {
  CHECK(assert(_N > 0)); return _keys[0];
}
/* get the value associated with the current minimum key */
template<class Key, class Val, class KeyLess> inline const Val&
bin_heap<Key,Val,KeyLess>::minv() const {
  CHECK(assert(_N > 0)); return _vals[0];
}

/* get the size of the heap */
template<class Key, class Val, class KeyLess> inline int
bin_heap<Key,Val,KeyLess>::size() const { return _N; }


/* remove all elements */
template<class Key, class Val, class KeyLess> inline void
bin_heap<Key,Val,KeyLess>::clear() { _N = 0; }


/* key enumeration */
template<class Key, class Val, class KeyLess> inline const Key*
bin_heap<Key,Val,KeyLess>::kbegin() const { return _keys; }
/* value enumeration */
template<class Key, class Val, class KeyLess> inline const Val*
bin_heap<Key,Val,KeyLess>::vbegin() const { return _vals; }






/* min/max binary heaps defined for convenience */

/* default less-than comparator */
template<class Key>
class bin_heap_key_lt {
public:
  static bool less(const Key& restrict k1,
                   const Key& restrict k2) const { return (k1 < k2); }
};

/* default greater-than comparator */
template<class Key>
class bin_heap_key_gt {
public:
  static bool less(const Key& restrict k1,
                   const Key& restrict k2) const { return (k1 > k2); }
};


/* basic min/max heaps */
template<class Key, class Val>
class bmin_heap : public bin_heap<Key,Val,bin_heap_key_lt<Key> > {
public:
  bmin_heap(const int initsize = BINHEAP_DEFSIZE)
    : bin_heap<Key,Val,bin_heap_key_lt<Key> >(initsize) {}
};

template<class Key, class Val>
class bmax_heap : public bin_heap<Key,Val,bin_heap_key_gt<Key> > {
public:
  bmax_heap(const int initsize = BINHEAP_DEFSIZE)
    : bin_heap<Key,Val,bin_heap_key_gt<Key> >(initsize) {}
};

#undef BINHEAP_DEFSIZE


#ifndef EXPLICIT_TEMPLATES
#include "BinHeap.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* EGSTRA_BINHEAP_H */
