#include "Macros.h"

/* we use an implicitly-defined heap array.  since the array is
   0-origin, the parent of index i is floor((i-1)/2), and the
   children of index i are 2*i + 1 and 2*i + 2. */


/* insert a new (key, value) pair into the heap */
template<class Key, class Val, class KeyLess> void
bin_heap<Key,Val,KeyLess>::ins(const Key& restrict k,
                               const Val& restrict v) {
  if(_N == _size) {
    _size <<= 1; /* array doubling */
    /* realloc() copies the existing array for us */
    _keys = (Key*)realloc(_keys, _size*sizeof(Key));
    _vals = (Val*)realloc(_vals, _size*sizeof(Val));
    assert(_keys != NULL); assert(_vals != NULL);
  }
  /* start from the end and sift upwards */
  int i = _N;       /* i will contain the destination of k,v */
  while(i != 0) {
    const int p = ((i - 1) >> 1); /* parent = floor((i-1)/2) */
    const Key& restrict pk = _keys[p];
    if(KeyLess::less(k, pk)) {     /* heap property violated */
      _keys[i] = pk;                        /* swap p into i */
      _vals[i] = _vals[p];
      i = p;                        /* recurse on the parent */
    } else {                      /* heap property satisfied */
      break;        /* k,v can safely be inserted at index i */
    }
  }
  /* i now contains the destination for k,v */
  _keys[i] = k;
  _vals[i] = v;
  ++_N;
}

/* remove the first (key, value) pair */
template<class Key, class Val, class KeyLess> void
bin_heap<Key,Val,KeyLess>::pop() {
  CHECK(assert(_N > 0));
  --_N;
  /* swap the last element into the root and sift downwards */
  const Key& restrict k = _keys[_N];
  int i = 0;       /* i will contain the destination of k,v */
  while(1) {
    const int c1 = ((i << 1) + 1);  /* leftchild  = 2*i + 1 */
    const int c2 = c1 + 1;          /* rightchild = 2*i + 2 */

    if(c1 >= _N) {  /* no children; heap property satisfied */
      break;       /* k,v can safely be inserted at index i */
    } else if(c2 >= _N) {          /* only one child exists */
      const Key& restrict k1 = _keys[c1];
      if(KeyLess::less(k1, k)) {  /* heap property violated */
        _keys[i] = k1;                    /* swap c1 into i */
        _vals[i] = _vals[c1];
        i = c1;              /* the final destination is c1 */
      }                          /* heap property satisfied */
      break;       /* k,v can safely be inserted at index i */
    } else {           /* general case: both children exist */
      const Key& restrict k1 = _keys[c1];
      const Key& restrict k2 = _keys[c2];
      if(KeyLess::less(k1, k) ||
         KeyLess::less(k2, k)) {  /* heap property violated */
        if(KeyLess::less(k1, k2)) { /* c1 is the new parent */
          _keys[i] = k1;                  /* swap c1 into i */
          _vals[i] = _vals[c1];
          i = c1;                          /* recurse on c1 */
        } else {                    /* c2 is the new parent */
          _keys[i] = k2;                  /* swap c2 into i */
          _vals[i] = _vals[c2];
          i = c2;                          /* recurse on c2 */
        }
      } else {                   /* heap property satisfied */
        break;     /* k,v can safely be inserted at index i */
      }
    }
  }
  /* i now contains the destination for (k,v) */
  _keys[i] = k;
  _vals[i] = _vals[_N];
}
