#include "Corpus.h"

#include "Timer.h"
#include "MathUtil.h"
#include "GZIO.h"
#include "Example.h"
#include "ActiveDeps.h"
#include "O3Oracle.h"
#include "O3Active.h"
#include "O1Parser.h"


#include <iostream>
#include <iomanip>

using std::cerr;
using std::endl;
using std::flush;
using std::fixed;
using std::setprecision;


/* types of data orderings */
#define DETERMINISTIC 0
#define UNIFORM       1
#define PERMUTATION   2


/* lifecycle */
corpus::corpus() : _data(), _i(0), _order(-1) {}
corpus::~corpus() {
  for(unsigned i = 0; i < _data.size(); ++i) {
    delete _data[i];
  }
  _data.clear();
}


/* set the example ordering */
void corpus::ordering(const string& type) {
  if(type == "det")       { _order = DETERMINISTIC; }
  else if(type == "unif") { _order = UNIFORM; }
  else if(type == "perm") { _order = PERMUTATION; }
  else {
    cerr << "Corpus : invalid ordering type " << type << endl;
    assert(0);
  }
}


/* return a descriptor for the example ordering */
string corpus::ostr() const {
  switch(_order) {
  case DETERMINISTIC: return string("D");
  case UNIFORM:       return string("R");
  case PERMUTATION:   return string("P");
  default:
    cerr << "Corpus : unknown ordering " << _order << endl;
    assert(0);
  }
  return string("???");
}

/* return a long descriptor for the example ordering */
string corpus::ostr_long() const {
  switch(_order) {
  case DETERMINISTIC: return string("deterministic");
  case UNIFORM:       return string("random with replacement");
  case PERMUTATION:   return string("random permutation");
  default:
    cerr << "Corpus : unknown ordering " << _order << endl;
    assert(0);
  }
  return string("???");
}


/* initialize the example index */
void corpus::start() { _i = 0; }

/* obtain the next example */
const ex* corpus::next() {
  const size_t Ntr = _data.size();
  if(_i >= Ntr) { return NULL; }

  ex* ret = NULL; /* the return value */
  switch(_order) {
  case DETERMINISTIC: ret = _data[_i]; break;
  case UNIFORM:       ret = _data[rand_int(Ntr)]; break;
  case PERMUTATION: {
    /* choose uniformly from remaining indices */
    const size_t idx = _i + rand_int(Ntr - _i);
    assert(idx < Ntr);
    /* swap into current i'th position */
    ret = _data[idx];
    _data[idx] = _data[_i];
    _data[_i] = ret;
    break;
  }
  default:
    cerr << "Corpus : unknown ordering " << _order << endl;
    assert(0);
  }
  assert(ret != NULL);

  /* advance and return */
  ++_i; return ret;
}


corpus* corpus::load(const string& fname, const int L) {
  cerr << "Corpus : loading data from \"" << fname << "\"" << flush;
  timer_start();
  corpus* const data = new corpus();
  assert(data != NULL);
  FILE* const f = gzio::gzopen(fname.c_str(), "r");
  ex* e = NULL;
  while((e = ex::fread(f, L)) != NULL) {
    /* set ids in ascending order */
    e->set_id(data->N());
    data->_data.push_back(e);
    if(((data->N()) & 0x3ff) == 0) {
      if(((data->N()) & 0x1fff) == 0) {
        const double kilos = ((double)data->N())/1000.0;
        cerr << setprecision(1) << fixed << "(" << kilos << ")" << flush;
      } else {
        cerr << "." << flush;
      }
    }
  }
  gzio::gzclose(fname.c_str(), f);
  timer_stop();
  cerr << " done [" << timestr(timediff) << "]" << endl;
  return data;
}


void corpus::relabel(const int L, const bool multiroot,
                     const active_deps* const restrict A,
                     FILE* const fout) {
  cerr << "Corpus : relabeling data..." << endl;
  o1parser P(L, multiroot);
  o3active act;
  const int Ntr = N();


  size_t nwords = 0;
  size_t ncorrect = 0;
  ticker_adjust_repint(4);
  ticker64_start();
  for(int i = 0; i < Ntr; ++i) {
    ex* const restrict e = _data[i];
    const int NW = (int)e->words.size();
    assert(NW == (int)e->labels.size());
    assert(NW == (int)e->words.size());
    assert(NW == (int)e->tags.size());

    /* oracle scorer that places weight 1 on the correct deps */
    o3oracle S(e->heads, e->labels);

    /* parse with or without pruning */
    vector<int> hhat, lhat;
    if(A != NULL) {
      assert(A->active_N(e->id()) == NW);
      act.set_active(false, L, NW,
                     A->active_num(e->id()),
                     A->active_heads(e->id()),
                     A->active_labels(e->id()));
      P.argmax(NW, S, act, hhat, lhat);
    } else {
      P.argmax(NW, S, hhat, lhat);
    }
    assert(NW == (int)hhat.size());
    assert(NW == (int)lhat.size());

    /* score the new tree */
    nwords += NW;
    for(int m = 0; m < NW; ++m) {
      const bool correct = ((hhat[m] == e->heads[m]) &&
                            (lhat[m] == e->labels[m]));
      // Follow-up: this sucks in practice
      //      /* EXPERIMENTAL: assign wrong edges a "garbage" label */
      //      if(!correct) { lhat[m] = L; }
      ncorrect += (correct ? 1 : 0);
    }

    /* save the new tree to disk if necessary */
    if(fout != NULL) { e->fwrite(fout, hhat, lhat); }

    /* copy the new tree; NB: the original labels are still saved in
       e->heads_orig and e->labels_orig */
    e->heads  = hhat;
    e->labels = lhat;

    ticker64(i, Ntr);
  }
  ticker64_finish(Ntr);
  ticker_adjust_repint(-4);

  cerr << "Corpus : relabeling accuracy = " << fixed << setprecision(2)
       << 100.0*((double)ncorrect)/((double)nwords)
       << " ; " << ncorrect << " / " << nwords << endl;
}
