/* -*-c++-*- */
#ifndef HASHTABLE_H
#define HASHTABLE_H
/* Hashtable : A hashtable with bucket lists to handle collisions.
   Templated on keys, values, and a hashfunction that hashes keys.
   Keys are compared using "==" so operator== should be redefined if
   required.  The hashfunction has a default implementation. */

#include "Macros.h"
#include "Hashfunction.h"

#include <stdlib.h>

#define DEF_BITLEN 8

template<class Key, class Val, class Hashf = hashfunction<Key> >
class hashtable {
public:
  hashtable(const int bitlen = DEF_BITLEN);
  ~hashtable();

  /* nested classes */
  class entry;          /* hashtable entry */
  /* implement the iterators as a template to avoid duplicated code */
  template<class Entry> class iter;
  /* read-write and read-only iterators */
  typedef iter<entry> iterator;
  typedef iter<const entry> const_iterator;

  /* enumeration of key/value pairs */
  iterator begin();
  iterator end();
  const_iterator cbegin() const;
  const_iterator cend() const;

  /* check whether a key exists in the table */
  bool has(const Key& restrict k) const;
  /* retrieve a value from the table.  it is an error if Key k does
     not already exist in the table */
  const Val& get(const Key& restrict k) const;
  /* retrieve a value from the table.  if Key k does not exist, then
     the value defval is returned instead */
  const Val& get(const Key& restrict k, const Val& restrict defval) const;
  /* retrieve a pointer to an entry in the table.  if Key k does not
     exist, then NULL is returned; otherwise a pointer to the relevant
     entry is returned.  note that this method allows both detection
     of key existence and value retrieval simultaneously. */
  const entry* pget(const Key& restrict k) const;
  /* retrieve a modifiable reference to a value in the table.  it is
     an error if Key k does not already */
  Val& mget(const Key& restrict k);
  /* insert/retrieve a modifiable reference to a value in the table.
     if Key k does not already exist, then it is inserted with value
     v, otherwise no changes are made.  a reference to the value
     (current or newly inserted) associated with k is returned */
  Val& mget(const Key& restrict k, const Val& restrict v);
  /* insert a value into the table.  it is an error if Key k already
     exists in the table */
  void ins(const Key& restrict k, const Val& restrict v);
  /* remove a value from the table.  it is an error if Key k does not
     already exist in the table */
  void del(const Key& restrict k);
  /* remove a value from the table.  it is *not* an error if Key k
     does not already exist in the table */
  void delq(const Key& restrict k);

  /* the number of elements in the table */
  uint64_t size() const;
  /* remove all entries from the table */
  void clear();
  /* remove all entries with matching value; return the number of
     entries deleted */
  uint64_t erasev(const Val& restrict v);
  /* set the size of the table to 2^{newbits}; may be more or less
     than current size */
  void setbits(const int newbits);

  /* the number of buckets (i.e., size of the table) */
  uint64_t buckets() const;
  /* compute hashtable statistics:
     num_collisions = #entries beyond the head of the bucket list
     max_collisions = longest bucket list (beyond the head)
     num_active = #nonempty buckets */
  void collision_stats(uint64_t& restrict num_collisions,
                       uint64_t& restrict max_collisions,
                       uint64_t& restrict num_active) const;

private:
  /* double the size of the current table */
  void doublesize();

  /* private data members */
  uint64_t _N;    /* the number of (active) entries in the table */
  uint64_t _Nmax; /* the maximum number of entries before doubling */
  uint64_t _size; /* the total number of entries in the table */
  hashcode _mask; /* a mask to convert a hashcode to a table index */
  hashcode _bits; /* the number of bits in the bit-mask */
  entry* _T;      /* the table of bucket entries; the first entry in
                     each bucket is stored inline in this table */
};




/*** nested classes ***/

/* a nested bucket entry in the table; next pointers have three
   separate semantics:
   (1) next == this : the current entry is not a valid entry
   (2) next == NULL : the current entry is the last entry
   (3) otherwise    : next holds the next (valid) entry */
template<class T> T* safe_new(const uint64_t n);
template<class Key, class Val, class Hashf>
class hashtable<Key,Val,Hashf>::entry {
private:
  entry* next; /* users can't walk the list */

public:
  Key key;
  Val val;

private:
  friend class hashtable<Key,Val,Hashf>;
  friend entry* safe_new<>(const uint64_t n);
  /* splice this in at the given point in the list */
  entry(const Key& restrict k, const Val& restrict v,
        entry* const restrict here);
  /* create an invalid entry */
  entry();
  ~entry();

  /* disable these routines */
  entry(const entry& that)            { assert(0); }
  entry& operator=(const entry& that) { assert(0); return *this; }
};


/* a nested iterator for hashtable entries */
template<class Key, class Val, class Hashf>
template<class Entry>
class hashtable<Key,Val,Hashf>::iter {
public:
  /* copy constructor, default constructor, destructor */
  iter(const iter& that);
  iter();
  ~iter();

  /* iterator operators */
  iter& operator=(const iter& that);
  Entry& operator*();
  Entry* operator->();
  void operator++();

  /* (sufficient) equality testing */
  bool operator==(const iter& that) const;
  bool operator!=(const iter& that) const;

private:
  friend class hashtable<Key,Val,Hashf>;
  /* create an iter starting at point beg in the table and ending
     at point end, which points one past-the-end of the table */
  iter(Entry* const beg, Entry* const eend);

  Entry* e;         /* pointer to current entry */
  Entry* bucket;    /* head of current bucket list (in table) */
  Entry* end; /* one past the end of the current table */
};




/*** inlined implementation ***/

/* main hashtable class */

/* size access */
template<class Key, class Val, class Hashf> inline uint64_t
hashtable<Key,Val,Hashf>::size() const { return _N; }

/* hashtable statistics */
template<class Key, class Val, class Hashf> inline uint64_t
hashtable<Key,Val,Hashf>::buckets() const { return _size; }

/* double the current table */
#ifdef CAUTIOUS
#include <iostream>
using std::cerr;
using std::flush;
#endif /* CAUTIOUS */
template<class Key, class Val, class Hashf>
inline void hashtable<Key,Val,Hashf>::doublesize() {
#ifdef CAUTIOUS
  // static uint64_t numcalls = 0;
  // static const char spin[4] = {'/', '-', '\\', '|'};
  // cerr << '\b' << spin[(++numcalls) & 0x3] << flush;
#endif /* CAUTIOUS */
  setbits(_bits + 1);
}



/* nested iterator class */

/* copy constructor, default constructor, destructor */
template<class Key, class Val, class Hashf> template<class Entry>
inline hashtable<Key,Val,Hashf>::iter<Entry>::
iter(const iter& that) : e(that.e), bucket(that.bucket), end(that.end) {}
template<class Key, class Val, class Hashf> template<class Entry>
inline hashtable<Key,Val,Hashf>::iter<Entry>::
iter() : e(NULL), bucket(NULL), end(NULL) {}
template<class Key, class Val, class Hashf> template<class Entry>
inline hashtable<Key,Val,Hashf>::iter<Entry>::~iter() {}

/* assignment operator */
template<class Key, class Val, class Hashf> template<class Entry>
inline hashtable<Key,Val,Hashf>::iter<Entry>&
hashtable<Key,Val,Hashf>::iter<Entry>::operator=(const iter& that) {
  e = that.e; bucket = that.bucket; end = that.end; return *this;
}

/* dereference member by pointer; members of this appear to be members
   of entry */
template<class Key, class Val, class Hashf> template<class Entry>
inline Entry& hashtable<Key,Val,Hashf>::iter<Entry>::
operator*() { return *e; }
template<class Key, class Val, class Hashf> template<class Entry>
inline Entry* hashtable<Key,Val,Hashf>::iter<Entry>::
operator->() { return e; }

/* preincrement (return whether there is a valid next element) */
template<class Key, class Val, class Hashf> template<class Entry>
inline void hashtable<Key,Val,Hashf>::iter<Entry>::operator++() {
  CHECK(assert(bucket < end)); /* check validity; before end */
  /* there are three possibilities:
     (1) there is a next element in the current bucket
     (2) there is a next element in a following bucket
     (3) there is no next element before the end */
  if(e->next != NULL) { /* bucket contains more elements */
    e = e->next;
  } else { /* bucket is empty; check following buckets */
    do {
      ++bucket; /* check the next bucket */
      /* repeat while the bucket is valid (before end) and
         the bucket is empty (self-loop) */
    } while(bucket < end && bucket->next == bucket);
    e = bucket; /* point to the first bucket */
  }
}

  /* (sufficient) equality testing */
template<class Key, class Val, class Hashf> template<class Entry>
inline bool hashtable<Key,Val,Hashf>::iter<Entry>::
operator==(const iter& that) const { return (e == that.e); }
template<class Key, class Val, class Hashf> template<class Entry>
inline bool hashtable<Key,Val,Hashf>::iter<Entry>::
operator!=(const iter& that) const { return (e != that.e); }


#undef DEF_BITLEN

#ifndef EXPLICIT_TEMPLATES
#include "Hashtable.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* HASHTABLE_H */
