#include "SafeNew.h"

#include <stdlib.h>
#include <assert.h>

#include <vector>
#include <limits>

using std::vector;


/* maximum value of 64-bit unsigned integer */
#define UINT64_MAX (std::numeric_limits<uint64_t>::max())

/* maximum size of the hashtable (log-base-2) */
#define MAX_BITLEN 32
/* maximum load factor (unless max-size is reached) */
#define MAX_LOADFAC 1.0


/* nested entry implementation */
template<class Key, class Val, class Hashf> inline
hashtable<Key,Val,Hashf>::entry::entry(const Key& restrict k,
                                       const Val& restrict v,
                                       entry* const restrict here)
  : next(here->next), key(k), val(v) {
  CHECK(assert(next != here)); /* check null list */
  here->next = this;
}

/* set next pointer to self-loop, indicating null */
template<class Key, class Val, class Hashf> inline
hashtable<Key,Val,Hashf>::entry::entry() : next(this), key(), val() {}

/* delete children if necessary */
template<class Key, class Val, class Hashf>
hashtable<Key,Val,Hashf>::entry::~entry() {
  if(next != this && next != NULL) { delete next; }
}


/* nested iterator implementation */

/* create an iter starting at point beg in the table and ending at
   point end, which points one past-the-end of the table */
template<class Key, class Val, class Hashf> template<class Entry>
inline hashtable<Key,Val,Hashf>::iter<Entry>::
iter(Entry* const beg, Entry* const eend)
  : e(beg), bucket(beg), end(eend) {
  CHECK(assert(bucket <= end));
  /* we need to find the first valid bucket; the bucket must be
     before the end and not be a self loop */
  while(bucket < end && bucket->next == bucket) { ++bucket; }
  e = bucket; /* point to the first element of this bucket */
}


/* hashtable implementation */

/* lifecycle */
template<class Key, class Val, class Hashf>
hashtable<Key,Val,Hashf>::hashtable(const int bitlen)
  : _N(0),
    _Nmax(0),
    _size(UINT64_C(1) << bitlen),
    _mask((uint32_t)((_size - 1) & 0xffffffff)),
    _bits(bitlen),
    _T(safe_new<entry>(_size)) {
  assert(_bits <= MAX_BITLEN);
  /* compute load factor threshold */
  _Nmax = (_bits == MAX_BITLEN ? UINT64_MAX /* can't double anymore */
           : ((uint64_t)(((double)_size)*MAX_LOADFAC)));
  CHECK(/* the constructor creates self-loops */
  for(uint64_t i = 0; i < _size; ++i) { assert(_T[i].next == (_T + i)); });
}

/* rely on entry destructor to delete children */
template<class Key, class Val, class Hashf>
hashtable<Key,Val,Hashf>::~hashtable() { delete [] _T; }



/* read-write iterator pointing to the start of the table */
template<class Key, class Val, class Hashf>
typename hashtable<Key,Val,Hashf>::iterator
hashtable<Key,Val,Hashf>::begin() {
  return iterator(_T, _T + _size);
}

/* read-write iterator pointing to the end of the table */
template<class Key, class Val, class Hashf>
typename hashtable<Key,Val,Hashf>::iterator
hashtable<Key,Val,Hashf>::end() {
  return iterator(_T + _size, _T + _size);
}

/* read-only iterator pointing to the start of the table */
template<class Key, class Val, class Hashf>
typename hashtable<Key,Val,Hashf>::const_iterator
hashtable<Key,Val,Hashf>::cbegin() const {
  return const_iterator(_T, _T + _size);
}

/* read-only iterator pointing to the end of the table */
template<class Key, class Val, class Hashf>
typename hashtable<Key,Val,Hashf>::const_iterator
hashtable<Key,Val,Hashf>::cend() const {
  return const_iterator(_T + _size, _T + _size);
}



/* hashtable access */

//#define INLINE inline
#define INLINE

/* check existence of a key */
template<class Key, class Val, class Hashf>
INLINE bool hashtable<Key,Val,Hashf>::
has(const Key& restrict k) const {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  const entry* restrict e = _T + idx;
  if(e->next == e) { return false; } /* empty bucket */
  /* search for an element with matching key */
  while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
  return (e != NULL);
}

/* retrieve a value from the hashtable */
template<class Key, class Val, class Hashf>
INLINE const Val& hashtable<Key,Val,Hashf>::
get(const Key& restrict k) const {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  const entry* restrict e = _T + idx;
  CHECK(assert(e->next != e)); /* bucket can't be empty */
  /* search for an element with matching key */
  while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
  CHECK(assert(e != NULL)); /* there must be a matching key */
  CHECK(assert(Hashf::eq(e->key, k)));
  return e->val;
}

/* retrieve a value with default return */
template<class Key, class Val, class Hashf>
INLINE const Val& hashtable<Key,Val,Hashf>::
get(const Key& restrict k, const Val& restrict defval) const {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  const entry* restrict e = _T + idx;
  if(e->next == e) { return defval; } /* empty bucket */
  /* search for an element with matching key */
  while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
  return (e == NULL ? defval : e->val);
}

/* retrieve a pointer to value */
template<class Key, class Val, class Hashf>
INLINE const typename hashtable<Key,Val,Hashf>::entry*
hashtable<Key,Val,Hashf>::
pget(const Key& restrict k) const {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  const entry* restrict e = _T + idx;
  if(e->next == e) { return NULL; } /* empty bucket */
  /* search for an element with matching key */
  while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
  return e;
}

/* retrieve a modifiable value */
template<class Key, class Val, class Hashf>
INLINE Val& hashtable<Key,Val,Hashf>::
mget(const Key& restrict k) {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  entry* restrict e = _T + idx;
  CHECK(assert(e->next != e)); /* bucket can't be empty */
  while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
  CHECK(assert(e != NULL)); /* there must be a matching key */
  CHECK(assert(Hashf::eq(e->key, k)));
  return e->val;
}

/* retrieve a value with default insertion */
template<class Key, class Val, class Hashf>
INLINE Val& hashtable<Key,Val,Hashf>::
mget(const Key& restrict k, const Val& restrict v) {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  /* e initially points to the head of the bucket list and is set to
     point to the entry with key k (newly-created if necessary) */
  entry* restrict e = _T + idx;
  if(e->next == e) { /* empty bucket, insert immediately */
    e->key = k; e->val = v; e->next = NULL;
    ++_N; /* additional entry */
  } else { /* nonempty bucket, check for duplicates */
    while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
    if(e == NULL) { /* no matching keys */
      e = new entry(k, v, _T + idx); /* create an entry for (k,v) */
      CHECK(assert(e == _T[idx].next)); /* e after the head */
      ++_N; /* additional entry */
    } /* else e points to an entry with matching key */
  }
  CHECK(assert(Hashf::eq(e->key, k)));
  /* check if doubling is necessary */
  if(_N > _Nmax) { /* must double the table */
    doublesize(); assert(_N <= _Nmax);
    /* e points into the old table; we must *re*-retrieve the
       reference to the appropriate value.  k is guaranteed to be in
       the table now and mget(k) will not cause any further doublings
       so the following suffices */
    return mget(k);
  } else { /* no doubling; e is a valid pointer */
    return e->val;
  }
}

/* insert a value into the hashtable */
template<class Key, class Val, class Hashf>
INLINE void hashtable<Key,Val,Hashf>::
ins(const Key& restrict k, const Val& restrict v) {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  entry* restrict e = _T + idx;
  if(e->next == e) { /* empty bucket, insert immediately */
    e->key = k; e->val = v; e->next = NULL;
  } else { /* nonempty bucket, check for duplicates */
    while(e != NULL && !Hashf::eq(e->key, k)) { e = e->next; }
    CHECK(assert(e == NULL)); /* no duplicates found */
    e = new entry(k, v, _T + idx);
    CHECK(assert(e == _T[idx].next));
  }
  ++_N; /* additional entry */
  /* check if doubling is necessary */
  if(_N > _Nmax) { doublesize(); assert(_N <= _Nmax); }
}

/* delete a value from the hashtable with error if k is not present */
template<class Key, class Val, class Hashf>
INLINE void hashtable<Key,Val,Hashf>::
del(const Key& restrict k) {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  entry* restrict e = _T + idx;
  CHECK(assert(e->next != e)); /* bucket can't be empty */
  entry* restrict pe = NULL;
  while(e != NULL && !Hashf::eq(e->key, k)) { pe = e; e = e->next; }
  CHECK(assert(e != NULL)); /* there must be a matching key */
  /* now delete the actual element */
  if(e == _T + idx) { /* e is the head of the bucket list */
    if(e->next == NULL) { /* single element; invalidate head */
      e->next = e; /* self-loop indicates null */
    } else { /* has a child; copy that onto the head */
      entry* const restrict ne = e->next;
      /* copy the values of the child; splicing it out */
      e->key  = ne->key;
      e->val  = ne->val;
      e->next = ne->next;
      /* delete the spliced-out child; NB avoid recursive deletion */
      ne->next = NULL; delete ne;
    }
  } else { /* beyond the head of the bucket; standard list splice */
    pe->next = e->next; /* e is spliced out of the list */
    /* delete the spliced-out child; NB avoid recursive deletion */
    e->next = NULL; delete e;
  }
  CHECK(assert(_N > 0));
  --_N; /* removed entry */
}

/* delete a value from the hashtable with no error checking */
template<class Key, class Val, class Hashf>
INLINE void hashtable<Key,Val,Hashf>::
delq(const Key& restrict k) {
  const hashcode idx = (Hashf::hash(k) & _mask);
  CHECK(assert(idx < _size));
  entry* restrict e = _T + idx;
  if(e->next == e) { return; } /* empty bucket; no deletion */
  entry* restrict pe = NULL;
  while(e != NULL && !Hashf::eq(e->key, k)) { pe = e; e = e->next; }
  if(e == NULL) { return; } /* no matching key; no deletion */
  /* now delete the actual element */
  if(e == _T + idx) { /* e is the head of the bucket list */
    if(e->next == NULL) { /* single element; invalidate head */
      e->next = e; /* self-loop indicates null */
    } else { /* has a child; copy that onto the head */
      entry* const restrict ne = e->next;
      /* copy the values of the child; splicing it out */
      e->key  = ne->key;
      e->val  = ne->val;
      e->next = ne->next;
      /* delete the spliced-out child; NB avoid recursive deletion */
      ne->next = NULL; delete ne;
    }
  } else { /* beyond the head of the bucket; standard list splice */
    pe->next = e->next; /* e is spliced out of the list */
    /* delete the spliced-out child; NB avoid recursive deletion */
    e->next = NULL; delete e;
  }
  CHECK(assert(_N > 0));
  --_N; /* removed entry */
}


/* remove all entries without resizing the underlying table */
template<class Key, class Val, class Hashf>
void hashtable<Key,Val,Hashf>::clear() {
  for(uint64_t i = 0; i < _size; ++i) {
    entry* const restrict e = _T + i;
    entry* const restrict n = e->next;
    if(n != e && n != NULL) { delete n; }
    e->next = e; /* invalidate the bucket */
  }
  _N = 0; /* size is now 0 */
}

/* remove all entries with matching value */
template<class Key, class Val, class Hashf>
uint64_t hashtable<Key,Val,Hashf>::erasev(const Val& restrict v) {
  /* save keys for deletion afterwards to avoid invalidating iterators */
  vector<Key> keys;

  /* find equal values */
  iterator it = begin();
  const iterator iend = end();
  for(; it != iend; ++it) {
    if(it->val == v) { keys.push_back(it->key); }
  }

  /* delete keys for equal values */
  const size_t Nkeys = keys.size();
  for(size_t i = 0; i < Nkeys; ++i) { del(keys[i]); }

  return keys.size();
}


/* set the table size to 2^{newbits} */
template<class Key, class Val, class Hashf>
void hashtable<Key,Val,Hashf>::setbits(const int newbits) {
  assert(newbits <= MAX_BITLEN);
  /* use iterator interface to enumerate old entries; save these now
     before changing the state of the table */
  const entry* oldT = _T;
  const uint64_t oldN = _N;
  const_iterator it = cbegin();
  const const_iterator iend = cend();

  /* reinitialize the table with double the size */
  _N = 0;
  _bits = newbits;
  _size = (UINT64_C(1) << _bits);
  _mask = ((uint32_t)((_size - 1) & 0xffffffff));
  /* compute load factor threshold */
  _Nmax = (_bits == MAX_BITLEN ? UINT64_MAX /* can't double anymore */
           : ((uint64_t)(((double)_size)*MAX_LOADFAC)));
  _T = safe_new<entry>(_size);
  CHECK(/* the constructor creates self-loops */
  for(uint64_t i = 0; i < _size; ++i) { assert(_T[i].next == (_T + i)); });

  /* insert old elements into the doubled table */
  for(; it != iend; ++it) { ins(it->key, it->val); }
  assert(_N == oldN);

  /* destroy the old table */
  delete [] oldT;
}


/* hashtable statistics */
template<class Key, class Val, class Hashf>
void hashtable<Key,Val,Hashf>::
collision_stats(uint64_t& restrict num_collisions,
                uint64_t& restrict max_collisions,
                uint64_t& restrict num_active) const {
  uint64_t sumc = 0, maxc = 0;
  const entry* restrict bucket = _T;
  const entry* const bucket_end = _T + _size;
  /* search for a nonempty bucket */
  uint64_t nactive = 0;
  for(; bucket < bucket_end; ++bucket) {
    if(bucket->next != bucket) {
      ++nactive; /* an active bucket */
      uint64_t len = 0;
      /* count starting from after the head */
      for(const entry* restrict e = bucket->next;
          e != NULL; e = e->next) { ++len; }
      if(len > maxc) { maxc = len; }
      sumc += len;
    }
  }
  assert(_N == 0 || sumc < _N);
  assert(sumc + nactive == _N);
  num_collisions = sumc;
  max_collisions = maxc;
  num_active = nactive;
}


#undef UINT64_MAX

#undef MAX_BITLEN
#undef MAX_LOADFAC
