#include "Mappings.h"

#include "GZIO.h"

#include <string.h>
#include <stdio.h>

#include <iostream>
#include <vector>

using std::cerr;
using std::endl;
using std::flush;
using std::vector;


/* static arrays of indicators */
int         mappings::_nw             = -1;
int         mappings::_nft            = -1;
int         mappings::_nct            = -1;
int         mappings::_nwft           = -1;
int         mappings::_nwct           = -1;
bool*       mappings::_tag_is_coord   = NULL;
bool*       mappings::_tag_is_prep    = NULL;
bool*       mappings::_tag_is_punc    = NULL;
bool*       mappings::_tag_is_verb    = NULL;
tag_t*      mappings::_coarse_tag     = NULL;
string*     mappings::_ftnames        = NULL;
string*     mappings::_ctnames        = NULL;
string*     mappings::_wnames         = NULL;
uint32_t*   mappings::_rev_wordftag   = NULL;
uint32_t*   mappings::_rev_wordctag   = NULL;
mappings::wtmap mappings::_wordftag;
mappings::wtmap mappings::_wordctag;
mappings::wbmap mappings::_wquote;
mappings::wbmap mappings::_wparen;
word_t      mappings::_wcomma         = WNULL;


void mappings::load(const char* const dir) {
  cerr << "Mappings : loading from \""
       << dir << "\"..." << flush;
  assert(_tag_is_coord == NULL);
  assert(_tag_is_prep  == NULL);
  assert(_tag_is_punc  == NULL);
  assert(_tag_is_verb  == NULL);
  assert(_coarse_tag   == NULL);

  map<string,int> fs2i;

  const string ftfname = string(dir) + "/ftags.map";
  FILE* f = gzio::gzopen(ftfname.c_str(), "r");
  char tag[1024];
  while(true) {
    int idx = -1;
    const int nscanned = fscanf(f, " %d %1023s", &idx, tag);
    if(nscanned == EOF && !ferror(f) && feof(f)) { break; }
    assert(nscanned == 2);
    assert(idx == (int)fs2i.size());
    assert(fs2i.find(tag) == fs2i.end());
    fs2i[tag] = idx;
  }
  gzio::gzclose(ftfname.c_str(), f);
  _nft = (int)fs2i.size();

  { /* copy the names */
    _ftnames = new string[_nft];
    map<string,int>::const_iterator it = fs2i.begin();
    const map<string,int>::const_iterator iend = fs2i.end();
    for(; it != iend; ++it) {
      assert(_ftnames[it->second].empty());
      _ftnames[it->second] = it->first;
      assert(!_ftnames[it->second].empty());
    }
  }

  cerr << " [" << _nft << " ftags]" << flush;



  /* create a fine tag to coarse tag mapping */
  map<string,tag_t> cs2i;
  _coarse_tag = new tag_t[_nft];
  for(int i = 0; i < _nft; ++i) { _coarse_tag[i] = TNULL; }

  /* load ftag->ctag mapping */
  const string f2cfname = string(dir) + "/f2ctag.map";
  f = gzio::gzopen(f2cfname.c_str(), "r");
  char tag2[1024];
  while(true) {
    const int nscanned = fscanf(f, " %1023s %1023s", tag, tag2);
    if(nscanned == EOF && !ferror(f) && feof(f)) { break; }
    assert(nscanned == 2);
    assert(fs2i.find(tag) != fs2i.end());
    const int fidx = fs2i[tag];
    if(cs2i.find(tag2) == cs2i.end()) {
      const int cidx = (int)cs2i.size();
      cs2i[tag2] = TCONV(cidx);
    }
    _coarse_tag[fidx] = cs2i[tag2];
  }
  gzio::gzclose(f2cfname.c_str(), f);
  _nct = (int)cs2i.size();

  { /* copy the names */
    _ctnames = new string[_nct];
    map<string,tag_t>::const_iterator it = cs2i.begin();
    const map<string,tag_t>::const_iterator iend = cs2i.end();
    for(; it != iend; ++it) {
      assert(_ctnames[it->second].empty());
      _ctnames[it->second] = it->first;
      assert(!_ctnames[it->second].empty());
    }
  }

  /* check that all coarse tags are set */
  for(int i = 0; i < _nft; ++i) { assert(_coarse_tag[i] != TNULL); }

  cerr << " [" << _nct << " ctags]" << flush;




  /* read tag-info */
  _tag_is_coord = new bool[_nft];
  _tag_is_prep  = new bool[_nft];
  _tag_is_punc  = new bool[_nft];
  _tag_is_verb  = new bool[_nft];
  bool* const restrict visited = new bool[_nft];
  for(int i = 0; i < _nft; ++i) {
    _tag_is_coord[i] = false;
    _tag_is_punc[i]  = false;
    _tag_is_prep[i]  = false;
    _tag_is_verb[i]  = false;
    visited[i] = false;
  }

  const string tifname = string(dir) + "/taginfo";
  f = gzio::gzopen(tifname.c_str(), "r");
  while(true) {
    const int nscanned = fscanf(f, " %1023s %1023s", tag, tag2);
    if(nscanned == EOF && !ferror(f) && feof(f)) { break; }
    assert(nscanned == 2);
    map<string,int>::const_iterator it = fs2i.find(tag);
    assert(it != fs2i.end());
    const int idx = it->second;
    assert(!visited[idx]);
    visited[idx] = true;
    const string flag(tag2);
    if(flag == "+coord") { /* coordination */
      _tag_is_coord[idx] = true;
    } else if(flag == "+prep") { /* preposition */
      _tag_is_prep[idx] = true;
    } else if(flag == "+punc") { /* punctuation */
      _tag_is_punc[idx] = true;
    } else if(flag == "+verb") { /* verb */
      _tag_is_verb[idx] = true;
    } else if(flag != "_") { /* no tag features */
      cerr << "invalid tag features: " << tag << " " << tag2 << endl;
      assert(0);
    }
  }
  gzio::gzclose(tifname.c_str(), f);
  for(int i = 0; i < _nft; ++i) { assert(visited[i]); }
  delete [] visited;

  cerr << " [" << _nft << " taginfo]" << flush;




  /* load the word mapping */
  map<string,word_t> ws2i;
  const string wfname = string(dir) + "/words.map";
  f = gzio::gzopen(wfname.c_str(), "r");
  char word[1024];
  while(true) {
    int idx = -1;
    const int nscanned = fscanf(f, " %d %1023s", &idx, word);
    if(nscanned == EOF && !ferror(f) && feof(f)) { break; }
    assert(nscanned == 2);
    assert(ws2i.find(word) == ws2i.end());
    assert(idx ==(int)ws2i.size());
    ws2i[word] = WCONV(idx);
  }
  gzio::gzclose(wfname.c_str(), f);
  _nw = (int)ws2i.size();

  { /* copy the names */
    _wnames = new string[_nw];
    map<string,word_t>::const_iterator it = ws2i.begin();
    const map<string,word_t>::const_iterator iend = ws2i.end();
    for(; it != iend; ++it) {
      assert(_wnames[it->second].empty());
      _wnames[it->second] = it->first;
      assert(!_wnames[it->second].empty());
    }
  }

  cerr << " [" << _nw << " words]" << flush;



  /* load the tag+word mapping */
  vector<uint32_t> rwft;
  vector<uint32_t> rwct;
  const string wftfname = string(dir) + "/wordftag.map";
  f = gzio::gzopen(wftfname.c_str(), "r");
  while(true) {
    int idx = -1, w = -1, t = -1;
    const int nscanned = fscanf(f, " %d %d+%d", &idx, &w, &t);
    if(nscanned == EOF && !ferror(f) && feof(f)) { break; }
    assert(nscanned == 3);
    assert(idx == (int)_wordftag.size());
    assert(w >= 0); assert(w < _nw);
    assert(t >= 0); assert(t < _nft);
    const word_t ww = WCONV(w);
    const tag_t  tt = TCONV(t);
    const uint32_t wwft = mkwt(ww, tt);
    assert(_wordftag.find(wwft) == _wordftag.end());
    _wordftag[wwft] = WCONV(idx);
    rwft.push_back(wwft);
    const uint32_t wwct = mkwt(ww, ctag(tt));
    if(_wordctag.find(wwct) == _wordctag.end()) {
      idx = (int)_wordctag.size();
      _wordctag[wwct] = WCONV(idx);
      rwct.push_back(wwct);
    }
  }
  gzio::gzclose(wftfname.c_str(), f);

  { /* add in pairs for the null word/tag combination */
    const uint32_t wtnull = mkwt(WNULL, TNULL);
    assert(_wordftag.find(wtnull) == _wordftag.end());
    assert(_wordctag.find(wtnull) == _wordctag.end());
    const int fidx = (int)_wordftag.size();
    const int cidx = (int)_wordctag.size();
    _wordftag[wtnull] = WCONV(fidx);
    _wordctag[wtnull] = WCONV(cidx);
    rwft.push_back(wtnull);
    rwct.push_back(wtnull);
  }
  assert(_wordftag.size() == rwft.size());
  assert(_wordctag.size() == rwct.size());
  assert(rwft.size() >= rwct.size());
  assert(rwft.size() < WMAX);
  assert(rwct.size() < WMAX);
  _nwft = (int)rwft.size();
  _nwct = (int)rwct.size();

  /* copy the reverse mappings */
  _rev_wordftag = new uint32_t[_nwft];
  for(int i = 0; i < _nwft; ++i) { _rev_wordftag[i] = rwft[i]; }
  _rev_wordctag = new uint32_t[_nwct];
  for(int i = 0; i < _nwct; ++i) { _rev_wordctag[i] = rwct[i]; }

  cerr << " [" << _nwft << " w+ft, " << _nwct << " w+ct]" << flush;



  /* load word feature information */
  const string winfname = string(dir) + "/wordinfo";
  f = gzio::gzopen(winfname.c_str(), "r");
  while(true) {
    const int nscanned = fscanf(f, " %1023s %1023s", word, tag);
    if(nscanned == EOF && !ferror(f) && feof(f)) { break; }
    assert(nscanned == 2);
    const map<string,word_t>::const_iterator it = ws2i.find(word);
    assert(it != ws2i.end());
    const word_t w = it->second;
    const string feat(tag);
    if(feat == "+openquote") {
      assert(_wquote.find(w) == _wquote.end());
      _wquote[w] = true;
    } else if(feat == "+closequote") {
      assert(_wquote.find(w) == _wquote.end());
      _wquote[w] = false;
    } else if(feat == "+openparen") {
      assert(_wparen.find(w) == _wparen.end());
      _wparen[w] = true;
    } else if(feat == "+closeparen") {
      assert(_wparen.find(w) == _wparen.end());
      _wparen[w] = false;
    } else if(feat == "+comma") {
      assert(_wcomma == WNULL);
      _wcomma = w;
    } else {
      cerr << "unrecognized word-feat: " << feat << endl;
      assert(0);
    }
  }
  gzio::gzclose(winfname.c_str(), f);

  cerr << " [" << _wquote.size() << " quote, " << _wparen.size()
       << " paren]" << flush;


  cerr << " done" << endl;
}
