/* -*-c++-*- */
#ifndef MAPPINGS_H
#define MAPPINGS_H
/* Mappings : Information about string-integer mappings and various
   forms of lexical or tag-based information. */

#include "Bits.h"
#include "Macros.h"

#include <string>
#include <map>

using std::map;
using std::string;

class mappings {
public:
  static void load(const char* const dir);

  static int  Nftags();                  /* how many tags total? */
  static int  Nctags();                  /* how many coarse tags? */

  static bool iscoord(const tag_t tag);  /* is tag a coordination? */
  static bool isprep(const tag_t tag);   /* is tag a preposition? */
  static bool ispunc(const tag_t tag);   /* is tag a punctuation? */
  static bool isverb(const tag_t tag);   /* is tag a verb? */

  static bool isoquote(const word_t w);  /* is word an open-quote? */
  static bool iscquote(const word_t w);  /* is word a close-quote? */
  static bool isoparen(const word_t w);  /* is word an open-paren? */
  static bool iscparen(const word_t w);  /* is word a close-paren? */
  static bool iscomma(const word_t w);   /* is word a comma? */

  /* transform the fine tag into a coarsened version */
  static tag_t ctag(const tag_t tag);
  /* transform the word+tag pair into a representative index */
  static word_t wft(const word_t w, const tag_t ft);
  static word_t wct(const word_t w, const tag_t ct);

private:
  typedef map<word_t,bool> wbmap;
  typedef map<uint32_t,word_t> wtmap;
  static int _nw;
  static int _nft;
  static int _nct;
  static int _nwft;
  static int _nwct;
  static bool* _tag_is_coord;
  static bool* _tag_is_prep;
  static bool* _tag_is_punc;
  static bool* _tag_is_verb;
  static tag_t* _coarse_tag;
  static string* _wnames;
  static string* _ftnames;
  static string* _ctnames;
  static uint32_t* _rev_wordftag;
  static uint32_t* _rev_wordctag;
  static wtmap _wordftag;
  static wtmap _wordctag;
  static wbmap _wquote;
  static wbmap _wparen;
  static word_t _wcomma;

  static uint32_t mkwt(const word_t w, const tag_t t);
};



/*** inlined implementation ***/

inline int mappings::Nftags() { return _nft; }
inline int mappings::Nctags() { return _nct; }

inline bool mappings::iscoord(const tag_t tag) {
  CHECK(assert(tag < _nft));
  return _tag_is_coord[tag];
}
inline bool mappings::isprep(const tag_t tag)  {
  CHECK(assert(tag < _nft));
  return _tag_is_prep[tag];
}
inline bool mappings::ispunc(const tag_t tag)  {
  CHECK(assert(tag < _nft));
  return _tag_is_punc[tag];
}
inline bool mappings::isverb(const tag_t tag)  {
  CHECK(assert(tag < _nft));
  return _tag_is_verb[tag];
}


/* the map records true for open-quote and false for close-quote; nil
   otherwise */
inline bool mappings::isoquote(const word_t w) {
  CHECK(assert(w < _nw));
  const wbmap::const_iterator it = _wquote.find(w);
  if(it == _wquote.end()) { return false; }
  return it->second;
}
inline bool mappings::iscquote(const word_t w) {
  CHECK(assert(w < _nw));
  const wbmap::const_iterator it = _wquote.find(w);
  if(it == _wquote.end()) { return false; }
  return !(it->second);
}

/* the map records true for open-paren and false for close-paren; nil
   otherwise */
inline bool mappings::isoparen(const word_t w) {
  CHECK(assert(w < _nw));
  const wbmap::const_iterator it = _wparen.find(w);
  if(it == _wparen.end()) { return false; }
  return it->second;
}
inline bool mappings::iscparen(const word_t w) {
  CHECK(assert(w < _nw));
  const wbmap::const_iterator it = _wparen.find(w);
  if(it == _wparen.end()) { return false; }
  return !(it->second);
}

/* is this equal to the comma-index? */
inline bool mappings::iscomma(const word_t w) {
  CHECK(assert(w < _nw));
  return (w != WNULL && w == _wcomma);
}


inline tag_t mappings::ctag(const tag_t tag)  {
  CHECK(assert(tag == TNULL || tag < _nft));
  const tag_t ret = (tag == TNULL ? TNULL : _coarse_tag[tag]);
  CHECK(assert(ret == TNULL || ret < _nct));
  return ret;
}

inline uint32_t mappings::mkwt(const word_t w, const tag_t ft) {
  return (((uint32_t)w) | (((uint32_t)ft) << WBITS));
}

inline word_t mappings::wft(const word_t w, const tag_t ft) {
  /* either both are null or both aren't */
  CHECK(assert((w == WNULL) == (ft == TNULL)));
  if(w == WNULL) { return WNULL; }
  wtmap::const_iterator it = _wordftag.find(mkwt(w, ft));
  return (it == _wordftag.end() ? WCONV(_nwft) : it->second);
}

inline word_t mappings::wct(const word_t w, const tag_t ct) {
  /* either both are null or both aren't */
  CHECK(assert((w == WNULL) == (ct == TNULL)));
  if(w == WNULL) { return WNULL; }
  wtmap::const_iterator it = _wordctag.find(mkwt(w, ct));
  return (it == _wordctag.end() ? WCONV(_nwct) : it->second);
}

#endif /* MAPPINGS_H */
