#include "O1Parser.h"

#include <assert.h>
#include <stdlib.h>
#include <math.h>

extern "C" {
#include "kdp.h"
}


/* nested score-backpointer class */
struct o1parser::scbp {
  double sc;
  int bp;
  scbp() : sc(-HUGE_VAL), bp(-1000) {}
};


/* lifecycle */
o1parser::o1parser(const int L,
                   const bool multiroot,
                   const bool projective)
  : _L(L),
    _multiroot(multiroot),
    _projective(projective),
    _N(-1),
    _partsc(NULL),
    _bestlab(NULL),
    _TC(NULL), _WC(NULL) {}
o1parser::~o1parser() { dealloc(); }

/* disallowed methods */
o1parser::o1parser(const o1parser& that)
  : _L(0), _multiroot(false), _projective(false) {
  assert(0);
}
o1parser& o1parser::operator=(const o1parser& that) {
  assert(0); return *this;
}


void o1parser::alloc(const int N) {
  assert(_partsc == NULL);
  assert(_bestlab == NULL);
  assert(_TC == NULL);
  assert(_WC == NULL);
  const int NN = N*N;
  _partsc = (double*)malloc(NN*sizeof(double));
  _bestlab = (lab_t*)malloc(NN*sizeof(lab_t));
  _TC = new scbp[NN];
  _WC = new scbp[NN];
  /* invalidate scores and labels */
  for(int i = 0; i < NN; ++i) {
    _partsc[i] = -HUGE_VAL;
    _bestlab[i] = LINV;
  }
  assert(_partsc != NULL);
  assert(_bestlab != NULL);
  assert(_TC != NULL);
  assert(_WC != NULL);
}
void o1parser::dealloc() {
  if(_partsc != NULL)  { free(_partsc);  _partsc = NULL; }
  if(_bestlab != NULL) { free(_bestlab); _bestlab = NULL; }
  if(_TC != NULL)      { delete [] _TC;  _TC = NULL; }
  if(_WC != NULL)      { delete [] _WC;  _WC = NULL; }
}



/* a helper that constructs an index into a matrix, allowing for the
   possibility of null h */
#define IDX(h, m) (_N*m + (h == -1 ? m : h))



double o1parser::argmax_mst(const int N,
                            vector<int>& restrict ret_heads,
                            vector<int>& restrict ret_labels) {
  int* const p = (int*)malloc(N*sizeof(int));
  const double score = mst1(p, (_multiroot ? 1 : 0), N, _partsc);
  /* copy the argmax tree */
  ret_heads.clear();
  for(int i = 0; i < N; ++i) { ret_heads.push_back(p[i]); }
  free(p);
  /* set the argmax labels */
  for(int i = 0; i < N; ++i) {
    ret_labels.push_back(_bestlab[IDX(ret_heads[i], i)]);
  }
  return score;
}


double o1parser::argmax_proj(const int N,
                             vector<int>& restrict ret_heads,
                             vector<int>& restrict ret_labels) {
  /* maximum span width to consider */
  const int max_width = (_multiroot ? N : (N - 1));
  /* consider spans starting from here */
  const int start_idx = (_multiroot ? -1 : 0);

  for(int w = 1; w <= max_width; ++w) { /* for small to large spans */
    const int end_idx = N - w;
    for(int s = start_idx; s < end_idx; ++s) {
      const int e = s + w; /* end index of the span */
      trap(s, e);
      if(s != -1) { trap(e, s); }
      wedge(s, e);
      if(s != -1) { wedge(e, s); }
    }
  }

  /* single-root constructions */
  if(!_multiroot) { sroot(N); }

  /* unravel the argmax tree */
  ret_heads.clear();
  for(int i = 0; i < N; ++i) { ret_heads.push_back(-2); }
  const int N1 = N - 1;
  if(_multiroot) { unravel_wedge(-1, N1, ret_heads); }
  else           { unravel_sroot(N, ret_heads); }
  /* set the argmax labels */
  for(int i = 0; i < N; ++i) {
    ret_labels.push_back(_bestlab[IDX(ret_heads[i], i)]);
  }

  return _WC[IDX(-1, N1)].sc;
}


void o1parser::sroot(const int N) {
  const int N1 = N - 1;
  scbp& restrict best = _WC[IDX(-1, N1)];
  for(int r = 0; r < N; ++r) {
    const double sc = _partsc[IDX(-1, r)] +
      (r == 0  ? 0 : _WC[IDX(r, 0)].sc) +
      (r == N1 ? 0 : _WC[IDX(r, N1)].sc);
    if(sc > best.sc) { best.sc = sc; best.bp = r; }
  }
}

void o1parser::unravel_sroot(const int N,
                             vector<int>& restrict ret_h) const {
  const int N1 = N - 1;
  const int r = _WC[IDX(-1, N1)].bp;
  assert(ret_h[r] == -2); ret_h[r] = -1;
  if(r != 0)  { unravel_wedge(r, 0, ret_h); }
  if(r != N1) { unravel_wedge(r, N1, ret_h); }
}



void o1parser::trap(const int h, const int m) {
  const int inc = (h < m ? 1 : -1);
  scbp& restrict best = _TC[IDX(h, m)];
  for(int rh = h; rh != m; rh += inc) {
    const int rm = rh + inc;
    const double sc = _partsc[IDX(h, m)] +
      (h == rh ? 0 : _WC[IDX(h, rh)].sc) +
      (m == rm ? 0 : _WC[IDX(m, rm)].sc);
    if(sc > best.sc) { best.sc = sc; best.bp = rm; }
  }
}

void o1parser::unravel_trap(const int h, const int m,
                            vector<int>& restrict ret_h) const {
  assert(h != m);
  const int rm = _TC[IDX(h, m)].bp;
  const int rh = rm + (h < m ? -1 : 1);
  assert(ret_h[m] == -2); ret_h[m] = h;
  if(h != rh) { unravel_wedge(h, rh, ret_h); }
  if(m != rm) { unravel_wedge(m, rm, ret_h); }
}



void o1parser::wedge(const int h, const int e) {
  const int inc = (h < e ? -1 : 1);
  scbp& restrict best = _WC[IDX(h, e)];
  for(int m = e; m != h; m += inc) {
    const double sc = _TC[IDX(h, m)].sc +
      (m == e ? 0 : _WC[IDX(m, e)].sc);
    if(sc > best.sc) { best.sc = sc; best.bp = m; }
  }
}

void o1parser::unravel_wedge(const int h, const int e,
                             vector<int>& restrict ret_h) const {
  assert(h != e);
  const int m = _WC[IDX(h, e)].bp;
  unravel_trap(h, m, ret_h);
  if(m != e) { unravel_wedge(m, e, ret_h); }
}
