/* -*-c++-*- */
#ifndef O3ABLATION_H
#define O3ABLATION_H
/* O3Ablation : Utilities for parsing with certain parts masked out or
   modified. */

#include "Macros.h"
#include "O3Part.h"
#include "O3Parser.h"

#include <string>
#include <vector>

using std::string;
using std::vector;


class o3ablation {
public:
  /* obtain an integer typecode */
  static int code(const string& name);

  /* obtain a string name */
  static string name(const int ablation);

  /* score a part-set with an ablated scorer */
  template<class PartScorer>
  static double score(const int ablation,
                      const o3part::bmap& restrict t,
                      const PartScorer& restrict S);

  /* parse with an ablated scorer */
  template<class PartScorer>
  static double argmax(const int ablation,
                       o3parser& restrict P,
                       const int N, const PartScorer& restrict S,
                       vector<int>& restrict ret_heads,
                       vector<int>& restrict ret_labels);


  /* ablated part-scorers */

  /* a pass-through; base class for the ablated scorers */
  template<class PartScorer> class ablate_none;
  /* ablate grandchild 4-grams */
  template<class PartScorer> class ablate_gsib;
  /* ablate sibling 4-grams */
  template<class PartScorer> class ablate_sibs;
  /* ablate all 4-grams */
  template<class PartScorer> class ablate_all4;
  /* ablate all grandchild-related scores */
  template<class PartScorer> class ablate_allg;
  /* ablate all sibling-related scores */
  template<class PartScorer> class ablate_alls;
  /* ablate all 4-grams, some grandchild features, and ignore
     positionality on other features.  this ablation is only meaningful
     when used with the sibling-style factorization, in which case it
     corresponds to a reimplementation of the higher-order factorization
     of Carreras (2007) */
  template<class PartScorer> class ablate_xc07;
  /* ablate all 4-grams, all grandchild parts, and ignore
     positionality on other features.  this ablation is meaningful for
     either style of parser, which case it corresponds to a
     sibling factorization (McDonald and Pereira, 2006) */
  template<class PartScorer> class ablate_mp06;
  /* ablate positionality of all features */
  template<class PartScorer> class ablate_posi;

private:
  static const string names[];

  /* types of part-ablation */
  enum ablation_type {
    ABLATE_NONE = 0,
    ABLATE_GSIB,
    ABLATE_SIBS,
    ABLATE_ALL4,
    ABLATE_ALLG,
    ABLATE_ALLS,
    ABLATE_XC07,
    ABLATE_MP06,
    ABLATE_POSI,
    NUM_TYPES
  };
};






/*** inlined implementation ***/

/* score a set of parts with an optionally-ablated scorer */
template<class PartScorer> inline double
o3ablation::score(const int ablation,
                  const o3part::bmap& restrict t,
                  const PartScorer& restrict S) {
  switch(ablation) {
  case ABLATE_NONE:
    return o3part::score(t, S);
  case ABLATE_GSIB:
    return o3part::score(t, ablate_gsib<const PartScorer>(S));
  case ABLATE_SIBS:
    return o3part::score(t, ablate_sibs<const PartScorer>(S));
  case ABLATE_ALL4:
    return o3part::score(t, ablate_all4<const PartScorer>(S));
  case ABLATE_ALLG:
    return o3part::score(t, ablate_allg<const PartScorer>(S));
  case ABLATE_ALLS:
    return o3part::score(t, ablate_alls<const PartScorer>(S));
  case ABLATE_XC07:
    return o3part::score(t, ablate_xc07<const PartScorer>(S));
  case ABLATE_MP06:
    return o3part::score(t, ablate_mp06<const PartScorer>(S));
  case ABLATE_POSI:
    return o3part::score(t, ablate_posi<const PartScorer>(S));
  default: assert(0);
  }
  return 0;
}

/* parse a sentence with an optionally-ablated scorer */
template<class PartScorer> inline double
o3ablation::argmax(const int ablation,
                   o3parser& restrict P,
                   const int N, const PartScorer& restrict S,
                   vector<int>& restrict ret_h,
                   vector<int>& restrict ret_l) {
  switch(ablation) {
  case ABLATE_NONE:
    return P.argmax(N, S,                                ret_h, ret_l);
  case ABLATE_GSIB:
    return P.argmax(N, ablate_gsib<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_SIBS:
    return P.argmax(N, ablate_sibs<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_ALL4:
    return P.argmax(N, ablate_all4<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_ALLG:
    return P.argmax(N, ablate_allg<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_ALLS:
    return P.argmax(N, ablate_alls<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_XC07:
    return P.argmax(N, ablate_xc07<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_MP06:
    return P.argmax(N, ablate_mp06<const PartScorer>(S), ret_h, ret_l);
  case ABLATE_POSI:
    return P.argmax(N, ablate_posi<const PartScorer>(S), ret_h, ret_l);
  default: assert(0);
  }
  return 0;
}






/*** ablated part-scorer implementation ***/

/* a base class that simply passes-through the underlying scorer */
template<class PartScorer>
class o3ablation::ablate_none {
public:
  double dep1(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return _S.dep1(i1, i2, l1);
  }
  double dep2(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return _S.dep2(i1, i2, l1);
  }
  double dep3(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return _S.dep3(i1, i2, l1);
  }
  double sib1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return _S.sib1(i1, i2, i3, l1, l2);
  }
  double sib2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return _S.sib2(i1, i2, i3, l1, l2);
  }
  double gci1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return _S.gci1(i1, i2, i3, l1, l2);
  }
  double gci2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return _S.gci2(i1, i2, i3, l1, l2);
  }
  double gco1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return _S.gco1(i1, i2, i3, l1, l2);
  }
  double gco2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return _S.gco2(i1, i2, i3, l1, l2);
  }
  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return _S.gcis(i1, i2, i3, i4, l1, l2, l3);
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return _S.gcos(i1, i2, i3, i4, l1, l2, l3);
  }
  double sibs(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return _S.sibs(i1, i2, i3, i4, l1, l2, l3);
  }

protected:
  PartScorer& restrict _S;
  ablate_none(PartScorer& restrict S) : _S(S) {}
};



/*** subclasses that mask out certain parts ***/

/* ablate grandchild 4-grams */
template<class PartScorer>
class o3ablation::ablate_gsib
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_gsib(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* convert these to NOPs */
  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
};


/* ablate sibling 4-grams */
template<class PartScorer>
class o3ablation::ablate_sibs
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_sibs(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* convert these to NOPs */
  double sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const {
    return 0;
  }
};


/* ablate all 4-grams */
template<class PartScorer>
class o3ablation::ablate_all4
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_all4(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* convert these to NOPs */
  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const {
    return 0;
  }
};


/* ablate all grandchild scores */
template<class PartScorer>
class o3ablation::ablate_allg
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_allg(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* convert these to NOPs */
  double gci1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gci2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gco1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gco2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }

  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
};


/* ablate all sibling scores */
template<class PartScorer>
class o3ablation::ablate_alls
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_alls(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* convert these to NOPs */
  double sib1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double sib2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }

  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const {
    return 0;
  }
};


/* reimplement Carreras (2007) features */
template<class PartScorer>
class o3ablation::ablate_xc07
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_xc07(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* ignore positionality of dependencies */
  double dep2(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return this->_S.dep1(i1, i2, l1);
  }
  double dep3(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return this->_S.dep1(i1, i2, l1);
  }

  /* ignore positionality of siblings */
  double sib2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return this->_S.sib1(i1, i2, i3, l1, l2);
  }

  /* ignore second-outer grandchildren */
  double gci2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gco2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }

  /* ignore 4-gram features */
  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double sibs(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
};


/* reimplement McDonald and Pereira (2006) features */
template<class PartScorer>
class o3ablation::ablate_mp06
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_mp06(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* ignore positionality of dependencies */
  double dep2(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return this->_S.dep1(i1, i2, l1);
  }
  double dep3(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return this->_S.dep1(i1, i2, l1);
  }

  /* ignore positionality of siblings */
  double sib2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return this->_S.sib1(i1, i2, i3, l1, l2);
  }

  /* ignore all grandchildren */
  double gci1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gci2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gco1(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }
  double gco2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return 0;
  }

  /* ignore 4-gram features */
  double gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
  double sibs(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
              const lab_t l1, const lab_t l2, const lab_t l3) const {
    return 0;
  }
};


/* ignore all positionality */
template<class PartScorer>
class o3ablation::ablate_posi
  : public o3ablation::ablate_none<PartScorer> {
public:
  ablate_posi(PartScorer& restrict S) : ablate_none<PartScorer>(S) {}

  /* ignore positionality of dependencies */
  double dep2(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return this->_S.dep1(i1, i2, l1);
  }
  double dep3(const idx_t i1, const idx_t i2, const lab_t l1) const {
    return this->_S.dep1(i1, i2, l1);
  }

  /* ignore positionality of siblings */
  double sib2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return this->_S.sib1(i1, i2, i3, l1, l2);
  }

  /* ignore positionality of grandchildren */
  double gci2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return this->_S.gci1(i1, i2, i3, l1, l2);
  }
  double gco2(const idx_t i1, const idx_t i2, const idx_t i3,
              const lab_t l1, const lab_t l2) const {
    return this->_S.gco1(i1, i2, i3, l1, l2);
  }
};

#endif /* O3ABLATION_H */
