#include "O3Active.h"

#include <map>

#include <assert.h>

using std::map;

o3active::o3active()
  : _N(-1),
    _activeN(0),
    _activemN(0),
    _activehN(0),
    _activehNidx(0),
    _activemL(NULL),
    _activemR(NULL),
    _activehL(NULL),
    _activehR(NULL),
    _activehLidx(NULL),
    _activehRidx(NULL),
    _activehPidx(NULL),
    _activehCidx(NULL) {}

o3active::~o3active() {
  if(_activemL != NULL) {
    free(_activemL[-1]);
    free(_activemL - 1);
    free(_activemR - 1);
  }
  if(_activehL != NULL) {
    free(_activehL[0]);
    free(_activehL);
    free(_activehR);
    free(_activehLidx[0]);
    free(_activehLidx);
    free(_activehRidx);
    free(_activehPidx - 1);
    free(_activehCidx - 1);
  }
}



typedef map<backp,bool> bpmap;

void o3active::set_active(const bool headlists,
                          const int L,
                          const int in_N,
                          const int* const active_num,
                          const int** const active_heads,
                          const int** const active_labels) {
  assert(in_N <= O3PARSER_MAX_SENTLEN);
  _N = in_N;
  bool resizeN = false;
  { /* save modifier lists */
    /* make room for N+1 words */
    if(_N > _activeN) {
      resizeN = true;
      _activemL =
        (backp**)(_activeN == 0 ? malloc((_N + 1)*sizeof(backp*))
                  : realloc(_activemL - 1, (_N + 1)*sizeof(backp*)));
      _activemR =
        (backp**)(_activeN == 0 ? malloc((_N + 1)*sizeof(backp*))
                  : realloc(_activemR - 1, (_N + 1)*sizeof(backp*)));
      ++_activemL; /* -1 is now a valid index */
      ++_activemR; /* -1 is now a valid index */
      _activeN = _N;
    }

    /* compute the size of the new active arrays */
    int newsize = 2*(_N + 1); /* #left and #right of each word + root */
    for(int m = 0; m < _N; ++m) { /* plus the actual #deps */
      assert(active_num[m] > 0);
      assert(active_num[m] <= O3PARSER_MAX_ACTIVE);
      newsize += active_num[m]; /* one backp per label+head */
    }
    /* reallocate the underlying array */
    if(newsize > _activemN) {
      _activemL[-1] =
        (backp*)(_activemN == 0 ? malloc(newsize*sizeof(backp))
                 : realloc(_activemL[-1], newsize*sizeof(backp)));
      _activemN = newsize;
    }

    /* modldeps and modrdeps are sorted lists of left and right
       dependencies for each head from -1 to N - 1 */
    bpmap* modldeps = new bpmap[_N + 1];
    bpmap* modrdeps = new bpmap[_N + 1];
    ++modldeps; /* -1 is now a valid index */
    ++modrdeps; /* -1 is now a valid index */

    /* collect the active deps in the left/right lists */
    for(int m = 0; m < _N; ++m) {
      const int nactive = active_num[m];
      const int* const heads_m  = active_heads[m];
      const int* const labels_m = active_labels[m];
      for(int i = 0; i < nactive; ++i) {
        const int h = heads_m[i];
        const int l = labels_m[i];
        assert(h >= -1); assert(h < _N);
        assert(l >= 0);  assert(l < L);
        /* save as rightward or leftward dependency */
        bpmap* deps = (h < m ? modrdeps : modldeps) + h;
        const backp b(ICONV(m), LCONV(l));
        const bool unique_pruned_dependency =
          deps->insert(bpmap::value_type(b, true)).second;
        assert(unique_pruned_dependency);
      }
    }

    /* transfer deps into the active arrays in head-outward order */
    backp* act = _activemL[-1]; /* start of the underlying block */
    for(int h = -1; h < _N; ++h) {
      { /* fill array of left deps */
        _activemL[h] = act; /* claim this chunk */
        const bpmap& deps = modldeps[h];
        const size_t nl = (int)deps.size();
        const uint16_t nnl = (uint16_t)nl;
        assert(nl == nnl);
        *act = backp(nnl); ++act;
        bpmap::const_reverse_iterator it = deps.rbegin();
        const bpmap::const_reverse_iterator iend = deps.rend();
        for(; it != iend; ++it, ++act) { *act = it->first; }
      }
      { /* fill array of right deps */
        _activemR[h] = act; /* claim this chunk */
        const bpmap& deps = modrdeps[h];
        const size_t nr = (int)deps.size();
        const uint16_t nnr = (uint16_t)nr;
        assert(nr == nnr);
        *act = backp(nnr); ++act;
        bpmap::const_iterator it = deps.begin();
        const bpmap::const_iterator iend = deps.end();
        for(; it != iend; ++it, ++act) { *act = it->first; }
      }
    }
    /* make sure we covered the claimed space */
    const long int nsaved = act - _activemL[-1];
    assert(nsaved == newsize);
    act = NULL;

    /* cleanup */
    --modldeps; /* undo the shift for deletion */
    --modrdeps; /* undo the shift for deletion */
    delete [] modldeps;
    delete [] modrdeps;

    /* validate the sorted orders */
    for(int h = -1; h < _N; ++h) {
      { /* left deps from high to low */
        const backp* const actl = _activemL[h];
        const uint16_t nactive = actl->raw();
        int last_m = h;
        for(int i = 1; i <= nactive; ++i) {
          const int m = actl[i].i();
          assert(m < h);
          assert(m <= last_m);
          assert(m >= 0);
          last_m = m;
        }
      }
      { /* right deps from low to high */
        const backp* const actr = _activemR[h];
        const uint16_t nactive = actr->raw();
        int last_m = h;
        for(int i = 1; i <= nactive; ++i) {
          const int m = actr[i].i();
          assert(m > h);
          assert(m >= last_m);
          assert(m < _N);
          last_m = m;
        }
      }
    }
  }
  if(headlists) { /* also save head-lists */
    /* make room for N words */
    const int N1 = _N + 1;
    const int NP = N1*_N;
    if(resizeN) {
      _activehPidx =
        (backp**)(_activehL == NULL ? malloc(NP*sizeof(backp*))
                  : realloc(_activehPidx - 1, NP*sizeof(backp*)));
      _activehCidx =
        (uint16_t*)(_activehL == NULL ? malloc(NP*sizeof(uint16_t))
                    : realloc(_activehCidx - 1, NP*sizeof(uint16_t)));
      ++_activehPidx; ++_activehCidx;
      _activehL = (backp**)realloc(_activehL, _N*sizeof(backp*));
      _activehR = (backp**)realloc(_activehR, _N*sizeof(backp*));
      _activehLidx = (int**)realloc(_activehLidx, _N*sizeof(int*));
      _activehRidx = (int**)realloc(_activehRidx, _N*sizeof(int*));
    }
    /* zero the matrix of pointers/counts */
    for(int i = NP - 2; i >= -1; --i) {
      _activehPidx[i] = NULL; _activehCidx[i] = 0;
    }

    /* compute the size of the new active arrays */
    int newsize = 2*_N; /* #left and #right of each word */
    for(int m = 0; m < _N; ++m) { /* plus the actual #deps */
      assert(active_num[m] > 0);
      assert(active_num[m] <= O3PARSER_MAX_ACTIVE);
      newsize += active_num[m]; /* one backp per label+head */
    }
    /* reallocate the underlying array */
    if(newsize > _activehN) {
      _activehL[0] =
        (backp*)(_activehN == 0 ? malloc(newsize*sizeof(backp))
                 : realloc(_activehL[0], newsize*sizeof(backp)));
      _activehN = newsize;
    }

    /* sorted lists of left and right heads for m = 0 to N - 1 */
    bpmap* lheads = new bpmap[_N];
    bpmap* rheads = new bpmap[_N];
    /* lists of root deps for all m */
    bpmap* roots = new bpmap[_N];
    /* sorted lists of left and right heads, index only */
    map<int,bool>* lhidx = new map<int,bool>[_N];
    map<int,bool>* rhidx = new map<int,bool>[_N];
    /* collect the active deps in the left/right/root lists */
    for(int m = 0; m < _N; ++m) {
      const int nactive = active_num[m];
      const int* const heads_m  = active_heads[m];
      const int* const labels_m = active_labels[m];
      for(int i = 0; i < nactive; ++i) {
        const int hh = heads_m[i];
        const idx_t h = ICONV_NULL(hh);
        const idx_t l = LCONV(labels_m[i]);
        assert(hh >= -1); assert(hh < _N);
        assert(labels_m[i] >= 0); assert(l < L);
        /* select proper save-list */
        bpmap* heads =
          (hh == -1 ? roots : (h < m ? lheads : rheads)) + m;
        /* insert and check uniqueness */
        const backp b(h, l);
        const bool unique_pruned_dependency =
          heads->insert(bpmap::value_type(b, true)).second;
        assert(unique_pruned_dependency);
        /* save the head-index */
        map<int,bool>* hidx = (hh < m ? lhidx : rhidx) + m;
        hidx->insert(map<int,bool>::value_type(hh, true));
      }
    }

    /* allocate arrays for the head-index lists */
    int newsize_idx = 2*_N; /* size left and right */
    for(int m = 0; m < _N; ++m) {
      newsize_idx += (int)(lhidx[m].size() + rhidx[m].size());
    }
    if(newsize_idx > _activehNidx) {
      _activehLidx[0] =
        (int*)(_activehNidx == 0 ? malloc(newsize_idx*sizeof(int))
               : realloc(_activehLidx[0], newsize_idx*sizeof(int)));
      _activehNidx = newsize_idx;
    }

    /* transfer heads into the active arrays in mod-outward order */
    backp* act = _activehL[0]; /* start of the underlying block */
    int* idx = _activehLidx[0]; /* start of the underlying block */
    for(int m = 0; m < _N; ++m) {
      backp**   hidxp = _activehPidx + m*N1; /* first instances of heads */
      uint16_t* hidxc = _activehCidx + m*N1; /* counts of heads */
      { /* fill array of left heads (including root) */
        _activehL[m] = act; /* claim this chunk */
        const bpmap& deps = lheads[m];
        const int nl = (int)(deps.size() + roots[m].size());
        const uint16_t nnl = (uint16_t)nl;
        *act = backp(nnl); ++act;
        bpmap::const_reverse_iterator it = deps.rbegin();
        bpmap::const_reverse_iterator iend = deps.rend();
        for(; it != iend; ++it, ++act) {
          const backp b = it->first;
          const int h = b.i();
          /* save the first occurrence and count of h */
          if(hidxp[h] == NULL) { hidxp[h] = act; }
          ++(hidxc[h]);
          *act = b;
        }
        /* insert root deps after normal heads, as they are the
           furthest to the left */
        it = roots[m].rbegin();
        iend = roots[m].rend();
        for(; it != iend; ++it, ++act) {
          /* save the first occurrence and count of h */
          if(hidxp[-1] == NULL) { hidxp[-1] = act; }
          ++(hidxc[-1]);
          *act = it->first;
        }
      }
      { /* fill array of right rheads */
        _activehR[m] = act; /* claim this chunk */
        const bpmap& deps = rheads[m];
        const int nr = (int)deps.size();
        const uint16_t nnr = (uint16_t)nr;
        *act = backp(nnr); ++act;
        bpmap::const_iterator it = deps.begin();
        const bpmap::const_iterator iend = deps.end();
        for(; it != iend; ++it, ++act) {
          const backp b = it->first;
          const int h = b.i();
          /* save the first occurrence and count of h */
          if(hidxp[h] == NULL) { hidxp[h] = act; }
          ++(hidxc[h]);
          *act = b;
        }
      }
      { /* fill array of left indices */
        _activehLidx[m] = idx; /* claim this chunk */
        const map<int,bool>& hidx = lhidx[m];
        *idx = (int)hidx.size(); ++idx;
        map<int,bool>::const_reverse_iterator it = hidx.rbegin();
        const map<int,bool>::const_reverse_iterator iend = hidx.rend();
        for(; it != iend; ++it, ++idx) { *idx = it->first; }
      }
      { /* fill array of right indices */
        _activehRidx[m] = idx; /* claim this chunk */
        const map<int,bool>& hidx = rhidx[m];
        *idx = (int)hidx.size(); ++idx;
        map<int,bool>::const_iterator it = hidx.begin();
        const map<int,bool>::const_iterator iend = hidx.end();
        for(; it != iend; ++it, ++idx) { *idx = it->first; }
      }
    }
    /* make sure we covered the claimed space */
    const long int nsaved = act - _activehL[0];
    assert(nsaved == newsize);
    const long int nsaved_idx = idx - _activehLidx[0];
    assert(nsaved_idx == newsize_idx);
    act = NULL;

    /* cleanup */
    delete [] lheads;
    delete [] rheads;
    delete [] roots;
    delete [] lhidx;
    delete [] rhidx;

    /* validate the sorted orders */
    for(int m = 0; m < _N; ++m) {
      { /* left heads from high to low */
        const backp* const actl = _activehL[m];
        const uint16_t nactive = actl->raw();
        int last_h = m;
        for(int i = 1; i <= nactive; ++i) {
          const int h = actl[i].i_null();
          assert(h < m);
          assert(h <= last_h);
          assert(h >= -1);
          last_h = h;
        }
      }
      { /* right heads from low to high */
        const backp* const actr = _activehR[m];
        const uint16_t nactive = actr->raw();
        int last_h = m;
        for(int i = 1; i <= nactive; ++i) {
          const int h = actr[i].i();
          assert(h > m);
          assert(h >= last_h);
          assert(h < _N);
          last_h = h;
        }
      }
      { /* left indices from high to low */
        const int* const actl = _activehLidx[m];
        const int nactive = *actl;
        int last_h = m;
        for(int i = 1; i <= nactive; ++i) {
          const int h = actl[i];
          assert(h < last_h);
          assert(h >= -1);
          last_h = h;
        }
      }
      { /* right indices from high to low */
        const int* const actr = _activehRidx[m];
        const int nactive = *actr;
        int last_h = m;
        for(int i = 1; i <= nactive; ++i) {
          const int h = actr[i];
          assert(h > last_h);
          assert(h < _N);
          last_h = h;
        }
      }
    }
  }
}
