/* -*-c++-*- */
#ifndef O3ACTIVE_H
#define O3ACTIVE_H
/* O3Active : State and helper methods for manipulating a set of
   active dependencies on a single sentence */

#include "Macros.h"
#include "Backpointer.h"

#include <assert.h>
#include <stdlib.h>

class o3tree; /* forward-declaration */

class o3active {
public:
  o3active();  /* create an empty set of dependencies */
  ~o3active();

  /* copy the active dependencies, deallocating any previous state.
     if headlists is true, head-lists will also be created, which are
     necessary for enumerating heads; as needed in bottom-up
     parsing */
  void set_active(const bool headlists,
                  const int L, const int N,
                  const int* const active_num,
                  const int** const active_heads,
                  const int** const active_labels);

  int N() const; /* number of words in the current sentence */

  /* nested enumerator classes.  these enumerate heads or modifiers
     and the concomitant labels in either head-outward or head-inward
     order.  head-inward enumerators start at the furthest dependency
     and walk inward; head-outward enumerators start at the closest
     dependency and walk outward.  the dep2 enumerators walk through
     heads that jointly head two words. */
  class lideps; /* enumerate left deps in head-inward order */
  class rideps; /* enumerate right deps in head-inward order */
  class lodeps; /* enumerate left deps in head-outward order */
  class rodeps; /* enumerate right deps in head-outward order */
  class lisibs; /* enumerate left sib-heads in head-inward order */
  class risibs; /* enumerate right sib-heads in head-inward order */
  class labels; /* enumerate labels of a single dependency */


  /* methods that return enumerators.  they can optionally be given an
     end index that defines a limited range of indices.  enumeration
     stops when the current word index is either at or beyond the
     given end index. */

#define LARGEIDX (IMAX + 10)
#define SMALLIDX (-10)
  /* enumerate left/right modifiers of h in inward/outward order. */
  lideps limods(const int h, const int end = LARGEIDX) const;
  lodeps lomods(const int h, const int end = SMALLIDX) const;
  rideps rimods(const int h, const int end = SMALLIDX) const;
  rodeps romods(const int h, const int end = LARGEIDX) const;
  /* enumerate left/right heads of m in inward/outward order. */
  lideps liheads(const int m, const int end = LARGEIDX) const;
  lodeps loheads(const int m, const int end = SMALLIDX) const;
  rideps riheads(const int m, const int end = SMALLIDX) const;
  rodeps roheads(const int m, const int end = LARGEIDX) const;
  /* enumerate left/right sib-heads of s, m */
  lisibs lisibheads(const int m, const int s) const;
  risibs risibheads(const int m, const int s) const;
  /* enumerate labels of a given dependency */
  labels deplabs(const int h, const int m) const;
#undef LARGEIDX
#undef SMALLIDX

private:
  int _N;                 /* size of the current sentence */
  int _activeN;           /* size of active arrays */
  int _activemN;          /* size of activemL/R arrays */
  int _activehN;          /* size of activehL/R arrays */
  int _activehNidx;       /* size of activehL/Ridx arrays */
  backp** _activemL;      /* list of allowable left modifiers */
  backp** _activemR;      /* list of allowable right modifiers */
  backp** _activehL;      /* list of allowable left heads */
  backp** _activehR;      /* list of allowable right heads */
  int**    _activehLidx;  /* list of allowable left head-words */
  int**    _activehRidx;  /* list of allowable right head-words */
  backp** _activehPidx;   /* matrix of pointers to first instances of
                             each head-word */
  uint16_t* _activehCidx; /* matrix of counts of instances of each
                             head-word */

  /* raw access, for brute force enumeration */
  friend class o3tree;
  const backp* lharr(const int m) const; /* raw access */
  const backp* rharr(const int m) const; /* raw access */
};


/* nested enumerator classes; the implicit copy constructor,
   destructor and assignment operators are fine.  the enumerators
   provide an interface with a single public next() method that
   returns false when the enumeration is complete.  the next() method
   takes as arguments references to integers that will be set with
   indices and label */

/* head-inward enumerators for left and right deps */
class o3active::lideps {
public:
  /* return false when no more deps are available */
  bool hasnext() const {
    return (_act != _actbeg) && (_act->i_null() < _eidx);
  }
  bool next(int& restrict idx, int& restrict lab) {
    if(_act == _actbeg) { return false; }
    const backp b = *_act; --_act; /* head-inward; decrement */
    idx = b.i_null();
    lab = b.l();
    /* left head-inward deps; don't cross to right of eidx */
    return (idx < _eidx);
  }

private:
  friend class o3active; /* can only be created by o3active */
  lideps(const backp* const arr, const int eidx, const int N)
    : _act(arr + arr[0].raw()), _actbeg(arr), _eidx(eidx), _N(N) {}
  const backp* _act;
  const backp* const _actbeg;
  const int _eidx;
  const int _N;
};

class o3active::rideps {
public:
  /* return false when no more deps are available */
  bool hasnext() const {
    return (_act != _actbeg) && (_act->i() > _eidx);
  }
  bool next(int& restrict idx, int& restrict lab) {
    if(_act == _actbeg) { return false; }
    const backp b = *_act; --_act; /* head-inward; decrement */
    idx = b.i();
    lab = b.l();
    /* right head-inward deps; don't cross to left of eidx */
    return (idx > _eidx);
  }

private:
  friend class o3active; /* can only be created by o3active */
  rideps(const backp* const arr, const int eidx)
    : _act(arr + arr[0].raw()), _actbeg(arr), _eidx(eidx) {}
  const backp* _act;
  const backp* const _actbeg;
  const int _eidx;
};


/* head-outward enumerators for left and right deps */
class o3active::lodeps {
public:
  /* return false when no more deps are available */
  bool hasnext() const {
    return (_act != _actend) && (_act->i() > _eidx);
  }
  bool next(int& restrict idx, int& restrict lab) {
    if(_act == _actend) { return false; }
    const backp b = *_act; ++_act; /* head-outward; increment */
    idx = b.i_null();
    lab = b.l();
    /* left head-outward deps; don't cross to left of eidx */
    return (idx > _eidx);
  }
  /* direct backpointer access */
  bool next(backp& restrict b) {
    if(_act == _actend) { return false; }
    b = *_act; ++_act; /* head-outward; increment */
    /* left head-outward deps; don't cross to left of eidx */
    return (b.i_null() > _eidx);
  }

private:
  friend class o3active; /* can only be created by o3active */
  lodeps(const backp* const arr, const int eidx, const int N)
    : _act(arr + 1), _actend(_act + arr[0].raw()), _eidx(eidx), _N(N) {}
  const backp* _act;
  const backp* const _actend;
  const int _eidx;
  const int _N;
};

class o3active::rodeps {
public:
  /* return false when no more deps are available */
  bool hasnext() const {
    return (_act != _actend) && (_act->i() < _eidx);
  }
  bool next(int& restrict idx, int& restrict lab) {
    if(_act == _actend) { return false; }
    const backp b = *_act; ++_act; /* head-outward; increment */
    idx = b.i();
    lab = b.l();
    /* right head-outward deps; don't cross to right of eidx */
    return (idx < _eidx);
  }
  /* direct backpointer access */
  bool next(backp& restrict b) {
    if(_act == _actend) { return false; }
    b = *_act; ++_act; /* head-outward; increment */
    /* right head-outward deps; don't cross to right of eidx */
    return (b.i() < _eidx);
  }

private:
  friend class o3active; /* can only be created by o3active */
  rodeps(const backp* const arr, const int eidx)
    : _act(arr + 1), _actend(_act + arr[0].raw()), _eidx(eidx) {}
  const backp* _act;
  const backp* const _actend;
  const int _eidx;
};


/* head-inward enumerators for sib-heads */
class o3active::lisibs {
public:
  /* return false when no more deps are available.  ordinarily, we
     would have two nested loops, which we must unroll into a single
     function.  the outer loop enumerates head-indices of s and m
     until the next matching head-index is found.  the inner loop
     enumerates possible label pairs for that matching head.  we
     unroll this nested loop as follows:

     1. we check if there are any more label-pairs to consider in the
        current label pair list.

        a. if there are more pairs, move on

        b. if there are no more pairs, then search for the next shared
           head-index, or quit if there are no other shared heads.

     2. step 1 ensures that, one way or another, there are more label
        pairs to choose from.  pick one and return it.

     thus we check the end condition of the inner loop first, execute
     the outer loop if necessary, and then execute the inner loop. */
  bool next(int& restrict head, int& restrict lab_hm, int& restrict lab_hs) {
    /* look for matching head-indices and set up the inner loop if a
       match is found */
#define next_sibhead(OUTSIDE) {                                         \
      while(1) {                                                        \
        /* stopping condition of outer loop */                          \
        if(_idxm == _idxmbeg || _idxs == _idxsbeg) { return false; }    \
        const int h_m = *_idxm;                                         \
        const int h_s = *_idxs;                                         \
        /* stop early if h_m is on the wrong side of s */               \
        if(!(h_m OUTSIDE _s)) { return false; }                         \
        if(h_s == h_m) { /* matching indices; consider labels */        \
          const uint16_t num_hs = _Cs[h_m];                             \
          const uint16_t num_hm = _Cm[h_m];                             \
          assert(num_hs > 0); assert(num_hm > 0);                       \
          head = h_m; /* save the head index */                         \
          /* set ranges for the inner loop */                           \
          _acts = _Ps[h_m];                                             \
          _l_hs = _acts->l(); /* pre-set current label */               \
          _actsend = _acts + num_hs;                                    \
          _actm = _actmbeg = _Pm[h_m];                                  \
          _actmend = _actmbeg + num_hm;                                 \
          break; /* handle the inner loop elsewhere */                  \
        } else if(h_m OUTSIDE h_s) { /* h_m is outside; advance it */   \
          --_idxm;                                                      \
        } else {                     /* h_s is outside; advance it */   \
          --_idxs;                                                      \
        }                                                               \
      }                                                                 \
    }

    /* obtain the next pair of sibling labels and advance to the next
       sibling label-pair */
#define next_siblab() {                                         \
      const backp bm = *_actm;                                  \
      lab_hm = bm.l(); /* save the label index */               \
      lab_hs = _l_hs;  /* save the label index */               \
      ++_actm; /* advance to next label */                      \
      if(_actm == _actmend) { /* finished m-loop */             \
        _actm = _actmbeg; /* restart loop */                    \
        ++_acts; /* advance to next label */                    \
        if(_acts == _actsend) { /* finished s-loop */           \
          --_idxs; --_idxm; /* advance outer loop */            \
        } else { /* set next s-label */                         \
          _l_hs = _acts->l(); /* save the label index */        \
        }                                                       \
      }                                                         \
    }

    /* put it together... */

    /* step 1: check if the inner loop is complete and advance the
       outer loop if necessary.  NB: upon initialization, _acts ==
       _actsend == NULL so this suffices for initialization as well */
    if(_acts == _actsend) {
      next_sibhead(<); /* left heads; "outside" means "left of" */
    }

    /* step 2: now _acts/_actm are set properly; get the labels and
       advance the loop */
    next_siblab();

    /* end-of-loop checking is handled in next_sibhead(); if we are at
       this point we've found a valid head and label pair. */
    return true;
  }

private:
  friend class o3active; /* can only be created by o3active */
  lisibs(const int* const idxm, const int* const idxs,
         const backp** const actP, const uint16_t* const actC,
         const int s, const int m, const int N)
    : _idxm(idxm + idxm[0]), _idxmbeg(idxm),
      _idxs(idxs + idxs[0]), _idxsbeg(idxs),
      _Ps(actP + (s*(N + 1))), _Pm(actP + (m*(N + 1))),
      _Cs(actC + (s*(N + 1))), _Cm(actC + (m*(N + 1))),
      _acts(NULL), _actsend(NULL),
      _actm(NULL), _actmbeg(NULL), _actmend(NULL),
      _s(s), _l_hs(-1) {}

  const int* _idxm;
  const int* const _idxmbeg;
  const int* _idxs;
  const int* const _idxsbeg;
  const backp** const _Ps;
  const backp** const _Pm;
  const uint16_t* const _Cs;
  const uint16_t* const _Cm;
  const backp* _acts;
  const backp* _actsend;
  const backp* _actm;
  const backp* _actmbeg;
  const backp* _actmend;
  const int _s;
  int _l_hs;
};

class o3active::risibs {
public:
  bool next(int& restrict head, int& restrict lab_hm, int& restrict lab_hs) {
    if(_acts == _actsend) {
      next_sibhead(>); /* right heads; "outside" means "right of" */
    }
    next_siblab();
    return true;
  }

private:
  friend class o3active; /* can only be created by o3active */
  risibs(const int* const idxm, const int* const idxs,
         const backp** const actP, const uint16_t* const actC,
         const int s, const int m, const int N)
    : _idxm(idxm + idxm[0]), _idxmbeg(idxm),
      _idxs(idxs + idxs[0]), _idxsbeg(idxs),
      _Ps(actP + (s*(N + 1))), _Pm(actP + (m*(N + 1))),
      _Cs(actC + (s*(N + 1))), _Cm(actC + (m*(N + 1))),
      _acts(NULL), _actsend(NULL),
      _actm(NULL), _actmbeg(NULL), _actmend(NULL),
      _s(s), _l_hs(-1) {}

  const int* _idxm;
  const int* const _idxmbeg;
  const int* _idxs;
  const int* const _idxsbeg;
  const backp** const _Ps;
  const backp** const _Pm;
  const uint16_t* const _Cs;
  const uint16_t* const _Cm;
  const backp* _acts;
  const backp* _actsend;
  const backp* _actm;
  const backp* _actmbeg;
  const backp* _actmend;
  const int _s;
  int _l_hs;
};

#undef next_sibhead
#undef next_siblab


/* label enumerator */
class o3active::labels {
public:
  /* return false when no more deps are available */
  bool next(int& restrict lab) {
    if(_act == _actend) { return false; }
    lab = _act->l(); ++_act;
    return true;
  }

private:
  friend class o3active; /* can only be created by o3active */
  labels(const backp* const arr, const int C)
    : _act(arr), _actend(arr + C) {}
  const backp* _act;
  const backp* const _actend;
};



/* inlined accessors */
inline int o3active::N() const { return _N; }


/* inlined enumerator access to active deps */
inline o3active::lodeps
o3active::lomods(const int h, const int end) const {
  CHECK(assert(h >= -1); assert(h < _N); assert(h > end));
  return lodeps(_activemL[h], end, _N);
}
inline o3active::rodeps
o3active::romods(const int h, const int end) const {
  CHECK(assert(h >= -1); assert(h < _N); assert(h < end));
  return rodeps(_activemR[h], end);
}

inline o3active::lideps
o3active::limods(const int h, const int end) const {
  CHECK(assert(h >= -1); assert(h < _N));
  return lideps(_activemL[h], end, _N);
}
inline o3active::rideps
o3active::rimods(const int h, const int end) const {
  CHECK(assert(h >= -1); assert(h < _N));
  return rideps(_activemR[h], end);
}

inline o3active::lodeps
o3active::loheads(const int m, const int end) const {
  CHECK(assert(m >= 0); assert(m < _N); assert(m > end));
  return lodeps(_activehL[m], end, _N);
}
inline o3active::rodeps
o3active::roheads(const int m, const int end) const {
  CHECK(assert(m >= 0); assert(m < _N); assert(m < end));
  return rodeps(_activehR[m], end);
}

inline o3active::lideps
o3active::liheads(const int m, const int end) const {
  CHECK(assert(m >= 0); assert(m < _N));
  return lideps(_activehL[m], end, _N);
}
inline o3active::rideps
o3active::riheads(const int m, const int end) const {
  CHECK(assert(m >= 0); assert(m < _N));
  return rideps(_activehR[m], end);
}

inline o3active::lisibs
o3active::lisibheads(const int m, const int s) const {
  CHECK(assert(s >= -1); assert(s < m); assert(m < _N));
  return lisibs(_activehLidx[m], _activehLidx[s],
                (const backp**)_activehPidx, _activehCidx,
                s, m, _N);
}
inline o3active::risibs
o3active::risibheads(const int m, const int s) const {
  CHECK(assert(m >= 0); assert(m < s); assert(s < _N));
  return risibs(_activehRidx[m], _activehRidx[s],
                (const backp**)_activehPidx, _activehCidx,
                s, m, _N);
}

inline o3active::labels
o3active::deplabs(const int h, const int m) const {
  CHECK(assert(h >= -1); assert(h < _N));
  CHECK(assert(m >= 0);  assert(m < _N));
  const int idx_hm = h + m*(_N + 1);
  return labels(_activehPidx[idx_hm], _activehCidx[idx_hm]);
}


/* inlined raw accessors */
inline const backp* o3active::lharr(const int m) const {
  CHECK(assert(m >= 0); assert(m < _N));
  return _activehL[m];
}
inline const backp* o3active::rharr(const int m) const {
  CHECK(assert(m >= 0); assert(m < _N));
  return _activehR[m];
}

#endif /* O3ACTIVE_H */
