#include "O3Decoder.h"
#include "O3Oracle.h"
#include "O3Part.h"
#include "GZIO.h"
#include "O3Ablation.h"

#include <stdio.h>
#include <stdlib.h>

#include <vector>
#include <sstream>
#include <iostream>

using std::cerr;
using std::endl;
using std::ostringstream;

/* lifecycle */
o3decoder::o3decoder()
  : o3trainer("Decoder"),
    _outfstem(),
    _outf(NULL),
    _nwords(0),
    _ncorrect(0) {}
o3decoder::~o3decoder() {}


/* print usage information */
void o3decoder::usage_impl() const {
  cerr
    << "O3Decoder options:" << endl
    << " --stem=<str>     : print decoded trees to <dir>/<stem>.NNN" << endl
    << " --pdir=<str>     : parameter directory" << endl
    << " --range=<ints>   : list of parameter iteration indices" << endl
    << " --ensemble=<int> : for parameter ensembles, number of models" << endl
    << " --timed=<int>    : log timing information, with <int> trials" << endl
    << endl;
}

/* initialize the trainer */
void o3decoder::init_impl(const o3trainer::strmap& opts) {
  /* recover output file */
  o3trainer::strmap::const_iterator it;

  if((it = opts.find("stem")) == opts.end()) {
    usage("please provide a decoding output stem --stem=...");
  }
  _outfstem = it->second;

  if((it = opts.find("pdir")) == opts.end()) {
    usage("please provide a parameter directory --pdir=...");
  }
  _pdir = it->second;

  if((it = opts.find("range")) == opts.end()) {
    usage("please provide a range of parameters --range=...");
  }
  _range.clear();

  /* process the range */
  const string rstr = it->second;
  const char* p = rstr.c_str();
  while(1) {
    char* nextp = NULL;
    const int next = (int)strtol(p, &nextp, 10);
    p = nextp;
    if(*p == '-') { /* insert a contiguous range */
      ++p;
      const int last = (int)strtol(p, &nextp, 10);
      p = nextp;
      for(int i = next; i <= last; ++i) { _range.push_back(i); }
    } else if(*p == ',' || *p == '\0') { /* insert a single number */
      _range.push_back(next);
    } else {
      cerr << "malformed range: \"" << rstr << "\"" << endl;
      assert(0);
    }
    if(*p == '\0') { break; }
    ++p;
  }


  _ensemble = 0;
  if((it = opts.find("ensemble")) != opts.end()) {
    _ensemble = atoi(it->second.c_str());
    assert(_ensemble > 0);
  }


  _timed = 0;
  if((it = opts.find("timed")) != opts.end()) {
    _timed = atoi(it->second.c_str());
    assert(_timed > 0);
  }
}


/* training interface */
void o3decoder::train_start() {
  _range_pos = 0;
}
void o3decoder::iter_start()  {
  if(_range_pos >= (int)_range.size()) {
    cerr << "O3Decoder : range exhausted; exiting..." << endl;
    exit(0);
  }
  char parnum[10];
  sprintf(parnum, "%03d", _range[_range_pos]);
  const string outfname = (_dir + "/" + _outfstem + "." + parnum);
  _outf = gzio::gzopen(outfname.c_str(), "w");
  _ncorrect = _nwords = 0;
  if(_ensemble == 0) { /* load single vector */
    const string parfname = (_pdir + "/parameters." + parnum + ".gz");
    cerr << "O3Decoder : loading parameters '" << parfname << "'" << flush;
    FILE* const fw = gzio::gzopen(parfname.c_str(), "r");
    if(_F->averaged()) { _F->Wavg().load(fw); }
    else               { _F->Wlin().load(fw); }
    gzio::gzclose(parfname.c_str(), fw);
  } else { /* load ensemble of vectors */
    if(_F->averaged()) { _F->Wavg().zero(); }
    else               { _F->Wlin().zero(); }
    for(int i = 1; i <= _ensemble; ++i) {
      char modnum[10];
      sprintf(modnum, "%03d", i);
      const string parfname =
        (_pdir + "/model-" + modnum + ".parameters." + parnum + ".gz");
      cerr << "O3Decoder : loading parameters '" << parfname << "'" << flush;
      o3features* const F2 = new o3features(_F->L(), _F->averaged());
      FILE* const f2 = gzio::gzopen(parfname.c_str(), "r");
      if(F2->averaged()) { F2->Wavg().load(f2); }
      else               { F2->Wlin().load(f2); }
      gzio::gzclose(parfname.c_str(), f2);
      if(_F->averaged()) { _F->Wavg().add(F2->Wavg()); }
      else               { _F->Wlin().add(F2->Wlin()); }
      delete F2;
    }
  }
}
void o3decoder::iter_end(string& logline, string& repline) {
  char padnum[10];
  sprintf(padnum, "%03d", _range[_range_pos]);
  const string outfname = (_dir + "/" + _outfstem + "." + padnum);
  gzio::gzclose(outfname.c_str(), _outf); _outf = NULL;
  /* record accuracy */
  const double percent = 100.0*((double)_ncorrect)/((double)_nwords);
  ostringstream os;
  os << fixed << setprecision(8)
     << _nwords << " " << _ncorrect << " " << percent;
  logline = os.str();
  os.str("");
  os << fixed << setprecision(2)
     << " acc = " << percent
     << " ; #words = " << _nwords
     << " ; #correct = " << _ncorrect;
  repline = os.str();
  ++_range_pos;
}


/* process an example */
void o3decoder::process(const ex* const restrict e) {
  struct timeval t0;
  if(_timed > 0) { gettimeofday(&t0, NULL); }

  const int trials = (_timed > 0 ? _timed : 1);
  const int N = (int)e->heads.size();
  vector<int> hhat, lhat;

  for(int trial = 0; trial < trials; ++trial) {
    if(_F->averaged()) {
      o3ablation::argmax(_ablation, *_P, N,
                         _F->avg_scavg(e->words, e->tags),
                         hhat, lhat);
    } else {
      o3ablation::argmax(_ablation, *_P, N,
                         _F->lin_scorer(e->words, e->tags),
                         hhat, lhat);
    }
    assert(e->heads.size()  == hhat.size());
    assert(e->labels.size() == lhat.size());
  }

  if(_timed > 0) {
    struct timeval tf;
    gettimeofday(&tf, NULL);
    const double secs = ((double)(tf.tv_sec - t0.tv_sec)) +
      ((double)(tf.tv_usec - t0.tv_usec))/1000000.0;
    fprintf(_outf, "TIME %d %d %.17g %.17g\n",
            e->id(), N, secs, secs/((double)_timed));
  }

  e->fwrite(_outf, hhat, lhat);

  _nwords += hhat.size();
  for(int i = 0; i < N; ++i) {
    _ncorrect += (((hhat[i] == e->heads[i]) &&
                   (lhat[i] == e->labels[i])) ? 1 : 0);
  }
}


/* process an example */
void o3decoder::classify(const ex* const restrict e,
                         vector<int>& restrict ret_h,
                         vector<int>& restrict ret_l) const {
  assert(0);
}
