/* -*-c++-*- */
#ifndef O3FGEN_H
#define O3FGEN_H
/* O3FGen : A class providing static methods for generating features
   and processing them.  The methods are templated on FeatFunc, a
   feature-processing functor that accepts a feature and does
   "something" with it. */

#include "PadSent.h"
#include "Bits.h"

#include <string>
#include <vector>

using std::string;
using std::vector;

class o3fgen {
public:
  /* generate dependency features */
  template<class FeatFunc>
  static double dep(FeatFunc& restrict proc,
                    const padsent& restrict x,
                    const int h, const int m);
  /* generate sibling features */
  template<class FeatFunc>
  static double sib(FeatFunc& restrict proc,
                    const padsent& restrict x,
                    const int h, const int m, const int c);
  /* generate grandchild features */
  template<class FeatFunc>
  static double gch(FeatFunc& restrict proc,
                    const padsent& restrict x,
                    const int h, const int m, const int c);
  /* generate tri-sibling features */
  template<class FeatFunc>
  static double sibs(FeatFunc& restrict proc, FeatFunc& restrict procw4,
                     FeatFunc& restrict ffunb4, FeatFunc& restrict ffunr4,
                     const padsent& restrict x,
                     const int h, const int m, const int c, const int s);
  /* generate grand-sibling features */
  template<class FeatFunc>
  static double gsib(FeatFunc& restrict proc, FeatFunc& restrict procw4,
                     FeatFunc& restrict ffunb4, FeatFunc& restrict ffunr4,
                     const padsent& restrict x,
                     const int h, const int m, const int c, const int s);

  /* generate auxiliary dependency features */
  template<class FeatFunc>
  static double aux(FeatFunc& restrict proc,
                    const padsent& restrict x,
                    const vector<bool>& restrict agree,
                    const int h, const int m);

  /* configure feature flags */
  static void config(const string& str);

  static bool USEFTAG;
  static bool USECTAG;
  static bool USECOORD;
  static bool USECOORD_BI;
  static bool USECOMMA;
  static bool DEPMASH;
  static double SCALE_4GRAM;

private:
  template<class FeatFunc>
  static double tok(FeatFunc& restrict proc,
                    const padsent& restrict x,
                    const int t, const tag_t tok_type);


  /* compute a reprentation of the given levels */
  static tag_t  levels2(const int hl, const int ml);
  static word_t levels4(const int hl, const int ml,
                        const int cl, const int sl);
};




/*** inlined implementation ***/
#define MAX_LEVEL   8
#define LEVEL_SHIFT 3U
inline tag_t o3fgen::levels2(const int hl, const int ml) {
  /* normalize so that levels are relative */
  const int lev_min = (hl < ml ? hl : ml);
  const int hl_norm = hl - lev_min;
  const int ml_norm = ml - lev_min;
  assert(hl_norm >= 0); assert(hl_norm < MAX_LEVEL);
  assert(ml_norm >= 0); assert(ml_norm < MAX_LEVEL);
  const unsigned uhl = (unsigned)(hl_norm);
  const unsigned uml = (unsigned)(ml_norm);
  const unsigned ret = ((uhl << LEVEL_SHIFT) | uml);
  return TCONV(ret);
}
inline word_t o3fgen::levels4(const int hl, const int ml,
                              const int cl, const int sl) {
  /* normalize so that levels are relative */
  int lev_min = (hl < ml ? hl : ml);
  lev_min = (lev_min < cl ? lev_min : cl);
  lev_min = (lev_min < sl ? lev_min : sl);
  const int hl_norm = hl - lev_min;
  const int ml_norm = ml - lev_min;
  const int cl_norm = cl - lev_min;
  const int sl_norm = sl - lev_min;
  assert(hl_norm >= 0); assert(hl_norm < MAX_LEVEL);
  assert(ml_norm >= 0); assert(ml_norm < MAX_LEVEL);
  assert(cl_norm >= 0); assert(cl_norm < MAX_LEVEL);
  assert(sl_norm >= 0); assert(sl_norm < MAX_LEVEL);
  const unsigned uhl = (unsigned)(hl_norm);
  const unsigned uml = (unsigned)(ml_norm);
  const unsigned ucl = (unsigned)(cl_norm);
  const unsigned usl = (unsigned)(sl_norm);
  const unsigned ret = ((uhl << (3U*LEVEL_SHIFT)) |
                        (uml << (2U*LEVEL_SHIFT)) |
                        (ucl << (LEVEL_SHIFT)) |
                        (usl));
  return WCONV(ret);
}
#undef MAX_LEVEL
#undef LEVEL_SHIFT

#ifndef EXPLICIT_TEMPLATES
#include "O3FGenTok.tcc"
#include "O3FGenDep.tcc"
#include "O3FGenSib.tcc"
#include "O3FGenGch.tcc"
#include "O3FGenSibs.tcc"
#include "O3FGenGsib.tcc"
#include "O3FGenAux.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3FGEN_H */
