#include "Mappings.h"
#include "FeatTypes.h"
#include "Feature.h"
#include "Macros.h"

#include <map>

using std::map;

/* dependency distance bins: a feature is generated if |h - m| > N */
#define DISTBINGT1 ((tag_t)2)
#define DISTBINGT2 ((tag_t)5)
#define DISTBINGT3 ((tag_t)10)
#define DISTBINGT4 ((tag_t)20)
#define DISTBINGT5 ((tag_t)30)
#define DISTBINGT6 ((tag_t)40)
/* maximum distance */
#define MAX_DIST   41
/* maximum number of crossed token-types */
#define MAX_CROSS  ((word_t)4)

template<class FeatFunc> double o3fgen::
dep(FeatFunc& restrict ffun, const padsent& restrict x,
    const int h, const int m) {
  /* sets for computing between features */
  static map<tag_t,bool>  CTAGFLAGS;
  static map<tag_t,bool>  FTAGFLAGS;
  static map<word_t,bool> WORDFLAGS;

  CHECK(assert(h >= -1); assert(h < x.N));
  CHECK(assert(m >= 0);  assert(m < x.N));
  double ret = 0;

  /* head and modifier token features */
  if(h == -1) {
    /*     root token                   features of root word */
    ret += ffun(fw(fbi::ROOT, WNULL)) + tok(ffun, x, m, 0xf0);
  } else {
    /*     head features                modifier features */
    ret += tok(ffun, x, h, 0)         + tok(ffun, x, m, 0xff);
  }


  // const int hqlev = x.quotelev[h];
  // const int mqlev = x.quotelev[m];
  // const int hplev = x.parenlev[h];
  // const int mplev = x.parenlev[m];

  /* preprocess the dependency for between features */
  tag_t dist = 0;
  word_t numVB = 0, numCC = 0, numIN = 0, numPU = 0;
  // bool cross_qlev_higher = false;
  // bool cross_qlev_lower = false;
  // bool cross_plev_higher = false;
  // bool cross_plev_lower = false;
  { /* collect counts of the tag types and set flags */
    const int start = (h < m ? h : m);
    const int end = (h < m ? m : h);
    const int tmp_dist = end - start;
    const int capped_dist = (tmp_dist > MAX_DIST ? MAX_DIST : tmp_dist);
    dist = TCONV(capped_dist);
    for(int b = start + 1; b < end; ++b) {
      const word_t bw = x.words[b];
      const tag_t  bt = x.tags[b];
      if(mappings::iscoord(bt)) { ++numCC; }
      if(mappings::isprep(bt))  { ++numIN; }
      if(mappings::ispunc(bt))  { ++numPU; }
      if(mappings::isverb(bt))  { ++numVB; }
      FTAGFLAGS[bt] = true;
      CTAGFLAGS[mappings::ctag(bt)] = true;
      WORDFLAGS[bw] = true;
      /* quote- and paren-levels */
      // const int bqlev = x.quotelev[b];
      // const int bplev = x.parenlev[b];
      // if(bqlev > hqlev && bqlev > mqlev) { cross_qlev_higher = true; }
      // if(bqlev < hqlev && bqlev < mqlev) { cross_qlev_lower  = true; }
      // if(bplev > hplev && bplev > mplev) { cross_plev_higher = true; }
      // if(bplev < hplev && bplev < mplev) { cross_plev_lower  = true; }
    }
    /* place a maximum on crossings */
    numCC = (numCC > MAX_CROSS ? MAX_CROSS : numCC);
    numIN = (numIN > MAX_CROSS ? MAX_CROSS : numIN);
    numPU = (numPU > MAX_CROSS ? MAX_CROSS : numPU);
    numVB = (numVB > MAX_CROSS ? MAX_CROSS : numVB);
  }


  /* head and mod words */
  const word_t hw   = x.words[h];
  const word_t mw   = x.words[m];
  const word_t hwp1 = x.words[h - 1];
  const word_t hwn1 = x.words[h + 1];
  const word_t mwp1 = x.words[m - 1];
  const word_t mwn1 = x.words[m + 1];
  { /* lexical-only features */
    ret +=
      /* normal dependency features */
      ffun(fww(fbi::HWMW, hw, mw)) +

      /* dependency context features */
      ffun(fwww(fbi::HWMW_HNW, hwn1, hw, mw)) +
      ffun(fwww(fbi::HWMW_HPW, hwp1, hw, mw)) +
      ffun(fwww(fbi::HWMW_MNW, mwn1, hw, mw)) +
      ffun(fwww(fbi::HWMW_MPW, mwp1, hw, mw)) +

      /* dependency distance features */
      ffun(fw(fbi::DISTEXACT, dist)) +
#define DIST(n)                                                         \
      ffun(ftt(fbi::DBIN,                                               \
               (dist <= (DISTBINGT ## n) ? ((tag_t)0) : TNULL),         \
               DISTBINGT ## n)) +                                       \
      ffun(fttww(fbi::DBIN_HWMW,                                        \
                 (dist <= (DISTBINGT ## n) ? ((tag_t)0) : TNULL),       \
                 DISTBINGT ## n, hw, mw))
    DIST(1) + DIST(2) + DIST(3) + DIST(4) + DIST(5) + DIST(6) +
#undef DIST

      /* dependency between features */
      ffun(fw(fbi::BCC, numCC)) +
      ffun(fw(fbi::BIN, numIN)) +
      ffun(fw(fbi::BPU, numPU)) +
      ffun(fw(fbi::BVB, numVB)) +

      ffun(fwww(fbi::HWMW_BCC, hw, mw, numCC)) +
      ffun(fwww(fbi::HWMW_BIN, hw, mw, numIN)) +
      ffun(fwww(fbi::HWMW_BPU, hw, mw, numPU)) +
      ffun(fwww(fbi::HWMW_BVB, hw, mw, numVB));

    /* word between features */
    map<word_t,bool>::const_iterator it = WORDFLAGS.begin();
    const map<word_t,bool>::const_iterator iend = WORDFLAGS.end();
    for(; it != iend; ++it) {
      const word_t bw = it->first;
      ret +=
        ffun(fw(fbi::BW, bw)) +
        ffun(fwww(fbi::HWMW_BW, hw, mw, bw));
    }
  }


  /* quote- and paren-level features */
  // const tag_t qlev_hm_norm = levels2(hqlev, mqlev);
  // const tag_t plev_hm_norm = levels2(hplev, mplev);
  /* add normalized level features */
  // ret +=
  //   ffun(ft(fbi::QUOTELEV, qlev_hm_norm)) +
  //   ffun(ft(fbi::PARENLEV, plev_hm_norm));
  // if(cross_qlev_higher) {
  //   ret += ffun(ft(fbi::QUOTELEV_CROSS, TMAX));
  // }
  // if(cross_qlev_lower)  {
  //   ret += ffun(ft(fbi::QUOTELEV_CROSS, 0));
  // }
  // if(cross_plev_higher) {
  //   ret += ffun(ft(fbi::PARENLEV_CROSS, TMAX));
  // }
  // if(cross_plev_lower)  {
  //   ret += ffun(ft(fbi::PARENLEV_CROSS, 0));
  // }



  if(USEFTAG) { /* head and mod tags with context */
    const tag_t ht   = x.tags[h];
    const tag_t htp1 = x.tags[h - 1];
    const tag_t htn1 = x.tags[h + 1];
    const tag_t mt   = x.tags[m];
    const tag_t mtp1 = x.tags[m - 1];
    const tag_t mtn1 = x.tags[m + 1];

    const word_t hb   = mappings::wft(hw,   ht);
    const word_t hbp1 = mappings::wft(hwp1, htp1);
    const word_t hbn1 = mappings::wft(hwn1, htn1);
    const word_t mb   = mappings::wft(mw,   mt);
    const word_t mbp1 = mappings::wft(mwp1, mtp1);
    const word_t mbn1 = mappings::wft(mwn1, mtn1);

    ret +=
      /* normal dependency features */
      ffun(ftt(fbi::HTMT, ht, mt)) +
      ffun(ftw(fbi::HWMT, mt, hw)) +
      ffun(ftw(fbi::HTMW, ht, mw)) +

      ffun(ftw(fbi::HTMB, ht, mb)) +
      ffun(fww(fbi::HWMB, hw, mb)) +
      ffun(ftw(fbi::HBMT, mt, hb)) +
      ffun(fww(fbi::HBMW, hb, mw)) +
      ffun(fww(fbi::HBMB, hb, mb)) +

      /* dependency context features */
      ffun(fttt(fbi::HTMT_HNT, htn1, ht, mt)) +
      ffun(fttt(fbi::HTMT_HPT, htp1, ht, mt)) +
      ffun(fttt(fbi::HTMT_MNT, mtn1, ht, mt)) +
      ffun(fttt(fbi::HTMT_MPT, mtp1, ht, mt)) +

      ffun(ftttt(fbi::HTMT_NNTT, htn1, mtn1, ht, mt)) +
      ffun(ftttt(fbi::HTMT_NPTT, htn1, mtp1, ht, mt)) +
      ffun(ftttt(fbi::HTMT_PNTT, htp1, mtn1, ht, mt)) +
      ffun(ftttt(fbi::HTMT_PPTT, htp1, mtp1, ht, mt)) +

      ffun(ftww(fbi::HWMW_HNT, htn1, hw, mw)) +
      ffun(ftww(fbi::HWMW_HPT, htp1, hw, mw)) +
      ffun(ftww(fbi::HWMW_MNT, mtn1, hw, mw)) +
      ffun(ftww(fbi::HWMW_MPT, mtp1, hw, mw)) +

      ffun(fttww(fbi::HWMW_NNTT, htn1, mtn1, hw, mw)) +
      ffun(fttww(fbi::HWMW_NPTT, htn1, mtp1, hw, mw)) +
      ffun(fttww(fbi::HWMW_PNTT, htp1, mtn1, hw, mw)) +
      ffun(fttww(fbi::HWMW_PPTT, htp1, mtp1, hw, mw)) +

      ffun(fwww(fbi::HBMB_HNB, hbn1, hb, mb)) +
      ffun(fwww(fbi::HBMB_HPB, hbp1, hb, mb)) +
      ffun(fwww(fbi::HBMB_MNB, mbn1, hb, mb)) +
      ffun(fwww(fbi::HBMB_MPB, mbp1, hb, mb)) +

      /* dependency distance features */
#define DIST(n)                                                         \
      ffun(ftttt(fbi::DBIN_HTMT,                                        \
                 (dist <= (DISTBINGT ## n) ? ((tag_t)0) : TNULL),       \
                 DISTBINGT ## n, ht, mt))
      DIST(1) + DIST(2) + DIST(3) + DIST(4) + DIST(5) + DIST(6) +
#undef DIST

    /* dependency between features */
    ffun(fttw(fbi::HTMT_BCC, ht, mt, numCC)) +
    ffun(fttw(fbi::HTMT_BIN, ht, mt, numIN)) +
    ffun(fttw(fbi::HTMT_BPU, ht, mt, numPU)) +
    ffun(fttw(fbi::HTMT_BVB, ht, mt, numVB));

    { /* word between features */
      map<word_t,bool>::const_iterator it         = WORDFLAGS.begin();
      const map<word_t,bool>::const_iterator iend = WORDFLAGS.end();
      for(; it != iend; ++it) {
        const word_t bw = it->first;
        ret += ffun(fttw(fbi::HTMT_BW, ht, mt, bw));
      }
    }
    { /* tag between features */
      map<tag_t,bool>::const_iterator it         = FTAGFLAGS.begin();
      const map<tag_t,bool>::const_iterator iend = FTAGFLAGS.end();
      for(; it != iend; ++it) {
        const tag_t bt = it->first;
        ret +=
          ffun(ft(fbi::BT, bt)) +
          ffun(fttt(fbi::HTMT_BT, ht, mt, bt)) +
          ffun(ftww(fbi::HWMW_BT, bt, hw, mw));
      }
    }

    /* add normalized level features */
    // ret +=
    //   ffun(fttt(fbi::QUOTELEV_HTMT, qlev_hm_norm, ht, mt)) +
    //   ffun(fttt(fbi::PARENLEV_HTMT, plev_hm_norm, ht, mt));
    // if(cross_qlev_higher) {
    //   ret += ffun(fttt(fbi::QUOTELEV_CROSS_HTMT, TMAX, ht, mt));
    // }
    // if(cross_qlev_lower)  {
    //   ret += ffun(fttt(fbi::QUOTELEV_CROSS_HTMT, 0, ht, mt));
    // }
    // if(cross_plev_higher) {
    //   ret += ffun(fttt(fbi::PARENLEV_CROSS_HTMT, TMAX, ht, mt));
    // }
    // if(cross_plev_lower)  {
    //   ret += ffun(fttt(fbi::PARENLEV_CROSS_HTMT, 0, ht, mt));
    // }
  }


  if(USECTAG) { /* head and mod ctags with context */
    const tag_t hq   = mappings::ctag(x.tags[h]);
    const tag_t hqp1 = mappings::ctag(x.tags[h - 1]);
    const tag_t hqn1 = mappings::ctag(x.tags[h + 1]);
    const tag_t mq   = mappings::ctag(x.tags[m]);
    const tag_t mqp1 = mappings::ctag(x.tags[m - 1]);
    const tag_t mqn1 = mappings::ctag(x.tags[m + 1]);

    const word_t hr   = mappings::wct(hw,   hq);
    const word_t hrp1 = mappings::wct(hwp1, hqp1);
    const word_t hrn1 = mappings::wct(hwn1, hqn1);
    const word_t mr   = mappings::wct(mw,   mq);
    const word_t mrp1 = mappings::wct(mwp1, mqp1);
    const word_t mrn1 = mappings::wct(mwn1, mqn1);

    ret +=
      /* normal dependency features */
      ffun(ftt(fbi::HQMQ, hq, mq)) +
      ffun(ftw(fbi::HWMQ, mq, hw)) +
      ffun(ftw(fbi::HQMW, hq, mw)) +

      ffun(ftw(fbi::HQMR, hq, mr)) +
      ffun(fww(fbi::HWMR, hw, mr)) +
      ffun(ftw(fbi::HRMQ, mq, hr)) +
      ffun(fww(fbi::HRMW, hr, mw)) +
      ffun(fww(fbi::HRMR, hr, mr)) +

      /* dependency context features */
      ffun(fttt(fbi::HQMQ_HNQ, hqn1, hq, mq)) +
      ffun(fttt(fbi::HQMQ_HPQ, hqp1, hq, mq)) +
      ffun(fttt(fbi::HQMQ_MNQ, mqn1, hq, mq)) +
      ffun(fttt(fbi::HQMQ_MPQ, mqp1, hq, mq)) +

      ffun(ftttt(fbi::HQMQ_NNQQ, hqn1, mqn1, hq, mq)) +
      ffun(ftttt(fbi::HQMQ_NPQQ, hqn1, mqp1, hq, mq)) +
      ffun(ftttt(fbi::HQMQ_PNQQ, hqp1, mqn1, hq, mq)) +
      ffun(ftttt(fbi::HQMQ_PPQQ, hqp1, mqp1, hq, mq)) +

      ffun(ftww(fbi::HWMW_HNQ, hqn1, hw, mw)) +
      ffun(ftww(fbi::HWMW_HPQ, hqp1, hw, mw)) +
      ffun(ftww(fbi::HWMW_MNQ, mqn1, hw, mw)) +
      ffun(ftww(fbi::HWMW_MPQ, mqp1, hw, mw)) +

      ffun(fttww(fbi::HWMW_NNQQ, hqn1, mqn1, hw, mw)) +
      ffun(fttww(fbi::HWMW_NPQQ, hqn1, mqp1, hw, mw)) +
      ffun(fttww(fbi::HWMW_PNQQ, hqp1, mqn1, hw, mw)) +
      ffun(fttww(fbi::HWMW_PPQQ, hqp1, mqp1, hw, mw)) +

      ffun(fwww(fbi::HRMR_HNR, hrn1, hr, mr)) +
      ffun(fwww(fbi::HRMR_HPR, hrp1, hr, mr)) +
      ffun(fwww(fbi::HRMR_MNR, mrn1, hr, mr)) +
      ffun(fwww(fbi::HRMR_MPR, mrp1, hr, mr)) +

      /* dependency distance features */
#define DIST(n)                                                         \
      ffun(ftttt(fbi::DBIN_HQMQ,                                        \
                 (dist <= (DISTBINGT ## n) ? ((tag_t)0) : TNULL),       \
                 DISTBINGT ## n, hq, mq))
      DIST(1) + DIST(2) + DIST(3) + DIST(4) + DIST(5) + DIST(6) +
#undef DIST

    /* dependency between features */
    ffun(fttw(fbi::HQMQ_BCC, hq, mq, numCC)) +
    ffun(fttw(fbi::HQMQ_BIN, hq, mq, numIN)) +
    ffun(fttw(fbi::HQMQ_BPU, hq, mq, numPU)) +
    ffun(fttw(fbi::HQMQ_BVB, hq, mq, numVB));

    { /* word between features */
      map<word_t,bool>::const_iterator it         = WORDFLAGS.begin();
      const map<word_t,bool>::const_iterator iend = WORDFLAGS.end();
      for(; it != iend; ++it) {
        const word_t bw = it->first;
        ret += ffun(fttw(fbi::HQMQ_BW, hq, mq, bw));
      }
    }
    { /* tag between features */
      map<tag_t,bool>::const_iterator it         = CTAGFLAGS.begin();
      const map<tag_t,bool>::const_iterator iend = CTAGFLAGS.end();
      for(; it != iend; ++it) {
        const tag_t bq = it->first;
        CHECK(assert(bq < mappings::Nctags()));
        ret +=
          ffun(ft(fbi::BQ, bq)) +
          ffun(fttt(fbi::HQMQ_BQ, hq, mq, bq)) +
          ffun(ftww(fbi::HWMW_BQ, bq, hw, mw));
      }
    }

    /* add normalized level features */
    // ret +=
    //   ffun(fttt(fbi::QUOTELEV_HQMQ, qlev_hm_norm, hq, mq)) +
    //   ffun(fttt(fbi::PARENLEV_HQMQ, plev_hm_norm, hq, mq));
    // if(cross_qlev_higher) {
    //   ret += ffun(fttt(fbi::QUOTELEV_CROSS_HQMQ, TMAX, hq, mq));
    // }
    // if(cross_qlev_lower)  {
    //   ret += ffun(fttt(fbi::QUOTELEV_CROSS_HQMQ, 0, hq, mq));
    // }
    // if(cross_plev_higher) {
    //   ret += ffun(fttt(fbi::PARENLEV_CROSS_HQMQ, TMAX, hq, mq));
    // }
    // if(cross_plev_lower)  {
    //   ret += ffun(fttt(fbi::PARENLEV_CROSS_HQMQ, 0, hq, mq));
    // }
  }


  /* clean up */
  CTAGFLAGS.clear();
  FTAGFLAGS.clear();
  WORDFLAGS.clear();

  return ret;
}


#undef DISTBINGT1
#undef DISTBINGT2
#undef DISTBINGT3
#undef DISTBINGT4
#undef DISTBINGT5
#undef DISTBINGT6

#undef MAX_DIST
#undef MAX_CROSS
