/* -*-c++-*- */
#ifndef O3FEATURES_H
#define O3FEATURES_H

/* O3Features : A feature mapping that performs on-the-fly generation
   of features.  The interface allows either performing inner products
   to compute scores or adding features to the parameters.  The
   o3features is a large wrapper with many sub-classes for tasks like
   feature generation and parameter manipulation. */


#include "Feature.h"
#include "FeatFunctors.h"
#include "O3ParamsAvg.h"
#include "O3ParamsLin.h"
#include "AuxDeps.h"

#include <assert.h>

#include <vector>

using std::vector;

class o3features {
public:
  o3features(const int inL, const bool avg);
  ~o3features();

  /* nested classes */
  /* feature generator implementing PartScorer interface */
  template<class FeatFunc> class fgen;
  /* feature scores augmented with loss scores */
  template<class FeatFunc> class fgen_loss;
  /* features with auxiliary dependencies */
  template<class FeatFunc> class fgen_aux;
  /* for convenience */
  typedef fgen<const featf_avg_dot>        favg_dot;
  typedef fgen<const featf_avg_adot>       favg_adot;
  typedef fgen<featf_avg_add>              favg_add;
  typedef fgen<const featf_lin_dot>        flin_dot;
  typedef fgen_loss<const featf_lin_dot>   flin_ldot;
  typedef fgen<featf_lin_add>              flin_add;
  typedef fgen_aux<const featf_avg_dot>    faux_dot;
  typedef fgen_aux<const featf_avg_adot>   faux_adot;
  typedef fgen_aux<featf_avg_add>          faux_add;

  /* functions that instantiate a feature-generator for a given
     sentence.  calling these will cause new objects to be created, so
     try to call each only once per sentence.  as a special case, the
     value to add to the features can be adjusted via
     fadd.handler().alpha(new_alpha) */
  const favg_dot& avg_scorer(const vector<int>& restrict words,
                             const vector<int>& restrict tags) const;
  const favg_adot& avg_scavg(const vector<int>& restrict words,
                             const vector<int>& restrict tags) const;
  favg_add& avg_adder(const vector<int>& restrict words,
                      const vector<int>& restrict tags);
  const flin_dot& lin_scorer(const vector<int>& restrict words,
                             const vector<int>& restrict tags) const;
  const flin_ldot& lin_scloss(const vector<int>& restrict words,
                              const vector<int>& restrict tags,
                              const vector<int>& restrict heads,
                              const vector<int>& restrict labels) const;
  flin_add& lin_adder(const vector<int>& restrict words,
                      const vector<int>& restrict tags);

  const faux_dot& aux_scorer(const vector<int>& restrict words,
                             const vector<int>& restrict tags,
                             const auxdeps& restrict aux) const;
  const faux_adot& aux_scavg(const vector<int>& restrict words,
                             const vector<int>& restrict tags,
                             const auxdeps& restrict aux) const;
  faux_add& aux_adder(const vector<int>& restrict words,
                      const vector<int>& restrict tags,
                      const auxdeps& restrict aux);

  /* parameter access */
  int L() const;
  bool averaged() const;
  o3params_avg& Wavg();
  o3params_lin& Wlin();
  const o3params_avg& Wavg() const;
  const o3params_lin& Wlin() const;


private:
  const int _L;                         /* the number of labels */
  o3params_avg* const _Wavg;            /* perceptron-style parameters */
  o3params_lin* const _Wlin;            /* normal linear parameters */

  mutable const favg_dot*  _FGavg_dot;  /* perceptron inner products */
  mutable const favg_adot* _FGavg_adot; /* averaged inner products */
  favg_add*                _FGavg_add;  /* add to perceptron params */
  mutable const flin_dot*  _FGlin_dot;  /* normal inner products */
  mutable const flin_ldot* _FGlin_ldot; /* loss-added inner products */
  flin_add*                _FGlin_add;  /* add to normal params */
  mutable const faux_dot*  _FGaux_dot;  /* perceptron inner products */
  mutable const faux_adot* _FGaux_adot; /* averaged inner products */
  faux_add*                _FGaux_add;  /* add to perceptron params */
};


/* inlined functions for o3features */
inline int o3features::L() const { return _L; }
inline bool o3features::averaged() const { return (_Wavg != NULL); }
inline o3params_avg& o3features::Wavg()             { return *_Wavg; }
inline o3params_lin& o3features::Wlin()             { return *_Wlin; }
inline const o3params_avg& o3features::Wavg() const { return *_Wavg; }
inline const o3params_lin& o3features::Wlin() const { return *_Wlin; }



/* include the templated feature generators */
#ifndef EXPLICIT_TEMPLATES
#include "O3FeaturesFGen.tcc"
#endif /* EXPLICIT_TEMPLATES */



/* inlined methods for o3features that depend on the feature generator
   (included after the definition of fgen and associated classes) */
inline const o3features::favg_dot&
o3features::avg_scorer(const vector<int>& restrict words,
                       const vector<int>& restrict tags) const {
  assert(_Wavg != NULL);
  if(_FGavg_dot != NULL) { delete _FGavg_dot; }
  _FGavg_dot = new favg_dot(words, tags, _Wavg);
  return *_FGavg_dot;
}

inline const o3features::favg_adot&
o3features::avg_scavg(const vector<int>& restrict words,
                      const vector<int>& restrict tags) const {
  assert(_Wavg != NULL);
  if(_FGavg_adot != NULL) { delete _FGavg_adot; }
  _FGavg_adot = new favg_adot(words, tags, _Wavg);
  return *_FGavg_adot;
}

inline o3features::favg_add&
o3features::avg_adder(const vector<int>& restrict words,
                      const vector<int>& restrict tags) {
  assert(_Wavg != NULL);
  if(_FGavg_add != NULL) { delete _FGavg_add; }
  _FGavg_add = new favg_add(words, tags, _Wavg);
  return *_FGavg_add;
}



/* methods that produce part-scorers for non-averaged parameters */
inline const o3features::flin_dot&
o3features::lin_scorer(const vector<int>& restrict words,
                       const vector<int>& restrict tags) const {
  assert(_Wlin != NULL);
  if(_FGlin_dot != NULL) { delete _FGlin_dot; }
  _FGlin_dot = new flin_dot(words, tags, _Wlin);
  return *_FGlin_dot;
}

inline const o3features::flin_ldot&
o3features::lin_scloss(const vector<int>& restrict words,
                       const vector<int>& restrict tags,
                       const vector<int>& restrict heads,
                       const vector<int>& restrict labels) const {
  assert(_Wlin != NULL);
  if(_FGlin_ldot != NULL) { delete _FGlin_ldot; }
  _FGlin_ldot = new flin_ldot(words, tags, heads, labels, _Wlin);
  return *_FGlin_ldot;
}

inline o3features::flin_add&
o3features::lin_adder(const vector<int>& restrict words,
                      const vector<int>& restrict tags) {
  assert(_Wlin != NULL);
  if(_FGlin_add != NULL) { delete _FGlin_add; }
  _FGlin_add = new flin_add(words, tags, _Wlin);
  return *_FGlin_add;
}



/* auxiliary-dependency scorers */
inline const o3features::faux_dot&
o3features::aux_scorer(const vector<int>& restrict words,
                       const vector<int>& restrict tags,
                       const auxdeps& restrict aux) const {
  assert(_Wavg != NULL);
  if(_FGaux_dot != NULL) { delete _FGaux_dot; }
  _FGaux_dot = new faux_dot(words, tags, aux, _Wavg);
  return *_FGaux_dot;
}

inline const o3features::faux_adot&
o3features::aux_scavg(const vector<int>& restrict words,
                      const vector<int>& restrict tags,
                      const auxdeps& restrict aux) const {
  assert(_Wavg != NULL);
  if(_FGaux_adot != NULL) { delete _FGaux_adot; }
  _FGaux_adot = new faux_adot(words, tags, aux, _Wavg);
  return *_FGaux_adot;
}

inline o3features::faux_add&
o3features::aux_adder(const vector<int>& restrict words,
                      const vector<int>& restrict tags,
                      const auxdeps& restrict aux) {
  assert(_Wavg != NULL);
  if(_FGaux_add != NULL) { delete _FGaux_add; }
  _FGaux_add = new faux_add(words, tags, aux, _Wavg);
  return *_FGaux_add;
}

#endif /* O3FEATURES_H */
