#include <assert.h>
#include <stdlib.h>
#include <math.h>

#include "PadSent.h"
#include "O3FGen.h"

template<class FeatFunc>
class o3features::fgen {
public:
  typedef typename FeatFunc::params  params;
  typedef typename FeatFunc::primals primals;
  typedef FeatFunc feat_func;

  /* create a feature generator for a given sentence */
  fgen(const vector<int>& restrict words,
       const vector<int>& restrict tags,
       params* const restrict W);
  ~fgen();

  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double sib1(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const;
  double sib2(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const;
  double gci1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gci2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gco1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gco2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const;
  double gcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const;
  double sibs(const idx_t h, const idx_t m, const idx_t s, const idx_t c,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const;

  void alpha(const double alpha);
  const padsent& x() const;

private:
  /* data members */
  const size_t _N;
  const padsent _x;
  params* const restrict _W;
  double _alpha;
};


/* inlined methods */
template<class FeatFunc> inline o3features::fgen<FeatFunc>::
fgen(const vector<int>& restrict words,
     const vector<int>& restrict tags,
     params* const restrict W)
  : _N(words.size()),
    _x(words, tags),
    _W(W),
    _alpha(-HUGE_VAL) {
  assert(words.size() == tags.size());
}

template<class FeatFunc> inline
o3features::fgen<FeatFunc>::~fgen() {}

template<class FeatFunc> inline void
o3features::fgen<FeatFunc>::alpha(const double a) { _alpha = a; }

template<class FeatFunc> inline const padsent&
o3features::fgen<FeatFunc>::x() const { return _x; }



/* version for loss-augmented scores */
template<class FeatFunc>
class o3features::fgen_loss : public o3features::fgen<FeatFunc> {
public:
  fgen_loss(const vector<int>& restrict words,
            const vector<int>& restrict tags,
            const vector<int>& restrict heads,
            const vector<int>& restrict labels,
            typename FeatFunc::params* const restrict W);
  ~fgen_loss();

  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const;

private:
  /* helper functions for construction */
  static const idx_t* copyheads(const vector<int>& restrict v);
  static const lab_t* copylabels(const vector<int>& restrict v);

  /* data members */
  const idx_t* const restrict _heads;
  const lab_t* const restrict _labels;
};

/* inlined methods */
template<class FeatFunc> inline
o3features::fgen_loss<FeatFunc>::
fgen_loss(const vector<int>& restrict words,
          const vector<int>& restrict tags,
          const vector<int>& restrict heads,
          const vector<int>& restrict labels,
          typename FeatFunc::params* const restrict W)
  : fgen<FeatFunc>(words, tags, W),
    _heads(copyheads(heads)),
    _labels(copylabels(labels)) {
  assert(words.size() == tags.size());
  assert(words.size() == heads.size());
  assert(words.size() == labels.size());
}

template<class FeatFunc> inline
o3features::fgen_loss<FeatFunc>::~fgen_loss() {
  free((idx_t*)_heads); free((lab_t*)_labels);
}

template<class FeatFunc> inline const idx_t*
o3features::fgen_loss<FeatFunc>::copyheads(const vector<int>& restrict v) {
  const size_t N = v.size();
  assert(N > 0);
  idx_t* const arr = (idx_t*)malloc(N*sizeof(idx_t));
  for(size_t i = 0; i < N; ++i) { arr[i] = ICONV_NULL(v[i]); }
  return arr;
}

template<class FeatFunc> inline const lab_t*
o3features::fgen_loss<FeatFunc>::copylabels(const vector<int>& restrict v) {
  const size_t N = v.size();
  assert(N > 0);
  lab_t* const arr = (lab_t*)malloc(N*sizeof(idx_t));
  for(size_t i = 0; i < N; ++i) { arr[i] = LCONV(v[i]); }
  return arr;
}






template<class FeatFunc>
class o3features::fgen_aux {
public:
  typedef typename FeatFunc::params  params;
  typedef typename FeatFunc::primals primals;
  typedef FeatFunc feat_func;

  /* create a feature generator for a given sentence */
  fgen_aux(const vector<int>& restrict words,
           const vector<int>& restrict tags,
           const auxdeps& restrict aux,
           params* const restrict W);
  ~fgen_aux();

  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const;

  /* deactivate all of these */
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
    assert(0); return 0;
  }
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
    assert(0); return 0;
  }
  double sib1(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const {
    assert(0); return 0;
  }
  double sib2(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const {
    assert(0); return 0;
  }
  double gci1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    assert(0); return 0;
  }
  double gci2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    assert(0); return 0;
  }
  double gco1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    assert(0); return 0;
  }
  double gco2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const {
    assert(0); return 0;
  }
  double gcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const {
    assert(0); return 0;
  }
  double gcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const {
    assert(0); return 0;
  }
  double sibs(const idx_t h, const idx_t m, const idx_t s, const idx_t c,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const {
    assert(0); return 0;
  }

  void alpha(const double alpha);

private:
  /* data members */
  const size_t _N;
  const padsent _x;
  const auxdeps& restrict _aux;
  params* const restrict _W;
  double _alpha;
};


/* inlined methods */
template<class FeatFunc> inline
o3features::fgen_aux<FeatFunc>::
fgen_aux(const vector<int>& restrict words,
         const vector<int>& restrict tags,
         const auxdeps& restrict aux,
         typename FeatFunc::params* const restrict W)
  : _N(words.size()),
    _x(words, tags),
    _aux(aux),
    _W(W),
    _alpha(-HUGE_VAL) {
  assert(words.size() == tags.size());
}

template<class FeatFunc> inline
o3features::fgen_aux<FeatFunc>::~fgen_aux() {}

template<class FeatFunc> inline void
o3features::fgen_aux<FeatFunc>::alpha(const double a) { _alpha = a; }






#define W1(left, name, l1)                                      \
  ((left) ? _W->l ## name(l1)         : _W->r ## name(l1))
#define W2(left, name, l1, l2)                                  \
  ((left) ? _W->l ## name(l1, l2)     : _W->r ## name(l1, l2))
#define W3(left, name, l1, l2, l3)                                      \
  ((left) ? _W->l ## name(l1, l2, l3) : _W->r ## name(l1, l2, l3))

#define selectbackW1(left, name, l1)                    \
  typename Params::primals& W1 =                        \
    ((left) ? _W->l ## name(l1) : _W->r ## name(l1));   \
  typename Params::primals& W0 =                        \
    ((left) ? _W->l ## name(LB) : _W->r ## name(LB))
#define selectbackW2(left, name, l1, l2)                        \
  typename Params::primals& W11 =                               \
    ((left) ? _W->l ## name(l1, l2) : _W->r ## name(l1, l2));   \
  typename Params::primals& W10 =                               \
    ((left) ? _W->l ## name(l1, LB) : _W->r ## name(l1, LB));   \
  typename Params::primals& W01 =                               \
    ((left) ? _W->l ## name(LB, l2) : _W->r ## name(LB, l2));   \
  typename Params::primals& W00 =                               \
    ((left) ? _W->l ## name(LB, LB) : _W->r ## name(LB, LB))
#define selectbackW3(left, name, l1, l2, l3)                            \
  typename Params::primals& W111 =                                      \
    ((left) ? _W->l ## name(l1, l2, l3) : _W->r ## name(l1, l2, l3));   \
  typename Params::primals& W110 =                                      \
    ((left) ? _W->l ## name(l1, l2, LB) : _W->r ## name(l1, l2, LB));   \
  typename Params::primals& W101 =                                      \
    ((left) ? _W->l ## name(l1, LB, l3) : _W->r ## name(l1, LB, l3));   \
  typename Params::primals& W100 =                                      \
    ((left) ? _W->l ## name(l1, LB, LB) : _W->r ## name(l1, LB, LB));   \
  typename Params::primals& W011 =                                      \
    ((left) ? _W->l ## name(LB, l2, l3) : _W->r ## name(LB, l2, l3));   \
  typename Params::primals& W010 =                                      \
    ((left) ? _W->l ## name(LB, l2, LB) : _W->r ## name(LB, l2, LB));   \
  typename Params::primals& W001 =                                      \
    ((left) ? _W->l ## name(LB, LB, l3) : _W->r ## name(LB, LB, l3));   \
  typename Params::primals& W000 =                                      \
    ((left) ? _W->l ## name(LB, LB, LB) : _W->r ## name(LB, LB, LB))
#undef selectbackW1
#undef selectbackW2
#undef selectbackW3


/* insert code for the feature generators themselves */
#include "O3FeaturesDep.tcc"
#include "O3FeaturesSib.tcc"
#include "O3FeaturesGCI.tcc"
#include "O3FeaturesGCO.tcc"
#include "O3FeaturesGCIS.tcc"
#include "O3FeaturesGCOS.tcc"
#include "O3FeaturesSibs.tcc"


#undef W1
#undef W2
#undef W3
