#include "O3MMSSGD.h"
#include "O3Part.h"
#include "O3Ablation.h"

#include <stdio.h>
#include <stdlib.h>

#include <vector>
#include <sstream>
#include <iostream>

using std::cerr;
using std::endl;
using std::ostringstream;

/* lifecycle */
o3mmssgd::o3mmssgd()
  : o3trainer("MMSSGD"), _mistakes(0), _num_processed(0), _C(-1) {}
o3mmssgd::~o3mmssgd() {}


/* print usage information */
void o3mmssgd::usage_impl() const {
  cerr << "O3MMSSGD options:" << endl
       << " --C=<real>    : regularization constant" << endl;
}

/* initialize the trainer */
void o3mmssgd::init_impl(const o3trainer::strmap& opts) {
  /* recover regularization constant */
  o3trainer::strmap::const_iterator it;

  if((it = opts.find("C")) == opts.end()) {
    usage("please provide a regularization constant --C=...");
  }
  _C = strtod(it->second.c_str(), NULL);
  assert(_C >= 0);
}


/* training interface */
void o3mmssgd::train_start() { _num_processed = 0; }
void o3mmssgd::iter_start()  { _mistakes = 0; }
void o3mmssgd::iter_end(string& logline, string& repline) {
  ostringstream os;
  os << _mistakes;
  logline = os.str();
  os.str("");
  os << "mistakes = " << _mistakes;
  repline = os.str();
  /* renormalize the parameters */
  _F->Wlin().rescale();
}


/* process an example */
void o3mmssgd::process(const ex* const restrict e) {
  const int N = (int)e->heads.size();
  assert(_F->L() == _P->L());
  assert(!(_F->averaged()));

  assert(_ablation == -1);

  /* compute the true part-vector */
  o3part::vec rstar;
  o3part::parts(_P->L(), _P->multiroot(), _P->sibparser(),
                e->heads, e->labels, rstar);

  /* compute the parsed part-vector */
  o3part::vec rhat;
  { /* argmax label */
    vector<int> hhat, lhat;
    _P->argmax(N, _F->lin_scloss(e->words, e->tags, e->heads, e->labels),
              hhat, lhat);
    o3part::parts(_P->L(), _P->multiroot(), _P->sibparser(),
                  hhat, lhat, rhat);
  }

  /* pegasos-style learning rate eta */
  const double eta_reg = 1.0/(double)(1 + _num_processed);
  const double eta = eta_reg/_C;

  /* rescale the parameters (regularizer gradient) */
  _F->Wlin().scale(1.0 - eta_reg); /* don't divide and mul by _C */

  /* collect the part-wise contributions */
  o3part::bmap rplus;
  o3part::bmap rminus;

  /* set initial positive and negative contributions */
  for(uint32_t i = 0; i < rstar.size(); ++i) { rplus.ins(rstar[i], true); }
  for(uint32_t i = 0; i < rhat.size();  ++i) { rminus.ins(rhat[i], true); }

  /* eliminate negative and positive contributions */
  for(uint32_t i = 0; i < rstar.size(); ++i) { rminus.delq(rstar[i]); }
  for(uint32_t i = 0; i < rhat.size();  ++i) { rplus.delq(rhat[i]); }

  if(rplus.size() > 0 || rminus.size() > 0) {
    ++_mistakes; /* part difference indicates mistake */
    /* create a scorer that adds eta to each positive feature */
    o3features::flin_add& restrict adder = _F->lin_adder(e->words, e->tags);
    adder.alpha(eta);
    o3part::bmap::iterator it = rplus.begin();
    o3part::bmap::iterator iend = rplus.end();
    for(; it != iend; ++it) { it->key.score_with(adder); }
    /* subtract eta from negative features */
    adder.alpha(-eta);
    it = rminus.begin();
    iend = rminus.end();
    for(; it != iend; ++it) { it->key.score_with(adder); }
  }

  /* update #processed */
  ++_num_processed;
}


/* classify an example */
void o3mmssgd::classify(const ex* const restrict e,
                        vector<int>& restrict ret_h,
                        vector<int>& restrict ret_l) const {
  o3ablation::argmax(_ablation, *_P, (int)(e->words.size()),
                     _F->lin_scorer(e->words, e->tags),
                     ret_h, ret_l);
}
