/* -*-c++-*- */
#ifndef O3ORACLE_H
#define O3ORACLE_H
/* O3Oracle : A part-scorer that computes scores according to an
   "oracle" that returns 1 for dependencies matching a reference tree
   and 0 otherwise.  This is useful for, e.g., computing the maximum
   achievable score subject to pruning or non-projectivity. */

#include "Bits.h"
#include "Macros.h"

#include <assert.h>

#include <vector>

using std::vector;

class o3oracle {
public:
  /* create an oracle with the provided reference tree */
  o3oracle(const vector<int>& restrict heads,
           const vector<int>& restrict labels);
  ~o3oracle();

  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double sib1(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const;
  double sib2(const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_hm, const lab_t l_hs) const;
  double gci1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gci2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gco1(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gco2(const idx_t g, const idx_t h, const idx_t m,
              const lab_t l_gh, const lab_t l_hm) const;
  double gcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const;
  double gcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
              const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const;
  double sibs(const idx_t h, const idx_t m, const idx_t s, const idx_t c,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const;

private:
  /* helper functions for construction */
  static const idx_t* copyheads(const vector<int>& restrict v);
  static const lab_t* copylabels(const vector<int>& restrict v);

  /* disallow use of these methods */
  o3oracle(const o3oracle& that);
  o3oracle& operator=(const o3oracle& that);

  /* data members */
  const idx_t* const restrict _heads;  /* array of parents */
  const lab_t* const restrict _labels; /* array of labels */
};


/* inlined part scorers */
inline double
o3oracle::dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  /* return an indicator for the reference dependency */
  return (((h == _heads[m]) && (l_hm == _labels[m])) ? 1 : 0);
}
inline double
o3oracle::dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return dep1(h, m, l_hm);
}
inline double
o3oracle::dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return dep1(h, m, l_hm);
}

/* the dependency indicators are sufficient to find a tree that
   recovers as many reference deps as possible; return 0 for all other
   parts */
inline double
o3oracle::sib1(const idx_t h, const idx_t m, const idx_t s,
               const lab_t l_hm, const lab_t l_hs) const { return 0; }
inline double
o3oracle::sib2(const idx_t h, const idx_t m, const idx_t s,
               const lab_t l_hm, const lab_t l_hs) const { return 0; }
inline double
o3oracle::gci1(const idx_t g, const idx_t h, const idx_t m,
               const lab_t l_gh, const lab_t l_hm) const { return 0; }
inline double
o3oracle::gci2(const idx_t g, const idx_t h, const idx_t m,
               const lab_t l_gh, const lab_t l_hm) const { return 0; }
inline double
o3oracle::gco1(const idx_t g, const idx_t h, const idx_t m,
               const lab_t l_gh, const lab_t l_hm) const { return 0; }
inline double
o3oracle::gco2(const idx_t g, const idx_t h, const idx_t m,
               const lab_t l_gh, const lab_t l_hm) const { return 0; }
inline double
o3oracle::gcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
               const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const {
  return 0;
}
inline double
o3oracle::gcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
               const lab_t l_gh, const lab_t l_hm, const lab_t l_hs) const {
  return 0;
}
inline double
o3oracle::sibs(const idx_t h, const idx_t m, const idx_t s, const idx_t c,
              const lab_t l_hm, const lab_t l_hs, const lab_t l_hc) const {
  return 0;
}


/* unused methods */
inline o3oracle::o3oracle(const o3oracle& that)
  : _heads(NULL), _labels(NULL) { assert(0); }
inline o3oracle& o3oracle::operator=(const o3oracle& that) {
  assert(0); return *this;
}

#endif /* O3ORACLE_H */
