/* -*-c++-*- */
#ifndef O3PARAMS_H
#define O3PARAMS_H
/* O3Params : A sparse parameter vector with support for labeled
   parameters for features of all types.  The class is templated on a
   value type Val and a utility class ValUtil providing the following
   methods:

   void ValUtil::fprint(FILE* f, const Val& v)
   Val  ValUtil::fscan(FILE* f)
   bool ValUtil::equals(const Val& v1, const Val& v2, const double tol)

   The tol parameter for equals() is a tolerance for floating-point
   error in equality tests. */


#include "Hashtable.h"
#include "Feature.h"
#include "Bits.h"

#include <stdio.h>

#include <map>
#include <vector>
#include <string>

using std::map;
using std::vector;
using std::string;

template<class Val, class ValUtil>
class o3params {
public:
  /* nested classes */
  typedef hashtable<fbits,Val> primals;  /* sparse primals */

  /* access special label indices */
  lab_t Lnull() const; /* null label index */
  lab_t Lback() const; /* backed-off label index */
  /* count the number of nonzero params */
  uint64_t nnz() const;

  /* save to/load from a file */
  void save(FILE* const f) const;
  void load(FILE* const f);
  void add(const o3params& restrict that);

  /* clear all values; setting the sparse vector to 0 */
  void zero();

  /* clear zero values from the hashtables */
  void erase_zeros();

  /* equality testing */
  bool equals(const o3params& that, const double tol = 0) const;

  /* methods for parallel training */
  /* write a representation of this to the file descriptor, scaling
     the parameters before sending them */
  void parallel_send(const int write_fd, const double scale) const;
  /* deserialize a set of parameters from the file descriptor and
     combine it with this */
  void parallel_recv(const int read_fd);


  /* print statistics about hashtables to stderr */
  void print_stats() const;


private:
  /* helpers for save/load */
  void save_primals(FILE* const f, const primals& restrict W) const;
  void load_primals(FILE* const f, primals& restrict W, const uint32_t n);

  /* data members */
  const int _L;        /* number of labels */
  const int _Lstride;  /* constant for matrix indexation */
  const int _LLstride; /* constant for matrix indexation */
  const int _Lnull;    /* null label index (for, e.g., -1 sibling) */
  const int _Lback;    /* backed-off label index */


protected: /* subclasses can only access these data members */
  const int _nW;         /* total number of parameter spaces */
  primals* const _W_arr; /* pointer to the underlying array */

  o3params(const int L);
  ~o3params();


private:
  /* maps from "names" of each parameter-space to: (1) the array of
     primals implementing it, (2) a list of label-names */
  map<string,primals*>        _name2arr;
  map<string,vector<string> > _name2lab;

  /* access the appropriate underlying maps for left-headed and
     right-headed deps and grandparents.  the higher-order features
     sib/gci/gco/gcis/gcos/gcisw4/gcosw4 may have null labels; null
     labels should be passed as _L.  labels may also be eliminated for
     reasons of backoff; i.e., conditioning on only one label in the
     case of sibling features.  the index _L + 1 should be used to
     indicate a backed-off label (one which is never consulted) as
     opposed to a null label (one which is consulted but does not
     exist).

     we have to define a ton of accessors here so we'll use macros to
     make it less painful.  the C... macros optionally check indices
     and compute the proper lookup index for some label combination.
     the DEF... macros define four accessors at a time (left/right and
     const/non-const) and also define the underlying arrays. */

#define DECL1(type, l1)                         \
public:                                         \
  primals& l ## type(const int l1);             \
  primals& r ## type(const int l1);             \
  const primals& l ## type(const int l1) const; \
  const primals& r ## type(const int l1) const; \
private:                                        \
  primals* const _W_ ## l ## type;              \
  primals* const _W_ ## r ## type
#define DECL2(type, l1, l2)                                     \
public:                                                         \
  primals& l ## type(const int l1, const int l2);               \
  primals& r ## type(const int l1, const int l2);               \
  const primals& l ## type(const int l1, const int l2) const;   \
  const primals& r ## type(const int l1, const int l2) const;   \
private:                                                        \
  primals* const _W_ ## l ## type;                              \
  primals* const _W_ ## r ## type
#define DECL3(type, l1, l2, l3)                                         \
public:                                                                 \
  primals& l ## type(const int l1, const int l2, const int l3);         \
  primals& r ## type(const int l1, const int l2, const int l3);         \
  const primals& l ## type(const int l1, const int l2, const int l3) const; \
  const primals& r ## type(const int l1, const int l2, const int l3) const; \
private:                                                                \
  primals* const _W_ ## l ## type;                                      \
  primals* const _W_ ## r ## type

  /* single-label parts */
  DECL1(dep, l_hm);
  DECL1(dep1, l_hm);
  DECL1(dep1adj, l_hm);
  DECL1(dep2, l_hm);
  DECL1(dep3, l_hm);
  /* double-label parts */
  DECL2(sib1, l_hm, l_hs);
  DECL2(sib2, l_hm, l_hs);
  DECL2(sib1adj, l_hm, l_hs);
  DECL2(sib2adj, l_hm, l_hs);
  DECL2(gci1, l_gh, l_hm);
  DECL2(gci2, l_gh, l_hm);
  DECL2(gco1, l_gh, l_hm);
  DECL2(gco2, l_gh, l_hm);
  /* triple-label parts */
  DECL3(sibs, l_hm, l_hs, l_hc);
  DECL3(gcis, l_gh, l_hm, l_hs);
  DECL3(gcos, l_gh, l_hm, l_hs);
  DECL3(sibsw4, l_hm, l_hs, l_hc);
  DECL3(gcisw4, l_gh, l_hm, l_hs);
  DECL3(gcosw4, l_gh, l_hm, l_hs);
  DECL3(sibsb4, l_hm, l_hs, l_hc);
  DECL3(gcisb4, l_gh, l_hm, l_hs);
  DECL3(gcosb4, l_gh, l_hm, l_hs);
  DECL3(sibsr4, l_hm, l_hs, l_hc);
  DECL3(gcisr4, l_gh, l_hm, l_hs);
  DECL3(gcosr4, l_gh, l_hm, l_hs);
  DECL3(sibsnadj, l_hm, l_hs, l_hc);
  DECL3(gcisnadj, l_gh, l_hm, l_hs);
  DECL3(gcosnadj, l_gh, l_hm, l_hs);
  DECL3(sibsnadjw4, l_hm, l_hs, l_hc);
  DECL3(gcisnadjw4, l_gh, l_hm, l_hs);
  DECL3(gcosnadjw4, l_gh, l_hm, l_hs);
  DECL3(sibsnadjb4, l_hm, l_hs, l_hc);
  DECL3(gcisnadjb4, l_gh, l_hm, l_hs);
  DECL3(gcosnadjb4, l_gh, l_hm, l_hs);
  DECL3(sibsnadjr4, l_hm, l_hs, l_hc);
  DECL3(gcisnadjr4, l_gh, l_hm, l_hs);
  DECL3(gcosnadjr4, l_gh, l_hm, l_hs);
  DECL3(sibsadj3, l_hm, l_hs, l_hc);
  DECL3(gcisadj3, l_gh, l_hm, l_hs);
  DECL3(gcosadj3, l_gh, l_hm, l_hs);
  DECL3(sibsadj3w4, l_hm, l_hs, l_hc);
  DECL3(gcisadj3w4, l_gh, l_hm, l_hs);
  DECL3(gcosadj3w4, l_gh, l_hm, l_hs);
  DECL3(sibsadj3b4, l_hm, l_hs, l_hc);
  DECL3(gcisadj3b4, l_gh, l_hm, l_hs);
  DECL3(gcosadj3b4, l_gh, l_hm, l_hs);
  DECL3(sibsadj3r4, l_hm, l_hs, l_hc);
  DECL3(gcisadj3r4, l_gh, l_hm, l_hs);
  DECL3(gcosadj3r4, l_gh, l_hm, l_hs);
  DECL3(sibsadj4, l_hm, l_hs, l_hc);
  DECL3(gcisadj4, l_gh, l_hm, l_hs);
  DECL3(gcosadj4, l_gh, l_hm, l_hs);
  DECL3(sibsadj4w4, l_hm, l_hs, l_hc);
  DECL3(gcisadj4w4, l_gh, l_hm, l_hs);
  DECL3(gcosadj4w4, l_gh, l_hm, l_hs);
  DECL3(sibsadj4b4, l_hm, l_hs, l_hc);
  DECL3(gcisadj4b4, l_gh, l_hm, l_hs);
  DECL3(gcosadj4b4, l_gh, l_hm, l_hs);
  DECL3(sibsadj4r4, l_hm, l_hs, l_hc);
  DECL3(gcisadj4r4, l_gh, l_hm, l_hs);
  DECL3(gcosadj4r4, l_gh, l_hm, l_hs);

#undef DECL1
#undef DECL2
#undef DECL3
};



/*** inlined methods ***/

template<class Val, class ValUtil> inline lab_t
o3params<Val,ValUtil>::Lnull() const { return LCONV(_Lnull); }

template<class Val, class ValUtil> inline lab_t
o3params<Val,ValUtil>::Lback() const { return LCONV(_Lback); }


#ifndef EXPLICIT_TEMPLATES
#include "O3Params.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3PARAMS_H */
