#include "Macros.h"
#include "Timer.h"
#include "ParallelUtil.h"

#include <assert.h>
#include <string.h>
#include <math.h>

#include <iostream>
#include <iomanip>

using std::cerr;
using std::endl;
using std::flush;
using std::fixed;
using std::setprecision;


/* lifecycle */
template<class Val, class ValUtil>
o3params<Val,ValUtil>::o3params(const int L)
  : _L(L),
    /* the stride length for skipping a single labels, with room for
       possible backoff */
    _Lstride(_L == 1 ? _L + 1 : _L + 2),
    /* stride length for for skipping a pair of labels, one of which
       may be null, with room for backoff */
    _LLstride(_Lstride*(_Lstride - 1)),
    /* index of the null label */
    _Lnull(_Lstride - 1),
    /* index of the backed-off label */
    _Lback(_L == 1 ? -2147483647 : _L),
    /* allocate a single array of primal vectors and establish
       pointers into this single array. */
    _nW(2*(5*_Lnull + 8*_LLstride + 48*_LLstride*_Lnull)),
    _W_arr(new primals[_nW]),
    /* the arrays are declared with const pointers so they must be
       declared in an initializer list.  doing this while walking over
       _W_arr is a pain, so we use macros to ensure consistency. */
#define SIZE1 _Lnull
#define SIZE2 _LLstride
#define SIZE3 _LLstride*_Lnull
#define LAST1 _W_arr
#define LAST2 _W_rdep3 + SIZE1
#define LAST3 _W_rgco2 + SIZE2
#define INIT(type, num)                         \
    _W_l ## type(LAST ## num),                  \
    _W_r ## type(_W_l ## type + SIZE ## num)
#define NEXT(type, prevtype, num)                       \
    _W_l ## type(_W_r ## prevtype + SIZE ## num),       \
    _W_r ## type(_W_l ## type + SIZE ## num)

    INIT(dep,              1),
    NEXT(dep1,    dep,     1),
    NEXT(dep1adj, dep1,    1),
    NEXT(dep2,    dep1adj, 1),
    NEXT(dep3,    dep2,    1),

    INIT(sib1,             2),
    NEXT(sib2,    sib1,    2),
    NEXT(sib1adj, sib2,    2),
    NEXT(sib2adj, sib1adj, 2),
    NEXT(gci1,    sib2adj, 2),
    NEXT(gci2,    gci1,    2),
    NEXT(gco1,    gci2,    2),
    NEXT(gco2,    gco1,    2),

    INIT(sibs,                    3),
    NEXT(gcis,        sibs,       3),
    NEXT(gcos,        gcis,       3),
    NEXT(sibsw4,      gcos,       3),
    NEXT(gcisw4,      sibsw4,     3),
    NEXT(gcosw4,      gcisw4,     3),
    NEXT(sibsb4,      gcosw4,     3),
    NEXT(gcisb4,      sibsb4,     3),
    NEXT(gcosb4,      gcisb4,     3),
    NEXT(sibsr4,      gcosb4,     3),
    NEXT(gcisr4,      sibsr4,     3),
    NEXT(gcosr4,      gcisr4,     3),
    NEXT(sibsnadj,    gcosr4,     3),
    NEXT(gcisnadj,    sibsnadj,   3),
    NEXT(gcosnadj,    gcisnadj,   3),
    NEXT(sibsnadjw4,  gcosnadj,   3),
    NEXT(gcisnadjw4,  sibsnadjw4, 3),
    NEXT(gcosnadjw4,  gcisnadjw4, 3),
    NEXT(sibsnadjb4,  gcosnadjw4, 3),
    NEXT(gcisnadjb4,  sibsnadjb4, 3),
    NEXT(gcosnadjb4,  gcisnadjb4, 3),
    NEXT(sibsnadjr4,  gcosnadjb4, 3),
    NEXT(gcisnadjr4,  sibsnadjr4, 3),
    NEXT(gcosnadjr4,  gcisnadjr4, 3),
    NEXT(sibsadj3,    gcosnadjr4, 3),
    NEXT(gcisadj3,    sibsadj3,   3),
    NEXT(gcosadj3,    gcisadj3,   3),
    NEXT(sibsadj3w4,  gcosadj3,   3),
    NEXT(gcisadj3w4,  sibsadj3w4, 3),
    NEXT(gcosadj3w4,  gcisadj3w4, 3),
    NEXT(sibsadj3b4,  gcosadj3w4, 3),
    NEXT(gcisadj3b4,  sibsadj3b4, 3),
    NEXT(gcosadj3b4,  gcisadj3b4, 3),
    NEXT(sibsadj3r4,  gcosadj3b4, 3),
    NEXT(gcisadj3r4,  sibsadj3r4, 3),
    NEXT(gcosadj3r4,  gcisadj3r4, 3),
    NEXT(sibsadj4,    gcosadj3r4, 3),
    NEXT(gcisadj4,    sibsadj4,   3),
    NEXT(gcosadj4,    gcisadj4,   3),
    NEXT(sibsadj4w4,  gcosadj4,   3),
    NEXT(gcisadj4w4,  sibsadj4w4, 3),
    NEXT(gcosadj4w4,  gcisadj4w4, 3),
    NEXT(sibsadj4b4,  gcosadj4w4, 3),
    NEXT(gcisadj4b4,  sibsadj4b4, 3),
    NEXT(gcosadj4b4,  gcisadj4b4, 3),
    NEXT(sibsadj4r4,  gcosadj4b4, 3),
    NEXT(gcisadj4r4,  sibsadj4r4, 3),
    NEXT(gcosadj4r4,  gcisadj4r4, 3) {

#undef SIZE1
#undef SIZE2
#undef SIZE3
#undef LAST1
#undef LAST2
#undef LAST3
#undef INIT
#undef NEXT

  assert(_L > 0);
  assert(_W_arr != NULL);
  const long int numclaimed = (_W_rgcosadj4r4 + _LLstride*_Lnull) - _W_arr;
  assert(numclaimed == _nW);

  /* populate the maps with named arrays and ranges */
#define MAP1(type, l1) {                        \
    _name2arr[STRCAT(l, type)] = _W_l ## type;  \
    _name2arr[STRCAT(r, type)] = _W_r ## type;  \
    vector<string> tmp;                         \
    tmp.push_back(#l1);                         \
    _name2lab[STRCAT(l, type)] = tmp;           \
    _name2lab[STRCAT(r, type)] = tmp;           \
  }
#define MAP2(type, l1, l2) {                    \
    _name2arr[STRCAT(l, type)] = _W_l ## type;  \
    _name2arr[STRCAT(r, type)] = _W_r ## type;  \
    vector<string> tmp;                         \
    tmp.push_back(#l1);                         \
    tmp.push_back(#l2);                         \
    _name2lab[STRCAT(l, type)] = tmp;           \
    _name2lab[STRCAT(r, type)] = tmp;           \
  }
#define MAP3(type, l1, l2, l3) {                \
    _name2arr[STRCAT(l, type)] = _W_l ## type;  \
    _name2arr[STRCAT(r, type)] = _W_r ## type;  \
    vector<string> tmp;                         \
    tmp.push_back(#l1);                         \
    tmp.push_back(#l2);                         \
    tmp.push_back(#l3);                         \
    _name2lab[STRCAT(l, type)] = tmp;           \
    _name2lab[STRCAT(r, type)] = tmp;           \
  }
  /* two levels are needed to ensure that token paste occurs before
     stringification */
#define STRCAT(a, b) STRIFY(a ## b)
#define STRIFY(a) #a

  /* single-label parts */
  MAP1(dep, l_hm);
  MAP1(dep1, l_hm);
  MAP1(dep1adj, l_hm);
  MAP1(dep2, l_hm);
  MAP1(dep3, l_hm);
  /* double-label parts */
  MAP2(sib1, l_hm, l_hs);
  MAP2(sib2, l_hm, l_hs);
  MAP2(sib1adj, l_hm, l_hs);
  MAP2(sib2adj, l_hm, l_hs);
  MAP2(gci1, l_gh, l_hm);
  MAP2(gci2, l_gh, l_hm);
  MAP2(gco1, l_gh, l_hm);
  MAP2(gco2, l_gh, l_hm);
  /* triple-label parts */
  MAP3(sibs, l_hm, l_hs, l_hc);
  MAP3(gcis, l_gh, l_hm, l_hs);
  MAP3(gcos, l_gh, l_hm, l_hs);
  MAP3(sibsw4, l_hm, l_hs, l_hc);
  MAP3(gcisw4, l_gh, l_hm, l_hs);
  MAP3(gcosw4, l_gh, l_hm, l_hs);
  MAP3(sibsb4, l_hm, l_hs, l_hc);
  MAP3(gcisb4, l_gh, l_hm, l_hs);
  MAP3(gcosb4, l_gh, l_hm, l_hs);
  MAP3(sibsr4, l_hm, l_hs, l_hc);
  MAP3(gcisr4, l_gh, l_hm, l_hs);
  MAP3(gcosr4, l_gh, l_hm, l_hs);
  MAP3(sibsnadj, l_hm, l_hs, l_hc);
  MAP3(gcisnadj, l_gh, l_hm, l_hs);
  MAP3(gcosnadj, l_gh, l_hm, l_hs);
  MAP3(sibsnadjw4, l_hm, l_hs, l_hc);
  MAP3(gcisnadjw4, l_gh, l_hm, l_hs);
  MAP3(gcosnadjw4, l_gh, l_hm, l_hs);
  MAP3(sibsnadjb4, l_hm, l_hs, l_hc);
  MAP3(gcisnadjb4, l_gh, l_hm, l_hs);
  MAP3(gcosnadjb4, l_gh, l_hm, l_hs);
  MAP3(sibsnadjr4, l_hm, l_hs, l_hc);
  MAP3(gcisnadjr4, l_gh, l_hm, l_hs);
  MAP3(gcosnadjr4, l_gh, l_hm, l_hs);
  MAP3(sibsadj3, l_hm, l_hs, l_hc);
  MAP3(gcisadj3, l_gh, l_hm, l_hs);
  MAP3(gcosadj3, l_gh, l_hm, l_hs);
  MAP3(sibsadj3w4, l_hm, l_hs, l_hc);
  MAP3(gcisadj3w4, l_gh, l_hm, l_hs);
  MAP3(gcosadj3w4, l_gh, l_hm, l_hs);
  MAP3(sibsadj3b4, l_hm, l_hs, l_hc);
  MAP3(gcisadj3b4, l_gh, l_hm, l_hs);
  MAP3(gcosadj3b4, l_gh, l_hm, l_hs);
  MAP3(sibsadj3r4, l_hm, l_hs, l_hc);
  MAP3(gcisadj3r4, l_gh, l_hm, l_hs);
  MAP3(gcosadj3r4, l_gh, l_hm, l_hs);
  MAP3(sibsadj4, l_hm, l_hs, l_hc);
  MAP3(gcisadj4, l_gh, l_hm, l_hs);
  MAP3(gcosadj4, l_gh, l_hm, l_hs);
  MAP3(sibsadj4w4, l_hm, l_hs, l_hc);
  MAP3(gcisadj4w4, l_gh, l_hm, l_hs);
  MAP3(gcosadj4w4, l_gh, l_hm, l_hs);
  MAP3(sibsadj4b4, l_hm, l_hs, l_hc);
  MAP3(gcisadj4b4, l_gh, l_hm, l_hs);
  MAP3(gcosadj4b4, l_gh, l_hm, l_hs);
  MAP3(sibsadj4r4, l_hm, l_hs, l_hc);
  MAP3(gcisadj4r4, l_gh, l_hm, l_hs);
  MAP3(gcosadj4r4, l_gh, l_hm, l_hs);

#undef MAP1
#undef MAP2
#undef MAP3
#undef STRCAT
#undef STRIFY
}

template<class Val, class ValUtil>
o3params<Val,ValUtil>::~o3params() { delete [] _W_arr; }



#define ENC_BACK(l) (l == _Lback ? -2 : l)
#define ENC_NULL(l) (l == _Lback ? -2 : (l == _Lnull ? -1 : l))

/* save to file */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
save(FILE* const f) const {
  timer_start();
  /* metadata */
  fprintf(f, "L=%d\n", _L);
  fprintf(f, "type=gch\n");

  assert(_name2arr.size() == _name2lab.size());
  typename map<string,primals*>::const_iterator it = _name2arr.begin();
  const typename map<string,primals*>::const_iterator iend = _name2arr.end();
  typename map<string,vector<string> >::const_iterator jt = _name2lab.begin();

  int cnt = 0;
  const int repint_tmp = _nW/20;
  const int repint = (repint_tmp == 0 ? 1 : repint_tmp);

  for(; it != iend; ++it, ++jt) {
    const string& name = it->first;
    assert(name == jt->first);
    const primals* const W = it->second;
    const vector<string> lnames = jt->second;
    assert(lnames.size() >= 1); assert(lnames.size() <= 3);
    if(lnames.size() == 1) {
      for(int l1 = 0; l1 < _Lnull; ++l1) {
        const primals& w = W[l1];
        fprintf(f, "%s n=%u %s=%d\n",
                name.c_str(), (uint32_t)w.size(),
                lnames[0].c_str(), ENC_BACK(l1));
        save_primals(f, w);
      }
    } else if(lnames.size() == 2) {
      for(int l1 = 0; l1 < _Lnull; ++l1) {
        for(int l2 = 0; l2 <= _Lnull; ++l2) {
          const primals& w = W[l1*_Lstride + l2];
          fprintf(f, "%s n=%u %s=%d %s=%d\n",
                  name.c_str(), (uint32_t)w.size(),
                  lnames[0].c_str(), ENC_BACK(l1),
                  lnames[1].c_str(), ENC_NULL(l2));
          save_primals(f, w);
        }
      }
    } else if(lnames.size() == 3) {
      for(int l1 = 0; l1 < _Lnull; ++l1) {
        for(int l2 = 0; l2 < _Lnull; ++l2) {
          for(int l3 = 0; l3 <= _Lnull; ++l3) {
            const primals& w = W[l1*_LLstride + l2*_Lstride + l3];
            fprintf(f, "%s n=%u %s=%d %s=%d %s=%d\n",
                    name.c_str(), (uint32_t)w.size(),
                    lnames[0].c_str(), ENC_BACK(l1),
                    lnames[1].c_str(), ENC_BACK(l2),
                    lnames[2].c_str(), ENC_NULL(l3));
            save_primals(f, w);
          }
        }
      }
    } else { assert(0); }
    /* tick to indicate progress */
    if(((++cnt) % repint) == 0) { cerr << "." << flush; }
  }
  timer_stop();
  cerr << " done [" << timestr(timediff) << "]" << endl;
  print_stats();
}

#undef ENC_BACK
#undef ENC_NULL

#define DEC_BACK(l) {                           \
    l = (l == -2 ? _Lback : l);                 \
    assert(l >= 0); assert(l < _Lnull);         \
  }
#define DEC_NULL(l) {                                   \
    l = (l == -2 ? _Lback : (l == -1 ? _Lnull : l));    \
    assert(l >= 0); assert(l <= _Lnull);                \
  }

/* load from file */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
load(FILE* const f) {
  timer_start();
  { /* metadata */
    int L = -1;
    fscanf(f, " L=%d", &L);
    assert(L >= 0);
    if(L != _L) {
      cerr << "O3Params : Label mismatch when loading: "
           << _L << " (this) vs " << L << " (file)" << endl;
      assert(0);
    }
    char buf[256];
    buf[0] = '\0';
    fscanf(f, " type=%255s", buf);
    const string mytype("gch");
    assert(mytype == buf);
  }
  /* clear the parameter vectors */
  for(int i = 0; i < _nW; ++i) { _W_arr[i].clear(); }
  /* flags for whether vectors have been loaded */
  bool* const loaded = new bool[_nW];
  for(int i = 0; i < _nW; ++i) { loaded[i] = false; }
  const int repint_tmp = _nW/20;
  const int repint = (repint_tmp == 0 ? 1 : repint_tmp);
  /* read each parameter vector in any order */
  char buf[256];
  for(int i = 0; i < _nW; ++i) {
    buf[0] = '\0';
    uint32_t n = 0;
    int nscanned = fscanf(f, " %255s n=%u", buf, &n);
    assert(nscanned == 2);
    assert(_name2arr.find(buf) != _name2arr.end());
    assert(_name2lab.find(buf) != _name2lab.end());
    primals* const W             = _name2arr.find(buf)->second;
    const vector<string>& lnames = _name2lab.find(buf)->second;
    assert(lnames.size() >= 1); assert(lnames.size() <= 3);
    /* recover all label-indices */
    int lidx[3];
    for(unsigned j = 0; j < lnames.size(); ++j) {
      lidx[j] = -10000;
      buf[0] = '\0';
      nscanned = fscanf(f, " %4c=%d", buf, lidx + j);
      buf[4] = '\0';
      assert(lnames[j] == buf);
    }
    DEC_BACK(lidx[0]);
    int w_idx = -10000;
    if(lnames.size() == 1) {
      w_idx = lidx[0];
    } else if(lnames.size() == 2) {
      DEC_NULL(lidx[1]);
      w_idx = lidx[0]*_Lstride + lidx[1];
    } else if(lnames.size() == 3) {
      DEC_BACK(lidx[1]); DEC_NULL(lidx[2]);
      w_idx = lidx[0]*_LLstride + lidx[1]*_Lstride + lidx[2];
    } else { assert(0); }
    primals& w = W[w_idx];
    /* load the primal vector */
    const long  int idx = (&w - _W_arr);
    assert(idx >= 0); assert(idx < _nW);
    assert(!loaded[idx]); loaded[idx] = true;
    load_primals(f, w, n);
    /* tick to indicate progress */
    if(((i + 1) % repint) == 0) { cerr << "." << flush; }
  }
  /* check that all vectors are loaded */
  for(int i = 0; i < _nW; ++i) { assert(loaded[i]); }
  delete [] loaded;
  timer_stop();
  cerr << " done [" << timestr(timediff) << "]" << endl;
  print_stats();
}

#undef DEC_BACK
#undef DEC_NULL


/* add another parameter vector to this */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
add(const o3params& restrict that) {
  for(int i = 0; i < _nW; ++i) {
    primals& restrict thisW = _W_arr[i];
    const primals& restrict thatW = that._W_arr[i];
    typename primals::const_iterator it = thatW.cbegin();
    const typename primals::const_iterator iend = thatW.cend();
    for(; it != iend; ++it) {
      if(thisW.has(it->key)) {
        ValUtil::add(thisW.mget(it->key), it->val);
      } else {
        thisW.ins(it->key, it->val);
      }
    }
  }
}



/* count the number of active params */
template<class Val, class ValUtil> uint64_t o3params<Val,ValUtil>::
nnz() const {
  uint64_t ret = 0;
  for(int i = 0; i < _nW; ++i) { ret += _W_arr[i].size(); }
  return ret;
}


/* clear all values from the hashtables */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
zero() {
  for(int i = 0; i < _nW; ++i) {
    primals& W = _W_arr[i];
    if(W.size() > 0) { W.clear(); }
  }
}


/* clear zero values from the hashtables */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
erase_zeros() {
  timer_start();
  cerr << "O3Params : removing zeros..." << flush;
  const Val z = Val();
  uint64_t num_del = 0;
  for(int i = 0; i < _nW; ++i) { num_del += _W_arr[i].erasev(z); }
  timer_stop();
  cerr << " " << num_del << " [" << timestr(timediff) << "]" << endl;
}



/* methods for parallel training */
/* serialize this and send it */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
parallel_send(const int write_fd, const double scale) const {
  for(int i = 0; i < _nW; ++i) {
    const primals& restrict W = _W_arr[i];
    /* write #elts */
    const uint64_t N = W.size();
    safe_write(write_fd, &N, sizeof(uint64_t));
    /* write mapping */
    fbits tmpf;
    typename primals::const_iterator it = W.cbegin();
    const typename primals::const_iterator iend = W.cend();
    for(; it != iend; ++it) {
      const fbits f = it->key;
      safe_write(write_fd, &f, sizeof(fbits));
      ValUtil::send(it->val, write_fd, scale);
    }
  }
}

/* deserialize parameters and merge */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
parallel_recv(const int read_fd) {
  uint64_t N = 0;
  fbits f = 0;
  for(int i = 0; i < _nW; ++i) {
    primals& restrict W = _W_arr[i];
    safe_read(read_fd, &N, sizeof(uint64_t));
    for(uint64_t j = 0; j < N; ++j) {
      safe_read(read_fd, &f, sizeof(fbits));
      /* get the current value (or insert one if necessary) */
      Val& restrict v = W.mget(f, Val());
      /* merge it with the value from the stream */
      ValUtil::recv(v, read_fd);
    }
  }
}





/* compute and report collision statistics */
template<class Val, class ValUtil> void o3params<Val,ValUtil>::
print_stats() const {
  uint64_t sumc = 0, maxc = 0, occup = 0, buckets = 0, elements = 0;
  for(int i = 0; i < _nW; ++i) {
    const primals& W = _W_arr[i];
    uint64_t tmp_sumc = 0, tmp_maxc = 0, tmp_active = 0;
    W.collision_stats(tmp_sumc, tmp_maxc, tmp_active);
    if(tmp_maxc > maxc) { maxc = tmp_maxc; }
    sumc += tmp_sumc; occup += tmp_active;
    buckets += W.buckets(); elements += W.size();
  }
  cerr << "O3Params : N = " << elements
       << " ; buckets = " << buckets
       << " ; active = " << occup << endl;
  cerr << fixed << setprecision(2);
  cerr << "O3Params"
       << " : col. = " << (100.0*((double)sumc)/((double)elements))
       << " ; occ. = " << (100.0*((double)occup)/((double)buckets))
       << " ; sng. = " << (100.0*((double)occup)/((double)elements))
       << " ; max = " << (maxc + 1) << endl;
}


template<class Val, class ValUtil> void o3params<Val,ValUtil>::
save_primals(FILE* const f, const primals& restrict W) const {
  typename primals::const_iterator it = W.cbegin();
  const typename primals::const_iterator iend = W.cend();
  for(; it != iend; ++it) {
    const fbits b = it->key;
    const uint32_t high = (uint32_t)(b >> 32);
    const uint32_t low = (uint32_t)(b & 0xffffffff);
    fprintf(f, "%08x%08x ", high, low);
    ValUtil::fprint(f, it->val);
    fprintf(f, "\n");
  }
  fprintf(f, "\n");
}

template<class Val, class ValUtil> void o3params<Val,ValUtil>::
load_primals(FILE* const f, primals& restrict W, const uint32_t n) {
  for(uint32_t i = 0; i < n; ++i) {
    uint32_t high = 0, low = 0;
    const int nscanned = fscanf(f, " %8x%8x", &high, &low);
    assert(nscanned == 2);
    const fbits b = ((((fbits)high) << 32) | ((fbits)low));
    W.ins(b, ValUtil::fscan(f));
  }
}


template<class Val, class ValUtil> bool o3params<Val,ValUtil>::
equals(const o3params& that, const double tol) const {
  if(_L != that._L) { return false; }
  assert(_nW == that._nW);
  for(int i = 0; i < _nW; ++i) {
    const primals& w1 = _W_arr[i];
    const primals& w2 = that._W_arr[i];
    if(w1.size() != w2.size()) { return false; }
    typename primals::const_iterator i1 = w1.cbegin();
    typename primals::const_iterator i2 = w2.cbegin();
    const typename primals::const_iterator e1 = w1.cend();
    const typename primals::const_iterator e2 = w2.cend();
    /* test equality of key set first */
    for(; i1 != e1; ++i1) { if(!w2.has(i1->key)) { return false; } }
    for(; i2 != e2; ++i2) { if(!w1.has(i2->key)) { return false; } }
    /* test equality of values */
    for(i1 = w1.cbegin(); i1 != e1; ++i1) {
      const Val& v1 = i1->val;
      const Val& v2 = w2.get(i1->key);
      if(!ValUtil::equals(v1, v2, tol)) { return false; }
    }
  }
  return true;
}



/* index checking */
#ifdef CAUTIOUS
#define CHK1(l) (assert(l >= 0), assert(l < _Lnull),  l)
#define CHK2(l) (assert(l >= 0), assert(l <= _Lnull), l)
#else /* CAUTIOUS */
#define CHK1(l) l
#define CHK2(l) l
#endif /* CAUTIOUS */

#define C1(l1)         (CHK1(l1))
#define C2(l1, l2)     (CHK1(l1)*_Lstride  + CHK2(l2))
#define C3(l1, l2, l3) (CHK1(l1)*_LLstride + CHK1(l2)*_Lstride + CHK2(l3))

#define PREAMB                                  \
  template<class Val, class ValUtil>            \
  typename o3params<Val,ValUtil>::primals&      \
  o3params<Val,ValUtil>
#define CPREAMB                                         \
  template<class Val, class ValUtil>                    \
  const typename o3params<Val,ValUtil>::primals&        \
  o3params<Val,ValUtil>

#define DEF1(type, l1)                          \
  PREAMB::l ## type(const int l1)               \
  { return _W_l ## type[C1(l1)]; }              \
  PREAMB::r ## type(const int l1)               \
  { return _W_r ## type[C1(l1)]; }              \
  CPREAMB::l ## type(const int l1) const        \
  { return _W_l ## type[C1(l1)]; }              \
  CPREAMB::r ## type(const int l1) const        \
  { return _W_r ## type[C1(l1)]; }
#define DEF2(type, l1, l2)                              \
  PREAMB::l ## type(const int l1, const int l2)         \
  { return _W_l ## type[C2(l1, l2)]; }                  \
  PREAMB::r ## type(const int l1, const int l2)         \
  { return _W_r ## type[C2(l1, l2)]; }                  \
  CPREAMB::l ## type(const int l1, const int l2) const  \
  { return _W_l ## type[C2(l1, l2)]; }                  \
  CPREAMB::r ## type(const int l1, const int l2) const  \
  { return _W_r ## type[C2(l1, l2)]; }
#define DEF3(type, l1, l2, l3)                                          \
  PREAMB::l ## type(const int l1, const int l2, const int l3)           \
  { return _W_l ## type[C3(l1, l2, l3)]; }                              \
  PREAMB::r ## type(const int l1, const int l2, const int l3)           \
  { return _W_r ## type[C3(l1, l2, l3)]; }                              \
  CPREAMB::l ## type(const int l1, const int l2, const int l3) const    \
  { return _W_l ## type[C3(l1, l2, l3)]; }                              \
  CPREAMB::r ## type(const int l1, const int l2, const int l3) const    \
  { return _W_r ## type[C3(l1, l2, l3)]; }

/* single-label parts */
DEF1(dep, l_hm)
DEF1(dep1, l_hm)
DEF1(dep1adj, l_hm)
DEF1(dep2, l_hm)
DEF1(dep3, l_hm)
/* double-label parts */
DEF2(sib1, l_hm, l_hs)
DEF2(sib2, l_hm, l_hs)
DEF2(sib1adj, l_hm, l_hs)
DEF2(sib2adj, l_hm, l_hs)
DEF2(gci1, l_gh, l_hm)
DEF2(gci2, l_gh, l_hm)
DEF2(gco1, l_gh, l_hm)
DEF2(gco2, l_gh, l_hm)
/* triple-label parts */
DEF3(sibs, l_hm, l_hs, l_hc)
DEF3(gcis, l_gh, l_hm, l_hs)
DEF3(gcos, l_gh, l_hm, l_hs)
DEF3(sibsw4, l_hm, l_hs, l_hc)
DEF3(gcisw4, l_gh, l_hm, l_hs)
DEF3(gcosw4, l_gh, l_hm, l_hs)
DEF3(sibsb4, l_hm, l_hs, l_hc)
DEF3(gcisb4, l_gh, l_hm, l_hs)
DEF3(gcosb4, l_gh, l_hm, l_hs)
DEF3(sibsr4, l_hm, l_hs, l_hc)
DEF3(gcisr4, l_gh, l_hm, l_hs)
DEF3(gcosr4, l_gh, l_hm, l_hs)
DEF3(sibsnadj, l_hm, l_hs, l_hc)
DEF3(gcisnadj, l_gh, l_hm, l_hs)
DEF3(gcosnadj, l_gh, l_hm, l_hs)
DEF3(sibsnadjw4, l_hm, l_hs, l_hc)
DEF3(gcisnadjw4, l_gh, l_hm, l_hs)
DEF3(gcosnadjw4, l_gh, l_hm, l_hs)
DEF3(sibsnadjb4, l_hm, l_hs, l_hc)
DEF3(gcisnadjb4, l_gh, l_hm, l_hs)
DEF3(gcosnadjb4, l_gh, l_hm, l_hs)
DEF3(sibsnadjr4, l_hm, l_hs, l_hc)
DEF3(gcisnadjr4, l_gh, l_hm, l_hs)
DEF3(gcosnadjr4, l_gh, l_hm, l_hs)
DEF3(sibsadj3, l_hm, l_hs, l_hc)
DEF3(gcisadj3, l_gh, l_hm, l_hs)
DEF3(gcosadj3, l_gh, l_hm, l_hs)
DEF3(sibsadj3w4, l_hm, l_hs, l_hc)
DEF3(gcisadj3w4, l_gh, l_hm, l_hs)
DEF3(gcosadj3w4, l_gh, l_hm, l_hs)
DEF3(sibsadj3b4, l_hm, l_hs, l_hc)
DEF3(gcisadj3b4, l_gh, l_hm, l_hs)
DEF3(gcosadj3b4, l_gh, l_hm, l_hs)
DEF3(sibsadj3r4, l_hm, l_hs, l_hc)
DEF3(gcisadj3r4, l_gh, l_hm, l_hs)
DEF3(gcosadj3r4, l_gh, l_hm, l_hs)
DEF3(sibsadj4, l_hm, l_hs, l_hc)
DEF3(gcisadj4, l_gh, l_hm, l_hs)
DEF3(gcosadj4, l_gh, l_hm, l_hs)
DEF3(sibsadj4w4, l_hm, l_hs, l_hc)
DEF3(gcisadj4w4, l_gh, l_hm, l_hs)
DEF3(gcosadj4w4, l_gh, l_hm, l_hs)
DEF3(sibsadj4b4, l_hm, l_hs, l_hc)
DEF3(gcisadj4b4, l_gh, l_hm, l_hs)
DEF3(gcosadj4b4, l_gh, l_hm, l_hs)
DEF3(sibsadj4r4, l_hm, l_hs, l_hc)
DEF3(gcisadj4r4, l_gh, l_hm, l_hs)
DEF3(gcosadj4r4, l_gh, l_hm, l_hs)

#undef PREAMB
#undef CPREAMB

#undef DEF1
#undef DEF2
#undef DEF3

#undef C1
#undef C2
#undef C3

#undef CHK1
#undef CHK2
