/* -*-c++-*- */
#ifndef O3PARSER_H
#define O3PARSER_H
/* O3Parser : A third-order dependency parser.  The parsing routines
   are templated on a PartScorer object, which provides scores for
   "parts" like dependencies, sibling trigrams, etc.  A PartScorer S
   should satisfy the following interface:

     S.dep(h, m, l_hm) = score of dependency h --> m with label l_hm.
         as a special case, h may be set to -1 to indicate a root
         dependency.  in this case the label l_hm should *not* be
         ignored.

     S.sib(h, m, s, l_hm, l_hs) = score of sibling trigram with head
         h, outer modifier m, inner modifier s, and labels l_hm and
         l_hs for the dependencies h --> m and h --> s respectively.
         here outer/inner means that either h < s < m or h > s > m, so
         that s is closer to h than m.  as a special case, s may be
         set to -1 to indicate the event that m is the innermost
         modifier of h; in this case the value of l_hs should be
         ignored.

     S.gci(h, m, c, l_hm, l_mc) = score of inner-grandchild trigram
         with head h, modifier m, grandchild c, and labels l_hm and
         l_mc for the dependencies h --> m and m --> c respectively.
         this scoring function is called when the trigram is arranged
         so that dependency h --> m is in the opposite direction from
         m --> c; i.e., h < c < m or h > c > m.  as a special case c
         may be set to -1 to indicate the event that m has no
         grandchild between h and m; in this case l_mc should be
         ignored.

     S.gco(h, m, c, l_hm, l_mc) = score of outer-grandchild trigram
         with head h, modifier m, grandchild c, and labels l_hm and
         l_mc as above.  the difference is that in this case the
         dependencies h --> m and m --> c point in the same direction;
         i.e., h < m < c or h > m > c.  as a special case c may be set
         to -1 to indicate that m has no grandchild in this direction.

     S.gcis(h, m, c, s, l_hm, l_mc, l_ms) = score of
         inner-grandchild-sibling 4-gram with head h, modifier m,
         grandchild c, and inner-sibling-of-grandchild s; l_hm, l_mc,
         and l_ms are the relevant labels.  the 4-gram is arranged so
         that c is the inner grandchild of m and s is the inner
         sibling of c; i.e., h < c < s < m or h > c > s > m.  as a
         special case, s may be set to -1 to indicate that c has no
         inner sibling; l_ms should be ignored in this case.

     S.gcos(h, m, c, s, l_hm, l_mc, l_ms) = score of
         outer-grandchild-sibling 4-gram with head h, modifier m,
         grandchild c, and inner-sibling-of-grandchild s; l_hm, l_mc,
         and l_ms are the relevant labels.  the 4-gram is arranged so
         that c is the outer grandchild of m and s is the inner
         sibling of c; i.e., h < m < s < c or h > m > s > c.  as a
         special case, s may be set to -1 to indicate that c has no
         inner sibling; l_ms should be ignored in this case.

   The PartScorer is intended to abstract away the issue of how the
   scores arise (i.e., a generative model's log-probability tables,
   inner products of a linear model, a precomputed table).  The
   PartScorer should ideally use short inlined methods to implement
   the interface above.  See "PartTest.h" for an example of what a
   part-scorer might look like. */


#include "Macros.h"
#include "O3Active.h"
#include "O3Chart.h"
#include "O3Part.h"

#include <vector>
#include <string>

using std::vector;
using std::string;

class o3parser {
public:
  o3parser(const int inL,
           const bool inmultiroot,
           const bool inpruned,
           const bool intopdown,
           const bool insibparser);
  ~o3parser();

  /* parse a N-word sentence using the given part-scorer, returning
     the score of the best parse and filling the head/label lists with
     the best parse as follows:

     ret_heads[m]  = head of m in the best parse, -1 being root
     ret_labels[m] = label of (ret_heads[m] --> m) in the best parse

     the "best" parse is the highest-scored tree, where the score of a
     tree is the sum of scores of its parts. */
  template<class PartScorer>
  double argmax(const int N, const PartScorer& restrict S,
                vector<int>& restrict ret_heads,
                vector<int>& restrict ret_labels);

  /* perform k-best parsing on a N-word sentence using the given
     part-scorer.

     kargmax_init()/kargmax_done() perform the necessary setup and
     teardown for k-best parsing; these should be called before and
     after each k-best parsing run.

     kargmax_next() retrieves the next-best parse.  the first call
     returns the best parse (the one ordinarily returned by argmax())
     and subsequent calls return parses in order of decreasing score.

     the Kmax value passed to kargmax_init() is a limit on the number
     of k-best trees that kargmax_next() will search for.

     possibility that all possible trees have been enumerated.  the
     return value of kargmax_next() should be examined in order to
     determine whether this has happened.  if a valid next-best parse
     was found, the function returns the rank of the parse (0-origin,
     so the argmax parse has rank 0), otherwise the function returns
     -1 and the caller should call kargmax_done() and move onto
     another example.

     the implementation of k-best parsing follows the efficient lazy
     k-best algorithm of Huang and Chiang (IWPT 2005).  retrieval of
     the first parse has the same complexity as normal argmax()
     parsing and subsequent parses cost closs to linear time.  the
     memory costs, may be significantly higher than the normal parser.
     parses that have equal score will be returned in an arbitrary
     order.  in addition, due to numerical errors it is possible,
     though unlikely, for parses to be returned out-of-order if their
     scores are very close. */
  void kargmax_init(const int N, const int Kmax = 64);
  void kargmax_done();
  template<class PartScorer>
  int  kargmax_next(const PartScorer& restrict S,
                    double& restrict ret_score,
                    vector<int>& restrict ret_heads,
                    vector<int>& restrict ret_labels);

  /* compute the log-partition function of the given sentence.
     specifically, assuming that the part-scores are log-weights,
     return:

     log[\sum_{tree y} \prod_{part p in y} exp(S(p))]

     for example, if the part-scores are inner products from a linear
     model, then partition() computes the normalization constant of a
     log-linear parser, or if the part-scores are log-probabilities of
     a generative model, then partition() computes the sentence
     probability */
  template<class PartScorer>
  double partition(const int N, const PartScorer& restrict S);

  /* compute the part-wise conditional marginals of the given
     sentence.  specifically, assuming that the part-scores are
     log-weights, compute

     marg(part p) = \sum_{tree y} (1/Z)\prod_{part q in y} exp(S(q))

     where Z is the partition function (i.e., exp(partition())).  due
     to the operation of the inside-outside algorithm, the marginals
     are computed incrementally rather than all at once.  the
     marginals are returned in M, indexed by o3part objects.  the
     o3part.score_with() function should be used to access the
     marginals. */
  template<class PartScorer>
  double marginals(const int N, const PartScorer& restrict S,
                   o3part::dmap& restrict M);


  /* the parser can have heavy time and space requirements, especially
     when parsing with labels.  the parser can therefore be created
     with pruning activated (see the constructor), in which only
     certain dependencies are considered.  if pruning is active, then
     this method must be called before performing inference
     (argmax/partition/marginals) on a new sentence.  it is not
     necessary to call prune() in between calling successive inference
     routines on the same sentence (e.g., calling marginals() and then
     calling argmax() on the same sentence) as the lists of
     dependencies are saved internally.  on the other hand, it is not
     an error to call prune() repeatedly with the same arguments; it
     is just inefficient.  prune() will copy the active lists
     internally so the caller can deallocate or reuse the memory after
     calling prune().

     the arguments have the following interpretation:

     N : number of words in the current sentence
     active_num[m] : number of pairs (h,l) such that (h,m,l) is an
         active dependency; must be >= 1
     active_heads[m] and active_labels[m] : these provide matched
         arrays such that: for i = 0 ... active_num[m] - 1,
         active_heads[m][i] and active_labels[m][i] name unique pairs
         (h,l) such that (h,m,l) is an active dependency.

     if pruning is active for this parser, then it is an error not to
     call prune() with the appropriate lists before performing
     inference on a new sentence.  specifically, the parser will
     continue to use the lists from the previous sentence, most likely
     resulting in serious error. */
  void prune(const int N,
             const int* const restrict active_num,
             const int** const restrict active_heads,
             const int** const restrict active_labels);

  /* accessors */
  int L() const;
  bool multiroot() const;
  bool pruned() const;
  bool topdown() const;
  bool sibparser() const;

  const o3active& act() const;

protected:
  const int _L;             /* number of labels */
  const bool _multiroot;    /* allow multiple roots? */
  const bool _pruned;       /* is pruning active? */
  const bool _topdown;      /* top-down parsing? */
  const bool _sibparser;    /* sibling-style parsing? */

  o3active _act;            /* set of active deps */

  int _N;                   /* size of the current sentence */
  int _K;                   /* rank-index for k-best enumeration */
  int _Kmax;                /* limit on k-best parsing */

  achart _C;                /* argmax chart */
  dchart _I;                /* inside chart */
  dchart _O;                /* outside chart */
  qchart _Q;                /* k-best queue chart */

  o3part::dmap _cache;      /* cache of part scores */


private:
  /* pre-empt the implicit versions of these */
  o3parser();
  o3parser(const o3parser& that);
  o3parser& operator=(const o3parser& that);
};



/* inlined methods */
inline int  o3parser::L()         const { return _L; }
inline bool o3parser::multiroot() const { return _multiroot; }
inline bool o3parser::pruned()    const { return _pruned; }
inline bool o3parser::topdown()   const { return _topdown; }
inline bool o3parser::sibparser() const { return _sibparser; }

/* pass arrays to the o3active object */
inline void
o3parser::prune(const int in_N,
                const int* const restrict active_num,
                const int** const restrict active_heads,
                const int** const restrict active_labels) {
  _act.set_active(!_topdown, _L, in_N,
                  active_num, active_heads, active_labels);
}

/* obtain the pruned arrays */
inline const o3active& o3parser::act() const { return _act; }



#ifndef EXPLICIT_TEMPLATES
#include "O3Parser.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3PARSER_H */
