/* -*-c++-*- */
#ifndef O3PART_H
#define O3PART_H
/* O3Part : A compact part representation.  Parts in the context of
   third-order parsing are:

   * dependencies        dep(h,m)       [ head h, mod m ]
   * siblings            sib(h,m,s)     [ s is between h and m ]
   * inner grandchildren gci(g,h,m)     [ g->h,h->m point opposite ]
   * outer grandchildren gco(g,h,m)     [ g->h,h->m in same direction ]
   * inner grandsiblings gcis(g,h,m,s)  [ (h,m,s) with grandparent ]
   * outer grandsiblings gcos(g,h,m,s)  [ (h,m,s) with grandparent ]
   * tri-siblings        sibs(h,m,s,c)  [ c is between h and s ]

   tri-siblings are only used in the sibling third-order parser.  Note
   that meaningful parts can only be created by static factory
   methods, which take care of the various checks and type codes. */

#include "Hashtable.h"
#include "Macros.h"
#include "Bits.h"

#include <assert.h>

#include <string>
#include <vector>
#include <limits>

using std::vector;
using std::string;


class o3part {
public:
  /* convenience types */
  typedef vector<o3part> vec;
  struct hashf; /* hashfunction for parts */
  typedef hashtable<o3part,double,hashf>  dmap;
  typedef hashtable<o3part,bool,hashf>    bmap;

  /* allow default construction, copy and assignment */
  o3part();
  o3part(const o3part& restrict p);
  o3part& operator=(const o3part& restrict p);

  /* score this part according to a part-scorer; this is the public
     interface of o3part.  score_with() will call exactly one of the
     scoring methods of the PartScorer S---the method corresponding to
     the kind of part represented by this---with the proper indices
     given as arguments.  the restricted interface guarantees that
     parts are handled properly and consistently. */
  template<class PartScorer>
  double score_with(PartScorer& restrict S) const;

  /* score a collection of parts (for convenience) */
  template<class PartScorer>
  static double score(const vec& restrict y, PartScorer& restrict S);
  template<class PartScorer>
  static double score(const bmap& restrict t, const PartScorer& restrict S);

  /* collect the parts of a dependency tree */
  static void parts(const int L,
                    const bool multiroot, const bool sibparser,
                    const vector<int>& restrict p,
                    const vector<int>& restrict l,
                    vec& restrict ret);

  /* for printing (i.e., debug messages) */
  string tostring() const;


protected:
  /* methods to create part objects for all parts.  to use these,
     create a trivial subclass of o3part and declare it friend of
     wherever it is that you need to construct parts.  this
     inconvenience is intended to discourage creation of parts outside
     the parser. */
  static o3part mkdep1(const idx_t h, const idx_t m, const lab_t l);
  static o3part mkdep2(const idx_t h, const idx_t m, const lab_t l);
  static o3part mkdep3(const idx_t h, const idx_t m, const lab_t l);
  static o3part mksib1(const idx_t h, const idx_t m, const idx_t s,
                       const lab_t lhm, const lab_t lhs);
  static o3part mksib2(const idx_t h, const idx_t m, const idx_t s,
                       const lab_t lhm, const lab_t lhs);
  static o3part mkgci1(const idx_t g, const idx_t h, const idx_t m,
                       const lab_t lgh, const lab_t lhm);
  static o3part mkgci2(const idx_t g, const idx_t h, const idx_t m,
                       const lab_t lgh, const lab_t lhm);
  static o3part mkgco1(const idx_t g, const idx_t h, const idx_t m,
                       const lab_t lgh, const lab_t lhm);
  static o3part mkgco2(const idx_t g, const idx_t h, const idx_t m,
                       const lab_t lgh, const lab_t lhm);
  static o3part mkgcis(const idx_t g, const idx_t h,
                       const idx_t m, const idx_t s,
                       const lab_t lgh, const lab_t lhm, const lab_t lhs);
  static o3part mkgcos(const idx_t g, const idx_t h,
                       const idx_t m, const idx_t s,
                       const lab_t lgh, const lab_t lhm, const lab_t lhs);
  static o3part mksibs(const idx_t h, const idx_t m,
                       const idx_t s, const idx_t c,
                       const lab_t lhm, const lab_t lhs, const lab_t lhc);


private:
  /* words and labels are 8 bits each; 4 indices + 3 labels gives 56
     bits with 8 bits remaining for type information */
  uint64_t _bits;

  /* this is used only by the static factory methods */
  o3part(const uint8_t tt,
         const idx_t i_1, const idx_t i_2, const idx_t i_3, const idx_t i_4,
         const lab_t l_1, const lab_t l_2, const lab_t l_3);

  /* internal convenience accessors */
  uint8_t type() const;
  idx_t i1() const;
  idx_t i2() const;
  idx_t i3() const;
  idx_t i4() const;
  lab_t l1() const;
  lab_t l2() const;
  lab_t l3() const;


  /* enumeration of part-types */
  enum types {
    O3PART_DEP1 = 0,
    O3PART_DEP2,
    O3PART_DEP3,
    O3PART_SIB1,
    O3PART_SIB2,
    O3PART_GCI1,
    O3PART_GCI2,
    O3PART_GCO1,
    O3PART_GCO2,
    O3PART_GCIS,
    O3PART_GCOS,
    O3PART_SIBS
  };
};




/*** inlined implementation ***/

/* score a list of parts */
template<class PartScorer>
inline double o3part::score(const vec& restrict y,
                            PartScorer& restrict S) {
  typename vec::const_iterator it = y.begin();
  const typename vec::const_iterator iend = y.end();
  double sc = 0;
  for(; it != iend; ++it) { sc += it->score_with(S); }
  return sc;
}
/* score a hashtable of parts (ignoring keys) */
template<class PartScorer>
inline double o3part::score(const bmap& restrict t,
                            const PartScorer& restrict S) {
  typename bmap::const_iterator it = t.cbegin();
  const typename bmap::const_iterator iend = t.cend();
  double sc = 0;
  for(; it != iend; ++it) { sc += it->key.score_with(S); }
  return sc;
}

/* construct parts */
inline o3part
o3part::mkdep1(const idx_t h, const idx_t m, const lab_t l) {
  return o3part(O3PART_DEP1, IINV, h, m, IINV, LINV, l, LINV);
}
inline o3part
o3part::mkdep2(const idx_t h, const idx_t m, const lab_t l) {
  return o3part(O3PART_DEP2, IINV, h, m, IINV, LINV, l, LINV);
}
inline o3part
o3part::mkdep3(const idx_t h, const idx_t m, const lab_t l) {
  return o3part(O3PART_DEP3, IINV, h, m, IINV, LINV, l, LINV);
}

inline o3part
o3part::mksib1(const idx_t h, const idx_t m, const idx_t s,
               const lab_t lhm, const lab_t lhs) {
  return o3part(O3PART_SIB1, IINV, h, m, s, LINV, lhm, lhs);
}
inline o3part
o3part::mksib2(const idx_t h, const idx_t m, const idx_t s,
               const lab_t lhm, const lab_t lhs) {
  return o3part(O3PART_SIB2, IINV, h, m, s, LINV, lhm, lhs);
}

inline o3part
o3part::mkgci1(const idx_t g, const idx_t h, const idx_t m,
               const lab_t lgh, const lab_t lhm) {
  return o3part(O3PART_GCI1, g, h, m, IINV, lgh,
                (m == INULL ? LNULL : lhm), LINV);
}
inline o3part
o3part::mkgci2(const idx_t g, const idx_t h, const idx_t m,
               const lab_t lgh, const lab_t lhm) {
  return o3part(O3PART_GCI2, g, h, m, IINV, lgh,
                (m == INULL ? LNULL : lhm), LINV);
}

inline o3part
o3part::mkgco1(const idx_t g, const idx_t h, const idx_t m,
               const lab_t lgh, const lab_t lhm) {
  return o3part(O3PART_GCO1, g, h, m, IINV, lgh,
                (m == INULL ? LNULL : lhm), LINV);
}
inline o3part
o3part::mkgco2(const idx_t g, const idx_t h, const idx_t m,
               const lab_t lgh, const lab_t lhm) {
  return o3part(O3PART_GCO2, g, h, m, IINV, lgh,
                (m == INULL ? LNULL : lhm), LINV);
}

inline o3part
o3part::mkgcis(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
               const lab_t lgh, const lab_t lhm, const lab_t lhs) {
  return o3part(O3PART_GCIS, g, h, m, s, lgh, lhm,
                (s == INULL ? LNULL : lhs));
}

inline o3part
o3part::mkgcos(const idx_t g, const idx_t h, const idx_t m, const idx_t s,
               const lab_t lgh, const lab_t lhm, const lab_t lhs) {
  return o3part(O3PART_GCOS, g, h, m, s, lgh, lhm,
                (s == INULL ? LNULL : lhs));
}

inline o3part
o3part::mksibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
               const lab_t lhm, const lab_t lhc, const lab_t lhs) {
  return o3part(O3PART_SIBS, h, m, c, s, lhm, lhc,
                (s == INULL ? LNULL : lhs));
}




/* nested hash function */
#include "Hashfunction.h"
struct o3part::hashf {
  static hashcode hash(const o3part& restrict p) {
    return hashfunction<uint64_t>::hash(p._bits);
  }
  static bool eq(const o3part& restrict p1, const o3part& restrict p2) {
    return p1._bits == p2._bits;
  }
};




/* constants for bit-packing */
#define I1OFF 8
#define I2OFF (I1OFF + IBITS)
#define I3OFF (I2OFF + IBITS)
#define I4OFF (I3OFF + IBITS)
#define L1OFF (I4OFF + IBITS)
#define L2OFF (L1OFF + LBITS)
#define L3OFF (L2OFF + LBITS)

#define INVALID (std::numeric_limits<uint64_t>::max())

/* methods must be declared inline to avoid linking errors */
inline o3part::
o3part(const uint8_t tt,
       const idx_t i_1, const idx_t i_2, const idx_t i_3, const idx_t i_4,
       const lab_t l_1, const lab_t l_2, const lab_t l_3)
  : _bits(tt |
          (((uint64_t)i_1) << I1OFF) |
          (((uint64_t)i_2) << I2OFF) |
          (((uint64_t)i_3) << I3OFF) |
          (((uint64_t)i_4) << I4OFF) |
          (((uint64_t)l_1) << L1OFF) |
          (((uint64_t)l_2) << L2OFF) |
          (((uint64_t)l_3) << L3OFF)) {
  CHECK(assert((L3OFF + LBITS) == 64));
}
inline o3part::o3part(const o3part& restrict p) : _bits(p._bits) {}
inline o3part::o3part()                         : _bits(INVALID) {}
inline o3part& o3part::operator=(const o3part& restrict p) {
  _bits = p._bits; return *this;
}


#define SEL(O, IL, type)                                \
  ((type ## _t)((_bits >> (O ## OFF)) & (IL ## MASK)))
inline uint8_t o3part::type() const { return (uint8_t)(_bits & 0xff); }
inline idx_t o3part::i1() const     { return SEL(I1, I, idx); }
inline idx_t o3part::i2() const     { return SEL(I2, I, idx); }
inline idx_t o3part::i3() const     { return SEL(I3, I, idx); }
inline idx_t o3part::i4() const     { return SEL(I4, I, idx); }
inline lab_t o3part::l1() const     { return SEL(L1, L, lab); }
inline lab_t o3part::l2() const     { return SEL(L2, L, lab); }
inline lab_t o3part::l3() const     { return SEL(L3, L, lab); }
#undef SEL

#undef I1OFF
#undef I2OFF
#undef I3OFF
#undef I4OFF
#undef L1OFF
#undef L2OFF
#undef L3OFF

#undef INVALID

#ifndef EXPLICIT_TEMPLATES
#include "O3Part.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3PART_H */
