#include "O3Perceptron.h"
#include "O3Part.h"
#include "O3Ablation.h"
#include "PartScoreLoss.h"

#include <stdio.h>

#include <vector>
#include <sstream>
#include <iostream>

using std::cerr;
using std::endl;
using std::vector;
using std::ostringstream;


/* lifecycle */
o3perceptron::o3perceptron() : o3trainer("Perc"), _mistakes(0) {}
o3perceptron::~o3perceptron() {}


/* print usage information */
void o3perceptron::usage_impl() const {
  cerr << "O3Perceptron options:" << endl
       << " --margin=<real>  : margin (Taskar-style)" << endl;
}
/* initialize the trainer */
void o3perceptron::init_impl(const o3trainer::strmap& opts) {
  /* recover margin */
  o3trainer::strmap::const_iterator it;

  _margin = 0;
  if((it = opts.find("margin")) != opts.end()) {
    _margin = strtod(it->second.c_str(), NULL);
    assert(_margin > 0);
  }
}


/* training interface */
void o3perceptron::train_start() {}
void o3perceptron::iter_start() { _mistakes = 0; }
void o3perceptron::iter_end(string& logline, string& repline) {
  ostringstream os;
  os << _mistakes;
  logline = os.str();
  os.str("");
  os << "mistakes = " << _mistakes;
  repline = os.str();
  /* flush averaged updates */
  _F->Wavg().avg_flush();
}


/* process an example */
void o3perceptron::process(const ex* const restrict e) {
  const int N = (int)e->heads.size();
  assert(_F->L() == _P->L());
  assert(_F->averaged());

  /* compute the true part-vector */
  o3part::vec rstar;
  o3part::parts(_P->L(), _P->multiroot(), _P->sibparser(),
                e->heads, e->labels, rstar);

  /* compute the parsed part-vector */
  o3part::vec rhat;
  { /* argmax label */
    vector<int> hhat, lhat;
    if(_margin == 0) {
      o3ablation::argmax(_ablation, *_P, N,
                         _F->avg_scorer(e->words, e->tags),
                         hhat, lhat);
    } else {
      o3ablation::argmax(_ablation, *_P, N,
                         addloss(_F->avg_scorer(e->words, e->tags),
                                 e, _margin),
                         hhat, lhat);
    }
    o3part::parts(_P->L(), _P->multiroot(), _P->sibparser(),
                  hhat, lhat, rhat);
  }

  /* collect the part-wise contributions */
  o3part::bmap rplus;
  o3part::bmap rminus;

  /* set initial positive and negative contributions */
  for(size_t i = 0; i < rstar.size(); ++i) { rplus.ins(rstar[i], true); }
  for(size_t i = 0; i < rhat.size();  ++i) { rminus.ins(rhat[i], true); }
  /* forbid duplicate parts */
  assert(rstar.size() == rplus.size());
  assert(rhat.size() == rminus.size());

  /* eliminate negative and positive contributions */
  for(size_t i = 0; i < rstar.size(); ++i) { rminus.delq(rstar[i]); }
  for(size_t i = 0; i < rhat.size();  ++i) { rplus.delq(rhat[i]); }

  if(rplus.size() > 0 || rminus.size() > 0) {
    ++_mistakes; /* part difference indicates mistake */
    /* create a scorer for adding values to the parameters */
    o3features::favg_add& adder = _F->avg_adder(e->words, e->tags);
    /* add 1 to the positive features */
    adder.alpha(1);
    o3ablation::score(_ablation, rplus, adder);
    /* subtract 1 from negative features */
    adder.alpha(-1);
    o3ablation::score(_ablation, rminus, adder);
  }

  /* advance the averaging timestamps */
  _F->Wavg().avg_tick();
}


/* classify an example */
void o3perceptron::classify(const ex* const restrict e,
                            vector<int>& restrict ret_h,
                            vector<int>& restrict ret_l) const {
  o3ablation::argmax(_ablation, *_P, (int)e->words.size(),
                     _F->avg_scavg(e->words, e->tags),
                     ret_h, ret_l);
}
