#include "O3Unravel.h"

#include "Backpointer.h"

#include <assert.h>


/* file-local recursive unraveling methods */
static void box(const achart& C, const o3box bb,
                vector<int>& ret_h, vector<int>& ret_l);
static void trip(const achart& C, const o3trip ii,
                 vector<int>& ret_h, vector<int>& ret_l);
static void trap(const achart& C, const o3trap tt,
                 vector<int>& ret_h, vector<int>& ret_l);
static void wedge(const achart& C, const o3wedge ww,
                  vector<int>& ret_h, vector<int>& ret_l);


/* the public interface */
void o3unravel::sibdeps(const int N, const bool multiroot,
                        const achart& C,
                        vector<int>& ret_h, vector<int>& ret_l) {
  /* clear and set the output vectors */
  ret_h.clear(); ret_l.clear();
  for(int i = 0; i < N; ++i) {
    ret_h.push_back(-2); ret_l.push_back(-2);
  }

  const int N1 = N - 1;
  const idx_t iN1 = ICONV(N1);
  const o3wedge sentspan(iN1);
  if(multiroot) { /* unravel from sentence spanning wedge */
    wedge(C, sentspan, ret_h, ret_l);
  } else { /* unravel from singleroot construction */
    /* the correct root is stored in the sentence-spanning wedge */
    const backp b = C.get(sentspan).bp;
    const lab_t l_r = b.l();
    const idx_t r   = b.i();
    /* claim the root dependency */
    assert(ret_h[r] == -2); ret_h[r] = -1;
    assert(ret_l[r] == -2); ret_l[r] = l_r;
    /* construct from a pair of wedges (unless null span) */
    if(r != N1) { wedge(C, o3wedge(b, iN1), ret_h, ret_l); }
    if(r != 0)  { wedge(C, o3wedge(b, 0),   ret_h, ret_l); }
  }

  /* check the output vectors */
  for(int i = 0; i < N; ++i) {
    assert(ret_h[i] != -2); assert(ret_l[i] != -2);
  }
}



/* recursive implementations */

/* unravel a box as a pair of facing wedges */
static void box(const achart& C, const o3box bb,
                vector<int>& ret_h, vector<int>& ret_l) {
  assert(bb.h() != bb.m()); assert(bb.h() != bb.s());
  assert(bb.s() != INULL);  assert(bb.m() != INULL);
  const backp b = C.get(bb).bp;
  const bool hbefm = (bb.h() == INULL || bb.h() < bb.m());
  /* recover endpoints of m-side and s-side wedges */
  const int tmp_rm = b.raw();
  const int tmp_rs = tmp_rm + (hbefm ? -1 : 1);
  const idx_t rm = ICONV(tmp_rm);
  const idx_t rs = ICONV(tmp_rs);
  if(hbefm) {
    assert(bb.h() == INULL || bb.h() < bb.s());
    assert(bb.s() <= rs);
    assert(rs < rm);
    assert(rm <= bb.m());
  } else {
    assert(bb.h() >= bb.s());
    assert(bb.s() >= rs);
    assert(rs > rm);
    assert(rm >= bb.m());
  }
  /* handle facing wedges (unless null span) */
  if(rs != bb.s()) { wedge(C, o3wedge(bb, rs), ret_h, ret_l); }
  if(rm != bb.m()) { wedge(C, o3wedge(rm, bb), ret_h, ret_l); }
}

/* unravel a trap as a trap and a box */
static void trip(const achart& C, const o3trip ii,
                 vector<int>& ret_h, vector<int>& ret_l) {
  assert(ii.h() != ii.m()); assert(ii.s() != ii.m());
  const backp b = C.get(ii).bp;
  const lab_t l_hc = b.l();
  const idx_t c    = b.i();
  assert(c != INULL);
  /* claim another dependency (NB: we only claim the the h->s dep
     here, the h->m is claimed by trap()) */
  assert(ret_h[ii.s()] == -2); ret_h[ii.s()] = IUNCONV_NULL(ii.h());
  assert(ret_l[ii.s()] == -2); ret_l[ii.s()] = ii.l_hs();
  /* the box part always exists */
  box(C, o3box(ii), ret_h, ret_l);
  /* the label is invalid iff the decomposition is a null sibling */
  if(l_hc == LINV) { /* null sibling, c = endpoint of wedge */
    if(c != ii.s()) {
      wedge(C, o3wedge(ii, c), ret_h, ret_l);
    }
  } else { /* sibling exists, given by c,l_hc */
    if(ii.h() == INULL) {
      assert(c < ii.s()); assert(ii.s() < ii.m());
    } else if(ii.h() < ii.m()) {
      assert(ii.h() < c); assert(c < ii.s()); assert(ii.s() < ii.m());
    } else {
      assert(ii.h() > c); assert(c > ii.s()); assert(ii.s() > ii.m());
    }
    trip(C, o3trip(ii, b), ret_h, ret_l);
  }
}

/* unravel a trap as a trap and a box */
static void trap(const achart& C, const o3trap tt,
                 vector<int>& ret_h, vector<int>& ret_l) {
  assert(tt.h() != tt.m());
  const backp b = C.get(tt).bp;
  const lab_t l_hs = b.l();
  const idx_t s    = b.i();
  assert(s != INULL);
  /* claim another dependency */
  assert(ret_h[tt.m()] == -2); ret_h[tt.m()] = IUNCONV_NULL(tt.h());
  assert(ret_l[tt.m()] == -2); ret_l[tt.m()] = tt.l_hm();
  /* the label is invalid iff the decomposition is a null sibling */
  if(l_hs == LINV) { /* null sibling, s = endpoint of wedge */
    if(s != tt.m()) {
      wedge(C, o3wedge(tt, s), ret_h, ret_l);
    }
  } else { /* sibling exists, given by s,l_hs */
    if(tt.h() == INULL) {
      assert(s < tt.m());
    } else if(tt.h() < tt.m()) {
      assert(tt.h() < s); assert(s < tt.m());
    } else {
      assert(tt.h() > s); assert(s > tt.m());
    }
    trip(C, o3trip(tt, b), ret_h, ret_l);
  }
}

/* unravel a wedge as a trap and wedge */
static void wedge(const achart& C, const o3wedge ww,
                  vector<int>& ret_h, vector<int>& ret_l) {
  assert(ww.h() != ww.e());
  const backp b = C.get(ww).bp;
  const idx_t m = b.i();
  if(ww.h() == INULL)      { assert(m <= ww.e()); }
  else if(ww.h() < ww.e()) { assert(m <= ww.e()); assert(ww.h() < m); }
  else                     { assert(m >= ww.e()); assert(ww.h() > m); }
  trap(C, o3trap(ww, b), ret_h, ret_l);
  if(m != ww.e()) { /* m has more grandchildren */
    wedge(C, o3wedge(ww, b), ret_h, ret_l);
  }
}
