/* -*-c++-*- */
#ifndef O3SIGNATURE_H
#define O3SIGNATURE_H
/* O3Signature : Chart signatures for third-order parsing. */

#include "Bits.h"
#include "Backpointer.h"


/* chart signatures */
class o3box;   /* boxes:                   h, m, s, l_hm, l_hs */
class o3trip;  /* inner-child trapezoids:  h, m, s, l_hm, l_hs */
class o3frap;  /* first-child trapezoids:  h, m, s, l_hm, l_hs */
class o3trap;  /* trapezoids:              g, h, m, l_gh, l_hm */
class o3wedge; /* wedges:                  g, h, e, l_gh */


/* underlying signature class */
class o3sig {
public:
  struct hashf; /* hashfunction for signatures */
  /* create an invalid signature */
  o3sig();

  /* index access */
  idx_t g() const;
  idx_t h() const;
  idx_t m() const;
  idx_t s() const;
  idx_t e() const;
  /* label access */
  lab_t l_gh() const;
  lab_t l_hm() const;
  lab_t l_hs() const;

protected:
  uint64_t _bits;
  explicit o3sig(const uint64_t raw);
};






/*** inlined implementation ***/

inline o3sig::o3sig() : _bits(UINT64_C(0xffffffffffffffff)) {}





/*** implementation of individual signatures ***/

/* the data layout is broken into bytes as follows

   byte 0:  e
   byte 1:  s
   byte 2:  l_hs
   byte 3:  m
   byte 4:  l_hm
   byte 5:  h
   byte 6:  l_gh
   byte 7:  g

   we use shift/mask to access various subfields.  different types of
   signatures are distinguished by the insertion of invalid values in
   the unused fields */
#define SHIFT_E  0
#define SHIFT_S  (SHIFT_E  + IBITS)
#define SHIFT_HS (SHIFT_S  + IBITS)
#define SHIFT_M  (SHIFT_HS + LBITS)
#define SHIFT_HM (SHIFT_M  + IBITS)
#define SHIFT_H  (SHIFT_HM + LBITS)
#define SHIFT_GH (SHIFT_H  + IBITS)
#define SHIFT_G  (SHIFT_GH + LBITS)
/* obtain the n'th slot of the raw bits */
#define SLOT(type) (_bits >> (SHIFT_ ## type))

inline o3sig::o3sig(const uint64_t raw) : _bits(raw) {
  CHECK(assert((SHIFT_G + IBITS) == 64));
  CHECK(assert(sizeof(o3sig) == sizeof(uint64_t)));
}

inline idx_t o3sig::g() const    { return (idx_t)SLOT(G); }
inline idx_t o3sig::h() const    { return (idx_t)SLOT(H); }
inline idx_t o3sig::m() const    { return (idx_t)SLOT(M); }
inline idx_t o3sig::s() const    { return (idx_t)SLOT(S); }
inline idx_t o3sig::e() const    { return (idx_t)SLOT(E); }

inline lab_t o3sig::l_gh() const { return (lab_t)SLOT(GH); }
inline lab_t o3sig::l_hm() const { return (lab_t)SLOT(HM); }
inline lab_t o3sig::l_hs() const { return (lab_t)SLOT(HS); }



/* hashfunction */
#include "Hashfunction.h"
struct o3sig::hashf {
  static hashcode hash(const o3sig& restrict p) {
    return hashfunction<uint64_t>::hash(p._bits);
  }
  static bool eq(const o3sig& restrict p1, const o3sig& restrict p2) {
    return p1._bits == p2._bits;
  }
};


/*** derived classes ***/

/* the exact layout of each signature is as follows:

   field:  g   l_gh  h  l_hm  m    l_hs  s    e
   ==============================================
   box:    INV INV   h  l_hm  m    l_hs  s    0
   frap:   INV 0     h  l_hm  m    l_hs  s    INV
   trip:   0   INV   h  l_hm  m    l_hs  s    INV
   trap:   g   l_gh  h  l_hm  m    INV   INV  INV
   wedge:  g   l_gh  h  INV   INV  INV   INV  e

   the inclusion of "INV" invalid tokens disambiguates the signatures.
   only wedges have INV l_hm,m; only traps have invalid e *and*
   l_hs,s; boxes/fraps/trips have invalid g,l_gh,e and are
   distinguished by a zero slotted in to one of the three positions.
   boxes and trips are quite similar, having the same indices, so they
   must be distinguished from each other by defining one of g,e as 0
   instead of INV. */

class o3box : public o3sig {
public:
  /* construct the box half of a trap */
  explicit o3box(const o3trap t, const backp s_lhs);
  /* construct the box half of a trip */
  explicit o3box(const o3trip t);
  /* construct a box directly */
  explicit o3box(const idx_t h, const idx_t m, const idx_t s,
                 const lab_t l_hm, const lab_t l_hs);
  /* disabled methods */
  idx_t g() const    { assert(0); return 0; }
  idx_t e() const    { assert(0); return 0; }
  lab_t l_gh() const { assert(0); return 0; }
  /* everyone's friends... */
  friend class o3wedge;
};

class o3trap : public o3sig {
public:
  /* construct the trap half of a larger trap */
  explicit o3trap(const o3trap t, const backp s_lhs);
  /* construct the trap half of a wedge */
  explicit o3trap(const o3wedge w, const backp m_lhm);
  /* construct a trap directly */
  explicit o3trap(const idx_t g, const idx_t h, const idx_t m,
                  const lab_t l_gh, const lab_t l_hm);
  /* disabled methods */
  idx_t s() const    { assert(0); return 0; }
  idx_t e() const    { assert(0); return 0; }
  lab_t l_hs() const { assert(0); return 0; }
  /* everyone's friends... */
  friend class o3box;
  friend class o3trip;
  friend class o3wedge;
};

class o3trip : public o3sig {
public:
  /* construct the trip half of a larger trip */
  explicit o3trip(const o3trip t, const backp c_lhc);
  /* construct the trip version of an existing trap */
  explicit o3trip(const o3trap t, const backp s_lhs);
  /* construct a trip directly */
  explicit o3trip(const idx_t h, const idx_t m, const idx_t s,
                  const lab_t l_hm, const lab_t l_hs);
  /* disabled methods */
  idx_t g() const    { assert(0); return 0; }
  idx_t e() const    { assert(0); return 0; }
  lab_t l_gh() const { assert(0); return 0; }
  /* everyone's friends... */
  friend class o3box;
  friend class o3wedge;
};

class o3wedge : public o3sig {
public:
  /* construct the s-side wedge half of a box */
  explicit o3wedge(const o3box b, const idx_t rs);
  /* construct the m-side wedge half of a box */
  explicit o3wedge(const idx_t rm, const o3box b);
  /* construct the wedge half of a wedge */
  explicit o3wedge(const o3wedge w, const backp m_lhm);
  /* construct a wedge that claims almost all of a trap (representing
     a trap with no sibling).  h1=h+1/h=h-1 depending on direction */
  explicit o3wedge(const o3trap t, const idx_t h1);
  /* construct a wedge that claims the near half of a trip
     (representing a trip with no sibling).  h1=h+1/h=h-1 */
  explicit o3wedge(const o3trip i, const idx_t h1);
  /* construct a sentence-spanning wedge */
  explicit o3wedge(const idx_t n1);
  /* construct half of a singleroot wedge */
  explicit o3wedge(const backp r_lr, const idx_t e);
  /* construct a wedge directly */
  explicit o3wedge(const idx_t g, const idx_t h, const idx_t e,
                   const lab_t l_gh);
  /* disabled methods */
  idx_t m() const    { assert(0); return 0; }
  idx_t s() const    { assert(0); return 0; }
  lab_t l_hm() const { assert(0); return 0; }
  lab_t l_hs() const { assert(0); return 0; }
  /* everyone's friends... */
  friend class o3trap;
};





/*** inlined implementation ***/

/* construct the box half of a trap */
inline o3box::o3box(const o3trap t, const backp s_lhs)
  : o3sig((t._bits &
           /* keep only h,l_hm,m */
           UINT64_C(0x0000ffffff000000)) |
          /* invalidate g,l_gh and keep e=0 */
          UINT64_C(0xffff000000000000) |
          /* insert new l_hs,s */
          (((uint64_t)s_lhs.raw()) << SHIFT_S)) {}

/* construct the box half of a trip */
inline o3box::o3box(const o3trip i)
  : o3sig((i._bits &
           /* keep h,l_hm,m,l_hs,s */
           UINT64_C(0x0000ffffffffff00)) |
          /* invalidate g,l_gh and keep e=0 */
          UINT64_C(0xffff000000000000)) {}

/* construct a box directly */
inline o3box::o3box(const idx_t hh, const idx_t mm, const idx_t ss,
                    const lab_t lhm, const lab_t lhs)
  : o3sig(/* insert all items */
          (((uint64_t)hh) << SHIFT_H) |
          (((uint64_t)mm) << SHIFT_M) |
          (((uint64_t)ss) << SHIFT_S) |
          (((uint64_t)lhm) << SHIFT_HM) |
          (((uint64_t)lhs) << SHIFT_HS) |
          /* invalidate g,l_gh and keep e=0 */
          UINT64_C(0xffff000000000000)) {}




/* construct the trap half of a larger trap */
inline o3trap::o3trap(const o3trap t, const backp s_lhs)
  : o3sig((t._bits &
           /* keep g,l_gh,h and invalid portions */
           UINT64_C(0xffffff0000ffffff)) |
          /* insert new l_hm,m */
          (((uint64_t)s_lhs.raw()) << SHIFT_M)) {}

/* construct the trap half of a wedge */
inline o3trap::o3trap(const o3wedge w, const backp m_lhm)
  : o3sig((w._bits &
           /* keep only g,l_gh,h */
           UINT64_C(0xffffff0000000000)) |
          /* invalidate l_hs,s,e */
          UINT64_C(0x0000000000ffffff) |
          /* insert new l_hm,m */
          (((uint64_t)m_lhm.raw()) << SHIFT_M)) {}

/* construct a trap directly */
inline o3trap::o3trap(const idx_t gg, const idx_t hh, const idx_t mm,
                      const lab_t lgh, const lab_t lhm)
  : o3sig(/* insert all items */
          (((uint64_t)gg) << SHIFT_G) |
          (((uint64_t)hh) << SHIFT_H) |
          (((uint64_t)mm) << SHIFT_M) |
          (((uint64_t)lgh) << SHIFT_GH) |
          (((uint64_t)lhm) << SHIFT_HM) |
          /* invalidate l_hs,s,e */
          UINT64_C(0x0000000000ffffff)) {}




/* construct the trip half of a larger trip */
inline o3trip::o3trip(const o3trip i, const backp c_lhc)
  : o3sig(((i._bits &
            /* keep only l_hs,s */
            UINT64_C(0x0000000000ffff00))
           /* shift l_hs,s into l_hm,m position */
           << (IBITS + LBITS)) |
          /* combine with other parts of t */
          (i._bits &
           /* keep only h and invalid portions */
           UINT64_C(0xffffff00000000ff)) |
          /* insert new l_hs,s */
          (((uint64_t)c_lhc.raw()) << SHIFT_S)) {}

/* construct the trip version of an existing trap */
inline o3trip::o3trip(const o3trap t, const backp s_lhs)
  : o3sig((t._bits &
           /* keep only h,l_hm,m */
           UINT64_C(0x0000ffffff000000)) |
          /* invalidate l_gh,e and keep g=0 */
          UINT64_C(0x00ff0000000000ff) |
          /* insert new l_hs,s */
          (((uint64_t)s_lhs.raw()) << SHIFT_S)) {}

/* construct a trip directly */
inline o3trip::o3trip(const idx_t hh, const idx_t mm, const idx_t ss,
                      const lab_t lhm, const lab_t lhs)
  : o3sig(/* insert all items */
          (((uint64_t)hh) << SHIFT_H) |
          (((uint64_t)mm) << SHIFT_M) |
          (((uint64_t)ss) << SHIFT_S) |
          (((uint64_t)lhm) << SHIFT_HM) |
          (((uint64_t)lhs) << SHIFT_HS) |
          /* invalidate l_gh,e and keep g=0 */
          UINT64_C(0x00ff0000000000ff)) {}




/* construct the s-side wedge half of a box */
inline o3wedge::o3wedge(const o3box b, const idx_t rs)
  : o3sig(((b._bits &
            /* keep only h */
            UINT64_C(0x0000ff0000000000))
           /* shift h into g position */
           << (IBITS + LBITS)) |
          /* combine with l_hs,s */
          ((b._bits &
            /* keep only l_hs,s */
            UINT64_C(0x0000000000ffff00))
           /* shift l_hs,s into l_gh,h position */
           << (IBITS + LBITS + IBITS + LBITS)) |
          /* invalidate l_hm,m,l_hs,s */
          UINT64_C(0x000000ffffffff00) |
          /* insert the new e */
          ((uint64_t)rs)) {}

/* construct the m-side wedge half of a box */
inline o3wedge::o3wedge(const idx_t rm, const o3box b)
  : o3sig(((b._bits &
            /* keep only h,l_hm,m */
            UINT64_C(0x0000ffffff000000))
           /* shift h,l_hm,m into g,l_gh,h position */
           << (IBITS + LBITS)) |
          /* invalidate l_hm,m,l_hs,s */
          UINT64_C(0x000000ffffffff00) |
          /* insert the new e */
          ((uint64_t)rm)) {}

/* construct the wedge half of a wedge */
inline o3wedge::o3wedge(const o3wedge w, const backp m_lhm)
  : o3sig(((w._bits &
            /* keep only h */
            UINT64_C(0x0000ff0000000000))
           /* shift h into g position */
           << (IBITS + LBITS)) |
          /* combine with e and invalid portions */
          (w._bits &
           /* keep only e and invalid portions */
           UINT64_C(0x000000ffffffffff)) |
          /* insert new l_gh,h */
          (((uint64_t)m_lhm.raw()) << SHIFT_H)) {}

/* construct a wedge that claims almost all of a trap (representing
   a trap with no sibling).  h1=h+1/h=h-1 depending on direction */
inline o3wedge::o3wedge(const o3trap t, const idx_t h1)
  : o3sig(((t._bits &
            /* keep only h,l_hm,m */
            UINT64_C(0x0000ffffff000000))
           /* shift h,l_hm,m into g,l_gh,h position */
           << (IBITS + LBITS)) |
          /* invalidate l_hm,m,l_hs,s */
          UINT64_C(0x000000ffffffff00) |
          /* insert the new e */
          ((uint64_t)h1)) {}

/* construct a wedge that claims the near half of a trip (representing
   a trip with no sibling).  h1=h+1/h=h-1 depending on direction */
inline o3wedge::o3wedge(const o3trip i, const idx_t h1)
  : o3sig(((i._bits &
            /* keep only h */
            UINT64_C(0x0000ff0000000000))
           /* shift h into g position */
           << (IBITS + LBITS)) |
          /* combine with l_hs,s */
          ((i._bits &
            /* keep only l_hs,s */
            UINT64_C(0x0000000000ffff00))
           /* shift l_hs,s into l_gh,h position */
           << (IBITS + LBITS + IBITS + LBITS)) |
          /* invalidate l_hm,m,l_hs,s */
          UINT64_C(0x000000ffffffff00) |
          /* insert the new e */
          ((uint64_t)h1)) {}

/* construct a sentence-spanning wedge */
inline o3wedge::o3wedge(const idx_t n1)
  : o3sig(/* initializer: invalid h,l_hm,m,l_hs; g,l_gh,h=-1 */
          UINT64_C(0xfefefeffffffff00) |
          /* insert (n-1) as e */
          ((uint64_t)n1)) {}

/* construct half of a singleroot wedge */
inline o3wedge::o3wedge(const backp r_lr, const idx_t ee)
  : o3sig(/* initializer: invalid h,l_hm,m,l_hs; g=-1 */
          UINT64_C(0xfe0000ffffffff00) |
          /* insert l_r,r as l_gh,h */
          (((uint64_t)r_lr.raw()) << SHIFT_H) |
          /* insert the new e */
          ((uint64_t)ee)) {}

/* construct a wedge directly */
inline o3wedge::o3wedge(const idx_t gg, const idx_t hh, const idx_t ee,
                        const lab_t lgh)
  : o3sig(/* insert all items */
          (((uint64_t)gg) << SHIFT_G) |
          (((uint64_t)hh) << SHIFT_H) |
          (((uint64_t)ee) << SHIFT_E) |
          (((uint64_t)lgh) << SHIFT_GH) |
          /* invalidate l_hm,m,l_hs,s */
          UINT64_C(0x000000ffffffff00)) {}


#undef SHIFT_E
#undef SHIFT_S
#undef SHIFT_HS
#undef SHIFT_M
#undef SHIFT_HM
#undef SHIFT_H
#undef SHIFT_GH
#undef SHIFT_G
#undef SLOT

#endif /* O3SIGNATURE_H */
