#include "O3Trainer.h"

#include "Timer.h"
#include "TestInf.h"
#include "Mappings.h"
#include "GZIO.h"
#include "O3FGen.h"
#include "O3Ablation.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <iostream>
#include <iomanip>

using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::fixed;
using std::scientific;
using std::setprecision;

/* key for storing executable name */
#define EXEC "#EXECFILE#"

/* reporting interval for parallel jobs */
#define PARREPINT 0x7f

/* buffer size for reading lines from files */
#define LINELEN 65536

#define MYNAME                                          \
  ((_trdata == NULL ? string("") : _trdata->ostr()) +   \
   _name + "[" +                                        \
   (_P == NULL ? "???" :                                \
    (string(_P->sibparser() ? "s" : "g") +              \
     (_P->topdown()         ? "t" : "b") +              \
     (_P->pruned()          ? "p" : "f") +              \
     (_P->multiroot()       ? "m" : "r") +              \
     (_P->L() == 1          ? "u" : "l"))) +            \
   (_ablation == -1 ? ""                                \
    : ",-" + o3ablation::name(_ablation)) +             \
   "]")


/* lifecycle */
o3trainer::o3trainer(const string& name)
  : _P(NULL),
    _F(NULL),
    _ablation(-1),
    _dir(),
    _trdata(NULL),
    _valdata(NULL),
    _Atr(NULL),
    _Aval(NULL),
    _name(name),
    _ordering(-1),
    _seed(0),
    _T(-1),
    _writeparams(false) {}

o3trainer::~o3trainer() {
  if(_trdata != NULL)  { delete _trdata; }
  if(_valdata != NULL) { delete _valdata; }
  if(_P != NULL)       { delete _P; }
  if(_F != NULL)       { delete _F; }
  if(_Atr != NULL)     { delete _Atr; }
  if(_Aval != NULL)    { delete _Aval; }
}



/* initialize the trainer */
void o3trainer::init(const vector<string>& args) {
  /* read a map of name-value options */
  map<string,string> opts;
  mkopts(args, opts);
  const map<string,string>& copts = opts;
  init(copts);

  /* pass options to the underlying implementation */
  init_impl(copts);
}



/* perform the actual training */
void o3trainer::train() {
  cerr << MYNAME << " : training started for max " << _T
       << " rounds ... " << endl;
  cerr << MYNAME << " : data order is " << _trdata->ostr_long() << endl;

  /* only report this for randomized enumeration */
  cerr << MYNAME << " : seeding the RNG with value " << _seed << endl;
  srandom(_seed);
  { /* save seed in model directory */
    FILE* const f = fopen((_dir + "/seed").c_str(), "w");
    assert(f != NULL);
    fprintf(f, "%d\n", _seed);
    fclose(f);
  }

  /* adjust the reporting interval */
  ticker_adjust_repint(-2);

  const int Ntr = _trdata->N();

  /* log file for primal and #pass information */
  const string objfname = _dir + "/obj.out";
  FILE* const obj_out = gzio::gzopen(objfname.c_str(), "w");

  /* log of example indices that are visited, for repeatability */
  const string eidxfname = _dir + "/example-ordering.gz";
  FILE* const eidx_out = gzio::gzopen(eidxfname.c_str(), "w");

  /* utilities for parallel training */
  int* pipe_read_fds = NULL; /* read ends for parent to rcv data */
  int pipe_write_fd = -1;    /* write end for worker to send data */
  int worker_id = -1;        /* worker id, for reporting */
  if(_parallel != 0) {
    assert(_trdata->ostr() != "R");
    assert(_parallel > 0);
    pipe_read_fds = new int[_parallel];
  }

  /* initialize the training implementation */
  train_start();
  double cpu_time = 0;
  const bool uniform = (_trdata->ostr() == "R");
  for(int t = 1; t <= _T; ++t) {
    /* NB: the first round of rand-uniform will be a permutation, to
       ensure that all examples are visited. */
    if(uniform) { _trdata->ordering(t == 1 ? "perm" : "unif"); }
    cerr << MYNAME << " : iteration " << t << "..." << endl;

    /* initialize the training specialization */
    iter_start();
    ticker64_start();
    const clock_t ct0 = clock();

    vector<const ex*> examples;
    { /* fill examples with the data points to process */
      const ex* e = NULL;
      _trdata->start();
      while((e = _trdata->next()) != NULL) {
        fprintf(eidx_out, "%d\n", e->id());
        examples.push_back(e);
      }
      assert(Ntr == (int)examples.size());
    }

    /* create start index and step size */
    int start_idx = 0, step_size = 1;
    map<pid_t,int> workers;
    if(_parallel != 0) {
      /* flush c++ output streams */
      cerr << flush;
      cout << flush;
      /* flush all open stdio output streams */
      fflush(NULL);
      /* workers process an interleaved partition of the examples */
      step_size = _parallel;
      /* initialize a null start index for the parent */
      start_idx = Ntr;
      /* initialize each parallel worker process */
      for(int worker = 1; worker <= _parallel; ++worker) {
        /* open a pipe for communication after the iteration */
        int readfd = -1, writefd = -1;
        safe_pipe(readfd, writefd);
        /* fork between parent and worker */
        const pid_t wpid = safe_fork();
        if(wpid == 0) { /* the worker executes this */
          /* claim the write end and discard the read end */
          pipe_write_fd = writefd;
          safe_close(readfd);
          /* redirect stderr to an appropriate log file (lest the
             diagnostic output get mixed) */
          char padnum[10];
          sprintf(padnum, "%03d", worker);
          safe_log_stderr((_dir + "/log.worker" + padnum).c_str(), false);
          /* discard data structures used by the parent */
          delete [] pipe_read_fds; pipe_read_fds = NULL;
          workers.clear();
          /* prepare the data partiton */
          worker_id = worker;
          start_idx = worker - 1;
          CHECK(cout << "worker " << worker << ": ready to work" << endl);
          break; /* begin working */
        } else { /* the parent executes this */
          CHECK(cerr << "parent: spawned worker " << worker << endl);
          /* claim the read end and discard the write end */
          pipe_read_fds[worker - 1] = readfd;
          safe_close(writefd);
          workers[wpid] = worker; /* save the worker pid */
        }
      }
    }


    /* stride across a subset of examples */
    int cnt = 0;
    for(int i = start_idx; i < Ntr; i += step_size) {
      const ex* const e = examples[i];
      /* prepare the active dependencies if necessary */
      if(_Atr != NULL) {
        assert(_Atr->active_N(e->id()) == (int)e->heads.size());
        _P->prune(_Atr->active_N(e->id()),
                  _Atr->active_num(e->id()),
                  _Atr->active_heads(e->id()),
                  _Atr->active_labels(e->id()));
      }

      /* update according to the training specialization */
      process(e);

      ticker64(cnt, Ntr);
      ++cnt;
      if(_parallel != 0) {
        if((cnt & PARREPINT) == 0) {
          const double percent = 100.0*((double)i)/((double)Ntr);
          cout << "worker " << worker_id << ": "
               << fixed << setprecision(2) << percent << "%" << endl;
        }
      }
    }
    if(pipe_read_fds == NULL) { ticker64_finish(Ntr); }


    /* gather results after the iteration */
    if(_parallel != 0) {
      if(pipe_read_fds == NULL) { /* child process */
        /* erase zeroed parameters */
        if(_F->averaged()) { _F->Wavg().erase_zeros(); }
        else               { _F->Wlin().erase_zeros(); }
        /* print number nonzero params */
        const uint64_t nnz =
          (_F->averaged() ? _F->Wavg().nnz() : _F->Wlin().nnz());
        cerr << MYNAME << " : nnz = " << nnz << " ("
             << fixed << setprecision(2) << ((double)nnz)/1000000.0
             << "m)" << endl;
        /* finish the iteration for the training specialization */
        string logline, repline;
        iter_end(logline, repline);
        /* send the trained parameters */
        CHECK(cout << flush << "worker " << worker_id
              << ": sending parameters..." << endl);
        /* scaling factor implements an average */
        const double scale = 1.0/_parallel;
        /* send the params to the parent via the pipe */
        if(_F->averaged()) {
          _F->Wavg().parallel_send(pipe_write_fd, scale);
        } else {
          _F->Wlin().parallel_send(pipe_write_fd, scale);
        }
        CHECK(cout << flush << "worker " << worker_id
              << ": params sent!" << endl);
        /* clean up file descriptors */
        safe_close(pipe_write_fd);
        CHECK(cout << flush << "worker " << worker_id
              << ": exiting cleanly" << endl);
        exit(0); /* work complete */
      } else { /* parent process */
        /* before starting, first erase all parameters; the result of
           the iteration will be the average of all of the workers'
           parameter vectors. */
        if(_F->averaged()) { _F->Wavg().zero(); }
        else               { _F->Wlin().zero(); }
        /* for each pipe read end, merge with its parameters.  we do
           this in order although this may be slightly inefficient,
           since workers may finish at different times and certainly
           not always in ascending order.  however, asynchronous
           handling would be more complex and the expected time
           savings minor */
        cerr << MYNAME << " : " << _parallel << " workers:" << flush;
        for(int worker = 1; worker <= _parallel; ++worker) {
          CHECK(cerr << flush << "parent: params from worker "
                << worker << "..." << endl);
          const int readfd = pipe_read_fds[worker - 1];
          /* read the params via the worker's pipe */
          if(_F->averaged()) {
            _F->Wavg().parallel_recv(readfd);
          } else {
            _F->Wlin().parallel_recv(readfd);
          }
          /* clean up file descriptors */
          safe_close(readfd);
          /* wait for the worker to prevent zombies.  NB workers may
             finish out of order due to the pipe buffer being large
             enough to hold the transfer */
          safe_wait(workers);
          cerr << " " << worker << flush;
        }
        /* clean up */
        CHECK(cerr << flush << "parent: finished" << endl);
        double splittime = 0;
        timer_split(splittime);
        cerr << " [" << timestr(splittime) << "]" << endl;
      }
    }

    cpu_time += ((double)(clock() - ct0))/(double)CLOCKS_PER_SEC;

    if(_F != NULL) { /* erase zeroed parameters */
      if(_F->averaged()) { _F->Wavg().erase_zeros(); }
      else               { _F->Wlin().erase_zeros(); }
      /* print number nonzero params */
      const uint64_t nnz =
        (_F->averaged() ? _F->Wavg().nnz() : _F->Wlin().nnz());
      cerr << MYNAME << " : nnz = " << nnz << " ("
           << fixed << setprecision(2) << ((double)nnz)/1000000.0
           << "m)" << endl;
    }
    /* finish the iteration for the training specialization */
    string logline, repline;
    iter_end(logline, repline);

    /* save the primals in the model directory */
    if(_writeparams && _F != NULL) {
      char buf[16];
      sprintf(buf, "%03d", t);
      const string fname(_dir + "/parameters." + buf + ".gz");
      cerr << MYNAME << " : saving to \"" << fname << "\" " << flush;
      { /* save the proper parameter vector */
        FILE* const f = gzio::gzopen(fname.c_str(), "w");
        if(_F->averaged()) { _F->Wavg().save(f); }
        else               { _F->Wlin().save(f); }
        gzio::gzclose(fname.c_str(), f);
      }
      // { /* check the parameters */
      //   o3features* const F2 = new o3features(_F->L(), _F->averaged());
      //   cerr << MYNAME << " : checking parameter loading" << flush;
      //   FILE* const f2 = gzio::gzopen(fname.c_str(), "r");
      //   if(F2->averaged()) { F2->Wavg().load(f2); }
      //   else               { F2->Wlin().load(f2); }
      //   gzio::gzclose(fname.c_str(), f2);
      //   if(!(_F->averaged()
      //        ? F2->Wavg().equals(_F->Wavg())
      //        : F2->Wlin().equals(_F->Wlin()))) {
      //     cerr << MYNAME << " : param load warning: W2 != W" << endl;
      //   }
      //   if(!(_F->averaged()
      //        ? _F->Wavg().equals(F2->Wavg())
      //        : _F->Wlin().equals(F2->Wlin()))) {
      //     cerr << MYNAME << " : param load warning: W != W2" << endl;
      //   }
      //   delete F2;
      // }
    }

    /* parse the validation data if provided */
    double val_accuracy = 0.0;
    if(_valdata != NULL) {
      char buf[10];
      sprintf(buf, "%03d", t);
      const string fname(_dir + "/val-labels." + buf);
      val_classify(_valdata, fname, val_accuracy);
    }

    /* update the objective log with the current iteration */
    fprintf(obj_out, "%d %.17g %s %.17g\n",
            t, cpu_time, logline.c_str(), val_accuracy);
    /* print a summary */
    cerr << MYNAME << " : finished it = " << t
         << scientific << setprecision(5)
         << " ; cpu = " << cpu_time
         << " ; " << repline << endl;
  }

  /* tidy up */
  if(_parallel != 0) { delete [] pipe_read_fds; }
  gzio::gzclose(objfname.c_str(),  obj_out);
  gzio::gzclose(eidxfname.c_str(), eidx_out);
  cerr << MYNAME << " : end of training (" << _T << ") iterations" << endl;
}






/* argument handling */
void o3trainer::usage(const string& mesg) const {
  cerr << "O3Trainer usage:" << endl
       << " Training options:" << endl
       << " --trdata=<path>   : training data file" << endl
       << " --valdata=<path>  : validation data file" << endl
       << " --dir=<path>      : model directory" << endl
       << " --T=<int>         : number of iterations" << endl
       << " --ordering=<str>  : data ordering [det,unif,perm]" << endl
       << " --seed=<int>      : random seed (def: 1)" << endl
       << " --relabel         : relabel the corpus using oracle decoding" << endl
       << " --noparams        : don't save parameter vectors" << endl
       << " --parallel=<int>  : use parallel training" << endl
       << " --fix-commas      : deterministically reattach commas" << endl
       << endl
       << " Parser options:" << endl
       << " --L=<int>      : number of dependency labels" << endl
       << " --multiroot    : multiroot parsing (def: single-root)" << endl
       << " --topdown      : parse top-down instead of bottom-up" << endl
       << " --sibparser    : sibling factorization (def: grandchild)" << endl
       << " --ablate=<str> : ignore certain parts, according to <str>:" << endl
       << "         'gsib' = ignore grand-sibling 4-grams" << endl
       << "         'sibs' = ignore tri-sibling 4-grams" << endl
       << "         'all4' = ignore all 4-grams" << endl
       << "         'allg' = ignore all grandchild scores" << endl
       << "         'alls' = ignore all sibling scores" << endl
       << "         'xc07' = in conjunction with --sibparser, reimplement" << endl
       << "                  Carreras (2007) factorization" << endl
       << "         'mp06' = reimplement McDonald and Pereira (2006)" << endl
       << endl
       << " Feature options:" << endl
       << " --fconf=<str>  : feature config; 0 or more of the following" << endl
       << "          '+ft' = use fine-tag features" << endl
       << "          '+ct' = use coarse-tag features" << endl
       << endl
       << " Pruning options:" << endl
       << " --prune-tmarg=<path> : load training marginals from file" << endl
       << " --prune-vmarg=<path> : load valdata marginals from file" << endl
       << " --prune-beam=<real>  : relative beam for pruned dependencies" << endl
       << " --prune-vbeam=<real> : beam for valdata (def: same as train)" << endl
       << " --prune-train-truth  : always include truth on training data" << endl
       << endl
       << " Testing options:" << endl
       << " --testinf-simple=<int>        : simple argmax tests" << endl
       << " --testinf-bf=<int>            : compare to brute-force enumeration" << endl
       << " --testinf-tdbu=<int>,<real>   : compare top-down vs bottom-up" << endl
       << " --testinf-pruned=<int>,<real> : compare to brute-force with pruning" << endl
       << endl << endl;

  usage_impl();

  cerr << endl << endl << endl << mesg << endl;
  exit(1);
}


/* initialize with processed options */
void o3trainer::init(const strmap& opts) {
  /*** process global training options ***/
  strmap::const_iterator it;

  /* load L first and load corpora */
  if((it = opts.find("L")) == opts.end()) {
    usage("please provide #labels --L...");
  }
  const int L = atoi(it->second.c_str());
  assert(L > 0);

  /* check for parallel training */
  _parallel = 0;
  if((it = opts.find("parallel")) != opts.end()) {
    const int njobs = atoi(it->second.c_str());
    assert(njobs > 0);
    _parallel = (unsigned)njobs;
  }


  /* training options */
  _dir = "";
  if((it = opts.find("dir")) == opts.end()) {
    usage("please provide model directory --dir=...");
  }
  _dir = it->second;
  assert(!_dir.empty());
  /* load mappings */
  mappings::load(_dir.c_str());
  /* save the current options to a file */
  write_opts(_dir + "/options", opts);

  if((it = opts.find("trdata")) == opts.end()) {
    usage("please provide training data --trdata=...");
  }
  assert(_trdata == NULL);
  _trdata = corpus::load(it->second.c_str(), L);
  _trdata->ordering("det");

  if((it = opts.find("valdata")) != opts.end()) {
    assert(_valdata == NULL);
    _valdata = corpus::load(it->second.c_str(), L);
    _valdata->ordering("det");
  }

  _T = -1;
  if((it = opts.find("T")) == opts.end()) {
    usage("please provide #iterations --T=...");
  }
  _T = atoi(it->second.c_str());
  assert(_T > 0);

  if((it = opts.find("ordering")) == opts.end()) {
    usage("please provide data ordering --ordering=...");
  }
  const string trorder = it->second;

  _seed = (unsigned int)time(NULL);
  if((it = opts.find("seed")) != opts.end()) {
    long int tmp = atol(it->second.c_str());
    assert(tmp >= 0);
    _seed = (unsigned int)tmp;
  }
  srand(_seed);

  _writeparams = true;
  if(opts.find("noparams") != opts.end()) { _writeparams = false; }


  /* parser options */
  bool multiroot = false;
  if(opts.find("multiroot") != opts.end()) { multiroot = true; }

  bool topdown = false;
  if(opts.find("topdown") != opts.end()) { topdown = true; }

  bool sibparser = false;
  if(opts.find("sibparser") != opts.end()) { sibparser = true; }

  _ablation = o3ablation::code("none");
  if((it = opts.find("ablate")) != opts.end()) {
    _ablation = o3ablation::code(it->second);
  }


  /* feature options */
  if((it = opts.find("params")) == opts.end()) {
    usage("please provide parameter type --params...");
  }
  const string params = it->second;
  assert(!params.empty());

  if((it = opts.find("fconf")) == opts.end()) {
    usage("please provide a feature configuration --fconf...");
  }
  o3fgen::config(it->second);


  /* pruning options */
  bool pruned = false;
  _Atr = NULL;
  _Aval = NULL;
  if((it = opts.find("prune-tmarg")) != opts.end()) {
    pruned = true;
    const bool reattach_commas = (opts.find("fix-commas") != opts.end());
    const string tmarg = it->second; /* save before clobbering it */
    /* look for the validation deps if valdata exists */
    string vmarg;
    if(_valdata != NULL) {
      if((it = opts.find("prune-vmarg")) == opts.end()) {
        usage("--prune-tmarg and --valdata require --prune-vmarg");
      }
      vmarg = it->second;
    }
    /* look for the pruning beam */
    if((it = opts.find("prune-beam")) == opts.end()) {
      usage("--prune-deps requires --prune-beam");
    }
    const double trbeam = strtod(it->second.c_str(), NULL);
    assert(trbeam > 0); assert(trbeam <= 1);
    /* check whether truth is included in training */
    const bool truth = (opts.find("prune-train-truth") != opts.end());
    /* create the active deps */
    _Atr = new active_deps(tmarg.c_str(), trbeam, L,
                           *_trdata, truth, reattach_commas);
    assert(_Atr != NULL);
    if(_valdata != NULL) {
      /* check for a different validation beam */
      double valbeam = trbeam;
      if((it = opts.find("prune-vbeam")) != opts.end()) {
        valbeam = strtod(it->second.c_str(), NULL);
        assert(valbeam > 0); assert(valbeam <= 1);
      }
      /* never include truth on validation data */
      _Aval = new active_deps(vmarg.c_str(), valbeam, L,
                              *_valdata, false, reattach_commas);
      assert(_Aval != NULL);
    } else {
      if((it = opts.find("prune-vbeam")) != opts.end()) {
        usage("--prune-vbeam requires --prune-vmarg");
      }
    }
  } else {
    if(opts.find("prune-beam") != opts.end()) {
      usage("--prune-beam requires --prune-tmarg");
    }
    if(opts.find("fix-commas") != opts.end()) {
      usage("--fix-commas requires --prune-tmarg");
    }
  }


  /* relabel the corpus if necessary */
  if(opts.find("relabel") != opts.end()) {
    const string relfn = _dir + "/trdata.relabeled.gz";
    FILE* const relabeled = gzio::gzopen(relfn.c_str(), "w");
    _trdata->relabel(L, multiroot, _Atr, relabeled);
    gzio::gzclose(relfn.c_str(), relabeled);
  }


  /* allocate the parser and feature generator */
  _P = new o3parser(L, multiroot, pruned, topdown, sibparser);
  assert(_P != NULL);
  _F = NULL;
  if(params != "none") {
    bool averaged = false;
    if(params == "avg")      { averaged = true; }
    else if(params == "lin") { averaged = false; }
    else { usage("unrecognized parameter type: " + params); }
    _F = new o3features(L, averaged);
    assert(_F != NULL);
  }

  _trdata->ordering(trorder);


  /* testing options; these immediately exit after testing */
  if((it = opts.find("testinf-simple")) != opts.end()) {
    testinf::simple(atoi(it->second.c_str()), _P); exit(0);
  }

  if((it = opts.find("testinf-bf")) != opts.end()) {
    testinf::bruteforce(atoi(it->second.c_str()), L); exit(0);
  }

  if((it = opts.find("testinf-tdbu")) != opts.end()) {
    const char* str = it->second.c_str();
    char* endptr = NULL;
    const int maxN = (int)strtoul(str, &endptr, 10);
    assert(endptr > str); assert(maxN > 0);
    assert(*endptr == ',');
    str = endptr + 1;
    const double ratio = strtod(str, NULL);
    testinf::tdbu(maxN, L, ratio); exit(0);
  }

  if((it = opts.find("testinf-pruned")) != opts.end()) {
    const char* str = it->second.c_str();
    char* endptr = NULL;
    const int maxN = (int)strtoul(str, &endptr, 10);
    assert(endptr > str); assert(maxN > 0);
    assert(*endptr == ',');
    str = endptr + 1;
    const double ratio = strtod(str, NULL);
    testinf::pruned(maxN, L, ratio); exit(0);
  }
}



/* construct mapped options */
void o3trainer::mkopts(const vector<string>& args,
                       strmap& opts) const {
  /* read options, some possibly coming from a file */
  const int argc = (int)args.size();
  assert(argc > 0);
  opts[EXEC] = args[0]; /* save the exec file's name */
  for(int i = 1; i < argc; ++i) {
    const char* arg = args[i].c_str();
    /* ignore zero-length arguments */
    if(arg[0] == '\0') { continue; }
    /* arguments are either command-line options beginning with '--'
       or they are filenames containing name=value pairs */
    if(arg[0] == '-' && arg[1] == '-') { /* an option */
      arg += 2;
      /* find the first equals sign */
      int eqidx = 0;
      for(; arg[eqidx] != '\0' && arg[eqidx] != '='; ++eqidx) {}
      if(arg[eqidx] == '\0') { /* no equals sign; unary option */
        opts[arg] = "1"; /* dummy value */
      } else { /* has equals sign; binary option */
        const string name(arg, arg + eqidx); /* the first half of arg */
        const char* const val = arg + eqidx + 1; /* second half */
        opts[name] = val;
      }
    } else { /* a filename */
      read_opts(arg, opts);
    }
  }

  /* print the options to stderr */
  cerr << "O3Trainer : options interpreted as" << endl;
  cerr << opts[EXEC];
  strmap::const_iterator it = opts.begin();
  const strmap::const_iterator iend = opts.end();
  for(; it != iend; ++it) {
    if(it->first != EXEC) {
      cerr << " \\" << endl
           << "  --" << it->first << "='" << it->second << "'";
    }
  }
  cerr << endl;
}



/* read options from a file */
void o3trainer::read_opts(const string& fname, strmap& opts) const {
  FILE* const f = fopen(fname.c_str(), "r");
  if(f == NULL) {
    string mesg("options file not valid: ");
    mesg += fname;
    mesg += " ; ";
    mesg += strerror(errno);
    usage(mesg.c_str());
  }
  char buf[LINELEN];
  buf[0] = '\0';
  while(fgets(buf, LINELEN, f) != NULL) {
    /* skip comments */
    if(buf[0] == '#') { continue; }
    { /* skip whitespace-only lines */
      bool allspace = true;
      for(const char* p = buf; *p != '\0'; ++p) {
        if(!isspace(*p)) { allspace = false; break; }
      }
      if(allspace) { continue; }
    }
    /* read a name=value pair */
    char* p = buf;
    while(isspace(*p)) { ++p; } /* skip leading whitespace */
    const char* const name = p; /* option name starts here */
    /* read characters of name: non-whitespace, ending at space or
       equals or null */
    while(*p != '\0' && *p != '=' && !isspace(*p)) { ++p; }
    assert(p > name); /* no null names */
    const string optname(name, p - name); /* option name ends here */
    while(isspace(*p)) { ++p; } /* skip whitespace */
    if(*p != '=') { /* unary option */
      assert(*p == '\0');
      opts[optname] = "1";
    } else { /* binary option */
      ++p; /* move past equals sign */
      while(isspace(*p)) { ++p; } /* skip whitespace */
      bool quoted = false; /* is the value quoted? */
      if(*p == '"') { quoted = true; ++p; }
      const char* const val = p; /* value starts here */
      while(1) {
        if(quoted) { /* quoted mode: read until another quote is found */
          if(*p == '"') {
            const string optval(val, p - val);
            opts[optname] = optval;
            break;
          } else if(*p == '\0') {
            cerr << "error: unterminated quote" << endl;
            assert(0);
          }
        } else {
          if(isspace(*p) || *p == '\0') {
            const string optval(val, p - val);
            opts[optname] = optval;
            break;
          }
        }
        ++p;
      }
    }
  }
  fclose(f);
}


/* write the options to a file */
void o3trainer::write_opts(const string& fname,
                           const strmap& opts) const {
  FILE* const f = fopen(fname.c_str(), "w");
  if(f == NULL) {
    string mesg("couldn't write to options file: ");
    mesg += fname;
    mesg += " ; ";
    mesg += strerror(errno);
    usage(mesg.c_str());
  }
  /* options file header */
  fprintf(f, "#!%s\n", opts.find(EXEC)->second.c_str());
  fprintf(f, "# Options file automatically generated by O3Trainer;\n");
  fprintf(f, "# passing the name of this file as an argument causes all\n");
  fprintf(f, "# name-value pairs below to be read\n\n");
  strmap::const_iterator it = opts.begin();
  const strmap::const_iterator iend = opts.end();
  for(; it != iend; ++it) {
    /* don't print the execfile name */
    if(it->first == EXEC) { continue; }
    fprintf(f, "%s = \"%s\"\n", it->first.c_str(), it->second.c_str());
  }
  fclose(f);
}



static inline void send_label(const int write_fd,
                              const vector<int>& restrict heads,
                              const vector<int>& restrict labels) {
  const int N = (int)heads.size();
  assert(N == (int)labels.size());
  assert(N > 0);
  safe_write(write_fd, &N, sizeof(int));
  for(int i = 0; i < N; ++i) {
    const int tmph = heads[i];
    const int tmpl = labels[i];
    safe_write(write_fd, &tmph, sizeof(int));
    safe_write(write_fd, &tmpl, sizeof(int));
  }
}

static inline void recv_label(const int read_fd,
                              vector<int>& restrict heads,
                              vector<int>& restrict labels) {
  heads.clear();
  labels.clear();
  int N = 0;
  safe_read(read_fd, &N, sizeof(int));
  assert(N > 0);
  for(int i = 0; i < N; ++i) {
    int tmp = 0;
    safe_read(read_fd, &tmp, sizeof(int));
    heads.push_back(tmp);
    safe_read(read_fd, &tmp, sizeof(int));
    labels.push_back(tmp);
  }
}


/* classify a validation dataset */
void o3trainer::val_classify(corpus* const val, const string& fname,
                             double& ret_accuracy) {
  cerr << MYNAME << " : classifying validation data..." << endl;
  const int Nval = val->N();

  vector<const ex*> examples;
  { /* fill examples with the data points to process */
    val->ordering("det");
    val->start();
    const ex* e = NULL;
    while((e = val->next()) != NULL) { examples.push_back(e); }
    assert(Nval == (int)examples.size());
  }

  int* pipe_read_fds = NULL; /* read ends for parent to rcv data */
  int pipe_write_fd = -1;    /* write end for worker to send data */
  int worker_id = -1;        /* worker id, for reporting */

  /* create start index and step size */
  int start_idx = 0, step_size = 1;
  map<pid_t,int> workers;
  if(_parallel != 0) {
    pipe_read_fds = new int[_parallel];
    /* flush c++ output streams */
    cerr << flush;
    cout << flush;
    /* flush all open stdio output streams */
    fflush(NULL);
    /* workers process an interleaved partition of the examples */
    step_size = _parallel;
    /* initialize a null start index for the parent */
    start_idx = Nval;
    /* initialize each parallel worker process */
    for(int worker = 1; worker <= _parallel; ++worker) {
      /* open a pipe for communication after the iteration */
      int readfd = -1, writefd = -1;
      safe_pipe(readfd, writefd);
      /* fork between parent and worker */
      const pid_t wpid = safe_fork();
      if(wpid == 0) { /* the worker executes this */
        /* claim the write end and discard the read end */
        pipe_write_fd = writefd;
        safe_close(readfd);
        /* redirect stderr to an appropriate log file (lest the
           diagnostic output get mixed) */
        char padnum[10];
        sprintf(padnum, "%03d", worker);
        safe_log_stderr((_dir + "/log.worker" + padnum).c_str(), false);
        /* discard data structures used by the parent */
        delete [] pipe_read_fds; pipe_read_fds = NULL;
        workers.clear();
        /* prepare the data partiton */
        worker_id = worker;
        start_idx = worker - 1;
        CHECK(cout << "worker " << worker << ": ready to work" << endl);
        break; /* begin working */
      } else { /* the parent executes this */
        CHECK(cerr << "parent: spawned worker " << worker << endl);
        /* claim the read end and discard the write end */
        pipe_read_fds[worker - 1] = readfd;
        safe_close(writefd);
        workers[wpid] = worker; /* save the worker pid */
      }
    }
  }



  /* lists of parsed heads and labels */
  vector<vector<int> > H, L;
  ticker64_start();
  /* stride across a subset of examples */
  int cnt = 0;
  for(int i = start_idx; i < Nval; i += step_size) {
    const ex* const e = examples[i];
    const int N = (int)e->heads.size();
    /* prepare the active dependencies if necessary */
    if(_Aval != NULL) {
      assert(_Aval->active_N(e->id()) == N);
      _P->prune(_Aval->active_N(e->id()),
                _Aval->active_num(e->id()),
                _Aval->active_heads(e->id()),
                _Aval->active_labels(e->id()));
    }
    /* add new empty vectors for the next sentence */
    const size_t idx = H.size();
    H.push_back(vector<int>());
    L.push_back(vector<int>());
    /* write the parser's label into these new vectors */
    classify(e, H[idx], L[idx]);
    ticker64(cnt, Nval);
    ++cnt;
    if(_parallel != 0) {
      if((cnt & PARREPINT) == 0) {
        const double percent = 100.0*((double)i)/((double)Nval);
        cout << "worker " << worker_id << ": "
             << fixed << setprecision(2) << percent << "%" << endl;
      }
    }
  }
  ticker64_finish(Nval);



  /* gather results after the iteration */
  if(_parallel != 0) {
    if(pipe_read_fds == NULL) { /* child process */
      /* send the labels */
      CHECK(cout << flush << "worker " << worker_id
            << ": sending labels..." << endl);
      /* send the labels to the parent via the pipe */
      const int nlabels = (int)H.size();
      assert(nlabels == (int)L.size());
      safe_write(pipe_write_fd, &nlabels, sizeof(int));
      for(int i = 0; i < nlabels; ++i) {
        send_label(pipe_write_fd, H[i], L[i]);
      }
      CHECK(cout << flush << "worker " << worker_id
            << ": labels sent!" << endl);
      /* clean up file descriptors */
      safe_close(pipe_write_fd);
      CHECK(cout << flush << "worker " << worker_id
            << ": exiting cleanly" << endl);
      exit(0); /* work complete */
    } else { /* parent process */
      /* before starting, first allocate space for all labels. */
      assert(H.size() == 0); assert(L.size() == 0);
      H.resize(Nval); L.resize(Nval);
      /* for each pipe read end, absorb its labels */
      cerr << MYNAME << " : " << _parallel << " workers:" << flush;
      for(int worker = 1; worker <= _parallel; ++worker) {
        CHECK(cerr << flush << "parent: labels from worker "
              << worker << "..." << endl);
        const int readfd = pipe_read_fds[worker - 1];
        int nlabels = -1;
        safe_read(readfd, &nlabels, sizeof(int));
        assert(nlabels >= 0);
        /* read the params via the worker's pipe */
        int labcnt = 0;
        for(int i = worker - 1; i < Nval; i += step_size) {
          assert(H[i].size() == 0); assert(L[i].size() == 0);
          recv_label(readfd, H[i], L[i]); ++labcnt;
        }
        assert(labcnt == nlabels);
        /* clean up file descriptors */
        safe_close(readfd);
        /* wait for the worker to prevent zombies */
        safe_wait(workers);
        cerr << " " << worker << flush;
      }
      /* clean up */
      CHECK(cerr << flush << "parent: finished" << endl);
      double splittime = 0;
      timer_split(splittime);
      cerr << " [" << timestr(splittime) << "]" << endl;
    }
  }
  if(_parallel != 0) { delete [] pipe_read_fds; }



  /* score the labels and write them to disk */
  double nwords = 0;
  double ncorrect = 0;
  FILE* const f = gzio::gzopen(fname.c_str(), "w");
  for(int i = 0; i < Nval; ++i) {
    const ex* const e = examples[i];
    const int N = (int)e->heads.size();

    /* recover parsed label */
    const vector<int>& restrict heads  = H[i];
    const vector<int>& restrict labels = L[i];

    /* write to disk */
    e->fwrite(f, heads, labels);

    { /* score the labels */
      const vector<int>& true_heads  = e->heads;
      const vector<int>& true_labels = e->labels;
      assert(N == (int)true_heads.size());
      assert(N == (int)true_labels.size());
      assert(N == (int)heads.size());
      assert(N == (int)labels.size());
      for(int j = 0; j < N; ++j) {
        if((true_heads[j]  == heads[j]) &&
           (true_labels[j] == labels[j])) {
          ++ncorrect;
        }
      }
      nwords += N;
    }
  }
  gzio::gzclose(fname.c_str(), f);

  ret_accuracy = 100.0*((double)ncorrect)/((double)nwords);
  cerr << MYNAME << " : validation accuracy = "
       << fixed << setprecision(2) << ret_accuracy << endl;
}
