/* -*-c++-*- */
#ifndef O3TRAINER_H
#define O3TRAINER_H
/* O3Trainer : A general-use training rig for training o3parsers */

#include "O3Parser.h"
#include "O3Features.h"
#include "ActiveDeps.h"
#include "Example.h"
#include "Corpus.h"

#include <stdio.h>

#include <map>
#include <vector>
#include <string>

using std::map;
using std::vector;
using std::string;

class o3trainer {
public:
  virtual ~o3trainer();

  /* nested classes */
  typedef map<string,string> strmap;

  /* initialize this with argv/argc, which are assumed to come from
     the program invocation. */
  void init(const vector<string>& args);

  /* train the model. */
  void train();


protected:
  o3parser* _P;      /* parser */
  o3features* _F;    /* feature generator */
  string _dir;       /* directory for storing model parameters */
  int _ablation;     /* part ablation, if any */

  corpus* _trdata;    /* training examples */
  corpus* _valdata;   /* validation examples */

  active_deps* _Atr;  /* active (pruned) dependencies on training */
  active_deps* _Aval; /* active (pruned) dependencies on valdata */

  o3trainer(const string& name);

  /* print usage information + error message to stderr and exit */
  void usage(const string& mesg = "") const;

  /* called when printing usage information */
  virtual void usage_impl() const = 0;
  /* called at initialization with a map of name-value options */
  virtual void init_impl(const strmap& opts) = 0;
  /* called at the beginning of training */
  virtual void train_start() = 0;
  /* called at the beginning of each iteration */
  virtual void iter_start() = 0;
  /* called at the end of each iteration.  logline and repline will be
     printed to a log file and to stderr, respectively */
  virtual void iter_end(string& logline, string& repline) = 0;
  /* update the model with the given example */
  virtual void process(const ex* const restrict e) = 0;
  /* compute a label for the given example */
  virtual void classify(const ex* const restrict e,
                        vector<int>& restrict ret_h,
                        vector<int>& restrict ret_l) const = 0;


private:
  string _name; /* name for this (for e.g. debug output) */

  int _ordering;      /* exaple ordering */
  unsigned _seed;     /* random seed */
  int _T;             /* number of iterations */
  int _parallel;      /* number of workers */
  bool _writeparams;  /* whether to save parameters to disk */

  /* initialize this using the name-value options in opts */
  void init(const strmap& opts);

  /* set up or merge parallel jobs */
  struct parallel_state;
  void parallel_setup(int& start_idx, int& step_size,
                      parallel_state* const X);
  void parallel_merge(int& start_idx, int& step_size,
                      parallel_state* const X);

  /* convert the argv/argc into map form */
  void mkopts(const vector<string>& args, strmap& opts) const;
  /* load options from/save options to a file */
  void read_opts(const string& fname, strmap& opts) const;
  void write_opts(const string& fname, const strmap& opts) const;

  /* label a corpus using the current model */
  void val_classify(corpus* const val, const string& fname,
                    double& ret_accuracy);
};

#endif /* O3TRAINER_H */
