#include "O3Tree.h"

#include "Macros.h"
#include "O3Active.h"
#include "O3Part.h"
#include "PartTest.h"

#include <string>
#include <iostream>

#include <stdlib.h>
#include <assert.h>

using std::cerr;
using std::endl;

/* lifecycle */
/* safe to declare inline since it's only ever used in this file */
inline o3tree::o3tree(const int ii, const int ll,
                      const int nl, const int nr)
  : _idx(ii), _label(ll), _nlc(nl), _nrc(nr), _p(NULL),
    _lc(nl == 0 ? NULL : new o3tree*[nl]),
    _rc(nr == 0 ? NULL : new o3tree*[nr]) {}

o3tree::~o3tree() {
  for(int i = 0; i < _nlc; ++i) { delete _lc[i]; }
  for(int i = 0; i < _nrc; ++i) { delete _rc[i]; }
  if(_lc != NULL) { delete [] _lc; }
  if(_rc != NULL) { delete [] _rc; }
}


/* pretty-print a tree */
void o3tree::print(const int indent) const {
  const string tab(indent, ' ');
  if(_nlc == 0 && _nrc == 0) {
    cerr << tab << _idx << endl;
    return;
  }
  if(_nlc > 0) {
    cerr << tab << _idx << " left" << endl;
    for(int i = 0; i < _nlc; ++i) {
      _lc[i]->print(indent + 2);
    }
  }
  if(_nrc > 0) {
    cerr << tab << _idx << " right" << endl;
    for(int i = 0; i < _nrc; ++i) {
      _rc[i]->print(indent + 2);
    }
  }
}



/* convert from head/label lists */
const o3tree* o3tree::convert(const vector<int>& restrict p) {
  vector<int> l(p.size(), 0); /* N copies of 0 */
  return convert(p, l);
}

const o3tree* o3tree::convert(const vector<int>& restrict p,
                              const vector<int>& restrict l) {
  assert(p.size() == l.size());
  const int n = (int)p.size();
  const int np1 = n + 1;
  const int nn = np1*np1;
  int* const lc = np1 + (int*)malloc(nn*sizeof(int));
  int* const rc = np1 + (int*)malloc(nn*sizeof(int));
  int* const nl = 1 + (int*)calloc(np1, sizeof(int));
  int* const nr = 1 + (int*)calloc(np1, sizeof(int));

  /* accumulate left and right children */
  for(int i = 0; i < n; ++i) {
    const int h = p[i];
    if(h > i) {
      int* const lch = lc + h*np1;
      const int nlc = nl[h];
      lch[nlc] = i;
      nl[h] = nlc + 1;
    } else {
      int* const rch = rc + h*np1;
      const int nrc = nr[h];
      rch[nrc] = i;
      nr[h] = nrc + 1;
    }
  }

  /* create dep tree nodes */
  o3tree** const arr = 1 + (o3tree**)malloc(np1*sizeof(o3tree*));
  arr[-1] = new o3tree(-1, -1, nl[-1], nr[-1]);
  for(int i = 0; i < n; ++i) {
    arr[i] = new o3tree(i, l[i], nl[i], nr[i]);
  }

  /* assign children in head-outward order */
  for(int i = -1; i < n; ++i) {
    o3tree* const t = arr[i];
    /* the right children are already in head-outward order */
    const int nrc = nr[i];
    int* const rch = rc + i*np1;
    o3tree** const trc = t->_rc;
    for(int j = 0; j < nrc; ++j) {
      o3tree* const c = arr[rch[j]];
      c->_p = t;
      trc[j] = c;
    }
    /* the left children must be reversed */
    const int last = nl[i] - 1;
    int* const lch = lc + i*np1;
    o3tree** const tlc = t->_lc;
    for(int j = last; j >= 0; --j) {
      o3tree* const c = arr[lch[j]];
      c->_p = t;
      tlc[last - j] = c;
    }
    arr[i] = t;
  }

  const o3tree* const root = arr[-1];

  free(lc - np1);
  free(rc - np1);
  free(nl - 1);
  free(nr - 1);
  free(arr - 1);

  return root;
}



/* tree enumeration */

/* check if a parent-list defines a cyclic stucture */
/* DFS states:
   white = unvisited,
   grey = being explored,
   black = completed */
#define WHITE 0
#define GREY  1
#define BLACK 2
bool o3tree::has_cycle(const vector<int>& restrict p) {
  const int N = (int)p.size();
  vector<unsigned char> state(N, WHITE);
  /* walk the path of ancestors starting from each child */
  for(int i = 0; i < N; ++i) {
    if(state[i] == WHITE) {
      /* grey in the path of ancestors, stopping either at root or
         if a non-fresh node is encountered */
      int node = i;
      for(; node != -1 && state[node] == WHITE; node = p[node]) {
        state[node] = GREY;
      }
      /* if the node is root or black, everything is fine */
      if(node == -1 || state[node] == BLACK) {
        /* blacken all nodes on the path */
        for(node = i; node != -1 && state[node] == GREY; node = p[node]) {
          state[node] = BLACK;
        }
      } else { /* the node must be grey */
        CHECK(assert(state[node] == GREY));
        /* encountering a grey node means the ancestor path returns
           to itself; a cycle */
        return true;
      }
    }
  }
  /* check that all nodes are complete */
  CHECK(for(int i = 0; i < N; ++i) { assert(state[i] == BLACK); });
  /* no cycles */
  return false;
}
#undef WHITE
#undef GREY
#undef BLACK

/* check if a parent-list defines a non-projective stucture */
bool o3tree::has_cross(const vector<int>& restrict p) {
  const int N = (int)p.size();
  /* a slow but simple O(n^2) test: for each dep (h,m), make sure
     that every word between h and m has a head between h and m
     (inclusive). */
  for(int m = 0; m < N; ++m) {
    const int h = p[m];
    CHECK(assert(h != m));
    /* redefine as s = start (min) and e = end (max) */
    const int s = (h < m ? h : m);
    const int e = (h < m ? m : h);
    /* enumerate possible descendents in (s,e), which must have
       parents within [s,e] */
    for(int d = s + 1; d < e; ++d) {
      const int hd = p[d];
      /* a demonstrably crossing dependency */
      if(hd < s || hd > e) { return true; }
    }
  }
  return false;
}


/* helper to advance a vector of heads/labels to the next valid
   tree, returning false if the enumeration wraps */
bool o3tree::next_tree_full(const int L, const bool multiroot,
                            vector<int>& restrict p,
                            vector<int>& restrict l,
                            int& restrict nroots) {
  const int N = (int)p.size();
  CHECK(if(multiroot) { assert(nroots >= 1); }
        else          { assert(nroots == 1); });
  /* first enumerate all labelings for the current structure */
  if(L > 1) {
    for(int i = 0; i < N; ++i) {
      ++(l[i]);
      /* if there is no carry, the current lists are okay */
      if(l[i] < L) { return true; }
      /* otherwise, we reset the current bit and the carry
         propagates forward */
      else { l[i] = 0; }
    }
    /* if this loop terminates, it means that we overflowed the
       most-significant digit of the labeling, and the entire
       labeling should be zero at this point */
    CHECK(for(int i = 0; i < N; ++i) { assert(l[i] == 0); });
  }
  /* if we got to this point, either it is an unlabeled model or the
     labels have been enumerated and we must advance the
     structure */
  while(true) {
    for(int i = 0; i < N; ++i) {
      /* keep track of a root that is eliminated */
      if(p[i] == -1) { --nroots; CHECK(assert(nroots >= 0)); }
      ++(p[i]);
      /* immediately step past self-loops */
      if(p[i] == i) { ++(p[i]); }
      /* if there is no carry, we can stop here */
      if(p[i] < N) { break; }
      /* else we must carry forward. */
      else {
        CHECK(assert(p[i] == N));
        /* if we are carrying on the most significant digit then we
           have enumerated every structure and labeling and we
           return false immediately */
        if(i == N - 1) { return false; }
        /* otherwise, reset this digit the lowest value and carry
           forward. */
        else { p[i] = -1; ++nroots; CHECK(assert(nroots <= N)); }
      }
    }
    /* at this point, we have advanced the array of heads but it may
       not define a proper tree.  stop only of various checks are
       satisfied */
    int tmp_nroots = 0;
    for(int i = 0; i < N; ++i) { if(p[i] == -1) { ++tmp_nroots; } }
    CHECK(assert(tmp_nroots == nroots));
    /* it is always an error to have no roots */
    if(nroots == 0) { continue; }
    /* keep trying for a single root if necessary */
    if(!multiroot && nroots > 1) { continue; }
    /* check for cycles */
    if(has_cycle(p)) { continue; }
    /* check for non-projectivity */
    if(has_cross(p)) { continue; }
    /* all tests pass; we can stop here */
    return true;
  }
  cerr << "tree_enum_next() reached unreachable area!" << endl;
  assert(0);
  return false;
}


/* helper to advance a vector of heads/labels to the next valid tree,
   returning false if the enumeration wraps; this version uses only
   active dependencies */
bool o3tree::next_tree_pruned(const int L, const bool multiroot,
                              const o3active& restrict A,
                              vector<int>& restrict p,
                              vector<int>& restrict l,
                              int& restrict nroots,
                              vector<int>& restrict state) {
  const int N = (int)p.size();
  { /* sanity checks */
    assert(N == A.N());
  }
  /* iterate through possible heads selected from pruned arrays */
  while(true) {
    if(state.empty()) { /* set initial state */
      nroots = 0;
      for(int i = 0; i < N; ++i) {
        const backp* const actL = A.lharr(i);
        const backp* const actR = A.rharr(i);
        CHECK(assert(actL[0].raw() > 0 || actR[0].raw() > 0));
        if(actL[0].raw() > 0) {
          state.push_back(1);
          const backp b = actL[1];
          p[i] = b.i_null();
          l[i] = b.l();
          if(p[i] == -1) { ++nroots; }
        } else {
          state.push_back(-1);
          const backp b = actR[1];
          p[i] = b.i_null();
          l[i] = b.l();
          if(p[i] == -1) { ++nroots; }
        }
      }
    } else { /* increment the current state */
      for(int i = 0; i < N; ++i) {
        const backp* const actL = A.lharr(i);
        const backp* const actR = A.rharr(i);
        const uint16_t nL = actL[0].raw();
        const uint16_t nR = actR[0].raw();
        /* advance to the next pruned dep via state */
        bool carry = false;
        int nextst = state[i];
        if(nextst > 0) {
          ++nextst;
          if(nextst > nL) {
            if(nR == 0) { nextst = 1; carry = true; }
            else        { nextst = -1; }
          }
        } else {
          --nextst;
          if(nextst < -nR) {
            if(nL == 0) { nextst = -1; }
            else        { nextst = 1; }
            carry = true;
          }
        }
        /* set the head and label of the next state */
        CHECK(assert(nextst != 0));
        state[i] = nextst; /* save updated state */
        const backp b = (nextst > 0 ? actL[nextst] : actR[-nextst]);
        /* keep track of a root that is eliminated */
        if(p[i] == -1) { --nroots; CHECK(assert(nroots >= 0)); }
        p[i] = b.i_null();
        l[i] = b.l();
        if(p[i] == -1) { ++nroots; CHECK(assert(nroots <= N)); }
        /* stop if no carrying */
        if(!carry) { break; }
        /* can't carry past the end */
        else if(i == N - 1) { return false; }
      }
    }
    /* at this point, we have advanced the array of heads but it may
       not define a proper tree.  stop only of various checks are
       satisfied */
    int tmp_nroots = 0;
    for(int i = 0; i < N; ++i) { if(p[i] == -1) { ++tmp_nroots; } }
    CHECK(assert(tmp_nroots == nroots));
    /* it is always an error to have no roots */
    if(nroots == 0) { continue; }
    /* keep trying for a single root if necessary */
    if(!multiroot && nroots > 1) { continue; }
    /* check for cycles */
    if(has_cycle(p)) { continue; }
    /* check for non-projectivity */
    if(has_cross(p)) { continue; }
    /* all tests pass; we can stop here */
    return true;
  }
  cerr << "tree_enum_next() reached unreachable area!" << endl;
  assert(0);
  return false;
}
