/* -*-c++-*- */
#ifndef O3TREE_H
#define O3TREE_H
/* O3Tree : A pointer-list representation of a labeled dependency
   tree.  Trees cannot be created normally, they can only be
   constructed by a static helper method that takes a head-list
   representation of a tree and converts it to a pointer-list tree.
   When finished, the deptree* should be deallocated by calling delete
   on its root node.  O3Tree also provides facilities for enumerating
   projective trees, scoring trees with a part-scorer, and creating
   lists of parts appearing in a tree. */

#include "Macros.h"

#include <vector>

using std::vector;

class o3active;
class o3part;

class o3tree {
public:
  ~o3tree(); /* public for deletion purposes */

  /* accessors: head/label/parent/children */
  int idx() const;
  int label() const;
  int nlc() const;
  int nrc() const;
  const o3tree* p() const;
  const o3tree* lc(const int i) const;
  const o3tree* rc(const int i) const;

  /* pretty-print a tree */
  void print(const int indent = 0) const;

  /* convert a labeled tree represented as head-list and label-list */
  static const o3tree* convert(const vector<int>& restrict p,
                               const vector<int>& restrict l);
  /* convert an unlabeled tree, setting l=0 for all deps */
  static const o3tree* convert(const vector<int>& restrict p);

  /* facilities for enumerating (projective) trees */
  static bool next_tree_full(const int L, const bool multiroot,
                             vector<int>& restrict p,
                             vector<int>& restrict l,
                             int& restrict nroots);
  static bool next_tree_pruned(const int L, const bool multiroot,
                               const o3active& restrict A,
                               vector<int>& restrict p,
                               vector<int>& restrict l,
                               int& restrict nroots,
                               vector<int>& restrict state);

  /* score the parts of a tree */
  template<class PartScorer>
  static double score(const bool multiroot, const bool sibparser,
                      const o3tree* const restrict root,
                      PartScorer& restrict S);
  template<class PartScorer>
  static double score(const bool multiroot, const bool sibparser,
                      const vector<int>& restrict p,
                      const vector<int>& restrict l,
                      PartScorer& restrict S);

private:
  const int _idx;     /* index of this in the sentence (-1 for root) */
  const int _label;   /* label between this and _p */
  const int _nlc;     /* number of left children */
  const int _nrc;     /* number of right children */
  o3tree* _p;         /* parent pointer (NULL for root) */
  o3tree** const _lc; /* array of left children */
  o3tree** const _rc; /* array of right children */

  /* private constructor; used by the conversion methods */
  o3tree(const int idx, const int lab, const int nl, const int nr);

  /* recursive tree-walking scoring procedures */
  template<class PartScorer>
  static double walk_gch(const idx_t g,
                         const o3tree* const restrict node,
                         PartScorer& restrict S);
  template<class PartScorer>
  static double walk_sib(const idx_t g,
                         const o3tree* const restrict node,
                         PartScorer& restrict S);

  /* helper predicates for tree enumeration */
  static bool has_cycle(const vector<int>& restrict p);
  static bool has_cross(const vector<int>& restrict p);
};


/* inlined accessors */
inline int o3tree::idx() const                     { return _idx; }
inline int o3tree::label() const                   { return _label; }
inline int o3tree::nlc() const                     { return _nlc; }
inline int o3tree::nrc() const                     { return _nrc; }
inline const o3tree* o3tree::p() const             { return _p; }
inline const o3tree* o3tree::lc(const int i) const { return _lc[i]; }
inline const o3tree* o3tree::rc(const int i) const { return _rc[i]; }

/* alternate invocation for convenience */
template<class PartScorer> inline double
o3tree::score(const bool multiroot, const bool sibparser,
              const vector<int>& restrict p,
              const vector<int>& restrict l,
              PartScorer& restrict S) {
  const o3tree* const restrict root = convert(p, l);
  const double sc = score(multiroot, sibparser, root, S);
  delete root;
  return sc;
}

#ifndef EXPLICIT_TEMPLATES
#include "O3Tree.tcc"
#endif /* EXPLICIT_TEMPLATES */

#endif /* O3TREE_H */
