#include "PadSent.h"

#include "Mappings.h"

#include <assert.h>
#include <stdlib.h>

static inline const word_t* copywords(const vector<int>& restrict v) {
  const int N = (int)v.size();
  assert(N > 0);
  /* allocate extra room for sentence-exterior null tokens */
  word_t* arr = (word_t*)malloc((N + 5)*sizeof(word_t));
  arr += 3; /* valid indices are now -3 ... N+1 */
  for(int i = 0; i < N; ++i) { arr[i] = WCONV(v[i]); }
  /* null word tokens */
  arr[-3] = arr[-2] = arr[-1] = arr[N] = arr[N + 1] = WNULL;
  return arr;
}

static inline const tag_t* copytags(const vector<int>& restrict v) {
  const int N = (int)v.size();
  assert(N > 0);
  /* allocate extra room for sentence-exterior null tokens */
  tag_t* arr = (tag_t*)malloc((N + 5)*sizeof(tag_t));
  arr += 3; /* valid indices are now -3 ... N+1 */
  for(int i = 0; i < N; ++i) { arr[i] = TCONV(v[i]); }
  /* null tag tokens */
  arr[-3] = arr[-2] = arr[-1] = arr[N] = arr[N + 1] = TNULL;
  return arr;
}


static inline const int* mkquotelev(const word_t* const restrict words,
                                    const int N) {
  assert(N > 0);
  /* allocate extra room for sentence-exterior null tokens */
  int* arr = (int*)malloc((N + 5)*sizeof(int));
  arr += 3; /* valid indices are now -3 ... N+1 */
  /* check whether each word opens or closes a level */
  int level = 0;
  for(int i = 0; i < N; ++i) {
    /* NB: brackets are included at the level they create */
    if(mappings::isoquote(words[i])) { ++level; }
    arr[i] = level;
    if(mappings::iscquote(words[i])) { --level; }
  }
  /* normalize the levels so they are non-negative (this can easily
     happen with multi-sentence quotations) */
  int minlevel = 0;
  for(int i = 0; i < N; ++i) {
    if(arr[i] < minlevel) { minlevel = arr[i]; }
  }
  for(int i = 0; i < N; ++i) { arr[i] -= minlevel; }
  /* the null tokens are always at the base level */
  arr[-3] = arr[-2] = arr[-1] = arr[N] = arr[N + 1] = 0;
  return arr;
}



static inline const int* mkparenlev(const word_t* const restrict words,
                                    const int N) {
  assert(N > 0);
  /* allocate extra room for sentence-exterior null tokens */
  int* arr = (int*)malloc((N + 5)*sizeof(int));
  arr += 3; /* valid indices are now -3 ... N+1 */
  /* check whether each word opens or closes a level */
  int level = 0;
  for(int i = 0; i < N; ++i) {
    /* NB: brackets are included at the level they create */
    if(mappings::isoparen(words[i])) { ++level; }
    arr[i] = level;
    if(mappings::iscparen(words[i])) { --level; }
  }
  /* normalize the levels so they are non-negative (this can easily
     happen with multi-sentence quotations) */
  int minlevel = 0;
  for(int i = 0; i < N; ++i) {
    if(arr[i] < minlevel) { minlevel = arr[i]; }
  }
  for(int i = 0; i < N; ++i) { arr[i] -= minlevel; }
  /* the null tokens are always at the base level */
  arr[-3] = arr[-2] = arr[-1] = arr[N] = arr[N + 1] = 0;
  return arr;
}



padsent::padsent(const vector<int>& restrict w,
                 const vector<int>& restrict t)
  : N((int)w.size()),
    words(copywords(w)),
    tags(copytags(t)),
    quotelev(mkquotelev(words, N)),
    parenlev(mkparenlev(words, N)) {
  assert(w.size() == t.size());
}
padsent::~padsent() {
  free((word_t*)(words - 3));
  free((tag_t*)(tags - 3));
  free((int*)(quotelev - 3));
  free((int*)(parenlev - 3));
}
