/* -*-c++-*- */
#ifndef PARALLEL_UTIL_H
#define PARALLEL_UTIL_H

/* error handling */
#include <errno.h>
#include <string.h>
/* fork/pipe/close/dup2/write/read */
#include <unistd.h>
/* wait */
#include <sys/wait.h>
#include <sys/types.h>
/* open/close */
#include <sys/stat.h>
#include <fcntl.h>
/* gettimeofday */
#include <sys/time.h>
#include <time.h>

#include <iostream>
#include <map>

using std::cerr;
using std::endl;



/* helper to fork with error checking */
static inline pid_t safe_fork() {
  const pid_t ret = fork();
  if(ret == -1) {
    cerr << "ParallelUtil : fork() failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  return ret;
}

/* helper to wait for a child with error checking */
static inline pid_t safe_wait(map<pid_t,int>& workers) {
  int status = -1;
  const pid_t ret = wait(&status);
  if(ret == -1) {
    cerr << "ParallelUtil : wait() failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  if(workers.find(ret) == workers.end()) {
    cerr << "ParallelUtil : wait() returned unknown PID "
         << ret << endl;
    assert(0);
  }
  const int worker = workers.find(ret)->second;
  if(!WIFEXITED(status)) {
    cerr << "ParallelUtil : worker " << worker << " (" << ret
         << ") did not exit normally" << endl;
    assert(0);
  }
  const int exitstatus = WEXITSTATUS(status);
  if(exitstatus != 0) {
    cerr << "ParallelUtil : worker " << worker << " (" << ret
         << ") exited with status " << exitstatus << endl;
    assert(0);
  }
  const size_t nerased = workers.erase(ret);
  assert(nerased == 1);
  return ret;
}

/* helper to redirect stderr to a log file with error checking */
static inline void safe_log_stderr(const char* const fname,
                                   const bool clobber) {
  /* make stdout <-- stderr */
  const int ret21 = dup2(2, 1);
  if(ret21 == -1) {
    cerr << "ParallelUtil : dup2(2 -> 1) failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  /* open an output file (mode "a") */
  const int fd = open(fname,
                      (O_WRONLY |
                       O_CREAT |
                       (clobber ? O_TRUNC : O_APPEND)),
                      0644);
  if(fd == -1) {
    cerr << "ParallelUtil : open(" << fname << ") failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  /* copy the output file onto stderr */
  const int ret = dup2(fd, 2);
  if(ret == -1) {
    cerr << "ParallelUtil : dup2(" << fname << ") failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  if(ret != 2) {
    cerr << "ParallelUtil : dup2(" << fname << ") returned: "
         << ret << " instead of 2; "
         << strerror(errno) << endl;
    assert(0);
  }
  /* close the original descriptor */
  const int retc = close(fd);
  if(retc == -1) {
    cerr << "ParallelUtil : close(" << fname << ") failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  /* now, stdout is the original stderr, and stderr redirects to a
     file */
}

/* helper to open a pipe with error checking */
static inline void safe_pipe(int& readfd, int& writefd) {
  int fds[2];
  fds[0] = fds[1] = -1;
  const int ret = pipe(fds);
  if(ret == -1) {
    cerr << "ParallelUtil : pipe() failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  readfd = fds[0];
  writefd = fds[1];
}

/* helper to close with error checking */
static inline void safe_close(const int fd) {
  const int ret = close(fd);
  if(ret == -1) {
    cerr << "ParallelUtil : close() failed: "
         << strerror(errno) << endl;
    assert(0);
  }
}

/* helper to read from a file with error checking */
static inline void safe_read(const int fd, void* const buf,
                             const size_t nbytes) {
  const ssize_t nread = read(fd, buf, nbytes);
  if(nread == -1) {
    cerr << "ParallelUtil : read() failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  if(nread != (ssize_t)nbytes) {
    cerr << "ParallelUtil : read() didn't get " << nbytes
         << " bytes: " << strerror(errno) << endl;
    assert(0);
  }
}

/* helper to write from a file with error checking */
static inline void safe_write(const int fd, const void* const buf,
                              const size_t nbytes) {
  const ssize_t nwrite = write(fd, buf, nbytes);
  if(nwrite == -1) {
    cerr << "ParallelUtil : write() failed: "
         << strerror(errno) << endl;
    assert(0);
  }
  if(nwrite != (ssize_t)nbytes) {
    cerr << "ParallelUtil : write() didn't put " << nbytes
         << " bytes: " << strerror(errno) << endl;
    assert(0);
  }
}

#endif /* PARALLEL_UTIL_H */
