/* -*-c++-*- */
#ifndef PART_SCORE_CACHE_H
#define PART_SCORE_CACHE_H
/* PartScoreCache : A utility that wraps a part-scorer with a
   hashtable to short-circuit score computation.  The cache is
   templated on the type of the underlying scorer. */

#include "Bits.h"
#include "Macros.h"
#include "O3Part.h"
#include "Hashtable.h"

#include <assert.h>

#include <string>

using std::string;

template<class PartScorer> class part_score_cache;


/* a non-templated base class */
class part_score_cache_base {
public:
  /* produce a human-readable summary of cache statistics */
  string report() const;

  /* the following are of interest for collecting statistics */
#ifdef CAUTIOUS
protected:
  part_score_cache_base();
  /* clear the counts */
  void clear_counts() const;

  /* define a count of the number of times each scorer is called and
     the number of cache hits */
#define VARS(type)                              \
  mutable uint32_t _Ncalls_ ## type;            \
  mutable uint32_t _Nhits_ ## type
  VARS(dep1); VARS(dep2); VARS(dep3);
  VARS(sib1); VARS(sib2);
  VARS(gci1); VARS(gci2);
  VARS(gco1); VARS(gco2);
  VARS(gcis); VARS(gcos); VARS(sibs);
#undef VARS
#endif /* CAUTIOUS */
};

template<class PartScorer>
class part_score_cache : public part_score_cache_base {
public:
  /* normal scoring methods */
  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double sib1(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_hc) const;
  double sib2(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_hc) const;
  double gci1(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gci2(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gco1(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gco2(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gcis(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const;
  double gcos(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const;
  double sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_hc, const lab_t l_hs) const;

  part_score_cache(const PartScorer& restrict S,
                   o3part::dmap& restrict cache);
  ~part_score_cache();

  void clear() const;

private:
  class fac;
  const PartScorer& restrict _S; /* the underlying scorer */
  o3part::dmap& restrict _cache; /* the cache of used scores */
};






/*** inlined implementation ***/

template<class PartScorer> inline part_score_cache<PartScorer>::
part_score_cache(const PartScorer& restrict S,
                 o3part::dmap& restrict cache)
  : _S(S), _cache(cache) {
  assert(_cache.size() == 0);
}
template<class PartScorer> inline part_score_cache<PartScorer>::
~part_score_cache() {}
// ~part_score_cache() { _cache.clear(); }

/* part-production factory */
template<class PartScorer>
class part_score_cache<PartScorer>::fac : protected o3part {
  friend class part_score_cache;
};


/* consult the cache and then return */
#define CACHED1(type)                                                   \
  CHECK(++_Ncalls_ ## type);                                            \
  const o3part r = fac::mk ## type(i1, i2, l1);                         \
  const o3part::dmap::entry* const restrict e = _cache.pget(r);         \
  CHECK(_Nhits_ ## type += ((e != NULL) ? 1 : 0));                      \
  return ((e != NULL) ? e->val                                          \
          : _cache.mget(r, _S.type(i1, i2, l1)))
template<class PartScorer> inline double part_score_cache<PartScorer>::
dep1(const idx_t i1, const idx_t i2, const lab_t l1) const {
  CACHED1(dep1);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
dep2(const idx_t i1, const idx_t i2, const lab_t l1) const {
  CACHED1(dep2);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
dep3(const idx_t i1, const idx_t i2, const lab_t l1) const {
  CACHED1(dep3);
}
#undef CACHED1

#define CACHED2(type)                                                   \
  CHECK(++_Ncalls_ ## type);                                            \
  const o3part r = fac::mk ## type(i1, i2, i3, l1, l2);                 \
  const o3part::dmap::entry* const restrict e = _cache.pget(r);         \
  CHECK(_Nhits_ ## type += ((e != NULL) ? 1 : 0));                      \
  return ((e != NULL) ? e->val                                          \
          : _cache.mget(r, _S.type(i1, i2, i3, l1, l2)))
template<class PartScorer> inline double part_score_cache<PartScorer>::
sib1(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  CACHED2(sib1);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
sib2(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  CACHED2(sib2);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
gci1(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  CACHED2(gci1);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
gci2(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  CACHED2(gci2);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
gco1(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  CACHED2(gco1);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
gco2(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  CACHED2(gco2);
}
#undef CACHED2

/* third-order parts are invoked once by the parser definition, so
   there's no need to cache them */
#ifdef CAUTIOUS
#define CACHED3(type)                                                   \
  CHECK(++_Ncalls_ ## type);                                            \
  const o3part r = fac::mk ## type(i1, i2, i3, i4, l1, l2, l3);         \
  const o3part::dmap::entry* const restrict e = _cache.pget(r);         \
  CHECK(_Nhits_ ## type += ((e != NULL) ? 1 : 0));                      \
  return ((e != NULL) ? e->val                                          \
          : _cache.mget(r, _S.type(i1, i2, i3, i4, l1, l2, l3)))
#else /* CAUTIOUS */
#define CACHED3(type) return _S.type(i1, i2, i3, i4, l1, l2, l3)
#endif /* CAUTIOUS */
template<class PartScorer> inline double part_score_cache<PartScorer>::
gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
     const lab_t l1, const lab_t l2, const lab_t l3) const {
  CACHED3(gcis);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
     const lab_t l1, const lab_t l2, const lab_t l3) const {
  CACHED3(gcos);
}
template<class PartScorer> inline double part_score_cache<PartScorer>::
sibs(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
     const lab_t l1, const lab_t l2, const lab_t l3) const {
  CACHED3(sibs);
}
#undef CACHED3


template<class PartScorer> inline void part_score_cache<PartScorer>::
clear() const {
#ifdef CAUTIOUS
  clear_counts();
#endif /* CAUTIOUS */
  _cache.clear();
}

#endif /* PART_SCORE_CACHE_H */
