/* -*-c++-*- */
#ifndef PART_SCORE_LOSS_H
#define PART_SCORE_LOSS_H
/* PartScoreLoss : A utility that augments a part-scorer with loss
   terms on the first-order parts, as in a max-margin markov network. */

#include "Bits.h"
#include "Macros.h"
#include "Example.h"

#include <assert.h>

/* forward declaration */
template<class PartScorer> class part_score_loss;
/* this is a convenient shorthand that wraps a scorer */
template<class PartScorer> static inline part_score_loss<PartScorer>
addloss(const PartScorer& restrict S,
        const ex* const restrict e,
        const double loss) {
  return part_score_loss<PartScorer>(S, e->heads, e->labels, loss);
}


/* the actual class itself */
template<class PartScorer>
class part_score_loss {
public:
  double dep1(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep2(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double dep3(const idx_t h, const idx_t m, const lab_t l_hm) const;
  double sib1(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_hc) const;
  double sib2(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_hc) const;
  double gci1(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gci2(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gco1(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gco2(const idx_t h, const idx_t m, const idx_t c,
              const lab_t l_hm, const lab_t l_mc) const;
  double gcis(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const;
  double gcos(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_mc, const lab_t l_ms) const;
  double sibs(const idx_t h, const idx_t m, const idx_t c, const idx_t s,
              const lab_t l_hm, const lab_t l_hc, const lab_t l_hs) const;

  part_score_loss(const PartScorer& restrict S,
                  const vector<int>& restrict heads,
                  const vector<int>& restrict labels,
                  const double loss);
  ~part_score_loss();

  void clear() const;

private:
  const PartScorer& restrict _S;  /* the underlying scorer */
  const vector<int>& restrict _h; /* the correct heads */
  const vector<int>& restrict _l; /* the correct labels */
  const double _loss;             /* the amount of loss to add */
};






/*** inlined implementation ***/

template<class PartScorer> inline part_score_loss<PartScorer>::
part_score_loss(const PartScorer& restrict S,
                const vector<int>& restrict heads,
                const vector<int>& restrict labels,
                const double loss)
  : _S(S), _h(heads), _l(labels), _loss(loss) {
  assert(loss > 0);
  assert(_h.size() == _l.size());
  assert(_h.size() > 0);
}
template<class PartScorer> inline part_score_loss<PartScorer>::
~part_score_loss() {}


template<class PartScorer> inline double part_score_loss<PartScorer>::
dep1(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return _S.dep1(h, m, l_hm) +
    (((_h[m] == h) && (_l[m] == l_hm)) ? 0.0 : _loss);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
dep2(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return _S.dep2(h, m, l_hm) +
    (((_h[m] == h) && (_l[m] == l_hm)) ? 0.0 : _loss);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
dep3(const idx_t h, const idx_t m, const lab_t l_hm) const {
  return _S.dep3(h, m, l_hm) +
    (((_h[m] == h) && (_l[m] == l_hm)) ? 0.0 : _loss);
}

template<class PartScorer> inline double part_score_loss<PartScorer>::
sib1(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  return _S.sib1(i1, i2, i3, l1, l2);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
sib2(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  return _S.sib2(i1, i2, i3, l1, l2);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
gci1(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  return _S.gci1(i1, i2, i3, l1, l2);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
gci2(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  return _S.gci2(i1, i2, i3, l1, l2);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
gco1(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  return _S.gco1(i1, i2, i3, l1, l2);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
gco2(const idx_t i1, const idx_t i2, const idx_t i3,
     const lab_t l1, const lab_t l2) const {
  return _S.gco2(i1, i2, i3, l1, l2);
}

template<class PartScorer> inline double part_score_loss<PartScorer>::
gcis(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
     const lab_t l1, const lab_t l2, const lab_t l3) const {
  return _S.gcis(i1, i2, i3, i4, l1, l2, l3);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
gcos(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
     const lab_t l1, const lab_t l2, const lab_t l3) const {
  return _S.gcos(i1, i2, i3, i4, l1, l2, l3);
}
template<class PartScorer> inline double part_score_loss<PartScorer>::
sibs(const idx_t i1, const idx_t i2, const idx_t i3, const idx_t i4,
     const lab_t l1, const lab_t l2, const lab_t l3) const {
  return _S.sibs(i1, i2, i3, i4, l1, l2, l3);
}

#endif /* PART_SCORE_LOSS_H */
