#include "TestInf.h"

#include "MathUtil.h"
#include "Timer.h"
#include "PartScoreTree.h"
#include "PartScoreRandom.h"
#include "PartScoreCache.h"
#include "O3Parser.h"
#include "O3BruteForce.h"
#include "O3Tree.h"

#include <iostream>
#include <iomanip>

using std::cerr;
using std::endl;
using std::flush;
using std::fixed;
using std::scientific;
using std::setprecision;

#define bfname(B)                               \
  "bf:L" << B->L() << ","                       \
  << (B->multiroot() ? "mr" : "sr") << ","      \
  << (B->sibparser() ? "S" : "G")
#define pname(P)                                                \
  "par:L" << P->L() << "," << (P->multiroot() ? "mr" : "sr")    \
  << "," << (P->pruned() ? "+p" : "-p")                         \
  << "," << (P->topdown() ? "td" : "bu")                        \
  << "," << (P->sibparser() ? "S" : "G")


template<class PartScorer>
static double check_tree(const vector<int>& restrict h1,
                         const vector<int>& restrict l1,
                         const vector<int>& restrict h2,
                         const vector<int>& restrict l2,
                         const o3parser* const restrict P,
                         PartScorer& restrict S) {
  const int N = (int)h1.size();
  bool correct = true;
  if(h1.size() == l1.size() &&
     h2.size() == l2.size() &&
     h1.size() == h2.size()) {
    for(int i = 0; i < N; ++i) {
      if(h1[i] != h2[i] || l1[i] != l2[i]) {
        correct = false; break;
      }
    }
  } else { correct = false; }
  if(!correct) {
    cerr << "tree mismatch:" << endl;
    for(int i = 0; i < N; ++i) {
      cerr << "  " << h1[i] << " " << l1[i]
           << " " << h2[i] << " " << l2[i] << endl;
    }
    assert(0);
  }
  o3part::vec y1;
  o3part::vec y2;
  o3part::parts(P->L(), P->multiroot(), P->sibparser(), h1, l1, y1);
  o3part::parts(P->L(), P->multiroot(), P->sibparser(), h2, l2, y2);
  const double scorey1 = o3part::score(y1, S);
  const double scorey2 = o3part::score(y2, S);
  /* check consistency of tree-scoring methods */
  assert(reldiff(scorey1, scorey2) <= O3PARSER_TOL);
  /* return average score */
  return (scorey1 + scorey2)*0.5;
}


static void check_parts(const vector<o3part>& restrict y1,
                        const vector<o3part>& restrict y2) {
  bool correct = true;
  const int YR = (int)y1.size();
  o3part::bmap parts;
  o3part::bmap::iterator it;
  if(y1.size() == y2.size()) {
    for(int i = 0; i < YR; ++i) {
      if(parts.has(y1[i])) {
        cerr << "check_parts(): duplicate part in y1" << endl;
        cerr << "  " << y1[i].tostring() << endl;
        correct = false; break;
      }
      parts.ins(y1[i], true);
    }
  } else { correct = false; }
  if(correct) {
    parts.clear();
    for(int i = 0; i < YR; ++i) {
      if(parts.has(y2[i])) {
        cerr << "check_parts(): duplicate part in y2" << endl;
        cerr << "  " << y2[i].tostring() << endl;
        correct = false; break;
      }
      parts.ins(y2[i], true);
    }
  }
  if(correct) {
    assert(parts.size() == y2.size());
    for(int i = 0; i < YR; ++i) {
      if(!parts.has(y1[i])) {
        cerr << "check_parts(): part mismatch between y1 and y2" << endl;
        correct = false; break;
      }
      parts.del(y1[i]);
    }
    if(correct) {
      assert(parts.size() == 0);
    }
  }
  if(!correct) {
    cerr << "y1 = " << endl;
    for(int i = 0; i < (int)y1.size(); ++i) {
      cerr << "  " << y1[i].tostring() << endl;
    }
    cerr << "y2 = " << endl;
    for(int i = 0; i < (int)y2.size(); ++i) {
      cerr << "  " << y2[i].tostring() << endl;
    }
    assert(0);
  }
}





void testinf::simple(const int Nmax, o3parser* const restrict P) {
  const int L = P->L();
  for(int N = 1; N <= Nmax; ++N) {
    { /* set the pruning arrays */
      int* const active_num = new int[N];
      const int** const active_heads = new const int*[N];
      const int** const active_labels = new const int*[N];
      for(int m = 0; m < N; ++m) {
        active_num[m] = N*L;
        int* const heads = new int[N*L];
        int* const labels = new int[N*L];
        int k = 0;
        for(int h = -1; h < N; ++h) {
          if(h == m) { continue; }
          for(int l = 0; l < L; ++l) {
            heads[k] = h;
            labels[k] = l;
            ++k;
          }
        }
        active_heads[m]  = heads;
        active_labels[m] = labels;
      }
      /* pass to the parser */
      if(P->pruned()) {
        P->prune(N, active_num, active_heads, active_labels);
      }
      /* safe to discard now */
      for(int m = 0; m < N; ++m) {
        delete [] active_heads[m];
        delete [] active_labels[m];
      }
      delete [] active_heads;
      delete [] active_labels;
      delete [] active_num;
    }

    cerr << "testinf::simple(): N=" << N << endl;
    cerr << pname(P) << endl;
    /* easy test cases */

    { /* linear left to right */
      vector<int> tgt_heads;
      vector<int> tgt_labels;
      for(int i = 0; i < N; ++i) {
        tgt_heads.push_back(i - 1);
        tgt_labels.push_back(i % L);
      }
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  lin -> score = " << score << endl;
    }

    { /* linear right to left */
      vector<int> tgt_heads;
      vector<int> tgt_labels;
      for(int i = 0; i < N; ++i) {
        tgt_heads.push_back(i + 1);
        tgt_labels.push_back((i + 1) % L);
      }
      tgt_heads[N - 1] = -1;
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  lin <- score = " << score << endl;
    }

    { /* left super root */
      vector<int> tgt_heads;
      vector<int> tgt_labels;
      for(int i = 0; i < N; ++i) {
        tgt_heads.push_back(0);
        tgt_labels.push_back((i + 2) % L);
      }
      tgt_heads[0] = -1;
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  root0 score = " << score << endl;
    }

    { /* right super root */
      vector<int> tgt_heads;
      vector<int> tgt_labels;
      for(int i = 0; i < N; ++i) {
        tgt_heads.push_back(N - 1);
        tgt_labels.push_back((i + 3) % L);
      }
      tgt_heads[N - 1] = -1;
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  rootN score = " << score << endl;
    }

    { /* center super root */
      vector<int> tgt_heads;
      vector<int> tgt_labels;
      const int C = N/2;
      for(int i = 0; i < N; ++i) {
        tgt_heads.push_back(C);
        tgt_labels.push_back((i + 4) % L);
      }
      tgt_heads[C] = -1;
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  rootC score = " << score << endl;
    }

    { /* groups of 5 */
      vector<int> tgt_heads;
      vector<int> tgt_labels;
      {
        int i;
        const int rem = (N % 5);
        for(i = 0; i < rem; ++i) {
          tgt_heads.push_back(N - 1);
        }
        for(; i < N; i += 5) {
          tgt_heads.push_back(i + 2);
          tgt_heads.push_back(i + 2);
          tgt_heads.push_back(N - 1);
          tgt_heads.push_back(i + 2);
          tgt_heads.push_back(i + 2);
        }
        assert(i == N);
        tgt_heads[N - 1] = -1;
        for(i = 0; i < N; ++i) {
          tgt_labels.push_back((i + 5) % L);
        }
      }
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  group5 score = " << score << endl;
    }

    { /* zigzag */
      vector<int> tgt_heads;

      vector<int> tgt_labels;
      for(int i = 0; i < N; ++i) {
        tgt_heads.push_back(-1);
        tgt_labels.push_back((i + 6) % L);
      }
      int start = 0;
      int end = N - 1;
      while(start < end) {
        tgt_heads[start] = end;
        --end;
        if(start == end) { break; }
        tgt_heads[end] = start;
        ++start;
      }
      const part_score_tree S(N, L, P->sibparser(), tgt_heads, tgt_labels);
      vector<int> heads;
      vector<int> labels;
      const double score = P->argmax(N, S, heads, labels);
      const double score2 =
        check_tree(heads, labels, tgt_heads, tgt_labels, P, S);
      assert(reldiff(score, score2) <= O3PARSER_TOL);
      cerr << "  zigzag score = " << score << endl;
    }
  }
}


void testinf::bruteforce(const int Nmax, const int L) {
  /* set up arrays of parsers */
  const int NP = 12;
  o3parser* parsers[NP] = {
    /* bottom-up grandchild: multi/single root; pruned/non-pruned */
    new o3parser(L, true,  true,  false, false),
    new o3parser(L, false, true,  false, false),
    new o3parser(L, true,  false, false, false),
    new o3parser(L, false, false, false, false),
    /* bottom-up sibling: multi/single root; pruned/non-pruned */
    new o3parser(L, true,  true,  false, true),
    new o3parser(L, false, true,  false, true),
    new o3parser(L, true,  false, false, true),
    new o3parser(L, false, false, false, true),
    /* top-down grandchild: multi/single root; always pruned */
    new o3parser(L, true,  true,  true,  false),
    new o3parser(L, false, true,  true,  false),
    /* top-down sibling: multi/single root; always pruned */
    new o3parser(L, true,  true,  true,  true),
    new o3parser(L, false, true,  true,  true)
  };
  /* brute-force parsers with matched characteristics */
  o3bruteforce* bfenums[NP] = {
    /* grandchild bruteforce matched to bottom-up */
    new o3bruteforce(L, true,  false, false),
    new o3bruteforce(L, false, false, false),
    new o3bruteforce(L, true,  false, false),
    new o3bruteforce(L, false, false, false),
    /* sibling bruteforce matched to bottom-up */
    new o3bruteforce(L, true,  false, true),
    new o3bruteforce(L, false, false, true),
    new o3bruteforce(L, true,  false, true),
    new o3bruteforce(L, false, false, true),
    /* same for top-down */
    new o3bruteforce(L, true,  false, false),
    new o3bruteforce(L, false, false, false),
    new o3bruteforce(L, true,  false, true),
    new o3bruteforce(L, false, false, true)
  };
  o3part::dmap score_cache;
  for(int N = 1; N <= Nmax; ++N) {
    cerr << "testinf::bruteforce(): N=" << N << endl;
    { /* set the pruning arrays */
      int* const active_num = new int[N];
      const int** const active_heads = new const int*[N];
      const int** const active_labels = new const int*[N];
      for(int m = 0; m < N; ++m) {
        active_num[m] = N*L;
        int* const heads = new int[N*L];
        int* const labels = new int[N*L];
        int k = 0;
        for(int h = -1; h < N; ++h) {
          if(h == m) { continue; }
          for(int l = 0; l < L; ++l) {
            heads[k] = h;
            labels[k] = l;
            ++k;
          }
        }
        active_heads[m]  = heads;
        active_labels[m] = labels;
      }
      /* pass to the parsers */
      for(int i = 0; i < NP; ++i) {
        if(parsers[i]->pruned()) {
          parsers[i]->prune(N, active_num, active_heads, active_labels);
        }
      }
      /* safe to discard now */
      for(int m = 0; m < N; ++m) {
        delete [] active_heads[m];
        delete [] active_labels[m];
      }
      delete [] active_heads;
      delete [] active_labels;
      delete [] active_num;
    }
    for(int i = 0; i < NP; ++i) {
      o3bruteforce* const bfp = bfenums[i];
      o3parser* const par = parsers[i];
      const part_score_random randS(N, L, par->sibparser(), -10, 10);
      if(!par->topdown() && par->pruned()) { continue; } /* FIXME */
      assert(bfp->sibparser() == par->sibparser());
      assert(bfp->multiroot() == par->multiroot());
      assert(bfp->L() == par->L());
      {  /* test argmax */
        cerr << "  argmax" << flush;
        vector<int> bfheads;
        vector<int> bflabels;
        const part_score_cache<part_score_random> Sbf(randS, score_cache);
        const double scorebf = bfp->argmax(N, Sbf, bfheads, bflabels);
        const string bfreport = Sbf.report(); Sbf.clear();
        const double scorebf2 =
          o3tree::score(bfp->multiroot(), bfp->sibparser(),
                        bfheads, bflabels, randS);
        o3part::vec ybf;
        o3part::parts(L, bfp->multiroot(), bfp->sibparser(),
                      bfheads, bflabels, ybf);
        const double scorebf3 = o3part::score(ybf, randS);
        cerr << " " << bfname(bfp) << " " << scorebf << flush;
        assert(reldiff(scorebf, scorebf2) <= O3PARSER_TOL);
        assert(reldiff(scorebf, scorebf3) <= O3PARSER_TOL);
        vector<int> heads;
        vector<int> labels;
        const part_score_cache<part_score_random> Spa(randS, score_cache);
        const double score = par->argmax(N, Spa, heads, labels);
        const string pareport = Spa.report(); Spa.clear();
        cerr << " " << pname(par) << " " << score << flush;
        o3part::vec y;
        o3part::parts(L, par->multiroot(), par->sibparser(),
                      heads, labels, y);
        const double score2 = o3part::score(y, randS);
        assert(reldiff(score, score2) <= O3PARSER_TOL);
        cerr << scientific << setprecision(7)
             << " diff " << reldiff(score, scorebf)
             << fixed << setprecision(2) << endl;
        if(0) {
          cerr << "bf part cache stats:" << endl << bfreport
               << "par part cache stats:" << endl << pareport << endl;
        }
        check_parts(ybf, y);
        assert(reldiff(score, scorebf) <= O3PARSER_TOL);
      }
      if(par->topdown()) { continue; }
      continue; /*** FIXME ***/
      { /* test partition */
        cerr << "  part" << flush;
        const part_score_cache<part_score_random> S(randS, score_cache);
        const double logZbf = bfp->partition(N, S);
        const string bfreport = S.report(); S.clear();
        cerr << " " << bfname(bfp) << " " << logZbf << flush;
        const double logZ = par->partition(N, S);
        const string pareport = S.report(); S.clear();
        cerr << " " << pname(par) << " " << logZ << flush;
        cerr << scientific << setprecision(7)
             << " diff " << reldiff(logZ, logZbf)
             << fixed << setprecision(2) << endl;
        if(0) {
          cerr << "bf part cache stats:" << endl << bfreport
               << "par part cache stats:" << endl << pareport << endl;
        }
        assert(reldiff(logZ, logZbf) <= O3PARSER_TOL);
      }
    }
  }
  for(int i = 0; i < NP; ++i) {
    delete parsers[i];
    delete bfenums[i];
  }
}


void testinf::tdbu(const int Nmax, const int L, const double pratio) {
  const int NP = 4;
  o3parser* td[NP] = {
    /* top-down parsers: single/multi root; sibling/grandchild */
    new o3parser(L, true,  true, true, false),
    new o3parser(L, false, true, true, false),
    new o3parser(L, true,  true, true, true),
    new o3parser(L, false, true, true, true)
  };
  o3parser* bu[NP] = {
    /* bottom-up parsers: single/multi root; sibling/grandchild */
    new o3parser(L, true,  (pratio < 1.0), false, false),
    new o3parser(L, false, (pratio < 1.0), false, false),
    /*** FIXME ***/
    new o3parser(L, true,  (pratio < 1.0), false, true),
    new o3parser(L, false, (pratio < 1.0), false, true)
  };
  double tdtime = 0;
  double butime = 0;
  for(int N = 1; N <= Nmax; ++N) {
    cerr << "testinf::tdbu(): N=" << N << endl;
    { /* set the pruning arrays */
      int* const active_num = new int[N];
      const int** const active_heads = new const int*[N];
      const int** const active_labels = new const int*[N];
      for(int m = 0; m < N; ++m) {
        int* const heads = new int[N*L];
        int* const labels = new int[N*L];
        int k = 0;
        for(int h = -1; h < N; ++h) {
          if(h == m) { continue; }
          for(int l = 0; l < L; ++l) {
            const double rand01 = (double)random()/(double)RAND_MAX;
            /* always admit a linear chain so the parser doesn't
               fail */
            if(rand01 <= pratio || (l == 0 && h == m - 1)) {
              heads[k] = h;
              labels[k] = l;
              ++k;
            }
          }
        }
        active_num[m]    = k;
        active_heads[m]  = heads;
        active_labels[m] = labels;
      }
      /* pass to the parser */
      for(int i = 0; i < NP; ++i) {
        td[i]->prune(N, active_num, active_heads, active_labels);
        if(bu[i]->pruned()) {
          bu[i]->prune(N, active_num, active_heads, active_labels);
        }
      }
      /* safe to discard now */
      for(int m = 0; m < N; ++m) {
        delete [] active_heads[m];
        delete [] active_labels[m];
      }
      delete [] active_heads;
      delete [] active_labels;
      delete [] active_num;
    }
    for(int i = 0; i < NP; ++i) {
      o3parser* const tdp = td[i];
      o3parser* const bup = bu[i];
      const part_score_random randS(N, L, tdp->sibparser(), -10, 10);
      { /* burn it in for timing */
        vector<int> tdheads;
        vector<int> tdlabels;
        tdp->argmax(N, randS, tdheads, tdlabels);
        bup->argmax(N, randS, tdheads, tdlabels);
      }
      assert(tdp->sibparser() == bup->sibparser());
      assert(tdp->multiroot() == bup->multiroot());
      assert(tdp->L() == bup->L());
      {  /* test argmax */
        cerr << "  argmax" << flush;


        o3part::vec tdy;
        o3part::vec buy;
        double tdscore = 0;
        double buscore = 0;

        { /* top-down */
          vector<int> tdheads;
          vector<int> tdlabels;
          /* timed run */
          timer_start();
          tdscore = tdp->argmax(N, randS, tdheads, tdlabels);
          timer_stop();
          tdtime += timediff;
          o3part::parts(L, tdp->multiroot(), tdp->sibparser(),
                        tdheads, tdlabels, tdy);
          const double tdscore2 = o3part::score(tdy, randS);
          cerr << " " << pname(tdp) << " " << tdscore << flush;
          assert(reldiff(tdscore, tdscore2) <= O3PARSER_TOL);
        }

        { /* bottom-up */
          vector<int> buheads;
          vector<int> bulabels;
          /* timed run */
          timer_start();
          buscore = bup->argmax(N, randS, buheads, bulabels);
          timer_stop();
          butime += timediff;
          o3part::parts(L, bup->multiroot(), bup->sibparser(),
                        buheads, bulabels, buy);
          const double buscore2 = o3part::score(buy, randS);
          cerr << " " << pname(bup) << " " << buscore << flush;
          assert(reldiff(buscore, buscore2) <= O3PARSER_TOL);
        }

        cerr << scientific << setprecision(7)
             << " diff " << reldiff(tdscore, buscore)
             << fixed << setprecision(2) << endl;
        check_parts(tdy, buy);
        assert(reldiff(tdscore, buscore) <= O3PARSER_TOL);
      }
    }
  }
  cerr << "testinf::tdbu(): top-down " << timestr(tdtime) << flush;
  cerr << " bottom-up " << timestr(butime) << endl;
  for(int i = 0; i < NP; ++i) {
    delete td[i];
    delete bu[i];
  }
}


void testinf::pruned(const int Nmax, const int L, const double pratio) {
  /* set up arrays of parsers */
  const int NP = 8;
  o3parser* parsers[NP] = {
    /* bottom-up parsers: multi/single root; gch/sib */
    new o3parser(L, true,  (pratio < 1.0),  false, false),
    new o3parser(L, false, (pratio < 1.0),  false, false),
    new o3parser(L, true,  (pratio < 1.0),  false, true),
    new o3parser(L, false, (pratio < 1.0),  false, true),
    /* top-down parsers: multi/single root; gch/sib */
    new o3parser(L, true,  true,  true,  false),
    new o3parser(L, false, true,  true,  false),
    new o3parser(L, true,  true,  true,  true),
    new o3parser(L, false, true,  true,  true)
  };
  /* brute-force parsers with matched characteristics */
  o3bruteforce* bfenums[NP] = {
    /* bruteforce matched to bottom-up */
    new o3bruteforce(L, true,  true, false),
    new o3bruteforce(L, false, true, false),
    new o3bruteforce(L, true,  true, true),
    new o3bruteforce(L, false, true, true),
    /* same for top-down */
    new o3bruteforce(L, true,  true, false),
    new o3bruteforce(L, false, true, false),
    new o3bruteforce(L, true,  true, true),
    new o3bruteforce(L, false, true, true)
  };
  o3part::dmap score_cache;
  for(int N = 1; N <= Nmax; ++N) {
    cerr << "testinf::pruned(): N=" << N << endl;
    { /* set the pruning arrays */
      int num_deps = 0;
      int num_kept = 0;
      int* const active_num = new int[N];
      const int** const active_heads = new const int*[N];
      const int** const active_labels = new const int*[N];
      for(int m = 0; m < N; ++m) {
        int* const heads = new int[N*L];
        int* const labels = new int[N*L];
        int k = 0;
        for(int h = -1; h < N; ++h) {
          if(h == m) { continue; }
          for(int l = 0; l < L; ++l) {
            const double rand01 = (double)random()/(double)RAND_MAX;
            /* always admit a linear chain so the parser doesn't
               fail */
            if(rand01 <= pratio || (l == 0 && h == m - 1)) {
              heads[k] = h;
              labels[k] = l;
              ++k; ++num_kept;
            }
            ++num_deps;
          }
        }
        active_num[m]    = k;
        active_heads[m]  = heads;
        active_labels[m] = labels;
      }
      cerr << "  pruned: " << num_kept << " / " << num_deps << " = "
           << 100.0*(((double)num_kept)/((double)num_deps)) << endl;
      /* pass to the parsers */
      for(int i = 0; i < NP; ++i) {
        if(parsers[i]->pruned()) {
          parsers[i]->prune(N, active_num, active_heads, active_labels);
        }
        bfenums[i]->prune(N, active_num, active_heads, active_labels);
      }
      /* safe to discard now */
      for(int m = 0; m < N; ++m) {
        delete [] active_heads[m];
        delete [] active_labels[m];
      }
      delete [] active_heads;
      delete [] active_labels;
      delete [] active_num;
    }
    for(int i = 0; i < NP; ++i) {
      o3bruteforce* const bfp = bfenums[i];
      o3parser* const par = parsers[i];
      const part_score_random randS(N, L, par->sibparser(), -10, 10);
      assert(bfp->sibparser() == par->sibparser());
      assert(bfp->multiroot() == par->multiroot());
      assert(bfp->L() == par->L());
      { /* test argmax */
        cerr << "  argmax" << flush;
        vector<int> bfheads;
        vector<int> bflabels;
        const part_score_cache<part_score_random> Sbf(randS, score_cache);
        const double scorebf = bfp->argmax(N, Sbf, bfheads, bflabels);
        const string bfreport = Sbf.report(); Sbf.clear();
        o3part::vec ybf;
        o3part::parts(L, bfp->multiroot(), bfp->sibparser(),
                      bfheads, bflabels, ybf);
        const double scorebf2 = o3part::score(ybf, randS);
        cerr << " " << bfname(bfp) << " " << scorebf << flush;
        assert(reldiff(scorebf, scorebf2) <= O3PARSER_TOL);
        vector<int> heads;
        vector<int> labels;
        const part_score_cache<part_score_random> Spa(randS, score_cache);
        const double score = par->argmax(N, Spa, heads, labels);
        const string pareport = Spa.report(); Spa.clear();
        cerr << " " << pname(par) << " " << score << flush;
        o3part::vec y;
        o3part::parts(L, par->multiroot(), par->sibparser(),
                      heads, labels, y);
        const double score2 = o3part::score(y, randS);
        assert(reldiff(score, score2) <= O3PARSER_TOL);
        cerr << scientific << setprecision(7)
             << " diff " << reldiff(score, scorebf)
             << fixed << setprecision(2) << endl;
        if(0) {
          cerr << "bf part cache stats:" << endl << bfreport
               << "par part cache stats:" << endl << pareport << endl;
        }
        check_parts(ybf, y);
        assert(reldiff(score, scorebf) <= O3PARSER_TOL);
      }
      if(par->topdown()) { continue; }
      continue; /*** FIXME ***/
      { /* test partition */
        cerr << "  part" << flush;
        const part_score_cache<part_score_random> S(randS, score_cache);
        const double logZbf = bfp->partition(N, S);
        const string bfreport = S.report(); S.clear();
        cerr << " " << bfname(bfp) << " " << logZbf << flush;
        const double logZ = par->partition(N, S);
        const string pareport = S.report(); S.clear();
        cerr << " " << pname(par) << " " << logZ << flush;
        cerr << scientific << setprecision(7)
             << " diff " << reldiff(logZ, logZbf)
             << fixed << setprecision(2) << endl;
        if(0) {
          cerr << "bf part cache stats:" << endl << bfreport
               << "par part cache stats:" << endl << pareport << endl;
        }
        assert(reldiff(logZ, logZbf) <= O3PARSER_TOL);
      }
    }
  }
  for(int i = 0; i < NP; ++i) {
    delete parsers[i];
    delete bfenums[i];
  }
}
