#ifndef BHEAP_H
#define BHEAP_H

#include <assert.h>
#include <stdlib.h>


/* Data structures and methods for a binary heap.  In tests, this
   binary heap implementation was found to outperform the Fibonacci
   heap in terms of both speed and memory consumption, presumably
   because of the efficient array representation of a binary tree.

   To use this file, define a key type heapkey, a data type heapval,
   and a key comparison function/macro keylt(), and then include
   "bheap.c".  Assumes that keys and values are fixed-length types and
   so they do not require dynamic allocation.  The keylt(key1, key2)
   procedure returns nonzero if key1 is "less than" (i.e., belongs
   before) key2.  The heap is a min-heap with respect to the total
   ordering defined by keylt() and does *not* detect duplicates.

   Author : Terry Koo
            maestro@csail.mit.edu */

struct bheap {
  heapkey* keys; /* array of keys */
  heapval* vals; /* array of values */
  unsigned size; /* sizes (in memory) of the above arrays */
  unsigned n;    /* number of elements in the heap */
};


/* initialize a heap */
static inline void bheap_init(struct bheap* const h,
                              const unsigned initsize) {
  h->size = initsize;
  h->keys = (heapkey*)malloc(initsize*sizeof(heapkey));
  h->vals = (heapval*)malloc(initsize*sizeof(heapval));
  h->n = 0;
}

/* create an emtpy heap with the given initial size */
static inline struct bheap* bheap_new(const unsigned initsize) {
  struct bheap* h = (struct bheap*)malloc(sizeof(struct bheap));
  bheap_init(h, initsize);
  return h;
}

/* create an array of empty heaps */
static inline struct bheap* bheap_arr_new(const unsigned n,
                                          const unsigned initsize) {
  struct bheap* H = (struct bheap*)malloc(n*sizeof(struct bheap));
  struct bheap* restrict h = H;
  struct bheap* const h_end = H + n;
  for(; h < h_end; ++h) { bheap_init(h, initsize); }
  return H;
}

/* deallocate a heap */
static inline void bheap_destroy(struct bheap* const h) {
  free(h->keys);
  free(h->vals);
  free(h);
}

/* deallocate an array of heaps */
static inline void bheap_arr_destroy(struct bheap* const H,
                                     const unsigned n) {
  struct bheap* restrict h = H;
  struct bheap* const h_end = H + n;
  for(; h < h_end; ++h) { free(h->keys); free(h->vals); }
  free(H);
}

/* we use an implicitly-defined heap array.  since the array is
   0-origin, the parent of index i is floor((i-1)/2), and the children
   of index i are 2*i + 1 and 2*i + 2. */

/* insert a new (key, value) pair into the heap */
static inline void bheap_ins(struct bheap* const h,
			     const heapkey key,
			     const heapval value) {
  const unsigned n = h->n;
  if(n == h->size) { /* array doubling */
    h->size *= 2;
    h->keys = (heapkey*)realloc(h->keys, h->size*sizeof(heapkey));
    h->vals = (heapval*)realloc(h->vals, h->size*sizeof(heapval));
  }
  heapkey* const restrict keys = h->keys;
  heapval* const restrict vals = h->vals;
  /* start with i pointing to the end of the heap and sift upwards to
     find the proper destination */
  unsigned i = n;
  while(i != 0) {
    const unsigned p = ((i - 1) >> 1); /* parent = floor((i-1)/2) */
    const heapkey pk = keys[p];
    if(keylt(key, pk)) { /* heap property violated */
      /* only do the downward swap since the new key and value are in
	 local variables */
      keys[i] = pk;
      vals[i] = vals[p];
      /* recurse on the parent position */
      i = p;
    } else {
      break;
    }
  }
  /* store the new pair at index i */
  keys[i] = key;
  vals[i] = value;
  h->n = n + 1;
}

/* get the current first (key, value) pair */
static inline heapkey bheap_get(const struct bheap* const h,
				heapval* const ret_val) {
  assert(h->n > 0);
  if(ret_val != NULL) { *ret_val = h->vals[0]; }
  return h->keys[0];
}

/* remove the current first (key, value) pair */
static inline heapkey bheap_pop(struct bheap* const h,
				heapval* const ret_val) {
  assert(h->n > 0);
  *ret_val = h->vals[0];
  const heapkey ret_key = h->keys[0];

  --(h->n);
  /* swap the last element into the root and sift downwards */
  const unsigned n = h->n;
  heapkey* const restrict keys = h->keys;
  heapval* const restrict vals = h->vals;
  const heapkey key = keys[n];
  const heapval val = vals[n];
  unsigned i = 0;
  while(1) {
    const unsigned c1 = ((i << 1) + 1); /* leftchild  = 2*i + 1 */
    const unsigned c2 = c1 + 1;         /* rightchild = 2*i + 2 */

    if(c1 >= n) { /* end of heap */
      break;
    } else if(c2 >= n) { /* only one key available */
      const heapkey k1 = keys[c1];
      if(keylt(k1, key)) { /* heap property violated */
	keys[i] = k1;
	vals[i] = vals[c1];
	/* the final destination is c1 */
	i = c1;
      } /* else the final destination is simply i */
      break; /* no need to recurse */
    } else { /* general case */
      const heapkey k1 = keys[c1];
      const heapkey k2 = keys[c2];
      if(keylt(k1, key) || keylt(k2, key)) { /* heap property violated */
	if(keylt(k1, k2)) { /* c1 is the new parent */
	  keys[i] = k1;
	  vals[i] = vals[c1];
	  /* recurse on c1 */
	  i = c1;
	} else {      /* c2 is the new parent */
	  keys[i] = k2;
	  vals[i] = vals[c2];
	  /* recurse on c2 */
	  i = c2;
	}
      } else {
	break;
      }
    }
  }
  keys[i] = key;
  vals[i] = val;

  return ret_key;
}


#endif /* BHEAP_H */
