#ifndef FHEAP_H
#define FHEAP_H

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

/* Data structures and methods for a Fibonacci heap.  The Fibonacci
   heap offers efficient implementation of the insert and merge
   operations.  In addition, this implementation allows one to apply a
   "adjustment" (i.e., add a constant value) to every element in a
   heap in amortized O(log(n)) time.  These adjustments remain
   consistent as heaps are merged and elements are popped.  The merge
   and additive-adjustment operations are essential for
   max-arborescence algorithms.

   To use this file, define a key type heapkey, a data type heapval,
   and a key comparison function/macro keylt(), two key adjustment
   function/macro keyadd() and keysub(), a constant keyzero, and then
   include "fheap2.c".  Assumes that keys and values are fixed-length
   types and so they do not require dynamic allocation.  The
   keylt(key1, key2) procedure returns nonzero if key1 is "less than"
   (i.e., belongs before) key2.  The heap is a min-heap with respect
   to the total ordering defined by keylt() and does *not* detect
   duplicates.  The keyadd(key, adj) takes two elements of type
   heapkey and returns "key + adj", and keysub(key, adj) returns "key
   - adj".  The constant keyzero is the additive identity under these
   adjustment functions (i.e., the value satisfying keyadd(key,
   keyzero) = keysub(key, keyzero) = key).

   Author : Terry Koo
            maestro@csail.mit.edu */

/* The heap is represented by a set of trees where each set of
   children is joined by a circular doubly-linked list.  The .adj
   field contains an adjustment is implicitly applied to all children
   of this node (such adjustments are applied explicitly only when
   child nodes reach the root level of the heap). */
struct fheap_node {
  struct fheap_node* lsib; /* left sibling */
  struct fheap_node* rsib; /* right sibling */
  struct fheap_node* ch;   /* children (doubly-linked list) */
  unsigned nch;            /* number of children */
  heapkey key;             /* priority level */
  heapkey adj;             /* adjustment */
  heapval val;             /* associated value */
};


struct fheap {
  /* pointer to the current minimum root */
  struct fheap_node* min_root;
  /* number of roots currently in the model */
  unsigned num_roots;
  /* total number of elements in this heap */
  unsigned n;
};

/* create a new heap node representing the given pair */
static inline struct fheap_node* fheap_node_new(const heapkey key,
						const heapval value) {
  struct fheap_node* n =
    (struct fheap_node*)malloc(sizeof(struct fheap_node));
  n->ch = NULL;
  n->nch = 0;
  n->key = key;
  n->adj = keyzero;
  n->val = value;
  return n;
}

/* recursively deallocate a heap node */
static void fheap_node_free(struct fheap_node* const node) {
  if(node == NULL) { return; }
  struct fheap_node* n = node;
  do {
    fheap_node_free(n->ch);
    struct fheap_node* const next = n->rsib;
    free(n);
    n = next;
  } while(n != node);
}




/* create an empty heap */
static inline struct fheap* fheap_new() {
  struct fheap* h = (struct fheap*)malloc(sizeof(struct fheap));
  h->min_root = NULL;
  h->num_roots = 0;
  h->n = 0;
  return h;
}

/* deallocate a heap */
static inline void fheap_destroy(struct fheap* const h) {
  fheap_node_free(h->min_root);
  free(h);
}


/* Helper function that recursively prints all subtrees of all
   the node and its siblings */
/* static void print_sibs(struct fheap_node* const node) { */
/*   if(node == NULL) { */
/*     printf("<null>"); */
/*   } else { */
/*     struct fheap_node* n = node; */
/*     do { */
/*       printf(" [%g %g", n->key, n->adj); */
/*       if(n->ch != NULL) { */
/* 	print_sibs(n->ch); */
/*       } */
/*       printf("]"); */
/*       n = n->rsib; */
/*     } while(n != node); */
/*   } */
/* } */



/* insert a new (key, value) pair into the heap */
static inline void fheap_ins(struct fheap* const h,
			     const heapkey key,
			     const heapval value) {
  struct fheap_node* const r = h->min_root;
  struct fheap_node* const n = fheap_node_new(key, value);
  if(r == NULL) { /* empty heap, so create a cycle */
    h->min_root = n->lsib = n->rsib = n;
  } else { /* add to current cycle */
    n->lsib = r->lsib;
    n->rsib = r;
    n->lsib->rsib = n;
    r->lsib = n;
    if(keylt(key, r->key)) { /* new minimizer */
      h->min_root = n;
    }
  }
  ++(h->num_roots);
  ++(h->n);
}


/* remove all elements from h2 and add them to h */
static inline void fheap_take(struct fheap* const h, struct fheap* const h2) {
/*   printf("merging: h = "); */
/*   print_sibs(h->min_root); */
/*   printf("\n"); */
/*   printf("merging: h2 = "); */
/*   print_sibs(h2->min_root); */
/*   printf("\n"); */

  /* check if there is anything to take */
  if(h2->num_roots == 0) { return; }

  /* special handling if h is empty */
  if(h->num_roots == 0) {
    h->num_roots = h2->num_roots;
    h->min_root = h2->min_root;
    h->n = h2->n;
    h2->min_root = NULL;
    h2->num_roots = h2->n = 0;
    return;
  }

  /* general case: both h and h2 are nonempty */


  /* find endpoints of h's list */
  struct fheap_node* const leftmost = h->min_root;
  struct fheap_node* const rightmost = leftmost->lsib;
  /* define endpoints of the other array */
  struct fheap_node* const leftmost2 = h2->min_root;
  struct fheap_node* const rightmost2 = leftmost2->lsib;
  /* join endpoints to each other */
  leftmost->lsib = rightmost2;
  rightmost2->rsib = leftmost;
  rightmost->rsib = leftmost2;
  leftmost2->lsib = rightmost;
  /* compare both min-roots to find new root pointer */
  h->min_root = (keylt(leftmost->key, leftmost2->key)
		 ? leftmost : leftmost2);
  /* update sizes */
  h->num_roots += h2->num_roots;
  h->n += h2->n;
  /* clear other heap */
  h2->min_root = NULL;
  h2->num_roots = h2->n = 0;

/*   printf("merged: h = "); */
/*   print_sibs(h->min_root); */
/*   printf("\n"); */
/*   printf("merged: h2 = "); */
/*   print_sibs(h2->min_root); */
/*   printf("\n"); */
}


/* get the current minimum (key, value) pair */
static inline heapkey fheap_get(const struct fheap* const h,
				heapval* const ret_val) {
  const struct fheap_node* const r = h->min_root;
  assert(r != NULL);
  if(ret_val != NULL) { *ret_val = r->val; }
  return r->key;
}

/* Helper function that applies a key-adjustment to all nodes in the
   linked list */
static inline void apply_adjustment(struct fheap_node* node,
				    const heapkey adj) {
  if(node != NULL) {
    struct fheap_node* n = node;
    do {
      /* update both the key itself and the adjustment (for
	 propagation to children) */
      n->key = keyadd(n->key, adj);
      n->adj = keyadd(n->adj, adj);
      n = n->rsib;
    } while(n != node);
  }
}


/* Helper function that compresses the root list into a list of size
   O(log(n)) by conjoining all roots of equal arity.  Here, we must be
   careful to maintain the key-adjustments when making one root a
   child of another.  If r1 becomes a child of r2, then we must update
   r1 as follows:

   r1.key -= r2.adj
   r1.adj -= r2.adj

   The reason for the first modification is clear; since every child
   of r2 has r2.adj implicitly applied to it, we must invert that
   adjustment on r1.  The second modification follows for a similar
   reason; all of r1's children implicitly ought to have r1.adj
   implicitly applied to them, and when we make r1 a child of r2,
   r2.adj is also implicitly applied; therefore we invert r2.adj on
   r1's children. */
static inline void compact_roots(struct fheap* const h) {
  if(h->num_roots == 0) { return; }
  /* for each arity, a pointer to the root that has that arity */
  unsigned size_a2r = 16;
  struct fheap_node** restrict a2r =
    (struct fheap_node**)malloc(size_a2r*sizeof(struct fheap_node*));
  unsigned i;
  for(i = 0; i < size_a2r; ++i) { a2r[i] = NULL; }
  /* walk through all root nodes and conjoin nodes that have equal
     arity, maintaining the heap property */
  struct fheap_node* n = h->min_root;
  while(1) { /* stopping criterion elsewhere */
    const unsigned arity = n->nch;
    /* resize array if necessary */
    if(arity >= size_a2r) {
      const unsigned oldsize = size_a2r;
      size_a2r = arity*2;
      a2r = (struct fheap_node**)
	realloc(a2r, size_a2r*sizeof(struct fheap_node*));
      for(i = oldsize; i < size_a2r; ++i) { a2r[i] = NULL; }
    }
    /* check for a root with identical arity */
    if(a2r[arity] != NULL) {
      struct fheap_node* const other = a2r[arity];
      /* stop when we see a repeat */
      if(n == other) { break; }
      /* find out which root must become a child */
      struct fheap_node* winner = NULL;
      struct fheap_node* loser = NULL;
      /* note that the root "other" was processed before the root
	 "n".  in order for the stopping criterion above to work, we
	 must maintain that the winner ends up in n's position in
	 the root list.  therefore, we swap the positions of other
	 and n if other is the winner */
      if(keylt(other->key, n->key)) {
	/* here, we must move other into n's position.  we have a
	   few special cases */
	if(h->num_roots == 2) {
	  other->lsib = other->rsib = other;
	} else if(other->rsib == n) { /* neighbor to right */
	  struct fheap_node* const rsib = n->rsib;
	  other->rsib = rsib;
	  rsib->lsib = other;
	} else if(other->lsib == n) { /* neighbor to left */
	  struct fheap_node* const lsib = n->lsib;
	  other->lsib = lsib;
	  lsib->rsib = other;
	} else { /* completely disjoint */
	  /* first, unlink other */
	  struct fheap_node* const oldlsib = other->lsib;
	  struct fheap_node* const oldrsib = other->rsib;
	  oldlsib->rsib = oldrsib;
	  oldrsib->lsib = oldlsib;
	  /* next, relink other into n's position */
	  struct fheap_node* const lsib = n->lsib;
	  struct fheap_node* const rsib = n->rsib;
	  other->lsib = lsib;
	  other->rsib = rsib;
	  lsib->rsib = other;
	  rsib->lsib = other;
	}
	winner = other; loser = n;
      } else {
	/* here, it suffices to just unlink the loser */
	struct fheap_node* const lsib = other->lsib;
	struct fheap_node* const rsib = other->rsib;
	lsib->rsib = rsib;
	rsib->lsib = lsib;
	winner = n; loser = other;
      }
      /* the above always removes one root */
      --(h->num_roots);
      /* the loser becomes a child of the winner, thus we must invert
	 the winner's adjustment on the loser */
      loser->key = keysub(loser->key, winner->adj);
      loser->adj = keysub(loser->adj, winner->adj);
      /* now, add the loser to the winner's child list */
      if(winner->nch == 0) { /* special case */
	winner->nch = 1;
	winner->ch = loser->lsib = loser->rsib = loser;
      } else {
	++(winner->nch);
	struct fheap_node* const leftmost = winner->ch;
	struct fheap_node* const rightmost = leftmost->lsib;
	loser->rsib = leftmost;
	leftmost->lsib = loser;
	loser->lsib = rightmost;
	rightmost->rsib = loser;
      }
      /* vacate the entry for the current arity */
      a2r[arity] = NULL;
      /* recurse on the winner node (which now has higher arity) */
      n = winner;
    } else { /* first root with this arity */
      a2r[arity] = n;
      /* advance to the next root in the root list */
      n = n->rsib;
    }
  }
  free(a2r);
  assert(h->num_roots > 0);

  /* after rearranging the roots, we still need to update the min-root
     pointer to point towards the smallest key */
  struct fheap_node* const first = n;
  struct fheap_node* min = first;
  double min_key = first->key;
  for(n = first->rsib; n != first; n = n->rsib) {
    const double key = n->key;
    if(keylt(key, min_key)) {
      min_key = key;
      min = n;
    }
  }
  h->min_root = min;
}


/* remove the current minimum (key, value) pair */
static inline heapkey fheap_pop(struct fheap* const h,
				heapval* const ret_val) {
  struct fheap_node* const r = h->min_root;
  assert(r != NULL);
  /* first, save the return values */
  const heapkey ret_key = r->key;
  if(ret_val != NULL) { *ret_val = r->val; }

  /* now, perform the pop in 3 phases.

     1. splice the child array into the current root list
     2. compact the root list by conjoining roots with identical arity
     3. find the minimum of the compacted root list
  */



  /* 1. splice the child array into the current root list */

  /* we begin by applying adjustments to all child nodes */
  apply_adjustment(r->ch, r->adj);

  /* now that the children are properly adjusted, we can safely splice
     them into the root-list */
  if(h->num_roots == 1) { /* special case: pass child array up */
    /* note that if r has no children, the below still works */
    h->min_root = r->ch;
    h->num_roots = r->nch;
  } else if(r->nch == 0) { /* just unlink the root */
    struct fheap_node* const lsib = r->lsib;
    struct fheap_node* const rsib = r->rsib;
    /* link r's siblings to each other */
    lsib->rsib = rsib;
    rsib->lsib = lsib;
    /* set new (invalid) root pointer */
    h->min_root = rsib;
    --(h->num_roots);
  } else { /* splice children into root list */
    struct fheap_node* const lsib = r->lsib;
    struct fheap_node* const rsib = r->rsib;
    /* define left/right ends of the child array */
    struct fheap_node* const leftmost = r->ch;
    struct fheap_node* const rightmost = leftmost->lsib;
    /* link left end to root's left sibling */
    lsib->rsib = leftmost;
    leftmost->lsib = lsib;
    /* link right end to root's right sibling */
    rsib->lsib = rightmost;
    rightmost->rsib = rsib;
    /* set new (invalid) root pointer */
    h->min_root = rsib;
    h->num_roots = h->num_roots - 1 + r->nch;
  }
  /* at this point, r has been removed from the root list and its
     children are taken care of, so we may discard r */
  free(r);


  /* 2. compact the root list by conjoining roots with identical
        arity
     3. find the minimum of the compacted root list */

  /* the function below handles both steps above and also keeps track
     of the key-adjustments */
  compact_roots(h);

  --(h->n);
  return ret_key;
}


/* Adjust all keys in this heap by some constant */
static inline void fheap_add(struct fheap* h, const heapkey adj) {
  /* perform the adjustment in 2 phases.

     1. compact the root list and update the min-root pointer
     2. adjust all roots */

  compact_roots(h);
  apply_adjustment(h->min_root, adj);

  /* this operation is O(log(n)) amortized for the same reason that
     pop() is O(log(n)).  compacting the root list costs O(log(n))
     amortized, and then adjusting the remaining O(log(n)) roots takes
     O(log(n)) time.  note that it is unnecessary to adjust the
     min-root pointer since the adjustment is applied to all nodes in
     the heap and thus cannot change the ordering of the nodes. */
}


#endif /* FHEAP_H */
