#ifndef KDP_H
#define KDP_H

/* Algorithms for performing k-best dependency parsing.  The parser
   must first be initialized with a given example, and then it can be
   repeatedly queried in an online fashion in order to retrieve the
   sequence of sorted k-best parses.  There are four parsers:

   kmpt1 is the O(n^3) edge-factored projective parser

   kmpt2 is the O(n^3) projective parser with sibling interactions

   kmpt3 is the O(n^4) second-order projective parser with sibling and
         grandparent interactions

   kmst1 is the O(n^2) spanning-tree parser

   Each parser may be run in a single-root or multi-root mode.

   The kmpt1, kmpt2, and kmpt3 algorithms are applications of the
   k-best parsing techniques of Huang and Chiang (2005) to the parsing
   frameworks of Eisner (2000), McDonald and Pereira (2006), and
   Carreras (2007) respectively.  The kmst1 algorithm is described in
   Camerini et al. (1980) and also Hall (2007).

   Author : Terry Koo
            maestro@csail.mit.edu */



/* Initialize the internal state to match the provided scores.  If the
   state is already initialized, it is cleared first.

   E is a n x n matrix of edge weights such that E[i + n*j] gives the
   score of edge i-->j, assuming i != j.  The entries E[r + n*r]
   contain root-selection scores.

   The score matrix is not needed after the initialization routines
   return, so they may be safely freed. */
void kmst1_init(const int multiroot,
		const int n,
		const double* const restrict E);

/* Clear the internal state (e.g., charts) of the parser */
void kmst1_clear();

/* Retrieve the k'th-best parse (counting from k=1 for argmax
   solution), represented as an array p of n integers, where p[i]
   gives the parent node of i; p[i] = -1 indicates that i is the root
   node (or a root node).  The score of the parse is returned in
   ret_score.  If the k'th-best parse does not exist (because there
   are not k different derivations), then zero is returned, otherwise
   nonzero is returned. */
int kmst1_get(const int k, int* const restrict p, double* const ret_score);

/* The k-best MST algorithm with k=1 can be 2 or 3 times slower than a
   single 1-best invocation, so this function is provided as a fast
   alternative. */
double mst1(int* const restrict p, const int multiroot, const int n,
	    const double* const restrict in_E);

#endif /* KDP_H */
