#include "kdp.h"

/* Implementation of first-order k-best max-spanning-tree (i.e.,
   nonprojective dependency parser)

   Author : Terry Koo
            maestro@csail.mit.edu */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>

/* We make use of several data structures in order to obtain an
   efficient implementation.

   * To order the in-edges of every node, we use a Fibonacci heap,
     which allows fast heap merging operations.  In addition, the heap
     is modified to allow an additional operation that adjusts every
     key in a heap by some constant.

   * To detect cycles quickly and also handle the node contractions,
     we use a disjoint-set forest structure, which is implicitly
     represented by an array.

   * To order the candidates for the next-best tree, we use a standard
     binary heap.

   * We represent the constraint sets Y and Z as singly-linked lists.
     Note that every time we create a new candidate to add to the
     next-best heap, we only increase the size of Y or Z.  Therefore,
     we can share the suffixes of the Y and Z lists, forming a
     singly-linked tree.  This allows all constraint sets Y and Z to
     be represented in O(m + K) memory.

     Note that the total memory requirements are actually O(m + nK)
     because the algorithm caches MSTs rather than reparse the best
     tree every time a next-best tree must be found (essentially
     trading a factor of 2 speedup for a factor of n in space, which
     is quite reasonable for the K and n that one encounters in
     parsing applications). */


/* tolerance parameter for differences between scores predicted by
   next() and scores actually found by best() */
#define KMST1_REL_TOL 1e-8


/* Fibonacci heap with special heap-adjust operation */
typedef double heapkey;
#define heapval int
#define keylt(k1, k2) (k1 > k2)
#define keyadd(k1, k2) (k1 + k2)
#define keysub(k1, k2) (k1 - k2)
#define keyzero 0
#include "fheap2.h"
#undef heapval
#undef keylt
#undef keyadd
#undef keysub
#undef keyzero

/* a node in a singly-linked upward tree.  the node also implements a
   singly-linked list that threads all allocated nodes together, for
   ease of freeing these structures. */
struct upnode {
  struct upnode* next; /* global linked list of allocated nodes */
  struct upnode* p;    /* parent of this node */
  int e;               /* an edge index */
};

/* information that allows construction of a spanning tree and its
   next-best neighbors */
struct partition {
  int e;            /* the pivot edge */
  const int* A;     /* the best tree satisfying Y,Z below */
  double sc;        /* the score of A */
  struct upnode* Y; /* edges which must be included */
  struct upnode* Z; /* edges which must be excluded */
};


/* binary heap on candidate structures */
typedef struct partition heapval;
#define keylt(k1, k2) (k1 > k2)
#include "bheap.h"
#undef keylt
#undef heapval

/* Helper function that creates a new upnode that hangs off of the
   given parent */
static inline struct upnode* upnode_new(struct upnode* const p,
					const int e) {
  struct upnode* const ret =
    (struct upnode*)malloc(sizeof(struct upnode));
  ret->p = p;
  ret->e = e;
  ret->next = NULL;
  return ret;
}

/* Helper function that deallocates an upnode */
#define upnode_destroy(p) free(p)



/* a weighted directed edge */
struct edge {
  int i;    /* head of edge */
  int j;    /* tail of edge */
  double s; /* score of edge */
};

/* set of all weighted edges in the graph */
static const struct edge* restrict E = NULL;
/* original root scores (in the case of single-root parsing) */
double* restrict RS = NULL;
/* number of nodes and edges */
static int N = -1;
static int M = -1;
static int K = -1;
/* whether parsing is single- or multi-root */
static int MULTIROOT = 0;
/* queue over candidate structures */
static struct bheap* treeQ = NULL;
/* pointer to the head of the list of allocated upnodes */
static struct upnode* YZ = NULL;

/* initialize the edge structures */
void kmst1_init(const int multiroot,
		const int n,
		const double* const restrict in_E) {
  assert(E == NULL);
  N = n;
  M = n*n;
  K = 0;
  MULTIROOT = multiroot;

  /* create a set of edges representing the edge-score matrix */
  struct edge* restrict tmpE =
    (struct edge*)malloc((unsigned)M*sizeof(struct edge));
  int cnt = 0;
  if(MULTIROOT) {
    int i;
    for(i = 1; i <= n; ++i) {
      int j, idx; /* idx = (i-1) + n*(j-1) */
      for(j = 1, idx = i - 1; j <= n; ++j, idx += n) {
	tmpE[cnt] = (struct edge){(i == j ? 0 : i), j, in_E[idx]};
	++cnt;
      }
    }
  } else {
    /* in the single-root case, we will offset the root edges by the
       sum-abs of all edge scores, to ensure that we enumerate all the
       single-root-edge trees before the multi-root-edge trees.  it is
       important to let the sum > 0 so that root edges will be
       discouraged even if all edge scores are identically zero (as is
       the case at the beginning of training) */
    double sum = 1.0;
    int i;
    for(i = 1; i <= n; ++i) {
      int j, idx; /* idx = (i-1) + n*(j-1) */
      for(j = 1, idx = i - 1; j <= n; ++j, idx += n) {
	const double score = in_E[idx];
	tmpE[cnt] = (struct edge){(i == j ? 0 : i), j, score};
        sum += (score < 0.0 ? -score : score);
	++cnt;
      }
    }
    /* now decrement the root edge scores (and save the originals) */
    assert(RS == NULL);
    RS = (double*)malloc((unsigned)n*sizeof(double));
    int idx;
    const int np1 = n + 1;
    for(i = idx = 0; i < n; ++i, idx += np1) {
      RS[i] = tmpE[idx].s;
      tmpE[idx].s -= sum;
      assert(tmpE[idx].i == 0);
      assert(tmpE[idx].j == i + 1);
    }
  }
  assert(cnt == M);
  E = tmpE;
  treeQ = bheap_new(16);
}


/* free the edge structures */
void kmst1_clear() {
  assert(E != NULL);
  /* the following violates the opaqueness of the data structure but
     is more efficient than e.g., popping all entries of the heap */
  const struct partition* restrict p = treeQ->vals;
  const struct partition* const restrict p_end = p + treeQ->n;
  for(; p < p_end; ++p) { free((void*)(p->A)); }
  bheap_destroy(treeQ);
  treeQ = NULL;
  free((void*)E);
  E = NULL;
  while(YZ != NULL) {
    struct upnode* next = YZ->next;
    upnode_destroy(YZ);
    YZ = next;
  }
  K = N = M = -1;
  if(MULTIROOT == 0) { free(RS); RS = NULL; }
}


/* Helper function that finds the set to which i belongs, in a
   disjoint-set forest data structure.  We perform path-compression
   after finding the set identity */
static inline int set_find(int* const restrict S, const int i) {
  /* a self-loop is a root */
  int k;
  for(k = i; k != S[k]; k = S[k]) {}
  /* k is now a pointer to the root */
  const int root = k;
  /* make all nodes along the path children of the root */
  k = i;
  while(1) {
    const int next = S[k];
    if(k == next) { break; }
    S[k] = root;
    k = next;
  }
  return root;
}

/* Helper function that adds i to j's set.  assumes that i and j are
   set-roots and not ordinary nodes */
static inline void set_add(int* const restrict S, const int i, const int j) {
  S[i] = j;
}


/* Function that finds the MST subject to the constraints Y and Z */
static int* best(const struct upnode* restrict Y,
		 const struct upnode* restrict Z) {
  /* Here and elsewhere, note that the cycle-contraction stages cannot
     create more than n-1 nodes.  Thus, the maximum number of nodes is
     2n (including the root). */
  const int maxn = 2*N;
  int i;

  /* queues of in-edges for every node */
  struct fheap** const restrict Q =
    (struct fheap**)malloc((unsigned)maxn*sizeof(struct fheap*));
  Q[0] = NULL; /* the root's queue */
  /* allocate queues */
  for(i = 1; i <= N; ++i) { Q[i] = fheap_new(); }
  for(; i < maxn; ++i)    { Q[i] = NULL; }

  /* this array of flags indicates whether a node has a Y constraint
     set on its in-edge */
  char* restrict fixed = (char*)calloc((unsigned)N + 1, sizeof(char));
  for(; Y != NULL; Y = Y->p) {
    const int e = Y->e;
    const struct edge* const restrict ed = E + e;
    const int tail = ed->j;
    fixed[tail] = 1;
    /* also insert the edge into this queue */
    const double sc = ed->s;
    fheap_ins(Q[tail], sc, e);
  }

  /* this array of flags indicates whether an edge has a Z constraint
     preventing it from being used */
  char* restrict ignore = (char*)calloc((unsigned)M, sizeof(char));
  for(; Z != NULL; Z = Z->p) { ignore[Z->e] = 1; }

  /* fill the queues */
  int e;
  for(e = 0; e < M; ++e) {
    if(ignore[e]) { continue; } /* skip Z-edges */
    const struct edge* const restrict ed = E + e;
    const int tail = ed->j;
    if(fixed[tail]) { continue; } /* skip Y-edges */
    fheap_ins(Q[tail], ed->s, e);
  }
  /* these are not used again in best() */
  free(fixed);  fixed = NULL;
  free(ignore); ignore = NULL;

  /* as cycles are contracted, node indices must be remapped.  this
     array is a disjoint-set representation of the remapping */
  int* const restrict D = (int*)malloc((unsigned)maxn*sizeof(int));
  /* fill with self-loops for original nodes */
  for(i = 0; i <= N; ++i) { D[i] = i; }
  for(; i < maxn; ++i)    { D[i] = INT_MIN; }

  /* the following array is an disjoint-set forest that allows
     detection of cycles */
  int* const restrict B = (int*)malloc((unsigned)maxn*sizeof(int));
  /* fill with self-loops for original nodes */
  for(i = 0; i <= N; ++i) { B[i] = i; }
  for(; i < maxn; ++i)    { B[i] = INT_MIN; }

  /* this array implicitly represents a branching indicating the
     cycle-contraction history */
  int* const restrict C = (int*)malloc((unsigned)maxn*sizeof(int));
  /* fill with self-loops for original nodes */
  C[0] = INT_MIN;
  for(i = 1; i <= N; ++i) { C[i] = i; }
  for(; i < maxn; ++i)    { C[i] = INT_MIN; }

  /* this array tracks the best in-edge for every node */
  int* const restrict beta = (int*)malloc((unsigned)(maxn + 1)*sizeof(int));
  double* const restrict betasc = (double*)malloc((unsigned)maxn*sizeof(double));
  for(i = 0; i < maxn; ++i) { beta[i] = INT_MIN; betasc[i] = HUGE_VAL; }
  /* include a final entry to signify the end of the list */
  beta[maxn] = INT_MIN;

  /* step 1: process all nodes and contract cycles until there are no
     unprocessed nodes */
  int nxt = 1; /* the next node to process */
  int n = N;   /* the current number of nodes in the graph, including
		  nodes created by cycle-contraction */
  while(nxt <= n) {
    const double sc = fheap_pop(Q[nxt], &e);
    assert(e >= 0); assert(e < M);
    const struct edge* const restrict ed = E + e;
    /* find the endpoints of this edge in the current graph */
    const int ii = set_find(D, ed->i);
    const int jj = set_find(D, ed->j);
    assert(jj == nxt);
    /* ignore self-loops */
    if(ii == jj) { continue; }
    /* set the best in-edge */
    assert(beta[jj] < 0);
    beta[jj] = e;
    betasc[jj] = sc;
    /* find the ancestor membership of each node */
    const int set_ii = set_find(B, ii);
    const int set_jj = set_find(B, jj);
    /* detect cycles */
    if(set_ii == set_jj) {
      /* create a new contracted node */
      const int newnode = n + 1;
      assert(newnode < maxn);
      assert(Q[newnode] == NULL);
      /* initialize data structures for this node */
      Q[newnode] = fheap_new();
      B[newnode] = C[newnode] = D[newnode] = newnode;
      /* anything which was a descendant of any node in the cycle is
	 now a descendant of the new node */
      set_add(B, set_ii, newnode);
      /* walk the nodes of the cycle and add them to the new node */
      int k = jj;
      do {
	/* make each node a child of the new node in C */
	C[k] = newnode;
	/* adjust the in-edges of this node to reflect the cost of
	   breaking the current best in-edge */
	const double collide_score = betasc[k];
	fheap_add(Q[k], -collide_score); /* shift all scores in heap */
	/* merge the adjusted heap into the new node's heap */
	fheap_take(Q[newnode], Q[k]);
	fheap_destroy(Q[k]); Q[k] = NULL;
	/* locate the next node in the cycle */
	const int in_e = beta[k];
	const int orig_head = E[in_e].i;
	const int next = set_find(D, orig_head);
	/* remap the current node to the contracted node */
	set_add(D, k, newnode);
	/* advance to the next node */
	k = next;
	/* when we get to a node we've already processed, the node
	   should be remapped to the contracted node.  thus it
	   suffices to check this condition to terminate the loop. */
      } while(k != newnode);
      /* increase size of graph */
      n = newnode;
    } else {
      /* no cycle, so just update the ancestry sets */
      set_add(B, set_jj, set_ii);
    }
    /* advance to the next node */
    ++nxt;
  }


  /* step 2: break cycles using the branching C and set the proper
     edges in the beta array */

  /* here, we need to be able to do two things: find roots in the
     branching C, and delete nodes from C.  as it was used above,
     roots are represented in C by self-loops.  we will delete nodes
     by setting their entries to -1.  thus, ther are two conditions
     that define roothood: either k == C[k], or C[C[k]] == -1.  below,
     we use the following process to identify the proper chains for
     deletion.

     * iterate all nodes i from 1 to N, and find the root r of i in C.

     * find the best in-edge e of the root r

     * find the original tail j of this edge

     * update beta[j] with beta[r] and delete the chain from j to r */
  for(i = 1; i <= N; ++i) {
    /* skip nodes that have already been deleted */
    if(C[i] == -1) { continue; }
    /* search for the root of i */
    int k;
    for(k = i; C[k] != k && C[C[k]] != -1; k = C[k]) {}
    const int root = k;
    /* in the trivial case, nothing needs to be done */
    if(i == root) { continue; }
    /* find the original tail of the in-edge */
    const int root_e = beta[root];
    const int j = E[root_e].j;
    assert(j != root);
    /* copy the edge from the root */
    beta[root] = beta[j]; /* save the original value */
    beta[j] = root_e;
    /* delete the chain from j to the root */
    k = j;
    while(1) {
      const int next = C[k];
      C[k] = -1;
      if(k == next || C[next] == -1) { break; }
      k = next;
    }
    /* if the tail j was not the same as i, then we still need to
       reprocess i */
    if(i != j) { --i; }
  }

  free(B);
  free(C);
  free(D);
  for(i = 1; i <= n; ++i) { if(Q[i] != NULL) { fheap_destroy(Q[i]); } }
  free(Q);
  free(betasc);

  return beta;
}


/* Helper function that returns the score of the edge "next to" the
   given edge b.  requires a 2n-element workspace to keep track of
   elements that have been popped.  the return value is zero if no
   "next" edge could be found; otherwise the function returns
   nonzero. */
static inline int seek(struct fheap* const q,
		       const int tail_b,
		       const int* const restrict pr,
		       const int* const restrict po,
		       const char* const restrict keep,
		       int* const restrict worki,
		       double* const restrict workd,
		       double* const ret_sc) {
  int n = 0;
  int found = 0;
  while(q->n > 0) {
    int e;
    const double sc = fheap_get(q, &e);
    const int ii = E[e].i;
    /* check if the tail of b is an ancestor of the head of e */
    if(pr[tail_b] <= pr[ii] && po[tail_b] >= po[ii]) {
      /* these edges are invalid and may be popped */
      fheap_pop(q, NULL);
      /* if the edge must be kept, then we save it in the workspace */
      if(keep[e]) {
	assert(n < 2*N);
	worki[n] = e;
	workd[n] = sc;
	++n;
      }
    } else {
      /* we found the edge that is next to b */
      *ret_sc = sc;
      found = 1;
      break;
    }
  }

  /* replace all edges which need to be replaced */
  int i;
  for(i = 0; i < n; ++i) { fheap_ins(q, workd[i], worki[i]); }

  return found;
}

/* Helper function that walks a given MST and finds the preorder and
   postorder numberings of the nodes */
static inline void make_pre_post(const int* const restrict A,
				 int* const restrict pr,
				 int* const restrict po) {
  /* convert the tree A to a child-array representation */
  int** const restrict ch = (int**)malloc((unsigned)(N + 1)*sizeof(int*));
  int* const restrict nch = (int*)calloc((unsigned)N + 1, sizeof(int));
  int i;
  for(i = 0; i <= N; ++i) {
    ch[i] = (int*)malloc((unsigned)N*sizeof(int));
  }
  for(i = 1; i <= N; ++i) {
    const int e = A[i];
    const struct edge* const restrict ed = E + e;
    assert(i == ed->j);
    const int par = ed->i;
    const int n = nch[par];
    int* const restrict charr = ch[par];
    charr[n] = i;
    nch[par] = n + 1;
  }

  /* create a worklist and fill it with the root's children */
  const int size = 2*N;
  int* const restrict stack = (int*)malloc((unsigned)size*sizeof(int));
  int n = nch[0];
  int* const restrict rootch = ch[0];
  for(i = 0; i < n; ++i) {
    stack[i] = rootch[i];
  }
  free(rootch); ch[0] = NULL;
  /* perform a DFS and record the preorder and postorder visitations */
  pr[0] = 0;
  int npr = 1;
  int npo = 0;
  while(n > 0) {
    --n;
    const int popped = stack[n];
    assert(popped != 0);
    if(popped < 0) { /* termination token for a node */
      po[-popped] = npo;
      ++npo;
    } else { /* expand this node */
      pr[popped] = npr;
      ++npr;
      /* push children onto stack */
      stack[n] = -popped; /* termination token */
      ++n;
      const int* const restrict charr = ch[popped];
      const int numch = nch[popped];
      for(i = 0; i < numch; ++i, ++n) {
	stack[n] = charr[i];
      }
      assert(n <= size);
      free((void*)charr); ch[popped] = NULL;
    }
  }
  assert(npr == N + 1);
  assert(npo == N);
  po[0] = N;

  free(stack);
  for(i = 0; i <= N; ++i) { assert(ch[i] == NULL); }
  free(ch);
  free(nch);
}


/* Function that finds an edge that, when excluded, results in a graph
   for which the MST is the next-best tree after A.  also returns the
   predicted difference in score between the MST and the next-best
   tree. */
static int next(const int* const restrict A,
		const struct upnode* restrict Y,
		const struct upnode* restrict Z,
		double* const ret_diff) {
  /* Here and elsewhere, note that the cycle-contraction stages cannot
     create more than n-1 nodes.  Thus, the maximum number of nodes is
     2n (including the root). */
  const int maxn = 2*N;
  int i;

  /* queues of in-edges for every node */
  struct fheap** const restrict Q =
    (struct fheap**)malloc((unsigned)maxn*sizeof(struct fheap*));
  Q[0] = NULL; /* the root's queue */
  /* allocate queues */
  for(i = 1; i <= N; ++i) { Q[i] = fheap_new(); }
  for(; i < maxn; ++i)    { Q[i] = NULL; }

  /* obtain the preorder and postorder numberings of the nodes based
     on A */
  int* const restrict pr = (int*)malloc((unsigned)(N + 1)*sizeof(int));
  int* const restrict po = (int*)malloc((unsigned)(N + 1)*sizeof(int));
  make_pre_post(A, pr, po);

  /* there is a slight flaw in the Camerini et al. algorithm where it
     is possible for seek() to pop an edge that is actually necessary.
     in the paper, the authors prove that seek() never pops edges that
     are possible next-edges.  however, it *is* possible for seek() to
     pop an edge that is not a possible next-edge but is necessary for
     reconstruction of the cycle-structure C created in best().  it is
     sufficient to modify seek() to reinsert certain edges which must
     be "kept".  these edges are exactly those which appear in the
     beta-array for all nodes (including cycle-contracted nodes). */
  char* const restrict keep = (char*)calloc((unsigned)M, sizeof(char));
  for(i = 1; A[i] >= 0; ++i) {
    assert(i < maxn);
    assert(keep[A[i]] == 0);
    keep[A[i]] = 1;
  }

  /* this array of flags indicates whether a node has a Y constraint
     set on its in-edge */
  char* const restrict fixed = (char*)calloc((unsigned)N + 1, sizeof(char));
  for(; Y != NULL; Y = Y->p) {
    const int e = Y->e;
    const struct edge* const restrict ed = E + e;
    const int tail = ed->j;
    fixed[tail] = 1;
    /* also insert the edge into this queue */
    const double sc = ed->s;
    fheap_ins(Q[tail], sc, e);
  }

  /* this array of flags indicates whether an edge has a Z constraint
     preventing it from being used */
  char* restrict ignore = (char*)calloc((unsigned)M, sizeof(char));
  for(; Z != NULL; Z = Z->p) { ignore[Z->e] = 1; }

  /* fill the queues */
  int e;
  for(e = 0; e < M; ++e) {
    if(ignore[e]) { continue; } /* skip Z-edges */
    const struct edge* const restrict ed = E + e;
    const int tail = ed->j;
    if(fixed[tail]) { continue; } /* skip Y-edges */
    fheap_ins(Q[tail], ed->s, e);
  }
  /* this is not used again in next() */
  free(ignore); ignore = NULL;

  /* as cycles are contracted, node indices must be remapped.  this
     array is a disjoint-set representation of the remapping */
  int* const restrict D = (int*)malloc((unsigned)maxn*sizeof(int));
  /* fill with self-loops for original nodes */
  for(i = 0; i <= N; ++i) { D[i] = i; }
  for(; i < maxn; ++i)    { D[i] = INT_MIN; }

  /* the following array is an disjoint-set forest that allows
     detection of cycles */
  int* const restrict B = (int*)malloc((unsigned)maxn*sizeof(int));
  /* fill with self-loops for original nodes */
  for(i = 0; i <= N; ++i) { B[i] = i; }
  for(; i < maxn; ++i)    { B[i] = INT_MIN; }

  /* NB: the branching C is unnecessary in next() */

  /* this array tracks the best in-edge for every node */
  int* const restrict beta = (int*)malloc((unsigned)maxn*sizeof(int));
  double* const restrict betasc = (double*)malloc((unsigned)maxn*sizeof(double));
  for(i = 0; i < maxn; ++i) { beta[i] = INT_MIN; betasc[i] = HUGE_VAL; }

  /* these arrays are workspaces for the seek() function */
  int* const restrict worki = (int*)malloc((unsigned)maxn*sizeof(int));
  double* const restrict workd = (double*)malloc((unsigned)maxn*sizeof(double));

  /* step 1: process all nodes and contract cycles until there are no
     unprocessed nodes */
  int nxt = 1; /* the next node to process */
  int n = N;   /* the current number of nodes in the graph, including
		  nodes created by cycle-contraction */
  int next_e = -1;          /* return value: next-edge index */
  double next_d = HUGE_VAL; /* return value: next-edge score diff */
  while(nxt <= n) {
    const double sc = fheap_pop(Q[nxt], &e);
    assert(e >= 0); assert(e < M);
    const struct edge* const restrict ed = E + e;
    /* find the endpoints of this edge in the current graph */
    const int ii = set_find(D, ed->i);
    const int jj = set_find(D, ed->j);
    assert(jj == nxt);
    /* ignore self-loops */
    if(ii == jj) { continue; }
    /* set the best in-edge */
    assert(beta[jj] < 0);
    beta[jj] = e;
    betasc[jj] = sc;

    /* NEW: look for next-edges if the current edge is one which
       actually appears in the MST and if the edge is not constrained
       by Y */
    if(e == A[ed->j] && !fixed[ed->j]) {
      double alt_sc = HUGE_VAL;
      if(seek(Q[jj], ed->j, pr, po, keep, worki, workd, &alt_sc)) {
	/* a valid alternate edge was found */
	const double delta = sc - alt_sc;
	if(next_e == -1 || delta < next_d) {
	  /* this edge beats the current next-edge */
	  next_e = e;
	  next_d = delta;
	}
      }
    }

    /* find the ancestor membership of each node */
    const int set_ii = set_find(B, ii);
    const int set_jj = set_find(B, jj);
    /* detect cycles */
    if(set_ii == set_jj) {
      /* create a new contracted node */
      const int newnode = n + 1;
      assert(newnode < maxn);
      assert(Q[newnode] == NULL);
      /* initialize data structures for this node */
      Q[newnode] = fheap_new();
      B[newnode] = D[newnode] = newnode;
      /* anything which was a descendant of any node in the cycle is
	 now a descendant of the new node */
      set_add(B, set_ii, newnode);
      /* walk the nodes of the cycle and add them to the new node */
      int k = jj;
      do {
	/* adjust the in-edges of this node to reflect the cost of
	   breaking the current best in-edge */
	const double collide_score = betasc[k];
	fheap_add(Q[k], -collide_score); /* shift all scores in heap */
	/* merge the adjusted heap into the new node's heap */
	fheap_take(Q[newnode], Q[k]);
	fheap_destroy(Q[k]); Q[k] = NULL;
	/* locate the next node in the cycle */
	const int in_e = beta[k];
	const int orig_head = E[in_e].i;
	const int next2 = set_find(D, orig_head);
	/* remap the current node to the contracted node */
	set_add(D, k, newnode);
	/* advance to the next node */
	k = next2;
	/* when we get to a node we've already processed, the node
	   should be remapped to the contracted node.  thus it
	   suffices to check this condition to terminate the loop. */
      } while(k != newnode);
      /* increase size of graph */
      n = newnode;
    } else {
      /* no cycle, so just update the ancestry sets */
      set_add(B, set_jj, set_ii);
    }
    /* advance to the next node */
    ++nxt;
  }

  /* the cycle-expansion step is unnecessary for next() */

  free(fixed);
  free(worki);
  free(workd);
  free(keep);
  free(pr);
  free(po);
  free(B);
  free(D);
  for(i = 1; i <= n; ++i) { if(Q[i] != NULL) { fheap_destroy(Q[i]); } }
  free(Q);
  free(betasc);
  free(beta);

  *ret_diff = next_d;
  return next_e;
}


/* Helper function that copies a tree from its edge-index
   representation to a parent-array representation.  the function also
   recovers the score of the tree.  as a special case, this function
   will set the variable INVALID to zero or nonzero depending on
   whether the tree in "beta" is a "valid" tree.  This is only
   relevant for the case of single-root parsing: the algorithm shifts
   the root-edge scores to discourage any tree with more than one root
   edge, but after a certain k, the algorithm will start to enumerate
   multi-root trees. */
static int INVALID = 0;
static inline double copytree(int* const restrict p,
			      const int* const restrict beta) {
  double sc = 0;
  int i;
  int num_roots = 0;
  for(i = 1; i <= N; ++i) {
    const int e = beta[i];
    assert(e >= 0);
    const struct edge* const restrict ed = E + e;
    const int head = ed->i - 1;
    const int mod = i - 1;
    p[mod] = head;
    if(head == -1) { ++num_roots; }
    /* in the single-root case use the original root scores */
    if(MULTIROOT == 0 && head == -1) {
      sc += RS[mod];
    } else {
      sc += ed->s;
    }
  }
  INVALID = 0;
  if(MULTIROOT == 0 && num_roots != 1) { INVALID = 1; }
  return sc;
}


/* Helper function that writes a directed graph to a file. */
static inline void write_digraph(const int id,
				 const struct upnode* restrict Y,
				 const struct upnode* restrict Z) {
  char buf[64];
  sprintf(buf, "digraph.%03d", id);
  FILE* f = fopen(buf, "w");
  if(f == NULL) {
    fprintf(stderr,
	    "*** error: couldn't open file \"%s\" to write digraph\n",
	    buf);
    return;
  }

  /* this array of flags indicates whether a node has a Y constraint
     set on its in-edge */
  char* const restrict fixed = (char*)calloc((unsigned)N + 1, sizeof(char));
  for(; Y != NULL; Y = Y->p) {
    const int e = Y->e;
    const struct edge* const restrict ed = E + e;
    const int tail = ed->j;
    fixed[tail] = 1;
    /* also write the edge */
    fprintf(f, "%d %d %.15g\n", ed->i, ed->j, ed->s);
  }

  /* this array of flags indicates whether an edge has a Z constraint
     preventing it from being used */
  char* const restrict ignore = (char*)calloc((unsigned)M, sizeof(char));
  for(; Z != NULL; Z = Z->p) { ignore[Z->e] = 1; }

  /* fill the queues */
  int e;
  for(e = 0; e < M; ++e) {
    if(ignore[e]) { continue; } /* skip Z-edges */
    const struct edge* const restrict ed = E + e;
    const int tail = ed->j;
    if(fixed[tail]) { continue; } /* skip Y-edges */
    fprintf(f, "%d %d %.15g\n", ed->i, ed->j, ed->s);
  }

  free(ignore);
  free(fixed);

  fclose(f);
}

/* alternate initialization routine for debugging purposes */
int kmst1_load(const char* const fname) {
  FILE* f = fopen(fname, "r");
  assert(f != NULL);
  int i = -1, j = -1, n = -1, m = 0, size = 16;
  double s = HUGE_VAL;
  struct edge* restrict tmpE = (struct edge*)malloc((unsigned)size*sizeof(struct edge));
  while(fscanf(f, "%d %d %lf", &i, &j, &s) == 3) {
    if(m == size) {
      size <<= 1;
      tmpE = (struct edge*)realloc(tmpE, (unsigned)size*sizeof(struct edge));
    }
    tmpE[m] = (struct edge){i, j, s};
    ++m;
    if(j > n) { n = j; }
  }
  fclose(f);
  tmpE = (struct edge*)realloc(tmpE, (unsigned)m*sizeof(struct edge));
  N = n;
  M = m;
  K = 0;
  MULTIROOT = 1;
  E = tmpE;
  treeQ = bheap_new(16);
  return n;
}


/* get the k'th best parse.  it is an error to request parses out of
   order */
int kmst1_get(const int k, int* const restrict p, double* const ret_score) {
  assert(k > K);
  if(k == 1) {
    K = 1;
    const int* const restrict beta = best(NULL, NULL);
    const double sc = copytree(p, beta);
    assert(INVALID == 0);
    *ret_score = sc;
    double next_d;
    const int next_e = next(beta, NULL, NULL, &next_d);
    if(next_e >= 0) {
      struct partition part = {
	next_e, /* pivot edge */
	beta,   /* current MST */
	sc,     /* score of MST */
	NULL,   /* Y constraints */
	NULL    /* Z constraints */
      };
      bheap_ins(treeQ, sc - next_d, part);
      assert(YZ == NULL);
    }
  } else {
    /* recover trees until we reach k */
    struct partition part;
    for(; K < k; ++K) {
      if(treeQ->n == 0) { return 0; } /* ran out of trees */
      const double sc = bheap_pop(treeQ, &part);

      /* create new constraint sets for the inclusion and exclusion of
	 the pivot */
      struct upnode* const Yprime = upnode_new(part.Y, part.e);
      struct upnode* const Zprime = upnode_new(part.Z, part.e);

      /* find the MST when excluding the pivot */
      const int* const restrict beta = best(part.Y, Zprime);
      const double real_sc = copytree(p, beta);
      if(INVALID) { return 0; }
      *ret_score = real_sc;

      /* check that the real score matches the predicted score */
      const double diff = (real_sc > sc ? real_sc - sc : sc - real_sc);
      const double sum =
	(real_sc < 0 ? -real_sc : real_sc) +
	(sc < 0 ? -sc : sc);
      const double rel_diff = diff/sum;
      if(rel_diff > KMST1_REL_TOL) {
	/*write_digraph(K + 1, part.Y, part.Z);*/
	fprintf(stderr, "\
*** error: score mismatch in parse %d:\n\
    actual    %.15g\n\
    predicted %.15g\n\
    rel-diff  %.15g\n",
	       K + 1, real_sc, sc, rel_diff);
	/*assert(0);*/
      }


      /* insert neighbors into the queue */

      /* neighbor of the Z-constrained MST */
      double Znext_d;
      const int Znext_e = next(beta, part.Y, Zprime, &Znext_d);
      if(Znext_e >= 0) {
	struct partition newpart = {
	  Znext_e, /* pivot edge */
	  beta,    /* current MST */
	  real_sc, /* score of MST */
	  part.Y,  /* Y constraints */
	  Zprime   /* Z constraints */
	};
	bheap_ins(treeQ, real_sc - Znext_d, newpart);
	/* add the constraint set to the global list */
	Zprime->next = YZ;
	YZ = Zprime;
      } else {
	/* this constraint set is unused, so deallocate it */
	upnode_destroy(Zprime);
      }

      /* neighbor of the Y-constrained MST */
      double Ynext_d;
      const int Ynext_e = next(part.A, Yprime, part.Z, &Ynext_d);
      if(Ynext_e >= 0) {
	/* just reuse the current partition */
	part.e = Ynext_e;
	part.Y = Yprime;
	bheap_ins(treeQ, part.sc - Ynext_d, part);
	/* add the constraint set to the global list */
	Yprime->next = YZ;
	YZ = Yprime;
      } else {
	/* this constraint set is unused, so deallocate it */
	upnode_destroy(Yprime);
      }
    }
  }
  return k;
}



/* get the best parse.  this function is provided since the k-best
   version generally takes 2 or 3 times as long as just the best()
   procedure  */
double mst1(int* const restrict p, const int multiroot, const int n,
	    const double* const restrict in_E) {
  kmst1_init(multiroot, n, in_E);
  const int* const restrict beta = best(NULL, NULL);
  const double sc = copytree(p, beta);
  assert(INVALID == 0);
  kmst1_clear();
  return sc;
}
